/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/blend.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** blend.c
** Blend Test.
**
** Description -
**    Two random colors are picked. They are verified to exist in
**    the color buffer by drawing points with these colors then
**    reading back the rendered colors. With blending disabled, a
**    point is now drawn at a random location using one of the
**    colors (destination color). Blending is then enabled and
**    another point is drawn to the same location using the other
**    color (source color). The resulting blended color is read
**    and compared to a correct result calculated by the test.
**
** Technical Specification -
**    Buffer requirements:
**	  Color buffer.
**    Color requirements:
**	  Full color.
**    States requirements:
**	  Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**	  Color and zero epsilons.
**    Paths:
**	  Allowed = Alpha, Depth, Fog, Shade, Stencil.
**	  Not allowed = Alias, Blend, Dither, Logicop.
*/

#include "conform.h"
#include "util.h"


static char errStr[320];
static char errStrFmt[][240] = {
    "Blend equation is %s, source blend op is %s, destination blend op is %s. Could not get a valid source color.",
    "Blend equation is %s, source blend op is %s, destination blend op is %s. Could not get a valid destination color.",
    "Blend equation is %s, source blend op is %s, destination blend op is %s. Source color is (%g, %g, %g, %g). Destination color is (%g, %g, %g, %g). Result is (%g, %g, %g, %g), should be (%g, %g, %g, %g)."
};

static void DrawPoint(GLftype x, GLftype y, const GLftype color[4]) 
{
    GLftype vertex[2];
    vertex[0] = x;
    vertex[1] = y;

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FTYPE, 0, &vertex[0]);

    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(4, GL_FTYPE, 0, &color[0]);

    glDrawArrays(GL_POINTS, 0, 1);
    
    glDisableClientState(GL_COLOR_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
}

/*
** Draw the specified color into the frame buffer and read back the
** result. If the result is too different from the specified color,
** return an error. Otherwise, return the result.
*/
static long ResultColor(GLftype *buf, GLftype *result)
{
    DrawPoint(ATOF("0.5"), ATOF("0.5"), buf);

    ReadScreen(0, 0, 1, 1, GL_RGBA, result);

    if (ABS(result[0]-buf[0]) > epsilon.color[0]) {
	return ERROR;
    }
    if (ABS(result[1]-buf[1]) > epsilon.color[1]) {
	return ERROR;
    }
    if (ABS(result[2]-buf[2]) > epsilon.color[2]) {
	return ERROR;
    }
    if (ABS(result[3]-buf[3]) > epsilon.color[3]) {
	if (buffer.colorBits[3] == 0) {
	    if (ABS(result[3]-ITOF(1)) > epsilon.zero) {
		return ERROR;
	    }
	} else {
	    return ERROR;
	}
    }
    return NO_ERROR;
}

/* Generate a random source or destination blend color */
static long GetColor(GLftype *result)
{
    GLftype buf[4];
    long i;

    for (i = 0; i < 4; i++) {
	buf[i] = OESRandom(ITOF(0), ITOF(1));
    }

    return ResultColor(buf, result);
}

static GLenum SrcOp(long index)
{
    switch (index) {
      case 0:
	return GL_ZERO;
      case 1:
	return GL_ONE;
      case 2:
	return GL_DST_COLOR;
      case 3:
	return GL_ONE_MINUS_DST_COLOR;
      case 4:
	return GL_SRC_ALPHA;
      case 5:
	return GL_ONE_MINUS_SRC_ALPHA;
      case 6:
	return GL_DST_ALPHA;
      case 7:
	return GL_ONE_MINUS_DST_ALPHA;
      case 8:
	return GL_SRC_ALPHA_SATURATE;
    }
    return (GLenum)-1;
}

static GLenum DestOp(long index)
{
    switch (index) {
      case 0:
	return GL_ZERO;
      case 1:
	return GL_ONE;
      case 2:
	return GL_SRC_COLOR;
      case 3:
	return GL_ONE_MINUS_SRC_COLOR;
      case 4:
	return GL_SRC_ALPHA;
      case 5:
	return GL_ONE_MINUS_SRC_ALPHA;
      case 6:
	return GL_DST_ALPHA;
      case 7:
	return GL_ONE_MINUS_DST_ALPHA;
    }
    return (GLenum)-1;
}

/*
** These errors account for arithmetic errors in blending, including cases
** where alpha has lower precision than color. One epsilon.color is
** assessed for each operation not involving destination alpha, and
** one max(epsilon.color,epsilon.alpha) for each operation that does
** involve destination alpha. This is extremely generous in the
** case of 1-bit alpha.
*/
static void GetOpError(GLenum srcOp, GLenum destOp,
		       GLftype *opError)
{
    GLint colorErr = 0, alphaErr = 0, i;

    /* Accumulated error for the source factor */
    switch (srcOp) {
	case GL_ZERO: case GL_ONE:
	    /* Identity and zero operations should incur no error */
	    break;
	case GL_DST_ALPHA: case GL_ONE_MINUS_DST_ALPHA:
	    alphaErr++;
	    break;
	case GL_SRC_ALPHA:
	case GL_ONE_MINUS_SRC_ALPHA:
	case GL_SRC_ALPHA_SATURATE:
	case GL_DST_COLOR: case GL_ONE_MINUS_DST_COLOR:
	    colorErr++;
	    break;
    }

    /* Accumulated error for the destination factor */
    switch (destOp) {
	case GL_ZERO: case GL_ONE:
	    /* Identity and zero operations should incur no error */
	    break;
	case GL_DST_ALPHA: case GL_ONE_MINUS_DST_ALPHA:
	    alphaErr++;
	    break;
	case GL_SRC_COLOR: case GL_ONE_MINUS_SRC_COLOR:
	case GL_SRC_ALPHA: case GL_ONE_MINUS_SRC_ALPHA:
	    colorErr++;
	    break;
    }

    /* Accumulated error for the blend operation */
    colorErr++;

    /* Compute total error expected for each component. If there
    ** are no destination alpha bits, it is assumed to be ONE,
    ** so no error is incurred.
    */
    for (i = 0; i < 4; i++) {
	if (buffer.colorBits[3] > 0) {
	    opError[i] = MUL(ITOF(colorErr), epsilon.color[i]) +
			 MUL(ITOF(alphaErr), epsilon.color[3]);
	} else {
	    opError[i] = MUL(ITOF(colorErr), epsilon.color[i]);
	}
    }
}

static void TrueResult(GLenum srcOp, GLenum destOp,
		       GLftype *src, GLftype *dest,
		       GLftype *result)
{
    GLftype srcResult[4], destResult[4], srcTerm[4], destTerm[4];
    long i;

    if (buffer.colorBits[3] == 0) {
	dest[3] = ITOF(1);
    }

    /* Default initialization (just to avoid compiler warnings). */
    for (i = 0; i < 4; i++) {
        destResult[i] = ITOF(0);
        srcResult[i] = ITOF(0);
    }

    switch (srcOp) {
      case GL_ZERO:
	for (i = 0; i < 4; i++) {
	    srcResult[i] = ITOF(0);
	}
	break;
      case GL_ONE:
	for (i = 0; i < 4; i++) {
	    srcResult[i] = ITOF(1);
	}
	break;
      case GL_DST_COLOR:
	for (i = 0; i < 4; i++) {
	    srcResult[i] = dest[i];
	}
	break;
      case GL_ONE_MINUS_DST_COLOR:
	for (i = 0; i < 4; i++) {
	    srcResult[i] = ITOF(1) - dest[i];
	}
	break;
      case GL_SRC_ALPHA:
	for (i = 0; i < 4; i++) {
	    srcResult[i] = src[3];
	}
	break;
      case GL_ONE_MINUS_SRC_ALPHA:
	for (i = 0; i < 4; i++) {
	    srcResult[i] = ITOF(1) - src[3];
	}
	break;
      case GL_DST_ALPHA:
	for (i = 0; i < 4; i++) {
	    srcResult[i] = dest[3];
	}
	break;
      case GL_ONE_MINUS_DST_ALPHA:
	for (i = 0; i < 4; i++) {
	    srcResult[i] = ITOF(1) - dest[3];
	}
	break;
      case GL_SRC_ALPHA_SATURATE:
	for (i = 0; i < 3; i++) {
	    srcResult[i] = (src[3] <= ITOF(1) - dest[3]) ? src[3] : (ITOF(1) - dest[3]);
	}
	srcResult[3] = ITOF(1);
	break;
    }

    switch (destOp) {
      case GL_ZERO:
	for (i = 0; i < 4; i++) {
	    destResult[i] = ITOF(0);
	}
	break;
      case GL_ONE:
	for (i = 0; i < 4; i++) {
	    destResult[i] = ITOF(1);
	}
	break;
      case GL_SRC_COLOR:
	for (i = 0; i < 4; i++) {
	    destResult[i] = src[i];
	}
	break;
      case GL_ONE_MINUS_SRC_COLOR:
	for (i = 0; i < 4; i++) {
	    destResult[i] = ITOF(1) - src[i];
	}
	break;
      case GL_SRC_ALPHA:
	for (i = 0; i < 4; i++) {
	    destResult[i] = src[3];
	}
	break;
      case GL_ONE_MINUS_SRC_ALPHA:
	for (i = 0; i < 4; i++) {
	    destResult[i] = ITOF(1) - src[3];
	}
	break;
      case GL_DST_ALPHA:
	for (i = 0; i < 4; i++) {
	    destResult[i] = dest[3];
	}
	break;
      case GL_ONE_MINUS_DST_ALPHA:
	for (i = 0; i < 4; i++) {
	    destResult[i] = ITOF(1) - dest[3];
	}
	break;
    }

    for (i = 0; i < 4; i++) {
	srcTerm[i] = MUL(srcResult[i], src[i]);
	destTerm[i] = MUL(destResult[i], dest[i]);
    }

    for (i = 0; i < 4; i++) {
	result[i] = srcTerm[i] + destTerm[i];
    }

    for (i = 0; i < 4; i++) {
	if (result[i] < ITOF(0)) {
	    result[i] = ITOF(0);
	}
	if (result[i] > ITOF(1)) {
	    result[i] = ITOF(1);
	}
    }
}

static long Test(GLftype *buf, GLftype *result, GLftype *opError)
{
    if (ABS(buf[0]-result[0]) > opError[0] + epsilon.color[0]) {
	return ERROR;
    }
    if (ABS(buf[1]-result[1]) > opError[1] + epsilon.color[1]) {
	return ERROR;
    }
    if (ABS(buf[2]-result[2]) > opError[2] + epsilon.color[2]) {
	return ERROR;
    }
    if (ABS(buf[3]-result[3]) > opError[3] + epsilon.color[3]) {
	if (buffer.colorBits[3] == 0) {
	    if (ABS(buf[3]-ITOF(1)) > epsilon.zero) {
		return ERROR;
	    }
	} else {
	    return ERROR;
	}
    }
    return NO_ERROR;
}

/*
** Test all combinations of source and destination blend factors.
** The blend equation and constant blend color are used in
** generating error messages.
*/
static long TestBlending(void)
{
    char tmp0[40], tmp1[40], tmp2[40];
    GLftype srcColor[4], destColor[4], buf[4], result[4], opError[4];
    GLint x, y, i, j;
    GLint numSrcModes, numDestModes;

    numSrcModes = 9;
    numDestModes = 8;
    STRCPY(tmp0, "default");

    for (i = 0; i < numSrcModes; i++) {
	for (j = 0; j < numDestModes; j++) {
	    GetEnumName(SrcOp(i), tmp1);
	    GetEnumName(DestOp(j), tmp2);

	    glDisable(GL_BLEND);

	    if (GetColor(srcColor) == ERROR) {
		StrMake(errStr, errStrFmt[0], tmp0, tmp1, tmp2);
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }

	    if (GetColor(destColor) == ERROR) {
		StrMake(errStr, errStrFmt[1], tmp0, tmp1, tmp2);
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }

	    TrueResult(SrcOp(i), DestOp(j), srcColor, destColor,
		       result);

	    x = FTOI(OESRandom(ITOF(0), ITOF(WINDSIZEX-1)));
	    y = FTOI(OESRandom(ITOF(0), ITOF(WINDSIZEY-1)));

	    if (buffer.colorBits[3] == 0) {
		destColor[3] = ITOF(1);
		DrawPoint(ITOF(x) + ATOF("0.5"), ITOF(y) + ATOF("0.5"), result);
		ReadScreen(x, y, 1, 1, GL_RGBA, result);
	    }

	    DrawPoint(ITOF(x) + ATOF("0.5"), ITOF(y) + ATOF("0.5"), destColor);

	    glEnable(GL_BLEND);
	    glBlendFunc(SrcOp(i), DestOp(j));

	    DrawPoint(ITOF(x) + ATOF("0.5"), ITOF(y) + ATOF("0.5"), srcColor);

	    ReadScreen(x, y, 1, 1, GL_RGBA, buf);

	    GetOpError(SrcOp(i), DestOp(j), opError);
	    if (Test(buf, result, opError) == ERROR) {
		StrMake(errStr, errStrFmt[2], tmp0, tmp1, tmp2,
			srcColor[0], srcColor[1], srcColor[2], srcColor[3],
			destColor[0], destColor[1], destColor[2], destColor[3],
			buf[0], buf[1], buf[2], buf[3],
			result[0], result[1], result[2], result[3]);
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	}
    }

    return NO_ERROR;
}

long BlendExec(void)
{
    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    if (TestBlending() == ERROR)
	return ERROR;

    return NO_ERROR;
}
