/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
**
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/bufobj.c,v 1.9 2005/07/08 09:58:11 artlab Exp $
*/

/* 
   bufobj.c
   Buffer object tests

   Description
   - Test that the buffer objects are properly implemented. The tests consist of:

   BOBuildUp & BOTearDown:
   - Buildup the test information and tear it down. The buildup will set up the GL to a
     state suitable for the BO tests. It also creates all the necessary buffer objects
     and ensures that the data contained in those objects cannot reside in the client
     memory.

   BOTestGenBuffers:
   - Tests the non-visual functionality of the buffer objects. These are not exactly
     similar to the primitive or coverage tests, since the error statuses are checked.
   - Generate buffer objects.
   - Bind them to all possible targets.
   - Check that the buffers are valid.
   - Free buffer binding.
   - Set buffer (arbitrary) data on invalid object.
   - Set buffer (arbitrary) data on invalid object.
   - Delete buffer objects.
   - Try deleting while bound, must fail.
   - Generate with incorrect count.
   - Delete with incorrect count.
   - Try deleting the "default" buffer object.

   BOTestTriangles:
   - Verify rendering output when using buffer objects as data source. The test verifies
     the following kinds of buffers in a single pass: Vertices, Normals, Colors, and
     texture coordinates. This is done by rendering triangles and checking colors matches,
     which should be according to the geometry, normals, colors and texture.

   BOTestMultiTexturing:
   - Test multiple texturing units (2, which is the minimum) work ok from buffer objects.
     A checker texture is rendered with two different repeat frequencies. The rendering
     output is validated by traversing the framebuffer and checking for color variation
     to appear in proper order and frequency.
   
   BOTestElements:
   - Test that DrawElements works from index and vertex data stored into buffer objects.
     A Set of triangles is rendered, and colors are sampled from the framebuffer to see
     that the output is correct.

   BOTestBufferSubData:
   - Tests that buffer server-side data is replaced correctly with BufferSubData.
     Again, a set of triangles is rendered, and the output is validated.

   BOTestPointSize:
   - Tests that point size is correctly read from a buffer object. The point size is
     tested by asserting that the resulting "objects" are increasing in size.

   BoTestMatrixPalette:
     Two matrices are defined the second of which causes the view to be "flipped" in
     the Y direction. The weights rendering is set up so that the flipping can be
     detected by using checkerboard textures.
   
*/

#include "conform.h"
#include "util.h"

#include "defines.h"

#include <assert.h>

#ifdef GL_OES_VERSION_1_1

static char errStr[320];

#define BUFOBJ_TEXTURES 2
GLuint gTextures[BUFOBJ_TEXTURES] = {0};

enum {
    BUFFER_TRIANGLES = 0,
    BUFFER_TRI_NORMALS,
    BUFFER_TRI_COLORS,
    BUFFER_POINTS,
    BUFFER_ELEMENT_VERTICES,
    BUFFER_ELEMENT_INDICES,
    BUFFER_TEXTURE0,
    BUFFER_TEXTURE1,
    BUFFER_POINT_SIZES,
    BUFFER_MATRIX_INDICES,
    BUFFER_MATRIX_WEIGHTS,
    BUFFER_LAST_ENUM
};

GLuint gBufferObjs[BUFFER_LAST_ENUM] = {0}; 

static const GLftype gTriVertexArray[] = {
    ITOF(WINDSIZEX) / 4, ITOF(WINDSIZEY) / 4,
    3 * ITOF(WINDSIZEX) / 4, ITOF(WINDSIZEY) / 4,
    3 * ITOF(WINDSIZEX) / 4, 3 * ITOF(WINDSIZEY) / 4,
    ITOF(WINDSIZEX) / 4, ITOF(WINDSIZEY) / 4,
    3 * ITOF(WINDSIZEX) / 4, 3 * ITOF(WINDSIZEY) / 4,
    ITOF(WINDSIZEX) / 4, 3 * ITOF(WINDSIZEY) / 4
};

static const GLftype gTriSubDataVertexArray[] = {
    ITOF(WINDSIZEX) / 2, ITOF(WINDSIZEY) / 2,
    3 * ITOF(WINDSIZEX) / 4, 3 * ITOF(WINDSIZEY) / 4,
    ITOF(WINDSIZEX) / 4, 3 * ITOF(WINDSIZEY) / 4
};

static const GLftype gTriNormalArray[] = {
    ITOF(0), ITOF(0), ITOF(1),
    ITOF(0), ITOF(0), ITOF(-1),
    ITOF(0), ITOF(0), ITOF(1),
    ITOF(0), ITOF(0), ITOF(1),
    ITOF(0), ITOF(0), ITOF(1),
    ITOF(0), ITOF(0), ITOF(1)
};

/* -- Color Data */
static const GLftype gTriColorArray[] = {
    ITOF(1), ITOF(0), ITOF(0), ITOF(1),
    ITOF(1), ITOF(0), ITOF(0), ITOF(1),
    ITOF(1), ITOF(0), ITOF(0), ITOF(1),
    ITOF(0), ITOF(1), ITOF(0), ITOF(1),
    ITOF(0), ITOF(1), ITOF(0), ITOF(1),
    ITOF(0), ITOF(1), ITOF(0), ITOF(1)
};

static const GLftype gTriTexCoordArray[] = {
    ITOF(0), ITOF(0),
    ITOF(1), ITOF(0),
    ITOF(1), ITOF(1),
    ITOF(0), ITOF(0),
    ITOF(1), ITOF(1),
    ITOF(0), ITOF(1)
};

#define CHECKER2_REPEATS 4
static const GLftype gTriTexCoordArray2[] = {
    ITOF(0), ITOF(0),
    ITOF(CHECKER2_REPEATS), ITOF(0),
    ITOF(CHECKER2_REPEATS), ITOF(CHECKER2_REPEATS),
    ITOF(0), ITOF(0),
    ITOF(CHECKER2_REPEATS), ITOF(CHECKER2_REPEATS),
    ITOF(0), ITOF(CHECKER2_REPEATS)
};

static const GLubyte gCheckerTextureData[] = {
    0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff 
};

static const GLubyte gChecker2TextureData[] = {
    0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0xff, 
    0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0xff, 0xff
};


/* - DrawElements test */

static const GLftype gElementVertices[] = {
    ITOF(WINDSIZEX) / 4, ITOF(WINDSIZEY) / 4,
    ITOF(WINDSIZEX) / 2, ITOF(WINDSIZEY) / 2,
    3 * ITOF(WINDSIZEX) / 4, ITOF(WINDSIZEY) / 4,
    3 * ITOF(WINDSIZEX) / 4, 3 * ITOF(WINDSIZEY) / 4,
    ITOF(WINDSIZEX) / 2, ITOF(WINDSIZEY) / 2, 
    ITOF(WINDSIZEX) / 4, 3 * ITOF(WINDSIZEY) / 4
};

static const GLushort gElementIndices[] = {
    0, 1, 5, 2, 3, 1
};

static const GLushort gElementIndexSubData[] = {
    3, 4, 5
};

static const GLftype gPointSizes[] = {
    3, 7, 11, 15
};

static const GLftype gPointVertices[] = {
    WINDSIZEX / 4, 3 * WINDSIZEY / 4,
    3 * WINDSIZEX / 4, 3 * WINDSIZEY / 4,
    WINDSIZEX / 4, WINDSIZEY / 4,
    3 * WINDSIZEX / 4, WINDSIZEY / 4
};

static const GLubyte gTriMatrixIndexArray[] = 
{
    0, 1,
    0, 1,
    0, 1,
    0, 1,
    0, 1,
    0, 1
};

static const GLftype gTriMatrixWeightArray[] =
{
    ITOF(1), ITOF(0),
    ITOF(1), ITOF(0),
    ITOF(1), ITOF(0),
    ITOF(1), ITOF(0),
    ITOF(1), ITOF(0),
    ITOF(1), ITOF(0),

    ITOF(0), ITOF(1),
    ITOF(0), ITOF(1),
    ITOF(0), ITOF(1),
    ITOF(0), ITOF(1),
    ITOF(0), ITOF(1),
    ITOF(0), ITOF(1)
};

int IsBlack(GLftype* pBuf)
{
    if ( ABS(pBuf[0] - ITOF(0)) > epsilon.color[0] || 
         ABS(pBuf[1] - ITOF(0)) > epsilon.color[1] || 
         ABS(pBuf[2] - ITOF(0)) > epsilon.color[2] )
    {
        return 0;
    }
    return 1;
}

int IsWhite(GLftype* pBuf)
{
    if ( ABS(pBuf[0] - ITOF(1)) > epsilon.color[0] || 
         ABS(pBuf[1] - ITOF(1)) > epsilon.color[1] || 
         ABS(pBuf[2] - ITOF(1)) > epsilon.color[2] )
    {
        return 0;
    }
    return 1;
}

long BOTestGenBuffers()
{
#define N_TESTED_BUFS 2

    static const GLint targets[] = { GL_ARRAY_BUFFER, GL_ELEMENT_ARRAY_BUFFER };
    static const GLint usages[] = { GL_STATIC_DRAW, GL_DYNAMIC_DRAW };

    GLuint bufs[N_TESTED_BUFS]; 
    GLuint twinBufs[2] = {0, 0};
    GLuint dummy0 = 0;

    int usg, tg, i;

    memset(bufs, 0, N_TESTED_BUFS * sizeof(GLuint));

    glGetError();

    glGenBuffers(N_TESTED_BUFS, bufs);
    if( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Could not create test buffers.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* Test different targets and different binds */
    for( usg = 0; usg < 2; usg++ ) /* Loop both usage types */
    {
        for( tg = 0; tg < 2; tg++ ) /* Loop both target types */
        {
            for( i = 0; i < N_TESTED_BUFS; i++)
            {
                glBindBuffer(targets[tg], bufs[i]);
                if( glGetError() != GL_NO_ERROR )
                {
                    StrMake(errStr, "Could not bind buffers.");
                    ErrorReport(__FILE__, __LINE__, errStr);
                    return ERROR;
                }
                if( !glIsBuffer(bufs[i]) )
                {
                    StrMake(errStr, "glIs buffer not functioning properly");
                    ErrorReport(__FILE__, __LINE__, errStr);
                    return ERROR;
                }
            }
            glBindBuffer(targets[tg], 0); /* Must free the target. */
            /* Test that the data cannot be set: glBufferData */
            glBufferData(targets[tg], sizeof(gTriVertexArray), gTriVertexArray, usages[usg]);
            if ( glGetError() == GL_NO_ERROR )
            {
                StrMake(errStr, "glBufferData not returning failure state.");
                ErrorReport(__FILE__, __LINE__, errStr);
                return ERROR;
            }
            /* glBufferSubData */
            glBufferSubData(targets[tg], 0, sizeof(gTriVertexArray), gTriVertexArray);
            if ( glGetError() == GL_NO_ERROR )
            {
                StrMake(errStr, "glBufferSubData not returning failure state.");
                ErrorReport(__FILE__, __LINE__, errStr);
                return ERROR;
            }
        }
    }

    glDeleteBuffers(N_TESTED_BUFS, bufs);

    /* Test the case, where object is still bound when trying to delete */
    glGenBuffers(2, twinBufs);

    glBindBuffer(GL_ARRAY_BUFFER, twinBufs[0]);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, twinBufs[1]);
    
    glDeleteBuffers(N_TESTED_BUFS, twinBufs);
    if( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "System allowed deleting bound buffers.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }
    
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glDeleteBuffers(N_TESTED_BUFS, bufs);

    glGenBuffers(-1, bufs);
    if( glGetError() == GL_NO_ERROR ) 
    {   
        StrMake(errStr, "No error on invalid number of generated buffers.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR; 
    }
    glDeleteBuffers(-1, bufs);
    if( glGetError() == GL_NO_ERROR ) 
    {   
        StrMake(errStr, "No error on invalid number of deleted buffers.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR; 
    }

    /* Test "0", should be a NOP */
    glDeleteBuffers(1, &dummy0); /* Quiet ignore */

    return NO_ERROR;

#undef N_TESTED_BUFS
}

long BOTestTriangles()
{
    /* Test triangle rendering with normals, texture, colors */
    static const char elementsErrorFmt[] = 
        "Incorrectly rasterized buffer object: expected [%.6f, %.6f, %.6f], got [%.6f, %.6f, %.6f]";
    GLftype buf[3];
    GLftype bufCmp[3];

    static const GLftype ltAmbient[] = {ITOF(0), ITOF(0), ITOF(0), ITOF(0)};
    static const GLftype ltDiffuse[] = {ITOF(1), ITOF(1), ITOF(1), ITOF(1)};
    static const GLftype ltSpecular[] = {ITOF(1), ITOF(1), ITOF(1), ITOF(1)};
    static const GLftype ltPosition[] = {ITOF(0), ITOF(0), ITOF(1), ITOF(0)};
    
    glClear(GL_COLOR_BUFFER_BIT);

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    glLighttv(GL_LIGHT0, GL_AMBIENT, ltAmbient);
    glLighttv(GL_LIGHT0, GL_DIFFUSE, ltDiffuse);
    glLighttv(GL_LIGHT0, GL_SPECULAR, ltSpecular);
    glLighttv(GL_LIGHT0, GL_POSITION, ltPosition);
    glLightt(GL_LIGHT0, GL_SPOT_CUTOFF, ITOF(180));
    glLightt(GL_LIGHT0, GL_SPOT_EXPONENT, ITOF(0));

    glClientActiveTexture(GL_TEXTURE0);
    glActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);

    glBindTexture(GL_TEXTURE_2D, gTextures[0]);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TRI_NORMALS]);
    glNormalPointer(GL_FTYPE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TRI_COLORS]);
    glColorPointer(4, GL_FTYPE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TEXTURE0]);
    glTexCoordPointer(2, GL_FTYPE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TRIANGLES]);
    glVertexPointer(2, GL_FTYPE, 0, 0);

    glDrawArrays(GL_TRIANGLES, 0, 6);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glDisable(GL_TEXTURE_2D);

    glDisable(GL_LIGHTING);
    glDisable(GL_COLOR_MATERIAL);

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    /* See that: 
       - The vertices are contained correctly and (at least) a triangle is rendered
       - The texture is applied correctly
       - The colors have been applied correctly.
    */
    ReadScreen(WINDSIZEX / 4 + 1, 3 * WINDSIZEY / 4 - 1, 1, 1, GL_RGB, buf);
    if ( ABS(buf[0] - ITOF(0)) > epsilon.color[0] || 
         ABS(buf[1] - ITOF(1)) > epsilon.color[1] || 
         ABS(buf[2] - ITOF(0)) > epsilon.color[2] )
    {
        StrMake(errStr, elementsErrorFmt, .0f, 1.0f, .0f, buf[0], buf[1], buf[2]);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    ReadScreen(3 * WINDSIZEX / 8 + 1, 3 * WINDSIZEY / 8 + 1, 1, 1, GL_RGB, buf);
    if ( ABS(buf[0] - ITOF(0)) > epsilon.color[0] || 
         ABS(buf[1] - ITOF(0)) > epsilon.color[1] || 
         ABS(buf[2] - ITOF(0)) > epsilon.color[2] )
    {
        StrMake(errStr, elementsErrorFmt, .0f, .0f, .0f, buf[0], buf[1], buf[2]);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* See that the normals are applied correctly. */
    ReadScreen(3 * WINDSIZEX / 4 - 1, WINDSIZEY / 4 + 1, 1, 1, GL_RGB, buf);
    ReadScreen(5 * WINDSIZEX / 8, 3 * WINDSIZEY / 8 + 1, 1, 1, GL_RGB, bufCmp);
    if ( (bufCmp[0] - buf[0]) < 2 * epsilon.color[0] || bufCmp[0] < 2 * epsilon.color[0] || 
         ABS(buf[1] - ITOF(0)) > epsilon.color[1] || 
         ABS(buf[2] - ITOF(0)) > epsilon.color[2] )
    {
        StrMake(errStr, "Buffer object incorrectly wraps a normal.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    return NO_ERROR;
}

long BOTestMultiTexturing()
{
    /*  Most of this must have been tested in the DrawArrays tests,
        Ensure that multitexturing "works" through objects also 
    */

    GLftype bufColors[WINDSIZEX / 2][3];

    int i;
    int prevPx = 0;
    int whiteCount = 0;
    int blackCount = 0;

    glClear(GL_COLOR_BUFFER_BIT);

    glEnableClientState(GL_VERTEX_ARRAY);

    /* ---  First multitexture test: Plain multitexturing. */

    glClientActiveTexture(GL_TEXTURE0);
    glActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);
    
    /*  Bind the buffer object containing the texture coords for the first
        checker texture. Set the ClientActiveTexture accordingly. Set the pointer.
    */
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TEXTURE0]); 
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindTexture(GL_TEXTURE_2D, gTextures[0]);
    glTexCoordPointer(2, GL_FTYPE, 0, 0);
     
    /*  Same for the second texturing unit. */
    glActiveTexture(GL_TEXTURE1);
    glEnable(GL_TEXTURE_2D);
    glClientActiveTexture(GL_TEXTURE1);

    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TEXTURE1]); 
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindTexture(GL_TEXTURE_2D, gTextures[1]);
    glTexCoordPointer(2, GL_FTYPE, 0, 0);

    /*  Finally setup the vertex buffer object. */
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TRIANGLES]);
    glVertexPointer(2, GL_FTYPE, 0, 0);
    glDrawArrays(GL_TRIANGLES, 0, 6);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glDisableClientState(GL_VERTEX_ARRAY);

    glActiveTexture(GL_TEXTURE1);
    glDisable(GL_TEXTURE_2D);
    glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    glActiveTexture(GL_TEXTURE0);
    glDisable(GL_TEXTURE_2D);
    glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    
    ReadScreen(WINDSIZEX / 4 + 1, 3 * WINDSIZEY / 4 - 1, WINDSIZEX / 2, 1, GL_RGB, (GLftype*)bufColors);

    prevPx = 1; /* We should start from a black pixel */
    for( i = 0; i < WINDSIZEX / 2; i++)
    {
        if(ABS(bufColors[i][0] - ITOF(0)) < epsilon.color[0] && prevPx != 0)
        {
            prevPx = 0;
            blackCount++;
        }
        if(ABS(bufColors[i][0] - ITOF(1)) < epsilon.color[0] && prevPx != 1)
        {
            prevPx = 1;
            whiteCount++;
        }
    }
    if( blackCount != 3 || whiteCount != 2)
    {
        StrMake(errStr, "Multitexturing with buffer objects failed.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    return NO_ERROR;
}

long BOTestElements()
{
    /* Test DrawElements on buffer objects. */
    static const char elementsErrorFmt[] = 
        "Incorrectly rasterized buffer object: expected [%.6f, %.6f, %.6f], got [%.6f, %.6f, %.6f]";
    GLftype buf[3];

    glClear(GL_COLOR_BUFFER_BIT);
    glColor4t(ITOF(1), ITOF(1), ITOF(1), ITOF(1));

    glEnableClientState(GL_VERTEX_ARRAY);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, gBufferObjs[BUFFER_ELEMENT_INDICES]);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_ELEMENT_VERTICES]);
    glVertexPointer(2, GL_FTYPE, 0, 0);
    glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, 0);

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glDisableClientState(GL_VERTEX_ARRAY);

    /* Test four pixels from the pattern, two of them should be white, and two
       black */
    ReadScreen(WINDSIZEX / 2, 3 * WINDSIZEY / 4 - 1, 1, 1, GL_RGB, buf);
    if ( ABS(buf[0] - 0) > epsilon.color[0] || 
         ABS(buf[1] - 0) > epsilon.color[1] || 
         ABS(buf[2] - 0) > epsilon.color[2] )
    {
        StrMake(errStr, elementsErrorFmt, .0f, .0f, .0f, buf[0], buf[1], buf[2]);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    ReadScreen(WINDSIZEX / 2, WINDSIZEY / 4, 1, 1, GL_RGB, buf);
    if ( ABS(buf[0] - 0) > epsilon.color[0] || 
         ABS(buf[1] - 0) > epsilon.color[1] || 
         ABS(buf[2] - 0) > epsilon.color[2] )
    {
        StrMake(errStr, elementsErrorFmt, .0f, .0f, .0f, buf[0], buf[1], buf[2]);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    ReadScreen(WINDSIZEX / 4 + 1, WINDSIZEY / 2, 1, 1, GL_RGB, buf);
    if ( ABS(buf[0] - ITOF(1)) > epsilon.color[0] || 
         ABS(buf[1] - ITOF(1)) > epsilon.color[1] || 
         ABS(buf[2] - ITOF(1)) > epsilon.color[2] )
    {
        StrMake(errStr, elementsErrorFmt, 1.0f, 1.0f, 1.0f, buf[0], buf[1], buf[2]);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    ReadScreen(3 * WINDSIZEX / 4 - 1, WINDSIZEY / 2, 1, 1, GL_RGB, buf);
    if ( ABS(buf[0] - ITOF(1)) > epsilon.color[0] || 
         ABS(buf[1] - ITOF(1)) > epsilon.color[1] || 
         ABS(buf[2] - ITOF(1)) > epsilon.color[2] )
    {
        StrMake(errStr, elementsErrorFmt, 1.0f, 1.0f, 1.0f, buf[0], buf[1], buf[2]);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    return NO_ERROR;
}

long BOTestBufferSubData(void)
{
    GLftype buf[3];
    GLuint tempObject = 0;

    glClear(GL_COLOR_BUFFER_BIT);

    glEnableClientState(GL_VERTEX_ARRAY);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, gBufferObjs[BUFFER_ELEMENT_INDICES]);

    glGenBuffers(1, &tempObject);
    glBindBuffer(GL_ARRAY_BUFFER, tempObject);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gElementVertices), gElementVertices, GL_STATIC_DRAW);

    glVertexPointer(2, GL_FTYPE, 0, 0);

    glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 3 * sizeof(GLushort), 3 * sizeof(GLushort), gElementIndexSubData);
    glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_SHORT, 0);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glDisableClientState(GL_VERTEX_ARRAY);

    glDeleteBuffers(1, &tempObject);

    ReadScreen(WINDSIZEX / 2, 3 * WINDSIZEY / 8, 1, 1, GL_RGB, buf);
    if ( !IsBlack(buf) )
    {
        StrMake(errStr, "BufferSubData did not replace buffer object data correctly.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    ReadScreen(WINDSIZEX / 2, 3 * WINDSIZEY / 4 - 1, 1, 1, GL_RGB, buf);
    if ( !IsWhite(buf) )
    {
        StrMake(errStr, "BufferSubData did not replace buffer object data correctly.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }
    return NO_ERROR;
}

long BOTestPointSize(void)
{
    /* This tests both point rendering, and the mandatory point size extension. */

#ifndef GL_OES_point_size_array
    return NO_ERROR;
#else
    GLftype buf[(15 + 2) * 3]; 
    int i, psz, x, y;
    int c;
    int lastC = -1;

    glClear(GL_COLOR_BUFFER_BIT);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_POINT_SIZE_ARRAY_OES);

    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_POINT_SIZES]);
    glPointSizePointerOES(GL_FTYPE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glVertexPointer(2, GL_FTYPE, 0, gPointVertices);
    glDrawArrays(GL_POINTS, 0, 4);

    glPointSizet(ITOF(1));

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_POINT_SIZE_ARRAY_OES);

    /* Verify that the rendered point sizes grow sequentially.*/
    for( i = 0; i < 4; i++ )
    {
        psz = gPointSizes[i]; 
        y = i < 2 ? (3 * WINDSIZEY / 4) : (WINDSIZEY / 4);
        ReadScreen(WINDSIZEX / 4 + ((i % 2) * WINDSIZEX / 2) - psz / 2 - 1, y, psz + 2, 1, GL_RGB, buf);
        
        c = 0;
        for( x = 0; x < psz + 2; x++ )
        {
            if(!IsBlack(buf + 3 * x))
            {
                c++;
            }
        }
        if( c <= lastC )
        {
            StrMake(errStr, "PointSizeArray not correctly read from a buffer object.");
            ErrorReport(__FILE__, __LINE__, errStr);
            return ERROR;
        }
    }
    return NO_ERROR;    

#endif
}

void BOgluOrtho2D(GLftype left, GLftype right, GLftype bottom, GLftype top)
{
    /*  Loads a 2D ortho-matrix as the current matrix. Differs from the utility
        function Ortho2D by not assuming everything goes into the projection
        matrix.
    */

    GLftype m[4][4];
    GLftype deltaX, deltaY;

    MakeIdentMatrix(&m[0][0]);
    deltaX = right - left;
    deltaY = top - bottom;
    m[0][0] = DIV(ITOF(2), deltaX);
    m[3][0] = DIV(-(right + left), deltaX);
    m[1][1] = DIV(ITOF(2), deltaY);
    m[3][1] = DIV(-(top + bottom), deltaY);
    m[2][2] = ITOF(-1);

    glLoadMatrixt(&m[0][0]);
}

long BOTestMatrixPalette(void)
{
#ifndef GL_OES_matrix_palette
    return NO_ERROR;
#else
    /* Test matrix palettes */
    GLftype buf[3];
    int err = NO_ERROR;

    GLint maxMatrices, maxUnits;

    if (!tkCheckExtension("GL_OES_matrix_palette"))
    {
        return NO_ERROR;
    }

    glGetIntegerv(GL_MAX_PALETTE_MATRICES_OES, &maxMatrices);
    glGetIntegerv(GL_MAX_VERTEX_UNITS_OES, &maxUnits);

    if (maxMatrices < 2 || maxUnits < 2)
    {
        return NO_ERROR;
    }

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glEnable(GL_MATRIX_PALETTE_OES);
    glMatrixMode(GL_MATRIX_PALETTE_OES);
    
    glCurrentPaletteMatrixOES(0);
    glLoadIdentity();
    /*Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));*/
    BOgluOrtho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glCurrentPaletteMatrixOES(1);
    glLoadIdentity();
    /*Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(WINDSIZEY), ITOF(0));*/
    BOgluOrtho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(WINDSIZEY), ITOF(0));

    glClear(GL_COLOR_BUFFER_BIT);

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_MATRIX_INDEX_ARRAY_OES);
    glEnableClientState(GL_WEIGHT_ARRAY_OES);

    glClientActiveTexture(GL_TEXTURE0);
    glActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);

    glBindTexture(GL_TEXTURE_2D, gTextures[0]);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TEXTURE0]);
    glTexCoordPointer(2, GL_FTYPE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TRIANGLES]);
    glVertexPointer(2, GL_FTYPE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_MATRIX_INDICES]);
    glMatrixIndexPointerOES(2, GL_UNSIGNED_BYTE, 0, 0);
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_MATRIX_WEIGHTS]);
    glWeightPointerOES(2, GL_FTYPE, 0, 0);

    glDrawArrays(GL_TRIANGLES, 0, 6);

    ReadScreen(WINDSIZEX / 4 + 1, 3 * WINDSIZEY / 4 - 1, 1, 1, GL_RGB, buf);
    if ( !IsWhite(buf) )
    {
        StrMake(errStr, "Matrix palettes not working properly from within buffer objects");
        ErrorReport(__FILE__, __LINE__, errStr);
        err = ERROR;
    }

    glWeightPointerOES(2, GL_FTYPE, 0, (void*)(2 * 6 * sizeof(GLftype)));

    glDrawArrays(GL_TRIANGLES, 0, 6);

    ReadScreen(WINDSIZEX / 4 + 1, 3 * WINDSIZEY / 4 - 1, 1, 1, GL_RGB, buf);
    if ( !IsBlack(buf) )
    {
        StrMake(errStr, "Matrix palettes not working properly from within buffer objects");
        ErrorReport(__FILE__, __LINE__, errStr);
        err = ERROR;
    }

    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glDisable(GL_COLOR_MATERIAL);

    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_MATRIX_INDEX_ARRAY_OES);
    glDisableClientState(GL_WEIGHT_ARRAY_OES);

    glDisable(GL_MATRIX_PALETTE_OES);

    glMatrixMode(GL_MODELVIEW);

    return err;
#endif
}

void* BOmemcpy(void *dst, const void* src, size_t count)
{
    char *c_dst = (char*)dst;
    const char *c_src = (const char*)src;

    int n = (int)count / sizeof(char);
    int i;

    assert((n * sizeof(char) == (int)count) && ("BOmemcpy failed."));

    for( i = n; i; i--)
    {
        *(c_dst++) = *(c_src++);
    }
    return c_dst;
}

long BOBuildUp(void)
{
    GLint i;
    static const GLftype ltmAmbient[] = {ITOF(0), ITOF(0), ITOF(0), ITOF(1)};
    void* tempBuffer = NULL;

    glGetError();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(1));
    glColor4t(ITOF(1), ITOF(1), ITOF(1), ITOF(1));

    glEnable(GL_NORMALIZE);
    glDisable(GL_TEXTURE_2D);
    
    glDisable(GL_LIGHTING);
    for( i = 0; i < 8; i++ )
    {
        glDisable(GL_LIGHT0 + i);
    }

    glShadeModel(GL_SMOOTH);

    glLightModeltv(GL_LIGHT_MODEL_AMBIENT, ltmAmbient);

    for( i = 0; i < BUFFER_LAST_ENUM; i++ )
    {
        gBufferObjs[i] = 0;
    }

    glGenBuffers(BUFFER_LAST_ENUM, gBufferObjs);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to generate required buffer objects.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* We'll use a temporary buffer to ensure, that local memory is not relied on. 
       That is, deleting the original data should have no effect on the buffer object. */
    
    /* Separate triangles */
    tempBuffer = MALLOC(sizeof(gTriVertexArray));
    BOmemcpy(tempBuffer, gTriVertexArray, sizeof(gTriVertexArray));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TRIANGLES]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gTriVertexArray), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }
    
    /* Triangle normals */
    tempBuffer = MALLOC(sizeof(gTriNormalArray));
    BOmemcpy(tempBuffer, gTriNormalArray, sizeof(gTriNormalArray));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TRI_NORMALS]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gTriNormalArray), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* Triangle colors */
    tempBuffer = MALLOC(sizeof(gTriColorArray));
    BOmemcpy(tempBuffer, gTriColorArray, sizeof(gTriColorArray));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TRI_COLORS]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gTriColorArray), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* --- */

    /* Texture coordinate array #1 */
    tempBuffer = MALLOC(sizeof(gTriTexCoordArray));
    BOmemcpy(tempBuffer, gTriTexCoordArray, sizeof(gTriTexCoordArray));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TEXTURE0]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gTriTexCoordArray), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* Texture coordinate array #2 */
    tempBuffer = MALLOC(sizeof(gTriTexCoordArray2));
    BOmemcpy(tempBuffer, gTriTexCoordArray2, sizeof(gTriTexCoordArray2));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_TEXTURE1]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gTriTexCoordArray2), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* --- */

    /* Vertices for the DrawElements calls */
    tempBuffer = MALLOC(sizeof(gElementVertices));
    BOmemcpy(tempBuffer, gElementVertices, sizeof(gElementVertices));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_ELEMENT_VERTICES]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gElementVertices), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* Indices for the DrawElements calls */
    tempBuffer = MALLOC(sizeof(gElementIndices));
    BOmemcpy(tempBuffer, gElementIndices, sizeof(gElementIndices));
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, gBufferObjs[BUFFER_ELEMENT_INDICES]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(gElementIndices), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* Point size buffer object */
    tempBuffer = MALLOC(sizeof(gPointSizes));
    BOmemcpy(tempBuffer, gPointSizes, sizeof(gPointSizes));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_POINT_SIZES]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gPointSizes), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* Matrix index buffer object */
    tempBuffer = MALLOC(sizeof(gTriMatrixIndexArray));
    BOmemcpy(tempBuffer, gTriMatrixIndexArray, sizeof(gTriMatrixIndexArray));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_MATRIX_INDICES]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gTriMatrixIndexArray), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    /* Matrix weight buffer object */
    tempBuffer = MALLOC(sizeof(gTriMatrixWeightArray));
    BOmemcpy(tempBuffer, gTriMatrixWeightArray, sizeof(gTriMatrixWeightArray));
    glBindBuffer(GL_ARRAY_BUFFER, gBufferObjs[BUFFER_MATRIX_WEIGHTS]);
    glBufferData(GL_ARRAY_BUFFER, sizeof(gTriMatrixWeightArray), tempBuffer, GL_STATIC_DRAW);
    FREE(tempBuffer);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to build a buffer object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    /* TEXTURES */

    glGenTextures(BUFOBJ_TEXTURES, gTextures);
    glBindTexture(GL_TEXTURE_2D, gTextures[0]);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 2, 2, 0, GL_RGBA, GL_UNSIGNED_BYTE, gCheckerTextureData);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glActiveTexture(GL_TEXTURE0); /* TODO: Assumes this is always bound to 0 */
    /* Modulate, will be used for lighting */
    glTexEnvx(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to construct a texture object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    glBindTexture(GL_TEXTURE_2D, gTextures[1]);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 2, 2, 0, GL_RGBA, GL_UNSIGNED_BYTE, gChecker2TextureData);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glActiveTexture(GL_TEXTURE1); /* TODO: Assumes this is always bound to 0 */
    /*  Modulate, will be used to multiply the color values
        in a simple check for the multitexturing */
    glTexEnvx(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    if ( glGetError() != GL_NO_ERROR )
    {
        StrMake(errStr, "Unable to construct a texture object.");
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }
    
    glActiveTexture(GL_TEXTURE0);
    /*glBindTexture(GL_TEXTURE_2D, 0);*/

    return NO_ERROR;
    /*return testGenBuffers();    */
}

long BOTearDown(void)
{
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    glDeleteBuffers(sizeof(gBufferObjs) / sizeof(GLint), gBufferObjs);  

    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(BUFOBJ_TEXTURES, gTextures);

    return NO_ERROR;
}

long BufferObjectExec(void)
{
    BOBuildUp();

    if ( BOTestGenBuffers() != NO_ERROR )
    {
        return ERROR;
    }
    if (BOTestTriangles() != NO_ERROR) 
    { 
        BOTearDown(); 
        return ERROR; 
    }
    if (BOTestElements() != NO_ERROR) 
    { 
        BOTearDown(); 
        return ERROR; 
    }
    if (BOTestMultiTexturing() != NO_ERROR) 
    { 
        BOTearDown(); 
        return ERROR; 
    }

    if (BOTestBufferSubData() != NO_ERROR)
    {
        BOTearDown();
        return ERROR;
    }

    if (BOTestPointSize() != NO_ERROR)
    {
        BOTearDown();
        return ERROR;
    }
#if 0
    if (BOTestMatrixPalette() != NO_ERROR)
    {
        BOTearDown();
        return ERROR;
    }
#endif
    BOTearDown();
    return NO_ERROR;
}

#endif /* GL_OES_VERSION_1_1 */
