/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/clip.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** xform.c
** Clip Test.
**
** Description -
**    Some common pitfalls expected in fixed-point clipping are tested.
**    modelview and projection matrices are set to identity.
**
**    This test consists of two sub tests:
**
**    Test 1:
**        To test the XY clipping, a quad that is larger than the viewport
**        is generated (i.e. outer edges do not intersect the viewport). 
**        The quad is scaled larger until fixed-point limits are reached
**        (both XY and Z coordinates are varied). glReadPixels() is used
**        to confirm that all pixels on the screen are drawn. 
**
**    Test 2:
**        XY rejection. QuadTris are generated to all eight surrounding
**        'octants' of the viewport, i.e., none of the resulting triangles
**        fall on the screen. glReadPixels() is used to confirm that all
**        pixels on the screen are black.
**
**        A QuadTri is a routine that generates a triangle inside a specified
**        quad such that all three triangle vertices touch some (possibly
**        different) edges of the quad.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilm.h"


static char errStr[240];
static char errStrFmt[] = "Number of pixels was %d, should be %d.\n";

long Compare(GLint component, GLftype observed, GLftype expected)
{
    if (buffer.colorBits[component] == 0) {
	expected = (component == 3) ? ITOF(1) : ITOF(0);
    }

    if (ABS(observed-expected) <= epsilon.zero) {
        return GL_TRUE;
    } else {
        return GL_FALSE;
    }
}

static long TestBuf(GLftype const *buf, GLftype color, long *pixel_count)
{
    long i, j, k, l, equal;

    l = 0;
    *pixel_count = 0;

    for (i = 0; i < WINDSIZEY; i++) {
	for (j = 0; j < WINDSIZEX; j++) {
	    equal = 1;
	    for (k = 0; k < 4; k++) {
		if (! Compare(k, buf[l], color)) {
		    equal = 0;
		}
		l++;
	    }
	    if (equal) {
		(*pixel_count)++;
	    }
	}
    }

    if (*pixel_count == WINDSIZEX * WINDSIZEY)
	return NO_ERROR;
    else
	return ERROR;
}

static void MakeQuadTris(GLftype vertices[3][2])
{
    static GLftype const vectors[8][2] = {
	{ITOF( 0), ITOF(+1)},
	{ITOF(+1), ITOF(+1)},
	{ITOF(+1), ITOF( 0)},
	{ITOF(+1), ITOF(-1)},
	{ITOF( 0), ITOF(-1)},
	{ITOF(-1), ITOF(-1)},
	{ITOF(-1), ITOF( 0)},
	{ITOF(-1), ITOF(+1)}
    };
    long i, j, octant;

    /* random within [-1, +1]^2 */
    for (i = 0; i < 3; i++) {
        for (j = 0; j < 2; j++) {
	    vertices[i][j] = OESRandom(ITOF(-1), ITOF(1));
	}
    }

    /* move random coordinate to either -1 or +1 */
    for (i = 0; i < 3; i++) {
	vertices[i][RAND() % 2] = (RAND() % 2 == 0) ? ITOF(-1) : ITOF(+1);
    }

    /* Move to random octant */
    octant = RAND() % 8;
    for (i = 0; i < 3; i++) {
	for (j = 0; j < 2; j++) {
            vertices[i][j] += vectors[octant][j] * 2;
	}
    }
}

static void DrawTriangle(GLftype vertices[3][2])
{
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FTYPE, 0, &vertices[0][0]);
    glDrawArrays(GL_TRIANGLES, 0, 3);
    glDisableClientState(GL_VERTEX_ARRAY);
}

long ClipExec(void)
{
    GLftype *buf;
    GLftype value, vertices[3][2];
    long i, max, pixel_count;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*4*sizeof(GLftype));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glColor4t(ITOF(1), ITOF(1), ITOF(1), ITOF(1));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    /*
    ** Test XY clipping.
    */

    max = (machine.pathLevel == 0) ? 40 : 2;
    for (i = 0; i < max; i++) {
	glClear(GL_COLOR_BUFFER_BIT);

	value = ITOF(1) + ITOF(32000 * i / max);
	DrawQuad2f(-value, -value, +value, +value);

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGBA, buf);

	if (TestBuf(buf, ITOF(1), &pixel_count) == ERROR) {
	    StrMake(errStr, errStrFmt, pixel_count, WINDSIZEX * WINDSIZEY); 
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}
    }

    DrawQuad2f(XTOF(0x80000000), XTOF(0x80000000),
	XTOF(0x7FFFFFFF), XTOF(0x7FFFFFFF));

    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGBA, buf);

    if (TestBuf(buf, ITOF(1), &pixel_count) == ERROR) {
	StrMake(errStr, errStrFmt, pixel_count, WINDSIZEX * WINDSIZEY); 
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    /*
    ** Test XY rejection.
    */

    max = (machine.pathLevel == 0) ? 100 : 2;
    for (i = 0; i < max; i++) {
	glClear(GL_COLOR_BUFFER_BIT);

	MakeQuadTris(vertices);
	DrawTriangle(vertices);

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGBA, buf);

	if (TestBuf(buf, ITOF(0), &pixel_count) == ERROR) {
	    StrMake(errStr, errStrFmt, pixel_count, WINDSIZEX * WINDSIZEY); 
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}
    }

    FREE(buf);
    return NO_ERROR;
}
