/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/colramp.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** colramp.c
** Color Ramp Test.
**
** Description -
**    Tests the RGB color ramp. Three tests are performed on each
**    color component (red, green and blue); monotonicity,
**    consistency between individual color shades and total
**    number of color shades. Given the number of bits for a
**    color component, the minimum increment from one color shade
**    to another can be calculated. By stepping from one color
**    shade to the next with an increment smaller then this
**    minimum increment, individual color shades can be isolated.
**    The monotonicity of the color component can be checked. The
**    granularity of the colors can be checked for consistency,
**    and the total number of individual color shades can be
**    counted.
**
** Error Explanation -
**    Failure occurs if any of the color components is not
**    monotonic increasing, if the color granularity is not
**    constant over the entire color range or if the total number
**    of individual color shades does not match the number
**    calculated from the given number of bits per color
**    component.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _colRampRec {
    long component;
    ftype step, lastValue, lastDelta;
    GLftype color[3];
    long maxColor, colorCount;
} colRampRec;


static void Init(void *data)
{
    colRampRec *ptr = (colRampRec *)data;
    ftype step;

    ptr->lastValue = ITOF(-1);
    ptr->lastDelta = ITOF(-2);

    ptr->colorCount = 0;
    ptr->maxColor = 1 << buffer.colorBits[ptr->component];

    step = (machine.pathLevel == 0) ? ITOF(16) : ITOF(2);
    ptr->step = DIV(DIV(ITOF(1), ITOF(ptr->maxColor - 1)), step);

    ptr->color[0] = ITOF(0);
    ptr->color[1] = ITOF(0);
    ptr->color[2] = ITOF(0);
}

static void Set(void *data, ftype *ref)
{
    colRampRec *ptr = (colRampRec *)data;

    ptr->color[ptr->component] = *ref;

    glColor4t(ptr->color[0], ptr->color[1], ptr->color[2], ITOF(1));
    DrawPoint2f(ATOF("0.5"), ATOF("0.5"));

    *ref += ptr->step;
}

static long Test(void *data, ftype ref)
{
    colRampRec *ptr = (colRampRec *)data;
    char str[240];
    GLftype buf[3]; 
    ftype value, delta;
    long i;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 0; i < 3; i++) {
	if (buf[i] != ITOF(0) && i != ptr->component) {
	    ColorError_RGBZero(str, ptr->component, buf);
	    ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
    }

    value = buf[ptr->component];
    if (ptr->lastDelta == ITOF(-2)) {
	ptr->lastDelta = ITOF(-1);
	ptr->lastValue = value;
	ptr->colorCount++;
    } else if (ptr->lastDelta == ITOF(-1)) {
	if (value < ptr->lastValue) {
	    ColorError_RGBMonotonic(str, ptr->component, value, ptr->lastValue);
	    ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    if (ABS(delta-ptr->step) > epsilon.color[ptr->component]) {
		ColorError_RGBStep(str, ptr->component, delta, ptr->step);
		ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    ptr->lastDelta = delta;
	    ptr->colorCount++;
	}
    } else {
	if (value < ptr->lastValue) {
	    ColorError_RGBMonotonic(str, ptr->component, value, ptr->lastValue);
	    ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    if (ABS(delta-ptr->step) > epsilon.color[ptr->component]) {
		ColorError_RGBStep(str, ptr->component, delta, epsilon.color[ptr->component]);
		ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    if (ABS(delta-ptr->lastDelta) > epsilon.zero) {
		ColorError_RGBDelta(str, ptr->component, delta, ptr->lastDelta);
		ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    ptr->lastDelta = delta;
	    ptr->colorCount++;
	}
    }

    if (ref >= ITOF(1) && value < (ITOF(1) - epsilon.zero)) {
	ColorError_RGBClamp(str, ptr->component, value);
	ErrorReport(__FILE__, __LINE__, str);
	return ERROR;
    }

    return NO_ERROR;
}

long ColRampExec(void)
{
    colRampRec data;
    char str[240];
    long max;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    max = (machine.pathLevel == 0) ? 3 : 1;
    for (data.component = 0; data.component < max; data.component++) {
	if (RampUtil(ITOF(0), ITOF(1), Init, Set, Test, &data) == ERROR) {
	    return ERROR;
	}
	if (data.colorCount != data.maxColor) {
	    ColorError_RGBCount(str, data.component, data.colorCount,
				data.maxColor);
	    ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
    }

    return NO_ERROR;
}
