/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/copytex.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** copytex.c
** Copy Texture Test.
**
** Description -
**    The test is executed for all allowed internal texture formats
**    for the given color buffer.
**
**    Draw a pixel with a random color. Use glCopyTexImage2D to
**    copy the rendered pixel into a texture. Draw a white point
**    with texturing enabled. Read back the rendered pixel and
**    compare against the expected color.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        None.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color epsilon. Zero epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/ 

#include "conform.h"
#include "util.h"

static char colorBufferFormatName[40];
static char textureFormatName[40];
static char errStr[320];
static const char errStrFmt[] =
    "Color Buffer Format is %s. "
    "Texture Format is %s. "
    "Color drawn is (%.3g, %.3g, %.3g, %.3g). "
    "Color read is (%.3g, %.3g, %.3g, %.3g). "
    "Expected color is (%.3g, %.3g, %.3g, %.3g). "
    "Epsilon is (%.3g, %.3g, %.3g, %.3g).";


static void Fill(GLsizei r, GLsizei g, GLsizei b, GLsizei a, GLsizei bits[4])
{
    bits[0] = r;
    bits[1] = g;
    bits[2] = b;
    bits[3] = a;
}

static void GetTextureBits(GLenum internalFormat, GLsizei bits[4])
{
    switch (internalFormat) {
    default:
        Fill(0, 0, 0, 0, bits);
	break;

    case GL_ALPHA:
        Fill(0, 0, 0, 4, bits);
	break;

    case GL_LUMINANCE:
        Fill(4, 0, 0, 0, bits);
	break;

    case GL_LUMINANCE_ALPHA:
        Fill(4, 0, 0, 4, bits);
	break;

    case GL_RGB:
        Fill(4, 4, 4, 0, bits);
	break;

    case GL_RGBA:
        Fill(4, 4, 4, 4, bits);
	break;
    }
}

static void CalcCopyTextureEpsilon(const GLsizei textureBits[4], GLftype textureEpsilon[4])
{
    GLint i;

    for (i = 0; i < 4; i++) {
	if (buffer.colorBits[i] == 0) {
	    textureEpsilon[i] = epsilon.zero;
	    continue;
	}

	if (textureBits[i] == 0) {
	    switch (i) {
	    case 0:
	    case 3:
	    	textureEpsilon[i] = epsilon.color[i];
		break;

	    case 1:
	    case 2:
		textureEpsilon[i] = MAX(textureEpsilon[0], epsilon.color[i]);
		break;
	    }

	    continue;
	}

        textureEpsilon[i] = MAX(CalcEpsilon(textureBits[i]), epsilon.color[i]);
    }
}

static void CalcExpectedColor(GLenum internalFormat, const GLftype ref[4], GLftype exp[4])
{
    GLsizei textureBits[4];
    long i;

    GetTextureBits(internalFormat, textureBits);

    for (i = 0; i < 4; i++) {
	if (buffer.colorBits[i] == 0) {
	    exp[i] = (i == 3) ? ITOF(1) : ITOF(0);
	} else {
	    if (textureBits[i] == 0) {
		switch (i) {
		case 0:
		case 3:
		    exp[i] = ITOF(1);
		    break;

		case 1:
		case 2:
		    exp[i] = exp[0];
		    break;
		}
	    } else {
		exp[i] = ref[i];
	    }
	}
    }
}

long CopyTexExec(void)
{
    GLenum colorBufferFormat, targetFormats[5];
    GLsizei numTargetFormats;
    GLftype ref[4], exp[4], buf[4];
    GLsizei textureBits[4];
    GLftype eps[4];
    long i, j, max;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glClear(GL_COLOR_BUFFER_BIT);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    colorBufferFormat = GuessColorBufferFormat();
    numTargetFormats = CalcCopyTextureTargetFormats(colorBufferFormat, targetFormats);

    max = (machine.pathLevel == 0) ? 4 : 1;

    for (i = 0; i < numTargetFormats; i++) {

        GetTextureBits(targetFormats[i], textureBits);
	CalcCopyTextureEpsilon(textureBits, eps);

	for (j = 0; j < max; j++) {
	    ref[0] = OESRandom(ITOF(0), ITOF(1));
	    ref[1] = OESRandom(ITOF(0), ITOF(1));
	    ref[2] = OESRandom(ITOF(0), ITOF(1));
	    ref[3] = OESRandom(ITOF(0), ITOF(1));

	    CalcExpectedColor(targetFormats[i], ref, exp);

            glColor4t(ref[0], ref[1], ref[2], ref[3]);
            DrawPoint2f(ITOF(1)/2, ITOF(1)/2);
            glCopyTexImage2D(GL_TEXTURE_2D, 0, targetFormats[i], 0, 0, 1, 1, 0);
	    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (GLftype)GL_NEAREST);
	    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, (GLftype)GL_NEAREST);
	    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (GLftype)GL_CLAMP_TO_EDGE);
	    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (GLftype)GL_CLAMP_TO_EDGE);

       glEnable(GL_TEXTURE_2D);
       glColor4t(ITOF(1), ITOF(1), ITOF(1), ITOF(1));
       /* pass q=1 instead of zero to avoid divide by zero */
	    glMultiTexCoord4t(GL_TEXTURE0, ITOF(0), ITOF(0), ITOF(0), ITOF(1));
       DrawPoint2f(ITOF(10) + ITOF(1)/2, ITOF(10) + ITOF(1)/2);
       glDisable(GL_TEXTURE_2D);

	    ReadScreen(10, 10, 1, 1, GL_RGBA, buf);

	    if (ABS(buf[0] - exp[0]) >= eps[0] ||
		ABS(buf[1] - exp[1]) >= eps[1] ||
		ABS(buf[2] - exp[2]) >= eps[2] ||
		ABS(buf[3] - exp[3]) >= eps[3])
	    {
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		GetEnumName(colorBufferFormat, colorBufferFormatName);
		GetEnumName(targetFormats[i], textureFormatName);
		StrMake(errStr, errStrFmt, colorBufferFormatName, textureFormatName,
			ref[0], ref[1], ref[2], ref[3],
			buf[0], buf[1], buf[2], buf[3],
			exp[0], exp[1], exp[2], exp[3],
			eps[0], eps[1], eps[2], eps[3]);
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	}
    }

    return NO_ERROR;
}
