/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
*/

/*
** drawtex.c
** Draw Texture Extension Test.
**
** Description -
**    Draws a random texture (random size & contents) to the screen using glDrawTex
**    Uses ReadScreen to pull the screen data and compares it to the texture data
**    Checks that glDrawTex doesn't affect any part of the screen that it shouldn't
**
**    Tests every glDrawTex call (with GLshort, GLint and GLfloat or GLfixed)
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        None.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
**	  
**	  
*/

#include "conform.h"
#include "util.h"
#include "utilm.h"

#ifdef GL_OES_draw_texture

/* Generate a random integer between lo and hi (inclusive) */
static GLint OESRandomInt(GLint lo, GLint hi)
{
    GLint   x = FTOI(OESRandom(ITOF(lo),ITOF(hi+1)));
    if (x > hi) x = hi;
    return x;
}


/* Produces an error string relevant for the fail */
static void CreateErrString(char* errStr, int screenWidth, int screenHeight, int texWidth, int texHeight, int widthSign, int heightSign, int drawMethod)
{
    char drawStr[20], widthOrient[10], heightOrient[10];
    switch (drawMethod)
    {
    case 1:
        sprintf(drawStr, "glDrawTexsOES"); break;
    case 2:
        sprintf(drawStr, "glDrawTexiOES"); break;
    case 3:
        sprintf(drawStr, "glDrawTextOES"); break;
    case 4:
        sprintf(drawStr, "glDrawTexsvOES"); break;
    case 5:
        sprintf(drawStr, "glDrawTexivOES"); break;
    case 6:
        sprintf(drawStr, "glDrawTextvOES"); break;
    default:
        sprintf(drawStr, "unknown"); break;
    }

    sprintf(widthOrient, (widthSign > 0) ? "normal" : "reverse");
    sprintf(heightOrient, (heightSign > 0) ? "normal" : "reverse");

    sprintf(errStr, "Texture drawn incorrectly using %s call. Screen is %i x %i. Tex is %i x %i. Width orientation is %s. Height orientation is %s", drawStr, screenWidth, screenHeight, texWidth, texHeight, widthOrient, heightOrient);
}

/* Test that the glDrawTex function throws an error if a negative width/height is used */
static long TestErrorValues(void)
{
    glGetError(); /* Reset error flags */
    glDrawTexiOES(0,0,0,-OESRandomInt(1,100),OESRandomInt(1,100));
    if (glGetError() != GL_INVALID_VALUE) 
    {
        ErrorReport(__FILE__, __LINE__, "Used glDrawTex with negative width. Expecting GL_INVALID_VALUE.");
        return ERROR;
    }

    glDrawTexiOES(0,0,0,OESRandomInt(1,100),-OESRandomInt(1,100));
    if (glGetError() != GL_INVALID_VALUE) 
    {
        ErrorReport(__FILE__, __LINE__, "Used glDrawTex with negative height. Expecting GL_INVALID_VALUE.");
        return ERROR;
    }

    glDrawTexiOES(0,0,0,-OESRandomInt(1,100),-OESRandomInt(1,100));
    if (glGetError() != GL_INVALID_VALUE) 
    {
        ErrorReport(__FILE__, __LINE__, "Used glDrawTex with negative width and height. Expecting GL_INVALID_VALUE.");
        return ERROR;
    }

    glDrawTexiOES(0,0,0,OESRandomInt(1,100),OESRandomInt(1,100));
    if (glGetError() != GL_NO_ERROR) 
    {
        ErrorReport(__FILE__, __LINE__, "Used glDrawTex with positive width and height. Expecting GL_NO_ERROR.");
        return ERROR;
    }
    
    return NO_ERROR;
}



/* Test the glDrawTex function by drawing a random texture on screen
 * widthSign and heightSign indicate whether the texture should be mirrored
 * widthSign > 0 and heightSign > 0 indicates that the texture should be drawn normally
 * drawMethod indicates which glDrawTex call should be used
 * depth is the bit-depth to use and indicates the format (hence 0 is GL_ALPHA and really means a bit depth of 1)
 */
static long TestValidDraw(int widthSign, int heightSign, int drawMethod, int depth)
{
    GLint texWidth, texHeight, screenWidth, screenHeight, texX, texY;
    GLenum tex, format;
    GLftype *texData, *origScreenData, *screenData;
    int i, size, screenSize;
    GLshort sParams[5]; GLint iParams[5];
    GLftype tParams[5];
    int thisScreenX, thisScreenY, thisDepth, screenI, thisTexX, thisTexY;
    int endX, endY, params[4];
    int alphaDepth;
    char errStr[200];
    GLubyte *texCharData;
	GLftype screenEpsilon[4];
    GLftype componentVal, epsilonVal;

    /************
     * GL SETUP *
     ************/
    
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    
    /*****************
     * RANDOM INPUTS *
     *****************/

    texWidth = 1 << OESRandomInt(0,5), texHeight = 1 << OESRandomInt(0,5);
    screenWidth = OESRandomInt(texWidth+1,48), screenHeight = OESRandomInt(texHeight+1,48);
    texX = OESRandomInt(-texWidth,screenWidth), texY = OESRandomInt(-texHeight,screenHeight);
    
    format = 0;
    alphaDepth = -1;
    if (depth == 0) {
        format = GL_ALPHA;
        alphaDepth = 0;
    }
    else if (depth == 1) {
        format = GL_LUMINANCE;
        alphaDepth = -1;
    }
    else if (depth == 2) {
        format = GL_LUMINANCE_ALPHA;
        alphaDepth = 1;
    }
    else if (depth == 3) {
        format = GL_RGB;
        alphaDepth = -1;
    }
    else if (depth == 4) {
        format = GL_RGBA;
        alphaDepth = 3;
    }
    
    if (depth == 0) depth = 1; /* GL_ALPHA really has a depth of 1 */

	for(i=0; i < 4; i++)
		screenEpsilon[i] = epsilon.color[i];

  	SwizzleTextureEpsilon(format, GL_UNSIGNED_BYTE, screenEpsilon);

    size = texWidth * texHeight * depth;
    screenSize = screenWidth * screenHeight * depth;
    texData = MALLOC(size * sizeof(GLftype));
    origScreenData = MALLOC(screenSize * sizeof(GLftype));
    screenData = MALLOC(screenSize * sizeof(GLftype));
    texCharData = MALLOC(size * sizeof(GLubyte));
    
    /**********************
     * TEXTURE GENERATION *
     **********************/
    
    glGenTextures(1, &tex);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex);
    
    for (i=0; i<size; i++)
        texCharData[i] = OESRandomInt(0, 255);
    
    for (i=0; i<size; i++)
        texData[i] = DIV(ITOF(texCharData[i]) , ITOF(255));
    
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    
    glTexImage2D(GL_TEXTURE_2D, 0, format, texWidth, texHeight, 0, format, GL_UNSIGNED_BYTE, texCharData);
    
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    
    /* Setup the texture crop settings (allowing for reversed textures using widthSign and heightSign) */
    params[0] = (widthSign > 0) ? 0 : texWidth;         /* Left   */ 
    params[1] = (heightSign > 0) ? 0 : texHeight;       /* Bottom */
    params[2] = widthSign * texWidth;                   /* Width  */
    params[3] = heightSign * texHeight;                 /* Height */
    
    glTexParameteriv(GL_TEXTURE_2D, GL_TEXTURE_CROP_RECT_OES, params);
    
    /*********************
     * DISPLAY ON SCREEN *
     *********************/
    
    glViewport(0,0,screenWidth,screenHeight);
    glClearColort(DIV(ITOF(1), ITOF(2)), DIV(ITOF(1), ITOF(2)), DIV(ITOF(1), ITOF(2)), ITOF(1));
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_TEXTURE_2D);
    
    ReadScreen(0,0,screenWidth,screenHeight,format,origScreenData);
    
    switch (drawMethod)
    {
        
    case 1:
        glDrawTexsOES((GLshort)texX,(GLshort)texY,(GLshort)0,(GLshort)texWidth,(GLshort)texHeight);
        break;
    case 2:
        glDrawTexiOES((GLint)texX,(GLint)texY,(GLint)0,(GLint)texWidth,(GLint)texHeight);
        break;
    case 3:
        glDrawTextOES(ITOF(texX),ITOF(texY),ITOF(0),ITOF(texWidth),ITOF(texHeight));
        break;
    case 4:
        sParams[0] = (GLshort)texX; sParams[1] = (GLshort)texY; sParams[2] = (GLshort)0;
        sParams[3] = (GLshort)texWidth; sParams[4] = (GLshort)texHeight;
        glDrawTexsvOES(&sParams[0]);
        break;
    case 5:
        iParams[0] = (GLint)texX; iParams[1] = (GLint)texY; iParams[2] = (GLint)0;
        iParams[3] = (GLint)texWidth; iParams[4] = (GLint)texHeight;
        glDrawTexivOES(&iParams[0]);
        break;
    case 6:
        tParams[0] = ITOF(texX); tParams[1] = ITOF(texY); tParams[2] = ITOF(0);
        tParams[3] = ITOF(texWidth); tParams[4] = ITOF(texHeight);
        glDrawTextvOES(&tParams[0]);
        break;
    }

    glDisable(GL_TEXTURE_2D);
    
    /*****************************
     * COMPARE SCREEN TO TEXTURE *
     *****************************/
    
    ReadScreen(0,0,screenWidth,screenHeight,format,screenData);
    
    endX = texX + texWidth; endY = texY + texHeight;
    
    /* Iterate over the entire screen */
    for (thisScreenY=0; thisScreenY<screenHeight; thisScreenY++)
    {
        for (thisScreenX=0; thisScreenX<screenWidth; thisScreenX++)
        {
            for (thisDepth = 0; thisDepth < depth; thisDepth++)
            {
                screenI = depth * (screenWidth * thisScreenY + thisScreenX) + thisDepth;

                /* If the render target doesn't have an alpha value, the read values should be 1 */
                if (thisDepth == alphaDepth && buffer.colorBits[3] == 0)
                {
                    componentVal = ITOF(1);
                    epsilonVal = epsilon.zero;
                }
                else if (thisScreenX >= texX && thisScreenX < endX
                          && thisScreenY >= texY && thisScreenY < endY) /* Inside the textured area */
                {
                    /* Get the texture coordinates, considering possible reverse textures */
                    thisTexX = (widthSign > 0) ? thisScreenX - texX : texX + texWidth - thisScreenX - 1; 
                    thisTexY = (heightSign > 0) ? thisScreenY - texY : texY + texHeight - thisScreenY - 1; 

                    componentVal = texData[depth * (texWidth * thisTexY + thisTexX) + thisDepth];
                    epsilonVal = screenEpsilon[thisDepth];
                }
                else /* Outside the textured area (shouldn't be altered) */
                {
                    componentVal = origScreenData[screenI];
                    epsilonVal = FTOX(0.01);
                }
                    
                if (ABS(SUB(screenData[screenI],componentVal)) > epsilonVal)
                {                     
                    CreateErrString(errStr, screenWidth, screenHeight, texWidth, texHeight, widthSign, heightSign, drawMethod);
                    ErrorReport(__FILE__, __LINE__, errStr);

                    goto DRAWTEX_ERROR;
                }
            }
        }
    }

    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, &tex);
    FREE(texData); FREE(origScreenData); FREE(screenData); FREE(texCharData);
    return NO_ERROR;

DRAWTEX_ERROR:
    
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, &tex);
    FREE(texData); FREE(origScreenData); FREE(screenData); FREE(texCharData);
    return ERROR;
}

long DrawTexExec(void)
{
#ifdef NULLWS
    return NO_ERROR;
#else
    if (tkCheckExtension("GL_OES_draw_texture"))
    {
        int widthS, heightS, drawM, depth;

        /* Checks that the correct errors are given with invalid input */
        if (TestErrorValues() == ERROR)
            return ERROR;
    
        /* Runs through various texture orientations, glDrawTex methods and bit depths
         * n.b. depth goes from 0 to 4 as GL_ALPHA and GL_LUMINANCE share bit-depth=1
	     */
        for (widthS=-1; widthS<2; widthS+=2)
            for (heightS=-1; heightS<2; heightS+=2)
                for(drawM=1; drawM<7; drawM++)
                    for (depth=0; depth<5; depth++)
                        if (TestValidDraw(widthS, heightS, drawM, depth) == ERROR)
                            return ERROR;
    }
    return NO_ERROR;
#endif
}

#endif /* GL_OES_draw_texture */
