/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/fogexp.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** fogexp.c
** Fog Exponential Test.
**
** Description -
**    Tests the exponential fog computation. Both GL_FOGEXP and GL_FOGEXP2 are
**    tested in this test. This test examines the effect which
**    varying the z eye coordinate and the density, d, has on the
**    color of the point. When the density value is being tested
**    z is set to -1.0 and the third coordinate of the vertex is
**    set to a referance value, r. When z is being examined it is
**    set to -r and d is set to 1.0. In other tests involving the
**    exponential function we use huge errors. This is for
**    exponents as large as 128. In this test the exponents are
**    bounded between 0 and 1. So error bounds are tighter.
**    
**    The test examines one color channel at a
**    time. First it examines the effect of the fog weighting on
**    the current color, and then it examines the effect on fog
**    color. When the current color is being examined the fog
**    color is set to black, and one channel of the current color
**    is set to 1.0. When the fog color is being examined the
**    current color is set to black, and one channel of the fog
**    color is set to 1.0.
**    
**    The test repeatedly renders a point in the lower left
**    corner of the window, increments r, and reads back the
**    color of the point. Since implementations could
**    conceivable use the distance of the object from the eye to
**    calculate f, the point is positioned so that it's distance
**    from the eye is equal to it's z coordinate.
**    
**    r is incremented in such a way that f, should theoretically
**    increment uniformly. The color should therefore also
**    increment uniformly, if we are testing the current color,
**    and decrement uniformly, if we are testing the fog color.
**    
**    The observed color of the point is examined. An
**    implementation fails if a color channel is non-zero, when
**    it should be zero, or if it is not 1.0, when it should be
**    1.0. The observed values are also examined for
**    monotonocity. An implementation fails if the results which
**    should increment or decrement monotonically fail to do so.
**    Finally an implementation will fail if the observed color
**    is more than 1.5 shades from the expected color.
**    
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**       Color and zero epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Logicop, Stencil.
**        Not allowed = Alias, Dither, Fog, Shade.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _fogExpRec {
    long color;
    ftype step, lastValue, lastDelta;
    long useFog, useZ, isSquared;
    ftype numColor, dz;
} fogExpRec;


static char errStr[320];
static char errStrFmt[][240] = {
    "RGB observed = (%g, %g, %g, %g). Expected %s to be 0.0, Fog_density = %g, z = %g, model = %s, fog RGBA = (%g, %g, %g, %g), color RGBA = (%g, %g, %g, %g).",
    "Observed RGB = (%g, %g, %g, %g), %s differs too much from expected, %g. Fog density = %g, z = %g, fog model = %s, fog RGBA = (%g, %g, %g, %g), color RGBA = (%g, %g, %g, %g).",
    "Non-monotonic change in %s color. Observed %s = %g, last value was %g. It should %s! Fog density = %g, z = %g, fog model = %s, fog RGBA = (%g, %g, %g, %g), color RGBA = (%g, %g, %g, %g).",
    "Expected fully saturated %s. Observed RGBA = (%g, %g, %g, %g), fog density = %g, z = %g, fog model = %s, fog RGBA = (%g, %g, %g, %g), color RGBA = (%g, %g, %g, %g)."
};


static char *ErrIndexToColorName(long index)
{

    switch (index) {
      case 0:
	return "RED";
      case 1:
	return "GREEN";
      case 2:
	return "BLUE";
      case 3:
	return "ALPHA";
      default:
	return "UNKNOWN COLOR";
    }
}

static char *MakeErrorString(long errType, long wrongColor, GLftype *buf,
			     fogExpRec *ptr, ftype ref)
{
    ftype fog[4], color[4];
    ftype d, z, lastValue; 
    char const *model, *colorDirection = "";

    color[0] = ITOF(0);
    color[1] = ITOF(0);
    color[2] = ITOF(0);
    color[3] = ITOF(1);

    fog[0] = ITOF(0);
    fog[1] = ITOF(0);
    fog[2] = ITOF(0);
    fog[3] = ITOF(1);

    if (ptr->useZ == GL_TRUE) {
	d = ITOF(1);
	z = -ptr->dz;
    } else {
	d = ptr->dz;
	z = -ITOF(1);
    }

    model = (ptr->isSquared == GL_TRUE) ? "GL_EXP2" : "GL_EXP";
    if (ptr->useFog == GL_TRUE) {
	ref = ITOF(1) - ref;
	lastValue = ITOF(1) - ptr->lastValue;
	fog[ptr->color] = ITOF(1);
	colorDirection = "decrease";
    } else { 
	color[ptr->color] = ITOF(1);
	lastValue = ptr->lastValue;
	colorDirection = "increase";
    }

    switch (errType) {
      case 0:
	StrMake(errStr, errStrFmt[errType], buf[0], buf[1], buf[2], ITOF(1),
		ErrIndexToColorName(wrongColor), d, z, model, color[0],
		color[1], color[2], color[3], fog[0], fog[1], fog[2], fog[3]);
	break;
      case 1:
	StrMake(errStr, errStrFmt[errType], buf[0], buf[1], buf[2], ITOF(1),
		ErrIndexToColorName(wrongColor), ref, d, z, model, color[0],
		color[1], color[2], color[3], fog[0], fog[1], fog[2], fog[3]);
	break;
      case 2:
	StrMake(errStr, errStrFmt[2], ErrIndexToColorName(wrongColor),
		ErrIndexToColorName(wrongColor), buf[ptr->color],
		lastValue, colorDirection, d, z, model, color[0],
		color[1], color[2], color[3], fog[0], fog[1], fog[2], fog[3]);
	break;
      case 3:
        StrMake(errStr, errStrFmt[3], ErrIndexToColorName(wrongColor), buf[0],
		buf[1], buf[2], ITOF(1), d, z, model, color[0], color[1], color[2],
		color[3], fog[0], fog[1], fog[2], fog[3]);
	break;
    }
    return errStr;
}

/*****************************************************************************/

static void Set(void *data, ftype *ref)
{
    static ftype one = ITOF(1), minusOne = -ITOF(1);
    fogExpRec *ptr = (fogExpRec *)data;

    ptr->dz = -LOG(*ref);
    if (ptr->isSquared == GL_TRUE) {
	ptr->dz = SQRT(ptr->dz);
    }

    glEnableClientState(GL_VERTEX_ARRAY);

    /*
    ** z value and distance in eye space are the same.
    */
    if (ptr->useZ == GL_TRUE) {
	GLftype vertex[3];
        vertex[0] = ITOF(0);
        vertex[1] = ITOF(0);
        vertex[2] = -ptr->dz;

	glFogt(GL_FOG_DENSITY, one);
	glVertexPointer(3, GL_FTYPE, 0, &vertex[0]);
	glDrawArrays(GL_POINTS, 0, 1);
    } else {
	GLftype vertex[3];
        vertex[0] = ITOF(0);
        vertex[1] = ITOF(0);
        vertex[2] = minusOne;

	glFogt(GL_FOG_DENSITY, ptr->dz);
	glVertexPointer(3, GL_FTYPE, 0, &vertex[0]);
	glDrawArrays(GL_POINTS, 0, 1);
    }

    glDisableClientState(GL_VERTEX_ARRAY);

    *ref += ptr->step;
}

/*****************************************************************************/

static void Init(void *data)
{
    fogExpRec *ptr = (fogExpRec *)data;
    GLftype buf[4];
    ftype tmp;

    ptr->lastValue = -ITOF(1);
    ptr->lastDelta = -ITOF(2);

    if (machine.pathLevel == 0) {
	ptr->step = DIV(ITOF(1)/2, ptr->numColor);
    } else {
	ptr->step = ITOF(1)/4;
    }

    /*
    ** If there were 16 bits of color, log() may be unstable enough 
    ** near zero to cause this test to be inaccurate.
    */
    tmp = LOG(ptr->step);
    if (ptr->isSquared == GL_TRUE) {
	tmp = SQRT(-tmp);
	tmp = -MUL(tmp, tmp);
    }
    if (ABS(ptr->step-EXP(tmp)) > epsilon.zero) {
	ptr->step = DIV(ITOF(1), ITOF(8096));
    }

    /*
    ** Both the fog and the current color must have alpha = 1.0;
    */
    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);

    if (ptr->useFog == GL_TRUE) {
	glColor4t(buf[0], buf[1], buf[2], buf[3]);
	buf[ptr->color] = ITOF(1);
	glFogtv(GL_FOG_COLOR, buf);
    } else {
	glFogtv(GL_FOG_COLOR, buf);
	buf[ptr->color] = ITOF(1);
	glColor4t(buf[0], buf[1], buf[2], buf[3]);
    }

    buf[0] = (GLftype)((ptr->isSquared == GL_TRUE) ? GL_EXP2 : GL_EXP);
    glFogtv(GL_FOG_MODE, buf);
}

static long Test(void *data, ftype ref)
{
    fogExpRec *ptr = (fogExpRec *)data;
    ftype value, delta;
    GLftype buf[3];
    GLftype MaxError;
    long i;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 0; i < 3; i++) {
	if (buf[i] != ITOF(0) && i != ptr->color) {
	    MakeErrorString(0, i, buf, ptr, ref);
            ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }

    value = buf[ptr->color];
    if (ptr->useFog == GL_TRUE) {
	value = ITOF(1) - value;
    }

    /*
    ** Exponentials are far more stable for values less than 0.0,
    ** so we do not allow as large an error as we do for lighting.
    */
    /*
    ** implementations with extreme visual depths may still 
    ** use a look up table as long as it has 128 entries or more.
    */
    MaxError = MAX(epsilon.color[ptr->color]*3/2, ITOF(1)/64);

    if (ABS(value-ref) > MaxError) {
	MakeErrorString(1, ptr->color, buf, ptr, ref);
        ErrorReport(__FILE__, __LINE__, errStr);
	return ERROR;
    }

    /*
    ** The linearity test is disabled, but due to variable sample spacing,
    ** delta should be approximately constant.
    */
    if (ptr->lastDelta == -ITOF(2)) {
	ptr->lastDelta = -ITOF(1);
	ptr->lastValue = value;
    } else if (ptr->lastDelta == -ITOF(1)) {
	if (value < ptr->lastValue) {
	    MakeErrorString(2, ptr->color, buf, ptr, ref); 
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	/*
	** This is only left here for people who want to
	** use it to debug.
	*/
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    } else {
	if (value < ptr->lastValue) {
	    MakeErrorString(2, ptr->color, buf, ptr, ref);
            ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	/*
	** This is only left here for people who want to
	** use it to debug.
	*/
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    }
    if (ref >= ITOF(1) && ABS(value-ITOF(1)) > epsilon.zero) {
	MakeErrorString(3, ptr->color, buf, ptr, ref);
        ErrorReport(__FILE__, __LINE__, errStr);
	return ERROR;
    }
    return NO_ERROR;
}

long FogExpExec(void)
{
    fogExpRec data;
    ftype start;
    GLftype m[4][4];

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    /*
    ** Since the eye coordinates are used to calculate the distance or zfog
    ** value, we have to do a little transformation hocus-pocus to allow for
    ** a large homogeneous clip coordinate and to allow z and distance to be
    ** equal in eye coordinates. 10.0 is chosen as it is bigger than ln(8196)
    ** which is likely to be the largest number of shades of any one color.
    */

    glMatrixMode(GL_PROJECTION);
    glTranslatet(ITOF(1)/2, ITOF(1)/2, ITOF(0));
    MakeIdentMatrix(&m[0][0]);
    m[0][0] = ITOF(10);
    m[1][1] = ITOF(10);
    m[2][2] =  ITOF(1);
    m[3][3] = ITOF(10);
    glMultMatrixt(&m[0][0]);
    glMatrixMode(GL_MODELVIEW);

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_FOG);

    for (data.useZ = GL_FALSE; data.useZ <= GL_TRUE; data.useZ++) {
	for (data.isSquared = GL_FALSE; data.isSquared <= GL_TRUE;
	     data.isSquared++) {
	    for (data.useFog = GL_FALSE; data.useFog <= GL_TRUE;
		 data.useFog++) {
		for (data.color = 0; data.color < 3; data.color++) {
		    data.numColor = LDEXP(ITOF(1), buffer.colorBits[data.color]) - ITOF(1);
		    start = DIV(ITOF(1)/2, data.numColor);
		    if (RampUtil(start, ITOF(1), Init, Set, Test,
			         &data) == ERROR) {
			return ERROR;
		    }
		}
	    }
	}
    }
    return NO_ERROR;
}
