/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/foglin.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** foglin.c
** Fog Linear Test.
**
** Description -
**    Tests the linear fog computation.
**    This test examines the effect which varying
**    the z eye coordinate has on the color of the point. The
**    test repeats itself with different fog start and fog end
**    values.
**    
**    The test examines one color channel at a time.
**    First it examines the effect of the fog weighting on the
**    current color, and then it examines the effect on fog
**    color. When the current color is being examined the fog
**    color is set to black, and one channel of the current color
**    is set to 1.0. When the fog color is being examined the
**    current color is set to black, and one channel of the fog
**    color is set to 1.0.
**    
**    The test repeatedly renders a point in the lower left
**    corner of the window and reads back the color of that
**    point. At each iteration the reference value is moved away
**    from the fogEnd value and toward the fogStart value. The z
**    value of the rendered point is the negative of this value.
**    Since implementations could conceivable use the distance of
**    the object from the eye to calculate the fog, the point is
**    positioned so that it's distance from the eye is equal to
**    it's z coordinate.
**    
**    By uniformly incrementing the point from the fogEnd value
**    to the fogStart value the fogWeighting, f, is uniformly
**    incremented. This should uniformly increment the color if
**    we are testing the current color and decrement the color if
**    we are testing the fog color.
**    
**    The observed color of the point is examined.
**    An implementation fails if a color channel is non-zero,
**    when it should be zero, or if it is not 1.0 when it should
**    be 1.0. The observed values are also examined for
**    monotonocity. An implementation fails if the results which
**    should increment/decrement monotonically fail to do so.
**    Finally an implementation will fail if the observed color
**    is more than 1.5 shades from the expected color.
**    
** Error Explanation -
**    other_color > 0.0.
**    value < lastValue when it should be greater.
**    maxObserved < 1.0.
**    (observed - expected) > 1.5*epsilon.color.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Logicop, Stencil.
**        Not allowed = Alias, Dither, Fog, Shade.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _fogLinRec {
    long color;
    ftype step, lastValue, lastDelta;
    long useFog;
    ftype numColor, e, s, z;
} fogLinRec;


static char errStr[320];
static char errStrFmt[][240] = {
    "Observed RGBA = (%g, %g, %g, %g). Expected %s to be 0.0. Fog start = %g, fog end = %g, z = %g fog RGBA = (%g, %g, %g, %g), color RGBA = (%g, %g, %g, %g).",
    "Observed RGB = (%g, %g, %g) differs too much from expected RGB = (%g, %g, %g). Fog start = %g, fog end = %g, z = %g, fog RGBA = (%g, %g, %g, %g), color RGBA = (%g, %g, %g, %g).",
    "Non-monotonic color. Fog start = %g, fog end = %g, z = %g, fog RGBA = (%g, %g, %g, %g), color RGBA = (%g, %g, %g, %g).",
    "Expected fully saturated color. Observed RGB = (%g, %g, %g), fog start = %g, fog end = %g, z = %g, fog RGBA = (%g, %g, %g, %g), color RGBA = (%g, %g, %g, %g)."
};


static char *ErrIndexToColorName(long i)
{

    switch (i) {
      case 0:
	return "RED";
      case 1:
	return "GREEN";
      case 2:
	return "BLUE";
      case 3:
	return "ALPHA";
    }
    return "NO_SUCH_COLOR";
}

/******************************************************************************/

static void Set(void *data, ftype *ref)
{
    GLftype vertex[3];
    fogLinRec *ptr = (fogLinRec *)data;

    ptr->z = MUL(ITOF(1) - *ref, ptr->e) + MUL(*ref, ptr->s);

    /*
    ** z value and distance in eye space are the same magnitude.
    */
    vertex[0] = ITOF(0);
    vertex[1] = ITOF(0);
    vertex[2] = -ptr->z;

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FTYPE, 0, &vertex[0]);
    glDrawArrays(GL_POINTS, 0, 1);
    glDisableClientState(GL_VERTEX_ARRAY);

    *ref += ptr->step;
}

/******************************************************************************/

static void InitToBlack(ftype color[4])
{

    color[0] = ITOF(0);
    color[1] = ITOF(0);
    color[2] = ITOF(0);
    color[3] = ITOF(1);
}

static void Init(void *data)
{
    fogLinRec *ptr = (fogLinRec *)data;
    GLftype buf[4];

    ptr->lastValue = -ITOF(1);
    ptr->lastDelta = -ITOF(2);

    if (machine.pathLevel == 0) {
	ptr->step = DIV(ITOF(1)/2, ptr->numColor);
    } else {
	ptr->step = ITOF(1)/4;
    }

    /*
    ** Both the fog and the current color must have alpha = 1.0;
    */
    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);

    if (ptr->useFog == GL_TRUE) {
	glColor4t(buf[0], buf[1], buf[2], buf[3]);
	buf[ptr->color] = ITOF(1);
	glFogtv(GL_FOG_COLOR, buf);
    } else {
	glFogtv(GL_FOG_COLOR, buf);
	buf[ptr->color] = ITOF(1);
	glColor4t(buf[0], buf[1], buf[2], buf[3]);
    }

    glFogt(GL_FOG_MODE, GL_LINEAR);
    glFogt(GL_FOG_END, ptr->e);
    glFogt(GL_FOG_START, ptr->s);
}

static long Test(void *data, ftype ref)
{
    fogLinRec *ptr = (fogLinRec *)data;
    GLftype buf[3];
    GLftype MaxError;
    ftype value, delta;
    long i;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 0; i < 3; i++) {
	if (buf[i] != ITOF(0) && i != ptr->color) {
	    /*
	    ** Only Error Mesg. stuff.
	    */
	    ftype fog[4], color[4];
	    InitToBlack(color);
	    InitToBlack(fog);
	    (ptr->useFog == GL_TRUE) ? (fog[ptr->color] = ITOF(1)) :
				       (color[ptr->color] = ITOF(1));
	    StrMake(errStr, errStrFmt[0], buf[0], buf[1], buf[2], ITOF(1),
		    ErrIndexToColorName(i), ptr->s, ptr->e, -ptr->z, color[0],
		    color[1], color[2], color[3], fog[0], fog[1], fog[2],
		    fog[3]);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }

    value = buf[ptr->color];
    if (ptr->useFog == GL_TRUE) {
	value = ITOF(1) - value;
    }

    /*
    ** implementations with extreme visual depths may still 
    ** use a look up table as long as it has 128 entries or more.
    */
    MaxError = MAX(epsilon.color[ptr->color]*3/2, LDEXP(ITOF(1), -6));

    if (ABS(value-ref) > MaxError) {
	/*
	** Only Error Mesg Stuff.
	*/
	ftype fog[4], color[4], expected[4];
	InitToBlack(color);
	InitToBlack(fog);
	InitToBlack(expected);
	(ptr->useFog == GL_TRUE) ? (fog[ptr->color] = ITOF(1)) :
				   (color[ptr->color] = ITOF(1));
	expected[ptr->color] = (ptr->useFog == GL_TRUE) ? (ITOF(1) - ref) : ref;
	StrMake(errStr, errStrFmt[1], buf[0], buf[1], buf[2], expected[0],
		expected[1], expected[2], ptr->s, ptr->e, -ptr->z, color[0],
		color[1], color[2], color[3], fog[0], fog[1], fog[2], fog[3]);
	ErrorReport(__FILE__, __LINE__, errStr);
	return ERROR;
    }

    /*
    ** The linearity test is disabled, but due to variable sample spacing
    ** delta should be approximately constant.
    */
    if (ptr->lastDelta == -ITOF(2)) {
	ptr->lastDelta = -ITOF(1);
	ptr->lastValue = value;
    } else if (ptr->lastDelta == -ITOF(1)) {
	if (value < ptr->lastValue) {
	    /*
	    ** Only Error Mesg Stuff.
	    */
	    ftype fog[4], color[4];
	    InitToBlack(color);
	    InitToBlack(fog);
	    (ptr->useFog == GL_TRUE) ? (fog[ptr->color] = ITOF(1) ) : 
				       (color[ptr->color] = ITOF(1));
	    StrMake(errStr, errStrFmt[2], ptr->s, ptr->e, -ptr->z, color[0],
		    color[1], color[2], color[3], fog[0], fog[1], fog[2],
		    fog[3]);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    } else {
	if (value < ptr->lastValue) {
	    /*
	    ** Only Error Mesg Stuff.
	    */
	    ftype fog[4], color[4];
	    InitToBlack(color);
	    InitToBlack(fog);
	    (ptr->useFog == GL_TRUE) ? (fog[ptr->color] = ITOF(1)) : 
				       (color[ptr->color] = ITOF(1));
	    StrMake(errStr, errStrFmt[2], ptr->s, ptr->e, -ptr->z, color[0],
		    color[1], color[2], color[3], fog[0], fog[1], fog[2],
		    fog[3]);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	/*
	** Enable this if you want to check uniformity of change.
	*/
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    }
    if (ref >= ITOF(1) && ABS(value-ITOF(1)) > epsilon.zero) {
	/*
	** Only Error Mesg Stuff.
	*/
	ftype fog[4], color[4];
	InitToBlack(color);
	InitToBlack(fog);
	(ptr->useFog == GL_TRUE) ? (fog[ptr->color] = ITOF(1)) : 
				   (color[ptr->color] = ITOF(1));
	StrMake(errStr, errStrFmt[3], buf[0], buf[1], buf[2], ptr->s, ptr->e,
		-ptr->z, color[0], color[1], color[2], color[3], fog[0],
		fog[1], fog[2], fog[3]);
	ErrorReport(__FILE__, __LINE__, errStr);
	return ERROR;
    }

    return NO_ERROR;
}

long FogLinExec(void)
{
    static const ftype endPoints[11][2] = {
	{
	    ITOF(0), ITOF(1)
	},
	{
	    ITOF(0), ITOF(10)
	},
	{
	    ITOF(0), ITOF(100)
	},
	{
	    ITOF(0), ITOF(1000)
	},
	{
	    ITOF(1), ITOF(2)
	},
	{
	    ITOF(1), ITOF(10)
	},
	{
	    ITOF(1), ITOF(1000)
	},
	{
	    ITOF(3), ITOF(7)/2
	},
	{
	    ITOF(7)/2, ITOF(1000)
	},
	{
	    ITOF(1), ITOF(0)
	},
	{
	    ITOF(10), ITOF(1)
	}
    };
    fogLinRec data;
    GLftype m[4][4];
    ftype start, unscale, tmp;
    long i;

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_FOG);

    for (i = 0; i < 11; i++) {
	data.e = endPoints[i][0];
	data.s = endPoints[i][1];

	/*
	** Since the eye-coordinates are used for calculating the distance
	** or zfog value, we have to do a little transformation hocus-pocus
	** to allow for a large homogeneous clip coordinate and to allow z
	** and distance to be equal in eye coordinates.
	*/
	unscale = ABS(data.e);
	if ((tmp = ABS(data.s)) > unscale) {
	    unscale = tmp;
	}

	Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));
	glMatrixMode(GL_PROJECTION);
	glTranslatet(ITOF(1)/2, ITOF(1)/2, ITOF(0));
	MakeIdentMatrix(&m[0][0]);
	m[0][0] = unscale;
	m[1][1] = unscale;
	m[3][3] = unscale;
	glMultMatrixt(&m[0][0]);
	glMatrixMode(GL_MODELVIEW);

	for (data.useFog = GL_FALSE; data.useFog <= GL_TRUE; data.useFog++) {
	    for (data.color = 0; data.color < 3; data.color++) {
		data.numColor = LDEXP(ITOF(1), buffer.colorBits[data.color]) - ITOF(1);
		start = DIV(ITOF(1)/2, data.numColor);
		if (RampUtil(start, ITOF(1), Init, Set, Test,
			     &data) == ERROR) {
		    return ERROR;
		}
	    }
	}
    }
    return NO_ERROR;
}
