/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
**
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/gets.c,v 1.12 2005/05/17 13:12:33 artlab Exp $
*/

#include "conform.h"
#include "util.h"
#include "utilm.h"

#ifdef GL_OES_VERSION_1_1

static char  errFmt[1024];
static char  errName[1024];
static char  errStr[1024];

/******************************************************************************/

static GLint OESRandomInt(GLint lo, GLint hi)
{
    GLint   x = FTOI(OESRandom(ITOF(lo),ITOF(hi+1)));
    if (x > hi) x = hi;
    return x;
}

/******************************************************************************/

/* Used to convert floating/fixed point values in the range [-1,+1] to 
**   integers in the range [-INT_MAX,+INT_MAX] for certain color and depth
**   settings.
*/
static GLint NormalizedFtypeToInt(GLftype inval)
{
#if GL_FTYPE == GL_FIXED
    return (GLint)MUL(0x7FFFFFFF, inval);
#else
    return FTOI((MUL(POW(2.0f,32.0f)-1.0f, inval)-1.0f)/2.0f);
#endif
}

/******************************************************************************/

static const char *enablestr[2]  = { "disabled", "enabled" };
static const char *booleanstr[2] = { "FALSE", "TRUE" };

typedef enum {
    Lookup_Enable  = 0,
    Lookup_Boolean,
    Lookup_Integer,
    Lookup_Ftype,
    Lookup_Pointer,
    LookupTypeCount
} LookupType_t;

static const char *typefmt[LookupTypeCount] = 
{ 
    "%s", 
    "%s",
    "0x%x",
#if GL_FTYPE == GL_FIXED
    "%f",
#else
    "%f",
#endif
    "%p",
};

/******************************************************************************/

typedef enum {
    Lookup_IsEnabled = 0,
    Lookup_GetBoolean,
    Lookup_GetInteger,
    Lookup_GetTexParameteri,
    Lookup_GetBufferParameteri,
    Lookup_GetTexEnvi,
    Lookup_GetFtype,
    Lookup_GetTexEnvt,
    Lookup_GetMaterialt,
    Lookup_GetLightt,
    Lookup_GetClipPlanet,
    Lookup_GetPointer,
    Lookup_IsBuffer,
    Lookup_IsTexture,
    LookupMethodCount
} LookupMethod_t;

typedef struct {
    LookupType_t    type;
    const char      *desc;
} LookupInfo_t;

static LookupInfo_t methodtable[LookupMethodCount] =
{
    { Lookup_Enable,  "enable setting" },
    { Lookup_Boolean, "boolean value" },
    { Lookup_Integer, "integer value" },
    { Lookup_Integer, "integer texture parameter" },
    { Lookup_Integer, "integer texture environment value" },
    { Lookup_Integer, "integer buffer parameter" },
#if GL_FTYPE == GL_FIXED
    { Lookup_Ftype,   "fixed point value" },
    { Lookup_Ftype,   "fixed point texture environment value" },
    { Lookup_Ftype,   "fixed point material value" },
    { Lookup_Ftype,   "fixed point light value" },
    { Lookup_Ftype,   "fixed point clip plane" },
#else
    { Lookup_Ftype,   "floating point value" },
    { Lookup_Ftype,   "floating point texture environment value" },
    { Lookup_Ftype,   "floating point material value" },
    { Lookup_Ftype,   "floating point light value" },
    { Lookup_Ftype,   "floating point clip plane" },
#endif
    { Lookup_Pointer, "pointer" },
    { Lookup_Boolean, "boolean value" },
    { Lookup_Boolean, "boolean value" }
};

/******************************************************************************/

typedef enum {
    Tolerance_Exact,
    Tolerance_Ftype,
    Tolerance_Color,
    Tolerance_Alpha,
    Tolerance_Depth
} Tolerance_t;

static GLftype  colortolerancet[4];
static GLint    colortolerancei[4];
static GLftype  alphatolerancet;
static GLftype  alphatolerancei;
static GLftype  depthtolerancet;
static GLint    depthtolerancei;

/******************************************************************************/

static long LookupValues(
    LookupMethod_t  method, 
    GLenum          target, 
    GLenum          pname, 
    void            *params, 
    char            *pnamefmt, 
    GLint           pnameind)
{
    GLenum  errCode;

    /* Perform the lookup */
    switch (method)
    {
        case Lookup_IsBuffer:
            *((GLboolean*)params) = (GLboolean)glIsBuffer((GLint)pname);
            break;
        case Lookup_IsTexture:
            *((GLboolean*)params) = (GLboolean)glIsTexture((GLint)pname);
            break;
        case Lookup_IsEnabled:
            *((GLboolean*)params) = (GLboolean)glIsEnabled(pname);
            break;
        case Lookup_GetBoolean:
            glGetBooleanv(pname, (GLboolean*)params);
            break;
        case Lookup_GetInteger:
            glGetIntegerv(pname, (GLint*)params);
            break;
        case Lookup_GetTexParameteri:
            glGetTexParameteriv(target, pname, (GLint*)params);
            break;
        case Lookup_GetTexEnvi:
            glGetTexEnviv(target, pname, (GLint*)params);
            break;
        case Lookup_GetBufferParameteri:
            glGetBufferParameteriv(target, pname, (GLint*)params);
            break;
        case Lookup_GetFtype:
            glGetFloatvt(pname, (GLftype*)params);
            break;
        case Lookup_GetTexEnvt:
            glGetTexEnvtv(target, pname, (GLftype*)params);
            break;
        case Lookup_GetMaterialt:
            glGetMaterialtv(target, pname, (GLftype*)params);
            break;
        case Lookup_GetLightt:
            glGetLighttv(target, pname, (GLftype*)params);
            break;
        case Lookup_GetClipPlanet:
            glGetClipPlanet(pname, (GLftype*)params);
            break;
        case Lookup_GetPointer:
            glGetPointerv(pname, (void**)params);
            break;
        default:
            StrMake(errStr, "Internal error: Unknown lookup method (%d)", method);
            ErrorReport(__FILE__, __LINE__, errStr);
            return ERROR;
            break;
    }

    /* Check for errors */
    errCode = glGetError();
    if (errCode != GL_NO_ERROR)
    {   StrMake(errName, pnamefmt, pnameind);
        StrMake(errStr, 
            "GL error (0x%08x) encountered while retrieving %s for %s", 
            errCode, methodtable[method].desc, errName);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    return NO_ERROR;
}

/******************************************************************************/

static void BuildErrFmt(
    LookupMethod_t  method, 
    GLint           numparams, 
    GLint           paramind, 
    Tolerance_t     tolerance,
    const char      *pnamefmt, 
    GLint           pnameind)
{
    char    pnameStr[256];
    char    indStr[256];
    char    tolStr[256];

    StrMake(pnameStr, pnamefmt, pnameind);

    if (numparams > 1)
        StrMake(indStr, "element %d of ", paramind);
    else
        indStr[0] = '\0';

    if (tolerance)
        StrMake(tolStr, " with a tolerance of %s", typefmt[methodtable[method].type]);
    else
        tolStr[0] = '\0';

    StrMake(errFmt,
        "Mismatch in %s %s%s.\nReturned value was %s.\nExpected value was %s%s.",
        methodtable[method].desc, indStr, pnameStr,
        typefmt[methodtable[method].type],
        typefmt[methodtable[method].type], tolStr);
}

/******************************************************************************/

static long VerifyValues(
    LookupMethod_t  method, 
    GLenum          target, 
    GLenum          pname, 
    GLint           numparams,
    void            *expect,
    Tolerance_t     tolerance,
    char            *pnamefmt, 
    GLint           pnameind)
{
    GLint   params[16];
    long    errCode;
    
    errCode = LookupValues(method, target, pname, params, pnamefmt, pnameind);
    if (errCode != NO_ERROR)
        return errCode;

    switch (methodtable[method].type)
    {
        case Lookup_Enable:
        {
            GLboolean   *expectb = (GLboolean*)expect;
            GLboolean   *paramsb = (GLboolean*)params;
            if (*expectb != *paramsb)
            {
                BuildErrFmt(method, 0, 0, Tolerance_Exact, pnamefmt, pnameind);
                StrMake(errStr, errFmt, enablestr[*paramsb], enablestr[*expectb]);
                ErrorReport(__FILE__, __LINE__, errStr);
                return ERROR;
            }
            break;
        }

        case Lookup_Boolean:
        {
            GLboolean   *expectb = (GLboolean*)expect;
            GLboolean   *paramsb = (GLboolean*)params;
            GLint       i;
            for (i=0; i<numparams; ++i)
            {
                if (expectb[i] != paramsb[i])
                {
                    BuildErrFmt(method, numparams, i, Tolerance_Exact, pnamefmt, pnameind);
                    StrMake(errStr, errFmt, booleanstr[paramsb[i]], booleanstr[expectb[i]]);
                    ErrorReport(__FILE__, __LINE__, errStr);
                    return ERROR;
                }
            }
            break;
        }

        case Lookup_Integer:
        {
            GLint   *expecti = (GLint*)expect;
            GLint   *paramsi = (GLint*)params;
            GLint   tolval;
            GLint   i;

            for (i=0; i<numparams; ++i)
            {
                GLint   delta = expecti[i] - paramsi[i];
                if (delta < 0) delta = -delta;

                switch (tolerance)
                {
                    case Tolerance_Color:
                        tolval = colortolerancei[i];
                        break;
                    case Tolerance_Alpha:
                        tolval = alphatolerancei;
                        break;
                    case Tolerance_Depth:
                        tolval = depthtolerancei;
                        break;
                    default:
                        tolval = 0;
                        break;
                }

                if (delta > tolval)
                {
                    BuildErrFmt(method, numparams, i, tolerance, pnamefmt, pnameind);
                    StrMake(errStr, errFmt, paramsi[i], expecti[i], tolval);
                    ErrorReport(__FILE__, __LINE__, errStr);
                    return ERROR;
                }
            }
            break;
        }

        case Lookup_Ftype:
        {
            GLftype *expectt = (GLftype*)expect;
            GLftype *paramst = (GLftype*)params;
            GLftype tolval;
            GLint   i;

            for (i=0; i<numparams; ++i)
            {
                GLftype delta = expectt[i] - paramst[i];
                if (delta < 0) delta = -delta;

                switch (tolerance)
                {
                    case Tolerance_Exact:
                        tolval = ITOF(0);
                        break;
                    case Tolerance_Ftype:
#if GL_FTYPE == GL_FIXED
                        tolval = 0x0004;
#else
                        if (expectt[i] >= 0.0f)
                            tolval = +expectt[i] / 100000.0f;
                        else
                            tolval = -expectt[i] / 100000.0f;
#endif
                        break;
                    case Tolerance_Color:
                        tolval = colortolerancet[i];
                        break;
                    case Tolerance_Alpha:
                        tolval = alphatolerancet;
                        break;
                    case Tolerance_Depth:
                        tolval = depthtolerancet;
                        break;
                    default:
                        ErrorReport(__FILE__, __LINE__, 
                            "VerifyValues(): Invalid value in switch");
                        return ERROR;
                }

                if (delta > tolval)
                {
                    BuildErrFmt(method, numparams, i, tolerance, pnamefmt, pnameind);
                    StrMake(errStr, errFmt, paramst[i], expectt[i], tolval);
                    ErrorReport(__FILE__, __LINE__, errStr);
                    return ERROR;
                }
            }
            break;
        }

        case Lookup_Pointer:
        {
            void    **expectp = (void**)expect;
            void    **paramsp = (void**)params;
            GLint       i;
            for (i=0; i<numparams; ++i)
            {
                if (expectp[i] != paramsp[i])
                {
                    BuildErrFmt(method, numparams, i, Tolerance_Exact, pnamefmt, pnameind);
                    StrMake(errStr, errFmt, paramsp[i], expectp[i]);
                    ErrorReport(__FILE__, __LINE__, errStr);
                    return ERROR;
                }
            }
            break;
        }
        case LookupTypeCount:
            ErrorReport(__FILE__, __LINE__, 
                "VerifyValues(): LookupTypeCount set in switch");
            return ERROR;
    }

    return NO_ERROR;
}

/******************************************************************************/

static long CheckErrors(const char *funcname)
{
    GLenum  errCode;

    /* Check for errors */
    errCode = glGetError();

    if (errCode != GL_NO_ERROR)
    {
        StrMake(errStr,
            "GL error (0x%08x) encountered after setting values for test function %s.", 
            errCode, funcname);
        printf("GL error (0x%08x) encountered after setting values for test function %s.\n", 
            errCode, funcname);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    return NO_ERROR;
}

/******************************************************************************/

static long TestImplementationState(void)
{
    GLint   maxclipplanes;
    GLint   maxmodelviewstack;
    GLint   maxprojectionstack;
    GLint   maxtexturestack;
    GLint   subpixelbits;
    GLint   maxtexturesize;
    GLint   maxviewportdims[2];

    GLftype aliasedpointsizerange[2];
    GLftype smoothpointsizerange[2];
    GLftype aliasedlinewidthrange[2];
    GLftype smoothlinewidthrange[2];

    GLint   maxtextureunits;
    GLint   samplebuffers;
    GLint   samples;
    GLint   *compressedtextureformats = NULL;
    GLint   numcompressedtextureformats;

    GLint   redbits;
    GLint   greenbits;
    GLint   bluebits;
    GLint   alphabits;
    GLint   depthbits;
    GLint   stencilbits;

    /* Retrieve the system settings */
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_CLIP_PLANES, &maxclipplanes, "GL_MAX_CLIP_PLANES", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_MODELVIEW_STACK_DEPTH, &maxmodelviewstack, "GL_MAX_MODELVIEW_STACK_DEPTH", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_PROJECTION_STACK_DEPTH, &maxprojectionstack, "GL_MAX_PROJECTION_STACK_DEPTH", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_STACK_DEPTH, &maxtexturestack, "GL_MAX_TEXTURE_STACK_DEPTH", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_SUBPIXEL_BITS, &subpixelbits, "GL_SUBPIXEL_BITS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_SIZE, &maxtexturesize, "GL_MAX_TEXTURE_SIZE", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_VIEWPORT_DIMS, maxviewportdims, "GL_MAX_VIEWPORT_DIMS", 0) == ERROR)
        goto fail;
    
    if (LookupValues(Lookup_GetFtype, 0, GL_ALIASED_POINT_SIZE_RANGE, aliasedpointsizerange, "GL_ALIASED_POINT_SIZE_RANGE", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetFtype, 0, GL_SMOOTH_POINT_SIZE_RANGE, smoothpointsizerange, "GL_SMOOTH_POINT_SIZE_RANGE", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetFtype, 0, GL_ALIASED_LINE_WIDTH_RANGE, aliasedlinewidthrange, "GL_ALIASED_LINE_WIDTH_RANGE", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetFtype, 0, GL_SMOOTH_LINE_WIDTH_RANGE, smoothlinewidthrange, "GL_SMOOTH_LINE_WIDTH_RANGE", 0) == ERROR)
        goto fail;

    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_UNITS, &maxtextureunits, "GL_MAX_TEXTURE_UNITS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_SAMPLE_BUFFERS, &samplebuffers, "GL_SAMPLE_BUFFERS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_SAMPLES, &samples, "GL_SAMPLES", 0) == ERROR)
        goto fail;

    if (LookupValues(Lookup_GetInteger, 0, GL_NUM_COMPRESSED_TEXTURE_FORMATS, &numcompressedtextureformats, "GL_NUM_COMPRESSED_TEXTURE_FORMATS", 0) == ERROR)
        goto fail;

    if (numcompressedtextureformats > 0)
    {
        compressedtextureformats = (GLint*)MALLOC(numcompressedtextureformats * sizeof(GLint));
        if (LookupValues(Lookup_GetInteger, 0, GL_COMPRESSED_TEXTURE_FORMATS, compressedtextureformats, "GL_COMPRESSED_TEXTURE_FORMATS", 0) == ERROR)
            goto fail;
    }

    if (LookupValues(Lookup_GetInteger, 0, GL_RED_BITS, &redbits, "GL_RED_BITS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_GREEN_BITS, &greenbits, "GL_GREEN_BITS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_BLUE_BITS, &bluebits, "GL_BLUE_BITS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_ALPHA_BITS, &alphabits, "GL_ALPHA_BITS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_DEPTH_BITS, &depthbits, "GL_DEPTH_BITS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_STENCIL_BITS, &stencilbits, "GL_STENCIL_BITS", 0) == ERROR)
        goto fail;

    /* Make sure values meet requirements, as relevant */
    if (maxclipplanes < 1)
    {
        StrMake(errStr, "GL_MAX_CLIP_PLANES must be at least 1.  Returned value was %d", maxclipplanes);
        ErrorReport(__FILE__, __LINE__, errStr);
        goto fail;
    }
    if (maxmodelviewstack < 16)
    {
        StrMake(errStr, "GL_MAX_MODELVIEW_STACK_DEPTH must be at least 16.  Returned value was %d", maxmodelviewstack);
        ErrorReport(__FILE__, __LINE__, errStr);
        goto fail;
    }
    if (maxprojectionstack < 2)
    {
        StrMake(errStr, "GL_MAX_PROJECTION_STACK_DEPTH must be at least 2.  Returned value was %d", maxprojectionstack);
        ErrorReport(__FILE__, __LINE__, errStr);
        goto fail;
    }
    if (maxtexturestack < 2)
    {
        StrMake(errStr, "GL_MAX_TEXTURE_STACK_DEPTH must be at least 2.  Returned value was %d", maxtexturestack);
        ErrorReport(__FILE__, __LINE__, errStr);
        goto fail;
    }
    if (maxtextureunits < 2)
    {
        StrMake(errStr, "GL_MAX_TEXTURE_UNITS must be at least 2.  Returned value was %d", maxtextureunits);
        ErrorReport(__FILE__, __LINE__, errStr);
        goto fail;
    }
    

    if (compressedtextureformats)
        FREE(compressedtextureformats);

    return NO_ERROR;

    fail:
    if (compressedtextureformats) 
        FREE(compressedtextureformats);

    return ERROR;
}
    

/******************************************************************************/

static long TestCurrentValueState(void)
{
    GLint   numtexunits;
    GLftype colort[4];
    GLint   colori[4];
    GLftype texcoord[32][4];
    GLftype normal[3];
    GLint   i;

    /* Get system parameters */
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_UNITS, &numtexunits, "GL_MAX_TEXTURE_UNITS", 0) == ERROR)
        return ERROR;

    /*
    ** Set random values
    */

    /* Set current color */
    colort[0] = OESRandom(ITOF(0), ITOF(1));
    colort[1] = OESRandom(ITOF(0), ITOF(1));
    colort[2] = OESRandom(ITOF(0), ITOF(1));
    colort[3] = OESRandom(ITOF(0), ITOF(1));
    colori[0] = NormalizedFtypeToInt(colort[0]);
    colori[1] = NormalizedFtypeToInt(colort[1]);
    colori[2] = NormalizedFtypeToInt(colort[2]);
    colori[3] = NormalizedFtypeToInt(colort[3]);
    glColor4t(colort[0], colort[1], colort[2], colort[3]);

    /* Set current texture coordinates */
    for (i=0; i<numtexunits; ++i)
    {
        texcoord[i][0] = OESRandom(ITOF(-16), ITOF(+16));
        texcoord[i][1] = OESRandom(ITOF(-16), ITOF(+16));
        texcoord[i][2] = OESRandom(ITOF(-16), ITOF(+16));
        texcoord[i][3] = OESRandom(DIV(ITOF(1),ITOF(16)), ITOF(16));
        glMultiTexCoord4t(GL_TEXTURE0+i, texcoord[i][0], texcoord[i][1], texcoord[i][2], texcoord[i][3]);
    }

    /* Set current normal */
    normal[0] = OESRandom(ITOF(-4),ITOF(+4));
    normal[1] = OESRandom(ITOF(-4),ITOF(+4));
    normal[2] = OESRandom(ITOF(-4),ITOF(+4));
    glNormal3t(normal[0], normal[1], normal[2]);

    /* Detect errors */
    if (CheckErrors("TestCurrentValueState") == ERROR)
        return ERROR;

    /*
    ** Read back and verify the values
    */

    if (VerifyValues(Lookup_GetInteger, 0, GL_CURRENT_COLOR, 4, colori, Tolerance_Color, "GL_CURRENT_COLOR", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_GetFtype, 0, GL_CURRENT_COLOR, 4, colort, Tolerance_Color, "GL_CURRENT_COLOR", 0) == ERROR)
        return ERROR;

    for (i=0; i<numtexunits; ++i)
    {
        glActiveTexture(GL_TEXTURE0+i);
        if (VerifyValues(Lookup_GetFtype, 0, GL_CURRENT_TEXTURE_COORDS, 4, texcoord[i], Tolerance_Ftype, "GL_CURRENT_TEXTURE_COORDS (texture %d)", i) == ERROR)
            return ERROR;
    }

    if (VerifyValues(Lookup_GetFtype, 0, GL_CURRENT_NORMAL, 3, normal, Tolerance_Ftype, "GL_CURRENT_NORMAL", 0) == ERROR)
        return ERROR;

    return NO_ERROR;
}

/******************************************************************************/

static long TestBufferObjectState(void)
{
    GLubyte dummy[1024];

    GLenum  target[16];
    GLenum  usage[16];
    GLsizei size[16];
    GLuint  i, genbuf;
    GLboolean expectb;
    
    
    /* Test special case of glIsBuffer(0) */
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsBuffer, 0, 0, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", 0) == ERROR)
        return ERROR;

    /* Binding to buffer 0 should still return false for glIsBuffer(0) */
    glBindBuffer(GL_ARRAY_BUFFER, 0);
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsBuffer, 0, 0, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", 0) == ERROR)
        return ERROR;    
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsBuffer, 0, 0, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", 0) == ERROR)
        return ERROR;

    for (i=1; i<16; ++i)
    {
        /* Test glIsBuffer(i) before binding buffer i */
        expectb = GL_FALSE;
        if (VerifyValues(Lookup_IsBuffer, 0, i, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", i) == ERROR)
            return ERROR;

        if (OESRandomInt(0,1))
        {
            glBindBuffer(GL_ARRAY_BUFFER, i);
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 16+OESRandomInt(0,15));
            target[i] = GL_ARRAY_BUFFER;
        }
        else
        {
            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, i);
            glBindBuffer(GL_ARRAY_BUFFER, 16+OESRandomInt(0,15));
            target[i] = GL_ELEMENT_ARRAY_BUFFER;
        }

        /* Test glIsBuffer(i) after binding buffer i */
        expectb = GL_TRUE;
        if (VerifyValues(Lookup_IsBuffer, 0, i, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", i) == ERROR)
            return ERROR;

        if (OESRandomInt(0,1))
            usage[i] = GL_STATIC_DRAW;
        else
            usage[i] = GL_DYNAMIC_DRAW;

        size[i] = OESRandomInt(1, 1024);

        glBufferData(target[i], size[i], dummy, usage[i]);

        /* Test glIsBuffer(i) after binding back to buffer 0 */
        glBindBuffer(target[i],0);
        expectb = GL_TRUE;
        if (VerifyValues(Lookup_IsBuffer, 0, i, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", i) == ERROR)
            return ERROR;

    }

    /* Detect errors */
    if (CheckErrors("TestBufferObjectState") == ERROR)
        return ERROR;

    for (i=1; i<16; ++i)
    {
        glBindBuffer(target[i], i);
        if (VerifyValues(Lookup_GetBufferParameteri, target[i], GL_BUFFER_SIZE, 1, &size[i], Tolerance_Exact, "GL_BUFFER_SIZE (buffer %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetBufferParameteri, target[i], GL_BUFFER_USAGE, 1, &usage[i], Tolerance_Exact, "GL_BUFFER_USAGE (buffer %d)", i) == ERROR)
            return ERROR;
    }

    for (i=1; i<16; ++i)
    {
        /* Test glIsBuffer(i) after we delete the buffers */
        glDeleteBuffers(1, &i);
        expectb = GL_FALSE;
        if (VerifyValues(Lookup_IsBuffer, 0, i, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", i) == ERROR)
            return ERROR;
    }

    /* Test glIsBuffer() on generated but not bound buffer */
    glGenBuffers(1, &genbuf);
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsBuffer, 0, genbuf, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", genbuf) == ERROR)
        return ERROR;
    
    /* Test glIsBuffer() on generated and bound buffer */
    glBindBuffer(GL_ARRAY_BUFFER, genbuf);
    expectb = GL_TRUE;
    if (VerifyValues(Lookup_IsBuffer, 0, genbuf, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", genbuf) == ERROR)
        return ERROR;

    /* Test glIsBuffer() on generated and deleted buffer */
    glDeleteBuffers(1, &genbuf);
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsBuffer, 0, genbuf, 1, &expectb, Tolerance_Exact, "GL_IS_BUFFER (buffer %d)", genbuf) == ERROR)
        return ERROR;

    return NO_ERROR;
}

/******************************************************************************/

#define BUFFER_OFFSET(i) ((char *)NULL + (i))

static long TestVertexArrayState(void)
{
    GLint       numtexunits;
    GLubyte     dummy[1024];
    GLboolean   getzerostrides;

    GLuint       arraybind;
    GLint       elementbind;
    GLint       activetexture;

    GLint       vertexbind;
    GLboolean   vertexenable;
    GLint       vertextype;
    GLint       vertexsize;
    GLint       vertexstride;
    void*       vertexpointer;

    GLint       normalbind;
    GLboolean   normalenable;
    GLint       normaltype;
    GLint       normalstride;
    void*       normalpointer;

    GLint       colorbind;
    GLboolean   colorenable;
    GLint       colortype;
    GLint       colorsize;
    GLint       colorstride;
    void*       colorpointer;

    GLint       texcoordbind[32];
    GLboolean   texcoordenable[32];
    GLint       texcoordtype[32];
    GLint       texcoordsize[32];
    GLint       texcoordstride[32];
    void*       texcoordpointer[32];

    GLboolean   enable;
    GLint       type;
    GLint       size;
    GLint       bytes;
    GLint       start;
    GLint       stride;
    GLint       i;

    /* Get system parameters */
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_UNITS, &numtexunits, "GL_MAX_TEXTURE_UNITS", 0) == ERROR)
        return ERROR;

    /* Determine how zero strides are read back */
    glVertexPointer(2, GL_FTYPE, 0, dummy);
    if (LookupValues(Lookup_GetInteger, 0, GL_VERTEX_ARRAY_STRIDE, &vertexstride, "GL_VERTEX_ARRAY_STRIDE", 0) == ERROR)
        return ERROR;
    getzerostrides = (vertexstride == 0) ? GL_TRUE : GL_FALSE;

    /*
    ** Set random array info for each stream
    */

    /* Vertex */
    enable = OESRandomInt(0,1);
    switch (OESRandomInt(0,2))
    {
        case 0:
            type  = GL_BYTE;
            bytes = sizeof(GLbyte);
            break;
        case 1:
            type  = GL_SHORT;
            bytes = sizeof(GLshort);
            break;
        case 2:
            type  = GL_FTYPE;
            bytes = sizeof(GLftype);
            break;
        default:
            type = -1;
            bytes = -1;
            return ERROR;
    }
    size   = OESRandomInt(2,4);
    start  = OESRandomInt(0,255) * bytes;
    stride = OESRandomInt(0,8) * bytes;
    if (stride < size * bytes) stride = 0;

    arraybind = OESRandomInt(0,3);
    glBindBuffer(GL_ARRAY_BUFFER, arraybind);
    if (enable)
        glEnableClientState(GL_VERTEX_ARRAY);
    else
        glDisableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(size, type, stride, arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start]);

    vertexbind     = arraybind;
    vertexenable   = enable;
    vertextype     = type;
    vertexsize     = size;
    vertexstride   = (stride || getzerostrides) ? stride : (size * bytes);
    vertexpointer  = arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start];

    /* Normal */
    enable = OESRandomInt(0,1);;
    switch (OESRandomInt(0,2))
    {
        case 0:
            type  = GL_BYTE;
            bytes = sizeof(GLbyte);
            break;
        case 1:
            type  = GL_SHORT;
            bytes = sizeof(GLshort);
            break;
        case 2:
            type  = GL_FTYPE;
            bytes = sizeof(GLftype);
            break;
    }
    start  = OESRandomInt(0,255) * bytes;
    stride = OESRandomInt(0,7) * bytes;
    if (stride < 3 * bytes) stride = 0;

    arraybind = OESRandomInt(0,3);
    glBindBuffer(GL_ARRAY_BUFFER, arraybind);
    if (enable)
        glEnableClientState(GL_NORMAL_ARRAY);
    else
        glDisableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(type, stride, arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start]);

    normalbind     = arraybind;
    normalenable   = enable;
    normaltype     = type;
    normalstride   = (stride || getzerostrides) ? stride : (3 * bytes);
    normalpointer  = arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start];

    /* Color */
    enable = OESRandomInt(0,1);;
    switch (OESRandomInt(0,1))
    {
        case 0:
            type  = GL_UNSIGNED_BYTE;
            bytes = sizeof(GLbyte);
            break;
        case 1:
            type  = GL_FTYPE;
            bytes = sizeof(GLftype);
            break;
    }
    size   = 4;
    start  = OESRandomInt(0,255) * bytes;
    stride = OESRandomInt(0,7) * bytes;
    if (stride < size * bytes) stride = 0;

    arraybind = OESRandomInt(0,3);
    glBindBuffer(GL_ARRAY_BUFFER, arraybind);
    if (enable)
        glEnableClientState(GL_COLOR_ARRAY);
    else
        glDisableClientState(GL_COLOR_ARRAY);
    glColorPointer(size, type, stride, arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start]);

    colorbind     = arraybind;
    colorenable   = enable;
    colortype     = type;
    colorsize     = size;
    colorstride   = (stride || getzerostrides) ? stride : (size * bytes);
    colorpointer  = arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start];

    /* Texture */
    for (i=0; i<numtexunits; ++i)
    {
        glClientActiveTexture(GL_TEXTURE0+i);

        enable = OESRandomInt(0,1);;
        switch (OESRandomInt(0,2))
        {
            case 0:
                type  = GL_BYTE;
                bytes = sizeof(GLbyte);
                break;
            case 1:
                type  = GL_SHORT;
                bytes = sizeof(GLshort);
                break;
            case 2:
                type  = GL_FTYPE;
                bytes = sizeof(GLftype);
                break;
        }
        size   = OESRandomInt(2,4);
        start  = OESRandomInt(0,255) * bytes;
        stride = OESRandomInt(0,7) * bytes;
        if (stride < size * bytes) stride = 0;

        arraybind = OESRandomInt(0,3);
        glBindBuffer(GL_ARRAY_BUFFER, arraybind);
        if (enable)
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        else
            glDisableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(size, type, stride, arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start]);

        texcoordbind[i]     = arraybind;
        texcoordenable[i]   = enable;
        texcoordtype[i]     = type;
        texcoordsize[i]     = size;
        texcoordstride[i]   = (stride || getzerostrides) ? stride : (size * bytes);
        texcoordpointer[i]  = arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start];
    }

    /* Active client texture */
    activetexture = GL_TEXTURE0+OESRandomInt(0,numtexunits-1);
    glClientActiveTexture(activetexture);

    /* Array binding */
    arraybind = OESRandomInt(0,3);
    glBindBuffer(GL_ARRAY_BUFFER, arraybind);

    /* Element array binding */
    elementbind = OESRandomInt(0,3);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, elementbind);

    /* Detect errors */
    if (CheckErrors("TestVertexArrayState") == ERROR)
        return ERROR;

    /*
    ** Read back and verify the values
    */

    /* Client active texture */
    if (VerifyValues(Lookup_GetInteger, 0, GL_CLIENT_ACTIVE_TEXTURE, 1, &activetexture, Tolerance_Exact, "GL_CLIENT_ACTIVE_TEXTURE", 0) == ERROR)
        return ERROR;

    /* Vertex */
    if (VerifyValues(Lookup_GetInteger, 0, GL_VERTEX_ARRAY_BUFFER_BINDING, 1, &vertexbind, Tolerance_Exact, "GL_VERTEX_ARRAY_BUFFER_BINDING", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_VERTEX_ARRAY, 1, &vertexenable, Tolerance_Exact, "GL_VERTEX_ARRAY", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_VERTEX_ARRAY_SIZE, 1, &vertexsize, Tolerance_Exact, "GL_VERTEX_ARRAY_SIZE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_VERTEX_ARRAY_STRIDE, 1, &vertexstride, Tolerance_Exact, "GL_VERTEX_ARRAY_STRIDE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_VERTEX_ARRAY_TYPE, 1, &vertextype, Tolerance_Exact, "GL_VERTEX_ARRAY_TYPE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetPointer, 0, GL_VERTEX_ARRAY_POINTER, 1, &vertexpointer, Tolerance_Exact, "GL_VERTEX_ARRAY_POINTER", 0) == ERROR)
        return ERROR;

    /* Normal */
    if (VerifyValues(Lookup_GetInteger, 0, GL_NORMAL_ARRAY_BUFFER_BINDING, 1, &normalbind, Tolerance_Exact, "GL_NORMAL_ARRAY_BUFFER_BINDING", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_NORMAL_ARRAY, 1, &normalenable, Tolerance_Exact, "GL_NORMAL_ARRAY", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_NORMAL_ARRAY_STRIDE, 1, &normalstride, Tolerance_Exact, "GL_NORMAL_ARRAY_STRIDE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_NORMAL_ARRAY_TYPE, 1, &normaltype, Tolerance_Exact, "GL_NORMAL_ARRAY_TYPE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetPointer, 0, GL_NORMAL_ARRAY_POINTER, 1, &normalpointer, Tolerance_Exact, "GL_NORMAL_ARRAY_POINTER", 0) == ERROR)
        return ERROR;

    /* Color */
    if (VerifyValues(Lookup_GetInteger, 0, GL_COLOR_ARRAY_BUFFER_BINDING, 1, &colorbind, Tolerance_Exact, "GL_COLOR_ARRAY_BUFFER_BINDING", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_COLOR_ARRAY, 1, &colorenable, Tolerance_Exact, "GL_COLOR_ARRAY", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_COLOR_ARRAY_SIZE, 1, &colorsize, Tolerance_Exact, "GL_COLOR_ARRAY_SIZE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_COLOR_ARRAY_STRIDE, 1, &colorstride, Tolerance_Exact, "GL_COLOR_ARRAY_STRIDE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_COLOR_ARRAY_TYPE, 1, &colortype, Tolerance_Exact, "GL_COLOR_ARRAY_TYPE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetPointer, 0, GL_COLOR_ARRAY_POINTER, 1, &colorpointer, Tolerance_Exact, "GL_COLOR_ARRAY_POINTER", 0) == ERROR)
        return ERROR;

    /* Texture */
    for (i=0; i<numtexunits; ++i)
    {
        glClientActiveTexture(GL_TEXTURE0+i);
        if (VerifyValues(Lookup_GetInteger, 0, GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING, 1, &texcoordbind[i], Tolerance_Exact, "GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_IsEnabled, 0, GL_TEXTURE_COORD_ARRAY, 1, &texcoordenable[i], Tolerance_Exact, "GL_TEXTURE_COORD_ARRAY (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_TEXTURE_COORD_ARRAY_SIZE, 1, &texcoordsize[i], Tolerance_Exact, "GL_TEXTURE_COORD_ARRAY_SIZE (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_TEXTURE_COORD_ARRAY_STRIDE, 1, &texcoordstride[i], Tolerance_Exact, "GL_TEXTURE_COORD_ARRAY_STRIDE (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_TEXTURE_COORD_ARRAY_TYPE, 1, &texcoordtype[i], Tolerance_Exact, "GL_TEXTURE_COORD_ARRAY_TYPE (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetPointer, 0, GL_TEXTURE_COORD_ARRAY_POINTER, 1, &texcoordpointer[i], Tolerance_Exact, "GL_TEXTURE_COORD_ARRAY_POINTER (texture %d)", i) == ERROR)
            return ERROR;
    }

    /* Array Binding */
    if (VerifyValues(Lookup_GetInteger, 0, GL_ARRAY_BUFFER_BINDING, 1, &arraybind, Tolerance_Exact, "GL_ARRAY_BUFFER_BINDING", 0) == ERROR)
        return ERROR;

    /* Element array Binding */
    if (VerifyValues(Lookup_GetInteger, 0, GL_ELEMENT_ARRAY_BUFFER_BINDING, 1, &elementbind, Tolerance_Exact, "GL_ELEMENT_ARRAY_BUFFER_BINDING", 0) == ERROR)
        return ERROR;

    /* Clean up any buffers used */
    for (arraybind=1; arraybind<=3; ++arraybind)
    {
        glDeleteBuffers(1, &arraybind);
    }

    return NO_ERROR; 
}

/******************************************************************************/

static long TestTransformationState(void)
{
    GLint       numtexunits;
    GLint       maxmodelviewdepth;
    GLint       maxprojectiondepth;
    GLint       maxtexturedepth;
    GLint       maxclipplanes;

    GLftype     *modelviewstack = 0;
    GLftype     *projectionstack = 0;
    GLftype     *texturestack = 0;
    GLint       modelviewdepth;
    GLint       projectiondepth;
    GLint       texturedepth[16];
    GLenum      matrixmode;
    GLint       viewport[4];
    GLftype     depthrange[2];
    GLboolean   normalize;
    GLboolean   rescalenormal;
    GLftype     clipplane[6][4];
    GLboolean   clipplaneenable[6];

    GLint       i, j, k;

    /* Get system parameters */
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_UNITS, &numtexunits, "GL_MAX_TEXTURE_UNITS", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_MODELVIEW_STACK_DEPTH, &maxmodelviewdepth, "GL_MAX_MODELVIEW_STACK_DEPTH", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_PROJECTION_STACK_DEPTH, &maxprojectiondepth, "GL_MAX_PROJECTION_STACK_DEPTH", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_STACK_DEPTH, &maxtexturedepth, "GL_MAX_TEXTURE_STACK_DEPTH", 0) == ERROR)
        goto fail;
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_CLIP_PLANES, &maxclipplanes, "GL_MAX_CLIP_PLANES", 0) == ERROR)
        goto fail;

    /*
    ** Limit the tests in case the implementation supports really
    ** enormous stacks
    */
    if (maxmodelviewdepth > 64)
        maxmodelviewdepth = 64;
    if (maxprojectiondepth > 64)
        maxprojectiondepth = 64;
    if (maxtexturedepth > 64)
        maxtexturedepth = 64;
    if (maxclipplanes > 6)
        maxclipplanes = 6;

    modelviewstack  = (GLftype*)MALLOC(sizeof(GLftype) * 16 * maxmodelviewdepth);
    projectionstack = (GLftype*)MALLOC(sizeof(GLftype) * 16 * maxprojectiondepth);
    texturestack    = (GLftype*)MALLOC(sizeof(GLftype) * 16 * maxtexturedepth * numtexunits);
    
    /* Choose random values */
    for (j=0; j<maxmodelviewdepth; ++j)
        for (i=0; i<16; ++i)
            modelviewstack[16*j+i] = OESRandom(ITOF(-128), ITOF(+128));
    for (j=0; j<maxprojectiondepth; ++j)
        for (i=0; i<16; ++i)
            projectionstack[16*j+i] = OESRandom(ITOF(-128), ITOF(+128));
    for (k=0; k<numtexunits; ++k)
        for (j=0; j<maxtexturedepth; ++j)
            for (i=0; i<16; ++i)
                texturestack[16*(maxtexturedepth*k+j)+i] = OESRandom(ITOF(-128), ITOF(+128));
    switch (OESRandomInt(0,2))
    {
        case 0:
            matrixmode = GL_MODELVIEW;
            break;
        case 1:
            matrixmode = GL_PROJECTION;
            break;
        case 2:
            matrixmode = GL_TEXTURE;
            break;
    }
    
    viewport[0] = OESRandomInt(0,WINDSIZEX/2-1);
    viewport[1] = OESRandomInt(0,WINDSIZEY/2-1);
    viewport[2] = OESRandomInt(WINDSIZEX/2,WINDSIZEX) - viewport[0];
    viewport[3] = OESRandomInt(WINDSIZEX/2,WINDSIZEY) - viewport[1];

    depthrange[0] = OESRandom(ITOF(0),ITOF(1));
    depthrange[1] = OESRandom(ITOF(0),ITOF(1));

    normalize     = OESRandomInt(0,1);
    rescalenormal = OESRandomInt(0,1);
    for (j=0; j<maxclipplanes; ++j)
    {
        clipplaneenable[j] = OESRandomInt(0,1);
        for (i=0; i<4; ++i)
            clipplane[j][i] = OESRandom(ITOF(-1),ITOF(+1));
    }

    /* Set values */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    for (j=0; j<maxclipplanes; ++j)
    {
        if (clipplaneenable[j])
            glEnable(GL_CLIP_PLANE0+j);
        else
            glDisable(GL_CLIP_PLANE0+j);

        glClipPlanet(GL_CLIP_PLANE0+j, clipplane[j]);
    }

    for (modelviewdepth=0; modelviewdepth<maxmodelviewdepth; ++modelviewdepth)
    {
        glLoadMatrixt(&modelviewstack[16*modelviewdepth]);
        if ((modelviewdepth+1)<maxmodelviewdepth)
            glPushMatrix();
    }
    glMatrixMode(GL_PROJECTION);
    for (projectiondepth=0; projectiondepth<maxprojectiondepth; ++projectiondepth)
    {
        glLoadMatrixt(&projectionstack[16*projectiondepth]);
        if ((projectiondepth+1)<maxprojectiondepth)
            glPushMatrix();
    }
    glMatrixMode(GL_TEXTURE);
    for (k=0; k<numtexunits; ++k)
    {
        glActiveTexture(GL_TEXTURE0+k);
        for (texturedepth[k]=0; texturedepth[k]<maxtexturedepth; ++texturedepth[k])
        {
            glLoadMatrixt(&texturestack[16*(maxtexturedepth*k+texturedepth[k])]);
            if ((texturedepth[k]+1)<maxtexturedepth)
                glPushMatrix();
        }
    }
    glMatrixMode(matrixmode);

    glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
    glDepthRanget(depthrange[0], depthrange[1]);

    if (normalize)
        glEnable(GL_NORMALIZE);
    else
        glDisable(GL_NORMALIZE);

    if (rescalenormal)
        glEnable(GL_RESCALE_NORMAL);
    else
        glDisable(GL_RESCALE_NORMAL);

    /* Test values */
    for (j=0; j<maxclipplanes; ++j)
    {
        if (VerifyValues(Lookup_IsEnabled, 0, GL_CLIP_PLANE0+j, 1, &clipplaneenable[j], Tolerance_Exact, "GL_CLIP_PLANE%d", j) == ERROR)
            goto fail;
        if (VerifyValues(Lookup_GetClipPlanet, 0, GL_CLIP_PLANE0+j, 4, clipplane[j], Tolerance_Ftype, "GL_CLIP_PLANE%d", j) == ERROR)
            goto fail;
    }

    if (VerifyValues(Lookup_GetInteger, 0, GL_MATRIX_MODE, 1, &matrixmode, Tolerance_Exact, "GL_MATRIX_MODE", 0) == ERROR)
        goto fail;

    glMatrixMode(GL_MODELVIEW);
    while (1)
    {
        if (VerifyValues(Lookup_GetInteger, 0, GL_MODELVIEW_STACK_DEPTH, 1, &modelviewdepth, Tolerance_Exact, "GL_MODELVIEW_STACK_DEPTH", 0) == ERROR)
            goto fail;
        if (VerifyValues(Lookup_GetFtype, 0, GL_MODELVIEW_MATRIX, 16, &modelviewstack[16*(modelviewdepth-1)], Tolerance_Ftype, "GL_MODELVIEW_MATRIX", 0) == ERROR)
            goto fail;
        if (VerifyValues(Lookup_GetInteger, 0, GL_PROJECTION_STACK_DEPTH, 1, &projectiondepth, Tolerance_Exact, "GL_MODELVIEW_STACK_DEPTH", 0) == ERROR)
            goto fail;
        if (VerifyValues(Lookup_GetFtype, 0, GL_PROJECTION_MATRIX, 16, &projectionstack[16*(projectiondepth-1)], Tolerance_Ftype, "GL_PROJECTION_MATRIX", 0) == ERROR)
            goto fail;
        for (k=0; k<numtexunits; ++k)
        {
            glActiveTexture(GL_TEXTURE0+k);
            if (VerifyValues(Lookup_GetInteger, 0, GL_TEXTURE_STACK_DEPTH, 1, &texturedepth[k], Tolerance_Exact, "GL_TEXTURE_STACK_DEPTH", 0) == ERROR)
                goto fail;
            if (VerifyValues(Lookup_GetFtype, 0, GL_TEXTURE_MATRIX, 16, &texturestack[16*(maxtexturedepth*k+(texturedepth[k]-1))], Tolerance_Ftype, "GL_TEXTURE_MATRIX", 0) == ERROR)
                goto fail;
        }
        if (modelviewdepth == 1)
            break;
        modelviewdepth--;
        glPopMatrix();
    }

    glMatrixMode(GL_PROJECTION);
    while (1)
    {
        if (VerifyValues(Lookup_GetInteger, 0, GL_MODELVIEW_STACK_DEPTH, 1, &modelviewdepth, Tolerance_Exact, "GL_MODELVIEW_STACK_DEPTH", 0) == ERROR)
            goto fail;
        if (VerifyValues(Lookup_GetFtype, 0, GL_MODELVIEW_MATRIX, 16, &modelviewstack[16*(modelviewdepth-1)], Tolerance_Ftype, "GL_MODELVIEW_MATRIX", 0) == ERROR)
            goto fail;
        if (VerifyValues(Lookup_GetInteger, 0, GL_PROJECTION_STACK_DEPTH, 1, &projectiondepth, Tolerance_Exact, "GL_MODELVIEW_STACK_DEPTH", 0) == ERROR)
            goto fail;
        if (VerifyValues(Lookup_GetFtype, 0, GL_PROJECTION_MATRIX, 16, &projectionstack[16*(projectiondepth-1)], Tolerance_Ftype, "GL_PROJECTION_MATRIX", 0) == ERROR)
            goto fail;
        for (k=0; k<numtexunits; ++k)
        {
            glActiveTexture(GL_TEXTURE0+k);
            if (VerifyValues(Lookup_GetInteger, 0, GL_TEXTURE_STACK_DEPTH, 1, &texturedepth[k], Tolerance_Exact, "GL_TEXTURE_STACK_DEPTH", 0) == ERROR)
                goto fail;
            if (VerifyValues(Lookup_GetFtype, 0, GL_TEXTURE_MATRIX, 16, &texturestack[16*(maxtexturedepth*k+(texturedepth[k]-1))], Tolerance_Ftype, "GL_TEXTURE_MATRIX", 0) == ERROR)
                goto fail;
        }
        if (projectiondepth == 1)
            break;
        projectiondepth--;
        glPopMatrix();
    }

    glMatrixMode(GL_TEXTURE);
    for (j=0; j<numtexunits; ++j)
    {
        while (1)
        {
            if (VerifyValues(Lookup_GetInteger, 0, GL_MODELVIEW_STACK_DEPTH, 1, &modelviewdepth, Tolerance_Exact, "GL_MODELVIEW_STACK_DEPTH", 0) == ERROR)
                goto fail;
            if (VerifyValues(Lookup_GetFtype, 0, GL_MODELVIEW_MATRIX, 16, &modelviewstack[16*(modelviewdepth-1)], Tolerance_Ftype, "GL_MODELVIEW_MATRIX", 0) == ERROR)
                goto fail;
            if (VerifyValues(Lookup_GetInteger, 0, GL_PROJECTION_STACK_DEPTH, 1, &projectiondepth, Tolerance_Exact, "GL_MODELVIEW_STACK_DEPTH", 0) == ERROR)
                goto fail;
            if (VerifyValues(Lookup_GetFtype, 0, GL_PROJECTION_MATRIX, 16, &projectionstack[16*(projectiondepth-1)], Tolerance_Ftype, "GL_PROJECTION_MATRIX", 0) == ERROR)
                goto fail;
            for (k=0; k<numtexunits; ++k)
            {
                glActiveTexture(GL_TEXTURE0+k);
                if (VerifyValues(Lookup_GetInteger, 0, GL_TEXTURE_STACK_DEPTH, 1, &texturedepth[k], Tolerance_Exact, "GL_TEXTURE_STACK_DEPTH", 0) == ERROR)
                    goto fail;
                if (VerifyValues(Lookup_GetFtype, 0, GL_TEXTURE_MATRIX, 16, &texturestack[16*(maxtexturedepth*k+(texturedepth[k]-1))], Tolerance_Ftype, "GL_TEXTURE_MATRIX", 0) == ERROR)
                    goto fail;
            }
            if (texturedepth[j] == 1)
                break;
            glActiveTexture(GL_TEXTURE0+j);
            texturedepth[j]--;
            glPopMatrix();
        }
    }

    if (VerifyValues(Lookup_GetInteger, 0, GL_VIEWPORT, 4, viewport, Tolerance_Exact, "GL_VIEWPORT", 0) == ERROR)
        goto fail;
    if (VerifyValues(Lookup_GetFtype, 0, GL_DEPTH_RANGE, 2, depthrange, Tolerance_Ftype, "GL_DEPTH_RANGE", 0) == ERROR)
        goto fail;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_NORMALIZE, 1, &normalize, Tolerance_Exact, "GL_NORMALIZE", 0) == ERROR)
        goto fail;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_RESCALE_NORMAL, 1, &rescalenormal, Tolerance_Exact, "GL_RESCALE_NORMAL", 0) == ERROR)
        goto fail;
    

    FREE(texturestack);
    FREE(projectionstack);
    FREE(modelviewstack);
    return NO_ERROR;

    fail:
    if(texturestack) FREE(texturestack);
    if(projectionstack) FREE(projectionstack);
    if(modelviewstack)  FREE(modelviewstack);

    return ERROR;
}


/******************************************************************************/

static long TestColoringState(void)
{
    GLftype     color[4];
    GLftype     density;
    GLftype     start;
    GLftype     end;
    GLenum      mode;
    GLboolean   enable;
    GLenum      shademodel;

    color[0] = OESRandom(ITOF(0),ITOF(1));
    color[1] = OESRandom(ITOF(0),ITOF(1));
    color[2] = OESRandom(ITOF(0),ITOF(1));
    color[3] = OESRandom(ITOF(0),ITOF(1));
    density  = OESRandom(ITOF(0),ITOF(2));
    start    = OESRandom(ITOF(0),ITOF(5));
    end      = OESRandom(ITOF(0),ITOF(5));
    switch (OESRandomInt(0,2))
    {
        case 0:
            mode = GL_LINEAR;
            break;
        case 1:
            mode = GL_EXP;
            break;
        case 2:
            mode = GL_EXP2;
            break;
    }
    enable = OESRandomInt(0,1);
    if (OESRandom(0,1))
        shademodel = GL_SMOOTH;
    else
        shademodel = GL_FLAT;

    glFogtv(GL_FOG_COLOR, color);
    glFogt(GL_FOG_DENSITY, density);
    glFogt(GL_FOG_START, start);
    glFogt(GL_FOG_END, end);
    glFogt(GL_FOG_MODE, (GLftype)mode);
    if (enable)
        glEnable(GL_FOG);
    else
        glDisable(GL_FOG);
    glShadeModel(shademodel);

    /* Detect errors */
    if (CheckErrors("TestColoringState") == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_GetFtype, 0, GL_FOG_COLOR, 4, color, Tolerance_Color, "GL_FOG_COLOR", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_FOG_DENSITY, 1, &density, Tolerance_Ftype, "GL_FOG_DENSITY", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_FOG_START, 1, &start, Tolerance_Ftype, "GL_FOG_START", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_FOG_END, 1, &end, Tolerance_Ftype, "GL_FOG_END", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_FOG_MODE, 1, &mode, Tolerance_Exact, "GL_FOG_MODE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_FOG, 1, &enable, Tolerance_Exact, "GL_FOG", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_SHADE_MODEL, 1, &shademodel, Tolerance_Exact, "GL_SHADE_MODEL", 0) == ERROR)
        return ERROR;

    return NO_ERROR;
}

/******************************************************************************/

static long TestLightingState(void)
{
    GLboolean   colormaterial;
    GLboolean   lightingenable;
    GLboolean   lightenable[8];
    GLboolean   lightmodeltwosided;
    GLftype     lightmodelambient[4];
    GLftype     mambient[4];
    GLftype     mdiffuse[4];
    GLftype     mspecular[4];
    GLftype     memission[4];
    GLftype     modelview[16];
    GLftype     mshininess;
    GLftype     lambient[8][4];
    GLftype     ldiffuse[8][4];
    GLftype     lspecular[8][4];
    GLftype     lposition[8][4];
    GLftype     lattenuation0[8];
    GLftype     lattenuation1[8];
    GLftype     lattenuation2[8];
    GLftype     lspotdirection[8][3];
    GLftype     lspotexponent[8];
    GLftype     lspotcutoff[8];
    GLftype     colort[4];
    GLint       i;

    /* Choose random values */

    lightingenable       = OESRandomInt(0,1);
    lightmodeltwosided   = OESRandomInt(0,1);
    lightmodelambient[0] = OESRandom(ITOF(0),ITOF(1));
    lightmodelambient[1] = OESRandom(ITOF(0),ITOF(1));
    lightmodelambient[2] = OESRandom(ITOF(0),ITOF(1));
    lightmodelambient[3] = OESRandom(ITOF(0),ITOF(1));
    colormaterial        = OESRandomInt(0,1);
    
    mambient[0]  = OESRandom(ITOF(0),ITOF(1));
    mambient[1]  = OESRandom(ITOF(0),ITOF(1));
    mambient[2]  = OESRandom(ITOF(0),ITOF(1));
    mambient[3]  = OESRandom(ITOF(0),ITOF(1));
    mdiffuse[0]  = OESRandom(ITOF(0),ITOF(1));
    mdiffuse[1]  = OESRandom(ITOF(0),ITOF(1));
    mdiffuse[2]  = OESRandom(ITOF(0),ITOF(1));
    mdiffuse[3]  = OESRandom(ITOF(0),ITOF(1));
    mspecular[0] = OESRandom(ITOF(0),ITOF(1));
    mspecular[1] = OESRandom(ITOF(0),ITOF(1));
    mspecular[2] = OESRandom(ITOF(0),ITOF(1));
    mspecular[3] = OESRandom(ITOF(0),ITOF(1));
    memission[0] = OESRandom(ITOF(0),ITOF(1));
    memission[1] = OESRandom(ITOF(0),ITOF(1));
    memission[2] = OESRandom(ITOF(0),ITOF(1));
    memission[3] = OESRandom(ITOF(0),ITOF(1));
    mshininess   = OESRandom(ITOF(0),ITOF(127));

    for (i=0; i<8; ++i)
    {
        lightenable[i]       = OESRandomInt(0,1);
        lambient[i][0]       = OESRandom(ITOF(0),ITOF(1));
        lambient[i][1]       = OESRandom(ITOF(0),ITOF(1));
        lambient[i][2]       = OESRandom(ITOF(0),ITOF(1));
        lambient[i][3]       = OESRandom(ITOF(0),ITOF(1));
        ldiffuse[i][0]       = OESRandom(ITOF(0),ITOF(1));
        ldiffuse[i][1]       = OESRandom(ITOF(0),ITOF(1));
        ldiffuse[i][2]       = OESRandom(ITOF(0),ITOF(1));
        ldiffuse[i][3]       = OESRandom(ITOF(0),ITOF(1));
        lspecular[i][0]      = OESRandom(ITOF(0),ITOF(1));
        lspecular[i][1]      = OESRandom(ITOF(0),ITOF(1));
        lspecular[i][2]      = OESRandom(ITOF(0),ITOF(1));
        lspecular[i][3]      = OESRandom(ITOF(0),ITOF(1));
        lposition[i][0]      = OESRandom(ITOF(-1),ITOF(+1));
        lposition[i][1]      = OESRandom(ITOF(-1),ITOF(+1));
        lposition[i][2]      = OESRandom(ITOF(-1),ITOF(+1));
        lposition[i][3]      = ITOF(1);
        lattenuation0[i]     = OESRandom(ITOF(0),ITOF(1));
        lattenuation1[i]     = OESRandom(ITOF(0),ITOF(1));
        lattenuation2[i]     = OESRandom(ITOF(0),ITOF(1));
        lspotdirection[i][0] = OESRandom(ITOF(0),ITOF(1));
        lspotdirection[i][1] = OESRandom(ITOF(0),ITOF(1));
        lspotdirection[i][2] = OESRandom(ITOF(0),ITOF(1));
        lspotexponent[i]     = OESRandom(ITOF(0),ITOF(127));
        lspotcutoff[i]       = OESRandom(ITOF(0),ITOF(90));
    }

    /* Per bugzilla 3122, set a well-defined matrix so light coordinates are returned in eyespace */
	modelview[0] = ITOF(2); modelview[1] = ITOF(1); modelview[2] = ITOF(0); modelview[3] = ITOF(0);
	modelview[4] = ITOF(0); modelview[5] = ITOF(1); modelview[6] = ITOF(0); modelview[7] = ITOF(0);
	modelview[8] = ITOF(0); modelview[9] = ITOF(0); modelview[10] = ITOF(1); modelview[11] = ITOF(0);
	modelview[12] = ITOF(0); modelview[13] = ITOF(0); modelview[14] = ITOF(0); modelview[15] = ITOF(1);

    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixt(modelview);

    /* Set values */
    if (colormaterial)
        glEnable(GL_COLOR_MATERIAL);
    else
        glDisable(GL_COLOR_MATERIAL);
    
    if (lightingenable)
        glEnable(GL_LIGHTING);
    else
        glDisable(GL_LIGHTING);

    glLightModelt(GL_LIGHT_MODEL_TWO_SIDE, (GLftype)lightmodeltwosided);
    glLightModeltv(GL_LIGHT_MODEL_AMBIENT, lightmodelambient);

    if (LookupValues(Lookup_GetFtype, 0, GL_CURRENT_COLOR, &colort, "GL_CURRENT_COLOR", 0) == ERROR)
        return ERROR;       

    glMaterialtv(GL_FRONT_AND_BACK, GL_AMBIENT, mambient);
    glMaterialtv(GL_FRONT_AND_BACK, GL_DIFFUSE, mdiffuse);
    glMaterialtv(GL_FRONT_AND_BACK, GL_SPECULAR, mspecular);
    glMaterialtv(GL_FRONT_AND_BACK, GL_EMISSION, memission);
    glMaterialt(GL_FRONT_AND_BACK, GL_SHININESS, mshininess);

    for (i=0; i<8; ++i)
    {
        if (lightenable[i])
            glEnable(GL_LIGHT0+i);
        else
            glDisable(GL_LIGHT0+i);

        glLighttv(GL_LIGHT0+i, GL_AMBIENT, lambient[i]);
        glLighttv(GL_LIGHT0+i, GL_DIFFUSE, ldiffuse[i]);
        glLighttv(GL_LIGHT0+i, GL_SPECULAR, lspecular[i]);
        glLighttv(GL_LIGHT0+i, GL_POSITION, lposition[i]);
        glLightt(GL_LIGHT0+i, GL_CONSTANT_ATTENUATION, lattenuation0[i]);
        glLightt(GL_LIGHT0+i, GL_LINEAR_ATTENUATION, lattenuation1[i]);
        glLightt(GL_LIGHT0+i, GL_QUADRATIC_ATTENUATION, lattenuation2[i]);
        glLighttv(GL_LIGHT0+i, GL_SPOT_DIRECTION, lspotdirection[i]);
        glLightt(GL_LIGHT0+i, GL_SPOT_EXPONENT, lspotexponent[i]);
        glLightt(GL_LIGHT0+i, GL_SPOT_CUTOFF, lspotcutoff[i]);
    }

    /* Detect errors */
    if (CheckErrors("TestLightingState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_IsEnabled, 0, GL_COLOR_MATERIAL, 1, &colormaterial, Tolerance_Exact, "GL_COLOR_MATERIAL", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_LIGHTING, 1, &lightingenable, Tolerance_Exact, "GL_LIGHTING", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_GetBoolean, 0, GL_LIGHT_MODEL_TWO_SIDE, 1, &lightmodeltwosided, Tolerance_Exact, "GL_LIGHT_MODEL_TWO_SIDE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_LIGHT_MODEL_AMBIENT, 4, lightmodelambient, Tolerance_Color, "GL_LIGHT_MODEL_AMBIENT", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_GetMaterialt, GL_FRONT, GL_AMBIENT, 4, colormaterial ? colort : mambient, Tolerance_Color, "GL_AMBIENT", 0) == ERROR)
         return ERROR;
    if (VerifyValues(Lookup_GetMaterialt, GL_FRONT, GL_DIFFUSE, 4, colormaterial ? colort : mdiffuse, Tolerance_Color, "GL_DIFFUSE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetMaterialt, GL_FRONT, GL_SPECULAR, 4, mspecular, Tolerance_Color, "GL_SPECULAR", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetMaterialt, GL_FRONT, GL_EMISSION, 4, memission, Tolerance_Color, "GL_EMISSION", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetMaterialt, GL_FRONT, GL_SHININESS, 1, &mshininess, Tolerance_Ftype, "GL_SHININESS", 0) == ERROR)
        return ERROR;

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    for (i=0; i<8; ++i)
    {
        XFormMul4x4(modelview, lposition[i], lposition[i]);
        XFormMul3x3(modelview, lspotdirection[i], lspotdirection[i]);

        if (VerifyValues(Lookup_IsEnabled, 0, GL_LIGHT0+i, 1, &lightenable[i], Tolerance_Exact, "GL_LIGHT%d", i) == ERROR)
            return ERROR;

        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_AMBIENT, 4, lambient[i], Tolerance_Color, "GL_AMBIENT (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_DIFFUSE, 4, ldiffuse[i], Tolerance_Color, "GL_DIFFUSE (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_SPECULAR, 4, lspecular[i], Tolerance_Color, "GL_SPECULAR (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_POSITION, 4, lposition[i], Tolerance_Ftype, "GL_POSITION (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_CONSTANT_ATTENUATION, 1, &lattenuation0[i], Tolerance_Ftype, "GL_CONSTANT_ATTENUATION (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_LINEAR_ATTENUATION, 1, &lattenuation1[i], Tolerance_Ftype, "GL_LINEAR_ATTENUATION (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_QUADRATIC_ATTENUATION, 1, &lattenuation2[i], Tolerance_Ftype, "GL_QUADRATIC_ATTENUATION (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_SPOT_DIRECTION, 3, lspotdirection[i], Tolerance_Ftype, "GL_SPOT_DIRECTION (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_SPOT_EXPONENT, 1, &lspotexponent[i], Tolerance_Ftype, "GL_SPOT_EXPONENT (light %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetLightt, GL_LIGHT0+i , GL_SPOT_CUTOFF, 1, &lspotcutoff[i], Tolerance_Ftype, "GL_SPOT_CUTOFF (light %d)", i) == ERROR)
            return ERROR;
    }

    return NO_ERROR;
}

/******************************************************************************/

static long TestRasterizationState(void)
{
    GLftype     hwpointsizerange[2];
    GLftype     hwlinewidthrange[2];
    GLboolean   pointsmooth;
    GLftype     pointsize;
    GLftype     pointsizemin;
    GLftype     pointsizemax;
    GLftype     pointfadethresh;
    GLftype     pointattenuation[3];
    GLftype     linewidth;
    GLboolean   linesmooth;
    GLboolean   cullenable;
    GLenum      cullmode;
    GLenum      frontface;
    GLboolean   polygonoffsetenable;
    GLftype     polygonoffsetfactor;
    GLftype     polygonoffsetunits;

    /* Get system limits */
    if (LookupValues(Lookup_GetFtype, 0, GL_SMOOTH_POINT_SIZE_RANGE, hwpointsizerange, "GL_SMOOTH_POINT_SIZE_RANGE", 0) == ERROR)
        return ERROR;
    if (LookupValues(Lookup_GetFtype, 0, GL_SMOOTH_LINE_WIDTH_RANGE, hwlinewidthrange, "GL_SMOOTH_LINE_WIDTH_RANGE", 0) == ERROR)
        return ERROR;

    /* Choose random values */
    pointsmooth  = OESRandomInt(0,1);
    pointsize    = OESRandom(hwpointsizerange[0],hwpointsizerange[1]);
    pointsizemin = OESRandom(hwpointsizerange[0],hwpointsizerange[1]);
    pointsizemax = OESRandom(pointsizemin,hwpointsizerange[1]);
    pointfadethresh     = OESRandom(ITOF(0),ITOF(1));
    pointattenuation[0] = OESRandom(ITOF(0),ITOF(1));
    pointattenuation[1] = OESRandom(ITOF(0),ITOF(1));
    pointattenuation[2] = OESRandom(ITOF(0),ITOF(1));

    linesmooth = OESRandomInt(0,1);
    linewidth  = OESRandom(hwlinewidthrange[0],hwlinewidthrange[1]);

    cullenable = OESRandomInt(0,1);
    switch (OESRandomInt(0,2))
    {
        case 0:
            cullmode = GL_FRONT;
            break;
        case 1:
            cullmode = GL_BACK;
            break;
        case 2:
            cullmode = GL_FRONT_AND_BACK;
            break;
    }
    frontface = OESRandomInt(0,1) ? GL_CW : GL_CCW;

    polygonoffsetenable = OESRandomInt(0,1);
    polygonoffsetfactor = OESRandom(ITOF(-1),ITOF(+1));
    polygonoffsetunits  = OESRandom(ITOF(-1),ITOF(+1));

    /* Set values */
    if (pointsmooth)
        glEnable(GL_POINT_SMOOTH);
    else
        glDisable(GL_POINT_SMOOTH);
    glPointSizet(pointsize);
    glPointParametert(GL_POINT_SIZE_MIN, pointsizemin);
    glPointParametert(GL_POINT_SIZE_MAX, pointsizemax);
    glPointParametert(GL_POINT_FADE_THRESHOLD_SIZE, pointfadethresh);
    glPointParametertv(GL_POINT_DISTANCE_ATTENUATION, pointattenuation);
    
    if (linesmooth)
        glEnable(GL_LINE_SMOOTH);
    else
        glDisable(GL_LINE_SMOOTH);
    glLineWidtht(linewidth);

    if (cullenable)
        glEnable(GL_CULL_FACE);
    else
        glDisable(GL_CULL_FACE);
    glCullFace(cullmode);
    glFrontFace(frontface);
    if (polygonoffsetenable)
        glEnable(GL_POLYGON_OFFSET_FILL);
    else
        glDisable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffsett(polygonoffsetfactor, polygonoffsetunits);

    /* Detect errors */
    if (CheckErrors("TestRasterizationState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_IsEnabled, 0, GL_POINT_SMOOTH, 1, &pointsmooth, Tolerance_Exact, "GL_POINT_SMOOTH", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_POINT_SIZE, 1, &pointsize, Tolerance_Ftype, "GL_POINT_SIZE", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_GetFtype, 0, GL_POINT_SIZE_MIN, 1, &pointsizemin, Tolerance_Ftype, "GL_POINT_SIZE_MIN", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_POINT_SIZE_MAX, 1, &pointsizemax, Tolerance_Ftype, "GL_POINT_SIZE_MAX", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_POINT_FADE_THRESHOLD_SIZE, 1, &pointfadethresh, Tolerance_Ftype, "GL_POINT_FADE_THRESHOLD_SIZE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_POINT_DISTANCE_ATTENUATION, 3, pointattenuation, Tolerance_Ftype, "GL_POINT_DISTANCE_ATTENUATION", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_IsEnabled, 0, GL_LINE_SMOOTH, 1, &linesmooth, Tolerance_Exact, "GL_LINE_SMOOTH", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_LINE_WIDTH, 1, &linewidth, Tolerance_Ftype, "GL_LINE_WIDTH", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_IsEnabled, 0, GL_CULL_FACE, 1, &cullenable, Tolerance_Exact, "GL_CULL_FACE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_CULL_FACE_MODE, 1, &cullmode, Tolerance_Exact, "GL_CULL_FACE_MODE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_FRONT_FACE, 1, &frontface, Tolerance_Exact, "GL_FRONT_FACE", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_IsEnabled, 0, GL_POLYGON_OFFSET_FILL, 1, &polygonoffsetenable, Tolerance_Exact, "GL_POLYGON_OFFSET_FILL", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_POLYGON_OFFSET_FACTOR, 1, &polygonoffsetfactor, Tolerance_Ftype, "GL_POLYGON_OFFSET_FACTOR", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_POLYGON_OFFSET_UNITS, 1, &polygonoffsetunits, Tolerance_Ftype, "GL_POLYGON_OFFSET_UNITS", 0) == ERROR)
        return ERROR;

    return NO_ERROR;
}

/******************************************************************************/

static long TestMultisamplingState(void)
{
    GLint       samplebuffers;
    GLboolean   multisample;
    GLboolean   alphatocoverage;
    GLboolean   alphatoone;
    GLboolean   samplecoverage;
    GLftype     samplecoveragevalue;
    GLboolean   samplecoverageinvert;

    /* Check whether multisampling is supported */
    if (LookupValues(Lookup_GetInteger, 0, GL_SAMPLE_BUFFERS, &samplebuffers, "GL_SAMPLE_BUFFERS", 0) == ERROR)
        return ERROR;
    if (samplebuffers == 0) return NO_ERROR;

    /* Choose values */
    multisample          = OESRandomInt(0,1);
    alphatocoverage      = OESRandomInt(0,1);
    alphatoone           = OESRandomInt(0,1);
    samplecoverage       = OESRandomInt(0,1);
    samplecoveragevalue  = OESRandom(ITOF(0),ITOF(1));
    samplecoverageinvert = OESRandomInt(0,1);

    /* Set values */
    if (multisample)
        glEnable(GL_MULTISAMPLE);
    else
        glDisable(GL_MULTISAMPLE);

    if (alphatocoverage)
        glEnable(GL_SAMPLE_ALPHA_TO_COVERAGE);
    else
        glDisable(GL_SAMPLE_ALPHA_TO_COVERAGE);

    if (alphatoone)
        glEnable(GL_SAMPLE_ALPHA_TO_ONE);
    else
        glDisable(GL_SAMPLE_ALPHA_TO_ONE);

    if (samplecoverage)
        glEnable(GL_SAMPLE_COVERAGE);
    else
        glDisable(GL_SAMPLE_COVERAGE);

    glSampleCoveraget(samplecoveragevalue, samplecoverageinvert);

    /* Detect errors */
    if (CheckErrors("TestMultisamplingState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_IsEnabled, 0, GL_MULTISAMPLE, 1, &multisample, Tolerance_Exact, "GL_MULTISAMPLE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_SAMPLE_ALPHA_TO_COVERAGE, 1, &alphatocoverage, Tolerance_Exact, "GL_SAMPLE_ALPHA_TO_COVERAGE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_SAMPLE_ALPHA_TO_ONE, 1, &alphatoone, Tolerance_Exact, "GL_SAMPLE_ALPHA_TO_ONE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_SAMPLE_COVERAGE, 1, &samplecoverage, Tolerance_Exact, "GL_SAMPLE_COVERAGE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_SAMPLE_COVERAGE_VALUE, 1, &samplecoveragevalue, Tolerance_Ftype, "GL_SAMPLE_COVERAGE_VALUE", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetBoolean, 0, GL_SAMPLE_COVERAGE_INVERT, 1, &samplecoverageinvert, Tolerance_Exact, "GL_SAMPLE_COVERAGE_INVERT", 0) == ERROR)
        return ERROR;

    return NO_ERROR;
}


/******************************************************************************/

static long TestTextureObjectState(void)
{
    GLboolean   texenabled;
    GLboolean   expectb;
    GLint       texbind;
    GLenum      texminfilter[16];
    GLenum      texmagfilter[16];
    GLenum      texwraps[16];
    GLenum      texwrapt[16];
    GLint       texgenmipmap[16];
    GLuint      i, gentex;

    /* Test special case of glIsTexture(0) */
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsTexture, 0, 0, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", 0) == ERROR)
        return ERROR;

    /* Binding to texture 0 should still return false for glIsTexture(0) */
    glBindTexture(GL_TEXTURE_2D, 0);
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsTexture, 0, 0, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", 0) == ERROR)
        return ERROR;    

    /* Choose values */
    texenabled = OESRandomInt(0,1);
    texbind    = OESRandomInt(0,15);
    for (i=0; i<16; ++i)
    {
        switch (OESRandomInt(0,5))
        {
            case 0:
                texminfilter[i] = GL_NEAREST;
                break;
            case 1:
                texminfilter[i] = GL_LINEAR;
                break;
            case 2:
                texminfilter[i] = GL_NEAREST_MIPMAP_NEAREST;
                break;
            case 3:
                texminfilter[i] = GL_NEAREST_MIPMAP_LINEAR;
                break;
            case 4:
                texminfilter[i] = GL_LINEAR_MIPMAP_NEAREST;
                break;
            case 5:
                texminfilter[i] = GL_LINEAR_MIPMAP_LINEAR;
                break;
        }
        switch (OESRandomInt(0,1))
        {
            case 0:
                texmagfilter[i] = GL_NEAREST;
                break;
            case 1:
                texmagfilter[i] = GL_LINEAR;
                break;
        }
        switch (OESRandomInt(0,1))
        {
            case 0:
                texwraps[i] = GL_CLAMP_TO_EDGE;
                break;
            case 1:
                texwraps[i] = GL_REPEAT;
                break;
        }
        switch (OESRandomInt(0,1))
        {
            case 0:
                texwrapt[i] = GL_CLAMP_TO_EDGE;
                break;
            case 1:
                texwrapt[i] = GL_REPEAT;
                break;
        }
        texgenmipmap[i] = OESRandom(0,1);
    }

    /* Set values */
    if (texenabled)
        glEnable(GL_TEXTURE_2D);
    else
        glDisable(GL_TEXTURE_2D);
    for (i=0; i<16; ++i)
    {
        /* Test glIsTexture(i) before binding texture i */
        expectb = GL_FALSE;
        if ( i != 0 && VerifyValues(Lookup_IsTexture, 0, i, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", i) == ERROR)
            return ERROR;

        glBindTexture(GL_TEXTURE_2D, i);
        glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (GLftype)texminfilter[i]);
        glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, (GLftype)texmagfilter[i]);
        glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (GLftype)texwraps[i]);
        glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (GLftype)texwrapt[i]);
        glTexParametert(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, (GLftype)texgenmipmap[i]);

        /* Test glIsTexture(i) after binding to texture i */
        expectb = GL_TRUE;
        if ( i != 0 && VerifyValues(Lookup_IsTexture, 0, i, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", i) == ERROR)
            return ERROR;

        /* Test glIsTexture(i) after binding back to texture 0 */
        glBindTexture(GL_TEXTURE_2D, 0);
        expectb = GL_TRUE;
        if ( i != 0 && VerifyValues(Lookup_IsTexture, 0, i, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", i) == ERROR)
            return ERROR;

    }
    glBindTexture(GL_TEXTURE_2D, texbind);

    /* Detect errors */
    if (CheckErrors("TestTextureObjectState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_IsEnabled, 0, GL_TEXTURE_2D, 1, &texenabled, Tolerance_Exact, "GL_TEXTURE_2D", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_TEXTURE_BINDING_2D, 1, &texbind, Tolerance_Exact, "GL_TEXTURE_BINDING_2D", 0) == ERROR)
        return ERROR;
    for (i=0; i<16; ++i)
    {
        glBindTexture(GL_TEXTURE_2D, i);
        if (VerifyValues(Lookup_GetTexParameteri, GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 1, &texminfilter[i], Tolerance_Exact, "GL_TEXTURE_MIN_FILTER (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexParameteri, GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, 1, &texmagfilter[i], Tolerance_Exact, "GL_TEXTURE_MAG_FILTER (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexParameteri, GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, 1, &texwraps[i], Tolerance_Exact, "GL_TEXTURE_WRAP_S (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexParameteri, GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, 1, &texwrapt[i], Tolerance_Exact, "GL_TEXTURE_WRAP_T (texture %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexParameteri, GL_TEXTURE_2D, GL_GENERATE_MIPMAP, 1, &texgenmipmap[i], Tolerance_Exact, "GL_GENERATE_MIPMAP (texture %d)", i) == ERROR)
            return ERROR;
    }

    for (i=1; i<16; ++i)
    {
        /* Test glIsTexture(i) after deleting texture */
        glDeleteTextures(1, &i);
        expectb = GL_FALSE;
        if ( VerifyValues(Lookup_IsTexture, 0, i, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", i) == ERROR)
            return ERROR;
    }

    /* Test glIsTexture() on generated but not bound texture */
    glGenTextures(1, &gentex);
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsTexture, 0, gentex, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", gentex) == ERROR)
        return ERROR;
    
    /* Test glIsTexture() on generated and bound texture */
    glBindTexture(GL_TEXTURE_2D, gentex);
    glBindTexture(GL_TEXTURE_2D, 0);
    expectb = GL_TRUE;
    if (VerifyValues(Lookup_IsTexture, 0, gentex, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", gentex) == ERROR)
        return ERROR;
    
    /* Test glIsTexture() after we delete the allocated texture */
    glDeleteTextures(1, &gentex);
    expectb = GL_FALSE;
    if (VerifyValues(Lookup_IsTexture, 0, gentex, 1, &expectb, Tolerance_Exact, "GL_IS_TEXTURE (texture %d)", gentex) == ERROR)
        return ERROR;

    return NO_ERROR;
}

/******************************************************************************/

static long TestTextureEnvironmentState(void)
{
    GLint   numtexunits;
    GLint   active;
    GLenum  envmode[16];
    GLftype envcolor[16][4];
    GLenum  combinergb[16];
    GLenum  combinealpha[16];
    GLenum  src[16][3][2];
    GLenum  operand[16][3][2];
    GLftype rgbscale[16];
    GLftype alphascale[16];
    GLint   i, j, k;

    /* Get system parameters */
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_UNITS, &numtexunits, "GL_MAX_TEXTURE_UNITS", 0) == ERROR)
        return ERROR;

    /* Choose values */
    active = OESRandomInt(0,numtexunits-1) + GL_TEXTURE0;
    for (i=0; i<numtexunits; ++i)
    {
        glActiveTexture(GL_TEXTURE0+i);

        switch (OESRandomInt(0,5))
        {
            case 0:
                envmode[i] = GL_REPLACE;
                break;
            case 1:
                envmode[i] = GL_MODULATE;
                break;
            case 2:
                envmode[i] = GL_DECAL;
                break;
            case 3:
                envmode[i] = GL_BLEND;
                break;
            case 4:
                envmode[i] = GL_ADD;
                break;
            case 5:
                envmode[i] = GL_COMBINE;
                break;
        }   
        
        envcolor[i][0] = OESRandom(ITOF(0),ITOF(1));
        envcolor[i][1] = OESRandom(ITOF(0),ITOF(1));
        envcolor[i][2] = OESRandom(ITOF(0),ITOF(1));
        envcolor[i][3] = OESRandom(ITOF(0),ITOF(1));

        switch (OESRandomInt(0,7))
        {
            case 0:
                combinergb[i] = GL_REPLACE;
                break;
            case 1:
                combinergb[i] = GL_MODULATE;
                break;
            case 2:
                combinergb[i] = GL_ADD;
                break;
            case 3:
                combinergb[i] = GL_ADD_SIGNED;
                break;
            case 4:
                combinergb[i] = GL_INTERPOLATE;
                break;
            case 5:
                combinergb[i] = GL_SUBTRACT;
                break;
            case 6:
                combinergb[i] = GL_DOT3_RGB;
                break;
            case 7:
                combinergb[i] = GL_DOT3_RGBA;
                break;
        }

        switch (OESRandomInt(0,5))
        {
            case 0:
                combinealpha[i] = GL_REPLACE;
                break;
            case 1:
                combinealpha[i] = GL_MODULATE;
                break;
            case 2:
                combinealpha[i] = GL_ADD;
                break;
            case 3:
                combinealpha[i] = GL_ADD_SIGNED;
                break;
            case 4:
                combinealpha[i] = GL_INTERPOLATE;
                break;
            case 5:
                combinealpha[i] = GL_SUBTRACT;
                break;
        }

        for (j=0; j<3; ++j)
        {
            for (k=0; k<2; ++k)
            {
                GLint   n = OESRandomInt(0,3);
                switch (n)
                {
                    case 0:
                        src[i][j][k] = GL_CONSTANT;
                        break;
                    case 1:
                        src[i][j][k] = GL_PRIMARY_COLOR;
                        break;
                    case 2:
                        src[i][j][k] = GL_PREVIOUS;
                        break;
                    case 3:
                        src[i][j][k] = GL_TEXTURE;
                        break;
                }
            }

            switch (OESRandomInt(0,3))
            {
                case 0:
                    operand[i][j][0] = GL_SRC_COLOR;
                    break;
                case 1:
                    operand[i][j][0] = GL_ONE_MINUS_SRC_COLOR;
                    break;
                case 2:
                    operand[i][j][0] = GL_SRC_ALPHA;
                    break;
                case 3:
                    operand[i][j][0] = GL_ONE_MINUS_SRC_ALPHA;
                    break;
            }

            switch (OESRandomInt(0,1))
            {
                case 0:
                    operand[i][j][1] = GL_SRC_ALPHA;
                    break;
                case 1:
                    operand[i][j][1] = GL_ONE_MINUS_SRC_ALPHA;
                    break;
            }
        }

        rgbscale[i]   = ITOF(OESRandomInt(1,2));
        alphascale[i] = ITOF(OESRandomInt(1,2));
    }

    /* Set values */
    for (i=0; i<numtexunits; ++i)
    {
        glActiveTexture(GL_TEXTURE0+i);

        glTexEnvt(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,   (GLftype)envmode[i]);
        glTexEnvtv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, envcolor[i]);
        glTexEnvt(GL_TEXTURE_ENV, GL_COMBINE_RGB,        (GLftype)combinergb[i]);
        glTexEnvt(GL_TEXTURE_ENV, GL_COMBINE_ALPHA,      (GLftype)combinealpha[i]);
        glTexEnvt(GL_TEXTURE_ENV, GL_SRC0_RGB,           (GLftype)src[i][0][0]);
        glTexEnvt(GL_TEXTURE_ENV, GL_SRC0_ALPHA,         (GLftype)src[i][0][1]);
        glTexEnvt(GL_TEXTURE_ENV, GL_SRC1_RGB,           (GLftype)src[i][1][0]);
        glTexEnvt(GL_TEXTURE_ENV, GL_SRC1_ALPHA,         (GLftype)src[i][1][1]);
        glTexEnvt(GL_TEXTURE_ENV, GL_SRC2_RGB,           (GLftype)src[i][2][0]);
        glTexEnvt(GL_TEXTURE_ENV, GL_SRC2_ALPHA,         (GLftype)src[i][2][1]);
        glTexEnvt(GL_TEXTURE_ENV, GL_OPERAND0_RGB,       (GLftype)operand[i][0][0]);
        glTexEnvt(GL_TEXTURE_ENV, GL_OPERAND0_ALPHA,     (GLftype)operand[i][0][1]);
        glTexEnvt(GL_TEXTURE_ENV, GL_OPERAND1_RGB,       (GLftype)operand[i][1][0]);
        glTexEnvt(GL_TEXTURE_ENV, GL_OPERAND1_ALPHA,     (GLftype)operand[i][1][1]);
        glTexEnvt(GL_TEXTURE_ENV, GL_OPERAND2_RGB,       (GLftype)operand[i][2][0]);
        glTexEnvt(GL_TEXTURE_ENV, GL_OPERAND2_ALPHA,     (GLftype)operand[i][2][1]);
        glTexEnvt(GL_TEXTURE_ENV, GL_RGB_SCALE,          rgbscale[i]);
        glTexEnvt(GL_TEXTURE_ENV, GL_ALPHA_SCALE,        alphascale[i]);
    }
    glActiveTexture(active);

    /* Detect errors */
    if (CheckErrors("TestTextureEnvironmentState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_GetInteger, 0, GL_ACTIVE_TEXTURE, 1, &active, Tolerance_Exact, "GL_ACTIVE_TEXTURE", 0) == ERROR)
        return ERROR;
    for (i=0; i<numtexunits; ++i)
    {
        glActiveTexture(GL_TEXTURE0+i);

        if (VerifyValues(Lookup_GetTexEnvt, GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, 4, envcolor[i], Tolerance_Color, "GL_TEXTURE_ENV_COLOR (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, 1, &envmode[i], Tolerance_Exact, "GL_TEXTURE_ENV_MODE (texture unit %d)", i) == ERROR)
            return ERROR;

        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_COMBINE_RGB, 1, &combinergb[i], Tolerance_Exact, "GL_COMBINE_RGB (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_COMBINE_ALPHA, 1, &combinealpha[i], Tolerance_Exact, "GL_COMBINE_ALPHA (texture unit %d)", i) == ERROR)
            return ERROR;

        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_SRC0_RGB, 1, &src[i][0][0], Tolerance_Exact, "GL_SRC0_RGB (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_SRC0_ALPHA, 1, &src[i][0][1], Tolerance_Exact, "GL_SRC0_ALPHA (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_SRC1_RGB, 1, &src[i][1][0], Tolerance_Exact, "GL_SRC1_RGB (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_SRC1_ALPHA, 1, &src[i][1][1], Tolerance_Exact, "GL_SRC1_ALPHA (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_SRC2_RGB, 1, &src[i][2][0], Tolerance_Exact, "GL_SRC2_RGB (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_SRC2_ALPHA, 1, &src[i][2][1], Tolerance_Exact, "GL_SRC2_ALPHA (texture unit %d)", i) == ERROR)
            return ERROR;

        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_OPERAND0_RGB, 1, &operand[i][0][0], Tolerance_Exact, "GL_OPERAND0_RGB (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_OPERAND0_ALPHA, 1, &operand[i][0][1], Tolerance_Exact, "GL_OPERAND0_ALPHA (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_OPERAND1_RGB, 1, &operand[i][1][0], Tolerance_Exact, "GL_OPERAND1_RGB (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_OPERAND1_ALPHA, 1, &operand[i][1][1], Tolerance_Exact, "GL_OPERAND1_ALPHA (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_OPERAND2_RGB, 1, &operand[i][2][0], Tolerance_Exact, "GL_OPERAND2_RGB (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvi, GL_TEXTURE_ENV, GL_OPERAND2_ALPHA, 1, &operand[i][2][1], Tolerance_Exact, "GL_OPERAND2_ALPHA (texture unit %d)", i) == ERROR)
            return ERROR;

        if (VerifyValues(Lookup_GetTexEnvt, GL_TEXTURE_ENV, GL_RGB_SCALE, 1, &rgbscale[i], Tolerance_Exact, "GL_RGB_SCALE (texture unit %d)", i) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetTexEnvt, GL_TEXTURE_ENV, GL_ALPHA_SCALE, 1, &alphascale[i], Tolerance_Exact, "GL_ALPHA_SCALE (texture unit %d)", i) == ERROR)
            return ERROR;
    }

    return NO_ERROR;
}

/******************************************************************************/

static long TestFramebufferControlState(void)
{
    GLint       depthbits;
    GLint       stencilbits;
    GLboolean   colormask[4];
    GLboolean   depthmask;
    GLint       stencilmask;
    GLftype     colorclear[4];
    GLftype     depthcleart = 0;
    GLint       depthcleari = 0;
    GLint       stencilclear;

    /* Get system limits */
    glGetIntegerv(GL_DEPTH_BITS, &depthbits);
    glGetIntegerv(GL_STENCIL_BITS, &stencilbits);

    /* Choose values */
    colormask[0]  = (GLboolean)OESRandomInt(0,1);
    colormask[1]  = (GLboolean)OESRandomInt(0,1);
    colormask[2]  = (GLboolean)OESRandomInt(0,1);
    colormask[3]  = (GLboolean)OESRandomInt(0,1);
    colorclear[0] = OESRandom(ITOF(0),ITOF(1));
    colorclear[1] = OESRandom(ITOF(0),ITOF(1));
    colorclear[2] = OESRandom(ITOF(0),ITOF(1));
    colorclear[3] = OESRandom(ITOF(0),ITOF(1));
    if (buffer.zBits > 0)
    {
        depthmask     = (GLboolean)OESRandomInt(0,1);
        depthcleart   = OESRandom(ITOF(0),ITOF(1));
        depthcleari   = NormalizedFtypeToInt(depthcleart);
    }
    if (buffer.stencilBits > 0)
    {
        stencilmask   = OESRandomInt(0,(1<<stencilbits)-1);
        stencilclear  = OESRandomInt(0,(1<<stencilbits)-1);
    }

    /* Set values */
    glColorMask(colormask[0], colormask[1], colormask[2], colormask[3]);
    glClearColort(colorclear[0], colorclear[1], colorclear[2], colorclear[3]);
    if (buffer.zBits > 0)
    {
        glDepthMask(depthmask);
        glClearDeptht(depthcleart);
    }
    if (buffer.stencilBits > 0)
    {
        glStencilMask(stencilmask);
        glClearStencil(stencilclear);
    }

    /* Detect errors */
    if (CheckErrors("TestFramebufferControlState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_GetBoolean, 0, GL_COLOR_WRITEMASK, 4, colormask, Tolerance_Exact, "GL_COLOR_WRITEMASK", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetFtype, 0, GL_COLOR_CLEAR_VALUE, 4, colorclear, Tolerance_Color, "GL_COLOR_CLEAR_VALUE", 0) == ERROR)
        return ERROR;
    if (buffer.zBits > 0)
    {
        if (VerifyValues(Lookup_GetBoolean, 0, GL_DEPTH_WRITEMASK, 1, &depthmask, Tolerance_Exact, "GL_DEPTH_WRITEMASK", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_DEPTH_CLEAR_VALUE, 1, &depthcleari, Tolerance_Depth, "GL_DEPTH_CLEAR_VALUE", 0) == ERROR)
            return ERROR;
    }
    if (buffer.stencilBits > 0)
    {
        if (VerifyValues(Lookup_GetInteger, 0, GL_STENCIL_WRITEMASK, 1, &stencilmask, Tolerance_Exact, "GL_STENCIL_WRITEMASK", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_STENCIL_CLEAR_VALUE, 1, &stencilclear, Tolerance_Exact, "GL_STENCIL_CLEAR_VALUE", 0) == ERROR)
            return ERROR;
    }

    return NO_ERROR;
}

/******************************************************************************/

static long TestPixelOperationsState(void)
{
    const GLenum testfunc[8] = 
    { 
        GL_NEVER, 
        GL_ALWAYS, 
        GL_LESS, 
        GL_LEQUAL, 
        GL_EQUAL, 
        GL_GEQUAL, 
        GL_GREATER, 
        GL_NOTEQUAL
    };
    const GLenum testop[6] = 
    { 
        GL_KEEP, 
        GL_ZERO, 
        GL_REPLACE, 
        GL_INCR, 
        GL_DECR, 
        GL_INVERT 
    };
    const GLenum srcblendfactor[] = 
    { 
        GL_ZERO, 
        GL_ONE, 
        GL_DST_COLOR, 
        GL_ONE_MINUS_DST_COLOR, 
        GL_SRC_ALPHA, 
        GL_ONE_MINUS_SRC_ALPHA, 
        GL_DST_ALPHA, 
        GL_ONE_MINUS_DST_ALPHA, 
        GL_SRC_ALPHA_SATURATE 
    };
    const GLenum dstblendfactor[] = 
    { 
        GL_ZERO, 
        GL_ONE, 
        GL_SRC_COLOR, 
        GL_ONE_MINUS_SRC_COLOR, 
        GL_SRC_ALPHA, 
        GL_ONE_MINUS_SRC_ALPHA, 
        GL_DST_ALPHA, 
        GL_ONE_MINUS_DST_ALPHA
    };
    const GLenum logic[16] = 
    { 
        GL_CLEAR, 
        GL_AND, 
        GL_AND_REVERSE, 
        GL_COPY, 
        GL_AND_INVERTED, 
        GL_NOOP, 
        GL_XOR, 
        GL_OR, 
        GL_NOR, 
        GL_EQUIV,
        GL_INVERT, 
        GL_OR_REVERSE, 
        GL_COPY_INVERTED, 
        GL_OR_INVERTED, 
        GL_NAND, 
        GL_SET 
    };


    GLint       stencilbits;
    GLboolean   scissorenable;
    GLint       scissorbox[4];
    GLboolean   alphatestenable;
    GLenum      alphatestfunc;
    GLftype     alphatestreft;
    GLint       alphatestrefi;
    GLboolean   stencilenable;
    GLenum      stencilfunc;
    GLint       stencilvaluemask;
    GLint       stencilref;
    GLenum      stencilfail;
    GLenum      stencildepthfail;
    GLenum      stencildepthpass;
    GLboolean   depthenable;
    GLenum      depthfunc;
    GLboolean   blendenable;
    GLenum      blendsrc;
    GLenum      blenddst;
    GLboolean   ditherenable;
    GLboolean   logicopenable;
    GLenum      logicopmode;

    /* Get system limits */
    glGetIntegerv(GL_STENCIL_BITS, &stencilbits);

    /* Choose values */
    scissorenable = OESRandomInt(0,1);
    scissorbox[0] = OESRandomInt(0,WINDSIZEX/2-1);
    scissorbox[1] = OESRandomInt(0,WINDSIZEY/2-1);
    scissorbox[2] = OESRandomInt(WINDSIZEX/2,WINDSIZEX-1);
    scissorbox[3] = OESRandomInt(WINDSIZEY/2,WINDSIZEY-1);

    alphatestenable = OESRandomInt(0,1);
    alphatestfunc   = testfunc[OESRandomInt(0,7)];
    alphatestreft   = OESRandom(ITOF(0),ITOF(1));
    alphatestrefi   = NormalizedFtypeToInt(alphatestreft);

    if (buffer.stencilBits > 0)
    {
        stencilenable    = OESRandomInt(0,1);
        stencilfunc      = testfunc[OESRandomInt(0,7)];
        stencilvaluemask = OESRandomInt(0,(1<<stencilbits)-1);
        stencilref       = OESRandomInt(0,(1<<stencilbits)-1);
        stencilfail      = testop[OESRandomInt(0,5)];
        stencildepthfail = testop[OESRandomInt(0,5)];
        stencildepthpass = testop[OESRandomInt(0,5)];;
    }

    if (buffer.zBits > 0)
    {
        depthenable = OESRandomInt(0,1);
        depthfunc   = testfunc[OESRandomInt(0,7)];
    }

    blendenable = OESRandomInt(0,1);
    blendsrc    = srcblendfactor[OESRandomInt(0,(sizeof(srcblendfactor)/sizeof(GLenum))-1)];
    blenddst    = dstblendfactor[OESRandomInt(0,(sizeof(dstblendfactor)/sizeof(GLenum))-1)];

    ditherenable = OESRandomInt(0,1);

    logicopenable = OESRandomInt(0,1);
    logicopmode   = logic[OESRandomInt(0,15)];

    /* Set values */
    if (scissorenable)
        glEnable(GL_SCISSOR_TEST);
    else
        glDisable(GL_SCISSOR_TEST);
    glScissor(scissorbox[0], scissorbox[1], scissorbox[2], scissorbox[3]);

    if (alphatestenable)
        glEnable(GL_ALPHA_TEST);
    else
        glDisable(GL_ALPHA_TEST);
    glAlphaFunct(alphatestfunc, alphatestreft);

    if (buffer.stencilBits > 0)
    {
        if (stencilenable)
            glEnable(GL_STENCIL_TEST);
        else
            glDisable(GL_STENCIL_TEST);
        glStencilFunc(stencilfunc, stencilref, stencilvaluemask);
        glStencilOp(stencilfail, stencildepthfail, stencildepthpass);
    }

    if (buffer.zBits > 0)
    {
        if (depthenable)
            glEnable(GL_DEPTH_TEST);
        else
            glDisable(GL_DEPTH_TEST);
        glDepthFunc(depthfunc);
    }

    if (blendenable)
        glEnable(GL_BLEND);
    else
        glDisable(GL_BLEND);
    glBlendFunc(blendsrc, blenddst);
 
    if (ditherenable)
        glEnable(GL_DITHER);
    else
        glDisable(GL_DITHER);

    if (logicopenable)
        glEnable(GL_COLOR_LOGIC_OP);
    else
        glDisable(GL_COLOR_LOGIC_OP);
    glLogicOp(logicopmode);
    
    /* Detect errors */
    if (CheckErrors("TestPixelOperationsState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_IsEnabled, 0, GL_SCISSOR_TEST, 1, &scissorenable, Tolerance_Exact, "GL_SCISSOR_TEST", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_SCISSOR_BOX, 4, scissorbox, Tolerance_Exact, "GL_SCISSOR_BOX", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_IsEnabled, 0, GL_ALPHA_TEST, 1, &alphatestenable, Tolerance_Exact, "GL_ALPHA_TEST", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_ALPHA_TEST_FUNC, 1, &alphatestfunc, Tolerance_Exact, "GL_ALPHA_TEST_FUNC", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_ALPHA_TEST_REF, 1, &alphatestrefi, Tolerance_Alpha, "GL_ALPHA_TEST_REF", 0) == ERROR)
        return ERROR;

    if (buffer.stencilBits > 0)
    {
        if (VerifyValues(Lookup_IsEnabled, 0, GL_STENCIL_TEST, 1, &stencilenable, Tolerance_Exact, "GL_STENCIL_TEST", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_STENCIL_FUNC, 1, &stencilfunc, Tolerance_Exact, "GL_STENCIL_FUNC", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_STENCIL_VALUE_MASK, 1, &stencilvaluemask, Tolerance_Exact, "GL_STENCIL_VALUE_MASK", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_STENCIL_REF, 1, &stencilref, Tolerance_Exact, "GL_STENCIL_REF", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_STENCIL_FAIL, 1, &stencilfail, Tolerance_Exact, "GL_STENCIL_FAIL", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_STENCIL_PASS_DEPTH_FAIL, 1, &stencildepthfail, Tolerance_Exact, "GL_STENCIL_PASS_DEPTH_FAIL", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_STENCIL_PASS_DEPTH_PASS, 1, &stencildepthpass, Tolerance_Exact, "GL_STENCIL_PASS_DEPTH_PASS", 0) == ERROR)
            return ERROR;
    }

    if (buffer.zBits > 0)
    {
        if (VerifyValues(Lookup_IsEnabled, 0, GL_DEPTH_TEST, 1, &depthenable, Tolerance_Exact, "GL_DEPTH_TEST", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_DEPTH_FUNC, 1, &depthfunc, Tolerance_Exact, "GL_DEPTH_FUNC", 0) == ERROR)
            return ERROR;
    }

    if (VerifyValues(Lookup_IsEnabled, 0, GL_BLEND, 1, &blendenable, Tolerance_Exact, "GL_BLEND", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_BLEND_SRC, 1, &blendsrc, Tolerance_Exact, "GL_BLEND_SRC", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_BLEND_DST, 1, &blenddst, Tolerance_Exact, "GL_BLEND_DST", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_IsEnabled, 0, GL_DITHER, 1, &ditherenable, Tolerance_Exact, "GL_DITHER", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_IsEnabled, 0, GL_COLOR_LOGIC_OP, 1, &logicopenable, Tolerance_Exact, "GL_COLOR_LOGIC_OP", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_LOGIC_OP_MODE, 1, &logicopmode, Tolerance_Exact, "GL_LOGIC_OP_MODE", 0) == ERROR)
        return ERROR;

    return NO_ERROR;
}

/******************************************************************************/

static long TestPixelsState(void)
{
    GLint   unpackalign;
    GLint   packalign;

    /* Choose values */
    unpackalign = 1 << OESRandomInt(0,3);
    packalign  = 1 << OESRandomInt(0,3);

    /* Set values */
    glPixelStorei(GL_UNPACK_ALIGNMENT, unpackalign);
    glPixelStorei(GL_PACK_ALIGNMENT, packalign);

    /* Detect errors */
    if (CheckErrors("TestPixelsState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_GetInteger, 0, GL_UNPACK_ALIGNMENT, 1, &unpackalign, Tolerance_Exact, "GL_UNPACK_ALIGNMENT", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_PACK_ALIGNMENT, 1, &packalign, Tolerance_Exact, "GL_PACK_ALIGNMENT", 0) == ERROR)
        return ERROR;

    return NO_ERROR;
}

/******************************************************************************/

static long TestHintState(void)
{
    GLenum  hints[3] = { GL_FASTEST, GL_NICEST, GL_DONT_CARE };

    GLenum  perspectivehint;
    GLenum  pointsmoothhint;
    GLenum  linesmoothhint;
    GLenum  foghint;
    GLenum  genmipmaphint;

    /* Choose values */
    perspectivehint = hints[OESRandomInt(0,2)];
    pointsmoothhint = hints[OESRandomInt(0,2)];
    linesmoothhint  = hints[OESRandomInt(0,2)];
    foghint         = hints[OESRandomInt(0,2)];
    genmipmaphint   = hints[OESRandomInt(0,2)];

    /* Set values */
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, perspectivehint);
    glHint(GL_POINT_SMOOTH_HINT, pointsmoothhint);
    glHint(GL_LINE_SMOOTH_HINT, linesmoothhint);
    glHint(GL_FOG_HINT, foghint);
    glHint(GL_GENERATE_MIPMAP_HINT, genmipmaphint);

    /* Detect errors */
    if (CheckErrors("TestHintState") == ERROR)
        return ERROR;

    /* Test values */
    if (VerifyValues(Lookup_GetInteger, 0, GL_PERSPECTIVE_CORRECTION_HINT, 1, &perspectivehint, Tolerance_Exact, "GL_PERSPECTIVE_CORRECTION_HINT", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_POINT_SMOOTH_HINT, 1, &pointsmoothhint, Tolerance_Exact, "GL_POINT_SMOOTH_HINT", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_LINE_SMOOTH_HINT, 1, &linesmoothhint, Tolerance_Exact, "GL_LINE_SMOOTH_HINT", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_FOG_HINT, 1, &foghint, Tolerance_Exact, "GL_FOG_HINT", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_GENERATE_MIPMAP_HINT, 1, &genmipmaphint, Tolerance_Exact, "GL_GENERATE_MIPMAP_HINT", 0) == ERROR)
        return ERROR;

    return NO_ERROR;
}

/******************************************************************************/

static long TestExtensionsState(void)
{
    GLubyte     dummy[4096];

    GLint       numtexunits;
    GLenum      colorreadtype;
    GLenum      colorreadformat;
    GLint       maxpalettematrices;
    GLint       maxvertexunits;

    GLboolean   matrixpalettesupported;

    GLint       vertexstride;
    GLboolean   getzerostrides;

    GLboolean   matrixpaletteenable;

    GLint       matrixindexbind;
    GLboolean   matrixindexenable;
    GLint       matrixindextype;
    GLint       matrixindexsize;
    GLint       matrixindexstride;
    void*       matrixindexpointer;

    GLint       weightbind;
    GLboolean   weightenable;
    GLint       weighttype;
    GLint       weightsize;
    GLint       weightstride;
    void*       weightpointer;

    GLint       pointsizebind;
    GLboolean   pointsizeenable;
    GLint       pointsizetype;
    GLint       pointsizestride;
    void*       pointsizepointer;

    GLboolean   pointspriteenable;
    GLint       coordreplaceenable[16];

    GLboolean   enable;
    GLint       type;
    GLint       size;
    GLint       bytes;
    GLint       start;
    GLint       stride;
    GLuint       arraybind;
    GLint       i;

    /* Determine if matrix palette is supported */
    matrixpalettesupported = tkCheckExtension("GL_OES_matrix_palette");

    /* Retrieve the system settings */
    if (LookupValues(Lookup_GetInteger, 0, GL_MAX_TEXTURE_UNITS, &numtexunits, "GL_MAX_TEXTURE_UNITS", 0) == ERROR)
        return ERROR;
    if (LookupValues(Lookup_GetInteger, 0, GL_IMPLEMENTATION_COLOR_READ_TYPE_OES, &colorreadtype, "GL_IMPLEMENTATION_COLOR_READ_TYPE_OES", 0) == ERROR)
        return ERROR;
    if (LookupValues(Lookup_GetInteger, 0, GL_IMPLEMENTATION_COLOR_READ_FORMAT_OES, &colorreadformat, "GL_IMPLEMENTATION_COLOR_READ_FORMAT_OES", 0) == ERROR)
        return ERROR;
    if (matrixpalettesupported)
    {
        if (LookupValues(Lookup_GetInteger, 0, GL_MAX_PALETTE_MATRICES_OES, &maxpalettematrices, "GL_MAX_PALETTE_MATRICES_OES", 0) == ERROR)
            return ERROR;
        if (LookupValues(Lookup_GetInteger, 0, GL_MAX_VERTEX_UNITS_OES, &maxvertexunits, "GL_MAX_VERTEX_UNITS_OES", 0) == ERROR)
            return ERROR;
    }

    /* Determine how zero strides are read back */
    glVertexPointer(2, GL_FTYPE, 0, dummy);
    if (LookupValues(Lookup_GetInteger, 0, GL_VERTEX_ARRAY_STRIDE, &vertexstride, "GL_VERTEX_ARRAY_STRIDE", 0) == ERROR)
        return ERROR;
    getzerostrides = (vertexstride == 0) ? GL_TRUE : GL_FALSE;

    /* Choose/set values */

    if (matrixpalettesupported)
    {
        matrixpaletteenable = OESRandomInt(0,1);
        if (matrixpaletteenable)
            glEnable(GL_MATRIX_PALETTE_OES);
        else
            glDisable(GL_MATRIX_PALETTE_OES);
    }

    /* Matrix index array */
    if (matrixpalettesupported)
    {
        enable = OESRandomInt(0,1);
        type  = GL_UNSIGNED_BYTE;
        bytes = sizeof(GLbyte);
        size   = OESRandomInt(1,maxvertexunits);
        start  = OESRandomInt(0,255) * bytes;
        stride = OESRandomInt(0,8) * bytes;
        if (stride < size * bytes) stride = 0;

        arraybind = OESRandomInt(0,3);
        glBindBuffer(GL_ARRAY_BUFFER, arraybind);
        if (enable)
            glEnableClientState(GL_MATRIX_INDEX_ARRAY_OES);
        else
            glDisableClientState(GL_MATRIX_INDEX_ARRAY_OES);
        glMatrixIndexPointerOES(size, type, stride, arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start]);

        matrixindexbind    = arraybind;
        matrixindexenable  = enable;
        matrixindextype    = type;
        matrixindexsize    = size;
        matrixindexstride  = (stride || getzerostrides) ? stride : (size * bytes);
        matrixindexpointer = arraybind ? BUFFER_OFFSET(start): (void*)&dummy[start];
    }

    /* Weight array */
    if (matrixpalettesupported)
    {
        enable = OESRandomInt(0,1);
        type   = GL_FTYPE;
        bytes  = sizeof(GLftype);
        size   = OESRandomInt(1,maxvertexunits);
        start  = OESRandomInt(0,255) * bytes;
        stride = OESRandomInt(0,8) * bytes;
        if (stride < size * bytes) stride = 0;

        arraybind = OESRandomInt(0,3);
        glBindBuffer(GL_ARRAY_BUFFER, arraybind);
        if (enable)
            glEnableClientState(GL_WEIGHT_ARRAY_OES);
        else
            glDisableClientState(GL_WEIGHT_ARRAY_OES);
        glWeightPointerOES(size, type, stride, arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start]);

        weightbind    = arraybind;
        weightenable  = enable;
        weighttype    = type;
        weightsize    = size;
        weightstride  = (stride || getzerostrides) ? stride : (size * bytes);
        weightpointer = arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start];
    }

    /* Point size array */
    enable = OESRandomInt(0,1);
    type   = GL_FTYPE;
    bytes  = sizeof(GLftype);
    size   = 1;
    start  = OESRandomInt(0,255) * bytes;
    stride = OESRandomInt(0,8) * bytes;
    if (stride < size * bytes) stride = 0;

    arraybind = OESRandomInt(0,3);
    glBindBuffer(GL_ARRAY_BUFFER, arraybind);
    if (enable)
        glEnableClientState(GL_POINT_SIZE_ARRAY_OES);
    else
        glDisableClientState(GL_POINT_SIZE_ARRAY_OES);
    glPointSizePointerOES(type, stride, arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start]);

    pointsizebind    = arraybind;
    pointsizeenable  = enable;
    pointsizetype    = type;
    pointsizestride  = (stride || getzerostrides) ? stride : (1 * bytes);
    pointsizepointer = arraybind ? BUFFER_OFFSET(start) : (void*)&dummy[start];

    pointspriteenable = OESRandomInt(0,1);
    if (pointspriteenable)
        glEnable(GL_POINT_SPRITE_OES);
    else
        glDisable(GL_POINT_SPRITE_OES);
    for (i=0; i<numtexunits; ++i)
    {
        coordreplaceenable[i] = OESRandomInt(0,1);
        glActiveTexture(GL_TEXTURE0+i);
        glTexEnvt(GL_POINT_SPRITE_OES, GL_COORD_REPLACE_OES, (GLftype)coordreplaceenable[i]);
    }

    /* Detect errors */
    if (CheckErrors("TestExtensionState") == ERROR)
        return ERROR;

    /* Test values */

    if (matrixpalettesupported)
    {
        if (VerifyValues(Lookup_IsEnabled, 0, GL_MATRIX_PALETTE_OES, 1, &matrixpaletteenable, Tolerance_Exact, "GL_MATRIX_PALETTE_OES", 0) == ERROR)
            return ERROR;
    }

    /* Matrix index array */
    if (matrixpalettesupported)
    {
        if (VerifyValues(Lookup_GetInteger, 0, GL_MATRIX_INDEX_ARRAY_BUFFER_BINDING_OES, 1, &matrixindexbind, Tolerance_Exact, "GL_MATRIX_INDEX_ARRAY_BUFFER_BINDING_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_IsEnabled, 0, GL_MATRIX_INDEX_ARRAY_OES, 1, &matrixindexenable, Tolerance_Exact, "GL_MATRIX_INDEX_ARRAY_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_MATRIX_INDEX_ARRAY_SIZE_OES, 1, &matrixindexsize, Tolerance_Exact, "GL_MATRIX_INDEX_ARRAY_SIZE_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_MATRIX_INDEX_ARRAY_STRIDE_OES, 1, &matrixindexstride, Tolerance_Exact, "GL_MATRIX_INDEX_ARRAY_STRIDE_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_MATRIX_INDEX_ARRAY_TYPE_OES, 1, &matrixindextype, Tolerance_Exact, "GL_MATRIX_INDEX_ARRAY_TYPE_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetPointer, 0, GL_MATRIX_INDEX_ARRAY_POINTER_OES, 1, &matrixindexpointer, Tolerance_Exact, "GL_MATRIX_INDEX_ARRAY_POINTER_OES", 0) == ERROR)
            return ERROR;
    }

    /* Weight array */
    if (matrixpalettesupported)
    {
        if (VerifyValues(Lookup_GetInteger, 0, GL_WEIGHT_ARRAY_BUFFER_BINDING_OES, 1, &weightbind, Tolerance_Exact, "GL_WEIGHT_ARRAY_BUFFER_BINDING_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_IsEnabled, 0, GL_WEIGHT_ARRAY_OES, 1, &weightenable, Tolerance_Exact, "GL_WEIGHT_ARRAY_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_WEIGHT_ARRAY_SIZE_OES, 1, &weightsize, Tolerance_Exact, "GL_WEIGHT_ARRAY_SIZE_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_WEIGHT_ARRAY_STRIDE_OES, 1, &weightstride, Tolerance_Exact, "GL_WEIGHT_ARRAY_STRIDE_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetInteger, 0, GL_WEIGHT_ARRAY_TYPE_OES, 1, &weighttype, Tolerance_Exact, "GL_WEIGHT_ARRAY_TYPE_OES", 0) == ERROR)
            return ERROR;
        if (VerifyValues(Lookup_GetPointer, 0, GL_WEIGHT_ARRAY_POINTER_OES, 1, &weightpointer, Tolerance_Exact, "GL_WEIGHT_ARRAY_POINTER_OES", 0) == ERROR)
            return ERROR;
    }

    /* Point size array */
    if (VerifyValues(Lookup_GetInteger, 0, GL_POINT_SIZE_ARRAY_BUFFER_BINDING_OES, 1, &pointsizebind, Tolerance_Exact, "GL_POINT_SIZE_ARRAY_BUFFER_BINDING_OES", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_IsEnabled, 0, GL_POINT_SIZE_ARRAY_OES, 1, &pointsizeenable, Tolerance_Exact, "GL_POINT_SIZE_ARRAY_OES", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_POINT_SIZE_ARRAY_STRIDE_OES, 1, &pointsizestride, Tolerance_Exact, "GL_POINT_SIZE_ARRAY_STRIDE_OES", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetInteger, 0, GL_POINT_SIZE_ARRAY_TYPE_OES, 1, &pointsizetype, Tolerance_Exact, "GL_POINT_SIZE_ARRAY_TYPE_OES", 0) == ERROR)
        return ERROR;
    if (VerifyValues(Lookup_GetPointer, 0, GL_POINT_SIZE_ARRAY_POINTER_OES, 1, &pointsizepointer, Tolerance_Exact, "GL_POINT_SIZE_ARRAY_POINTER_OES", 0) == ERROR)
        return ERROR;

    if (VerifyValues(Lookup_IsEnabled, 0, GL_POINT_SPRITE_OES, 1, &pointspriteenable, Tolerance_Exact, "GL_POINT_SPRITE_OES", 0) == ERROR)
        return ERROR;
    for (i=0; i<numtexunits; ++i)
    {
        glActiveTexture(GL_TEXTURE0+i);
        if (VerifyValues(Lookup_GetTexEnvi, GL_POINT_SPRITE_OES, GL_COORD_REPLACE_OES, 1, &coordreplaceenable[i], Tolerance_Exact, "GL_COORD_REPLACE_OES (unit %d)", i) == ERROR)
            return ERROR;
    }

    if (matrixpalettesupported)
    {
        glDisable(GL_MATRIX_PALETTE_OES);
    }

    /* Clean up any buffers used */
    for (arraybind=1; arraybind<=3; ++arraybind)
    {
        glDeleteBuffers(1, &arraybind);
    }

    return NO_ERROR;
}

/******************************************************************************/

long GetsExec(void)
{
    colortolerancet[0] = buffer.colorBits[0] ? CalcEpsilon(buffer.colorBits[0]) : ITOF(1);
    colortolerancet[1] = buffer.colorBits[1] ? CalcEpsilon(buffer.colorBits[1]) : ITOF(1);
    colortolerancet[2] = buffer.colorBits[2] ? CalcEpsilon(buffer.colorBits[2]) : ITOF(1);
    colortolerancet[3] = buffer.colorBits[3] ? CalcEpsilon(buffer.colorBits[3]) : ITOF(1);
    alphatolerancet = colortolerancet[0];
    if (colortolerancet[1] > alphatolerancet)
        alphatolerancet = colortolerancet[1];
    if (colortolerancet[2] > alphatolerancet)
        alphatolerancet = colortolerancet[2];
    depthtolerancet = buffer.zBits ? CalcEpsilon(buffer.zBits) : ITOF(1);

#if GL_FTYPE == GL_FIXED
    colortolerancei[0] = (colortolerancet[0] < ITOF(1)) ? (GLint)MUL(0x7FFFFFFF, colortolerancet[0]) : 0x7FFFFFFF;
    colortolerancei[1] = (colortolerancet[1] < ITOF(1)) ? (GLint)MUL(0x7FFFFFFF, colortolerancet[1]) : 0x7FFFFFFF;
    colortolerancei[2] = (colortolerancet[2] < ITOF(1)) ? (GLint)MUL(0x7FFFFFFF, colortolerancet[2]) : 0x7FFFFFFF;
    colortolerancei[3] = (colortolerancet[3] < ITOF(1)) ? (GLint)MUL(0x7FFFFFFF, colortolerancet[3]) : 0x7FFFFFFF;
    alphatolerancei    = (alphatolerancet < ITOF(1))    ? (GLint)MUL(0x7FFFFFFF, alphatolerancet) : 0x7FFFFFFF;
    depthtolerancei    = (depthtolerancet < ITOF(1))    ? (GLint)MUL(0x7FFFFFFF, depthtolerancet) : 0x7FFFFFFF;
#else
    colortolerancei[0] = (colortolerancet[0] < ITOF(1)) ? FTOI(MUL(ITOF(0x7FFFFFFF), colortolerancet[0])) : 0x7FFFFFFF;
    colortolerancei[1] = (colortolerancet[1] < ITOF(1)) ? FTOI(MUL(ITOF(0x7FFFFFFF), colortolerancet[1])) : 0x7FFFFFFF;
    colortolerancei[2] = (colortolerancet[2] < ITOF(1)) ? FTOI(MUL(ITOF(0x7FFFFFFF), colortolerancet[2])) : 0x7FFFFFFF;
    colortolerancei[3] = (colortolerancet[3] < ITOF(1)) ? FTOI(MUL(ITOF(0x7FFFFFFF), colortolerancet[3])) : 0x7FFFFFFF;
    alphatolerancei    = (alphatolerancet < ITOF(1))    ? FTOI(MUL(ITOF(0x7FFFFFFF), alphatolerancet)) : 0x7FFFFFFF;
    depthtolerancei    = (depthtolerancet < ITOF(1))    ? FTOI(MUL(ITOF(0x7FFFFFFF), depthtolerancet)) : 0x7FFFFFFF;
#endif

    if (TestImplementationState() == ERROR)
        return ERROR;

    if (TestCurrentValueState() == ERROR)
        return ERROR;

    if (TestBufferObjectState() == ERROR)
        return ERROR;

    if (TestVertexArrayState() == ERROR)
        return ERROR;

    if (TestTransformationState() == ERROR)
        return ERROR;

    if (TestColoringState() == ERROR)
        return ERROR;

    if (TestLightingState() == ERROR)
        return ERROR;

    if (TestRasterizationState() == ERROR)
        return ERROR;

    if (TestMultisamplingState() == ERROR)
        return ERROR;

    if (TestTextureObjectState() == ERROR)
        return ERROR;

    if (TestTextureEnvironmentState() == ERROR)
        return ERROR;

    if (TestFramebufferControlState() == ERROR)
        return ERROR;

    if (TestPixelOperationsState() == ERROR)
        return ERROR;

    if (TestPixelsState() == ERROR)
        return ERROR;

    if (TestHintState() == ERROR)
        return ERROR;

#if 0
    if (TestExtensionsState() == ERROR)
        return ERROR;
#endif
    return NO_ERROR;
}

#endif /* GL_OES_VERSION_1_1 */


/******************************************************************************/
