/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/l_ac.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** l_ac.c
** Light - Attenuation Constant Test.
**
** Description -
**    This test examines the effect which varying the
**    Attenuation_Constants has on the color produced by
**    lighting. 8 lights and all 3
**    Attenuation_Constants are tested one at a time. The
**    distance of the object from the origin in eye-coordinates
**    is exactly 1.0. At each iteration the reciprocal of an
**    Attenuation_Constant is uniformly incremented, and the red
**    component of a lit point is read and stored.
**    
**    Ambient, Emissive, and Specular summands of the
**    equation are set to black, and the diffuse parameters are
**    set to fully saturated red. Successive results should
**    differ by 1.0 / nR where nR is the number of red shades.
**    Theoretically the differance between successive results
**    should be constant. If successive observed differences
**    differ by more than 0.25 / nR, the implementation is judged
**    to differ by too much from the correct graph of the
**    attenuation function.
**    
** Error Explanation -
**    This test issues an error if the blue or green
**    channels are not completely "off", or if an expected
**    lighting result of (1, 0, 0) does not result in a
**    fully-saturated red object. An error is also issued if a
**    smaller Attenuation_Constant results in a less saturated
**    red, or if the differences between the results of two
**    successive attenuation settings differ by more than 0.25 /
**    nR.
**    
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilons, 0.25*color_epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Logicop, Stencil.
**        Not allowed = Alias, Dither, Fog, Shade.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _atnConstRec {
    ftype step, numColor, lastValue, lastDelta, delta;
    GLenum light;
    long kIndex;
    GLftype k[3], position[4];
} atnConstRec;


/******************************************************************************/

static void Set(void *data, ftype *ref)
{
    atnConstRec *ptr = (atnConstRec *)data;
    long i;

    for (i = 0; i < 3; i++) {
	ptr->k[i] = ITOF(0);
    }
    ptr->k[ptr->kIndex] = (*ref == ITOF(0)) ? DIV(ITOF(1), ptr->step) : DIV(ITOF(1), *ref);

    glLighttv(ptr->light, GL_CONSTANT_ATTENUATION, &ptr->k[0]);
    glLighttv(ptr->light, GL_LINEAR_ATTENUATION, &ptr->k[1]);
    glLighttv(ptr->light, GL_QUADRATIC_ATTENUATION, &ptr->k[2]);

    DrawPoint2f(ITOF(1)/2, ITOF(1)/2);

    *ref += ptr->step;
}

/******************************************************************************/

static void Init(void *data)
{
    atnConstRec *ptr = (atnConstRec *)data;
    GLftype buf[4];
    ftype step;
    GLenum i;

    ptr->lastValue = -ITOF(1);
    ptr->lastDelta = -ITOF(2);

    step = (machine.pathLevel == 0) ? ITOF(4) : ITOF(2);
    ptr->delta = DIV(ITOF(1), (LDEXP(ITOF(1), buffer.colorBits[0]) - ITOF(1)));
    ptr->step = DIV(ptr->delta, step);

    for (i = GL_LIGHT0; i <= GL_LIGHT7; i++) {
	glDisable(i);
    }
    glEnable(ptr->light);

    ptr->position[0] = ITOF(1)/2;
    ptr->position[1] = ITOF(1)/2;
    ptr->position[2] = ITOF(1);
    ptr->position[3] = ITOF(1);
    glLighttv(ptr->light, GL_POSITION, ptr->position);

    ptr->k[0] = ITOF(0);
    ptr->k[1] = ITOF(0);
    ptr->k[2] = ITOF(0);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(1);
    glNormal3t(buf[0], buf[1], buf[2]);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_AMBIENT, buf);

    buf[0] = ITOF(1);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_DIFFUSE, buf);

    buf[0] = ITOF(1);
    buf[1] = ITOF(1);
    buf[2] = ITOF(1);
    buf[3] = ITOF(1);
    glLighttv(ptr->light, GL_DIFFUSE, buf);
}

static long Test(void *data, ftype ref)
{
    atnConstRec *ptr = (atnConstRec *)data;
    char str[240];
    GLftype buf[3]; 
    ftype value, delta;
    long i;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 1; i < 3; i++) {
	if (buf[i] != ITOF(0)) {
	    ColorError_RGBZero(str, 0, buf);
	    ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
    }

    value = buf[0];
    if (ptr->lastDelta == -ITOF(2)) {
	ptr->lastDelta = -ITOF(1);
	ptr->lastValue = value;
    } else if (ptr->lastDelta == -ITOF(1)) {
	if (value < ptr->lastValue) {
	    ColorError_RGBMonotonic(str, 0, value, ptr->lastValue);
	    ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    if (ABS(delta-ptr->delta) > epsilon.color[0]/4) {
		ColorError_RGBStep(str, 0, delta, ptr->delta);
		ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    ptr->lastDelta = delta;
	}
    } else {
	if (value < ptr->lastValue) {
	    ColorError_RGBMonotonic(str, 0, value, ptr->lastValue);
	    ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    if (ABS(delta-ptr->delta) > epsilon.color[0]/4) {
		ColorError_RGBStep(str, 0, delta, ptr->delta);
		ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    if (ABS(delta-ptr->lastDelta) > epsilon.zero) {
		ColorError_RGBDelta(str, 0, delta, ptr->lastDelta);
		ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    ptr->lastDelta = delta;
	}
    }

    if (ref >= ITOF(1) && ABS(value-ITOF(1)) > epsilon.zero) {
	ColorError_RGBClamp(str, 0, value);
	ErrorReport(__FILE__, __LINE__, str);
	return ERROR;
    }

    return NO_ERROR;
}

long AtnConstExec(void)
{
    atnConstRec data;
    long max;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_LIGHTING);

    max = (machine.pathLevel == 0) ? GL_LIGHT7 : GL_LIGHT0;
    for (data.light = GL_LIGHT0; (long)data.light <= max; data.light++) {
	for (data.kIndex = 0; data.kIndex < 3; data.kIndex++) {
	    if (RampUtil(ITOF(0), ITOF(1), Init, Set, Test, &data) == ERROR) {
		return ERROR;
	    }
	}
    }
    return NO_ERROR;
}
