/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/l_dm.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** l_dm.c
** Light - Diffuse Material Test.
**
** Description -
**    This test analyses the effect that varying the
**    Diffuse_Material parameter has on the color produced by
**    lighting. To analyse only this one
**    parameter, parameter values are chosen so that the
**    Attenuation factor, the SpotLight factor, and the
**    Light_Position factor are all 1.0. The
**    Specular, Emissive, and Ambient summands in the lighting
**    equation are 0.0, and the Diffuse_Light is set to (1, 1, 1,
**    1).
**    
**    This forms a restriction of the lighting equation so that
**    the output color is equal to a multiple of the
**    Diffuse_Material setting (color ~= Diffuse_Material). This
**    function's output is analysed for monotonicity, endpoints,
**    and variation. One light is enabled at a time.
**    All individual lights are tested.
**    
**    The test repeatedly renders and reads the same point in the
**    lower left corner of the screen, uniformly incrementing
**    Diffuse_Material. The uniformity of the change
**    in observed color is tested, and the implementation fails
**    if consecutive changes are too different. (thisDelta -
**    lastDelta) > 1/4 shade. The observed results are also
**    checked for monotonicity. When Diffuse_Material reaches
**    1.0, the observed color is checked to insure that it is
**    rendered with a saturation of 1.0. All color channels are
**    tested one at a time, and the implementation fails if any
**    color channel, besides the one being tested, is not 0.0.
**    
** Error Explanation -
**    other_color > 0.0.
**    value < lastValue.
**    maxObserved < 1.0.
**    (thisDelta - lastDelta) > 1/4 shade.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilons, 0.25*color_epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Logicop, Stencil.
**        Not allowed = Alias, Dither, Fog, Shade.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _difMatRec {
    long color;
    ftype step, numColor, delta, lastValue, lastDelta;
    GLftype material_diffuse[4], position[4];
} difMatRec;


/****************************************************************************/

static void Init(void *data)
{
    difMatRec *ptr = (difMatRec *)data;
    GLftype buf[4];
    ftype step;
    GLenum i;

    ptr->lastValue = -ITOF(1);
    ptr->lastDelta = -ITOF(2);

    step = (machine.pathLevel == 0) ? ITOF(4) : ITOF(2);
    ptr->delta = DIV(ITOF(1), LDEXP(ITOF(1), buffer.colorBits[ptr->color]) - ITOF(1));
    ptr->step = DIV(ptr->delta, step);

    for (i = GL_LIGHT0; i <= GL_LIGHT7; i++) {
	glEnable(i);
    }
    glNormal3t(ITOF(0), ITOF(0), ITOF(1));

    ptr->material_diffuse[0] = ITOF(0);
    ptr->material_diffuse[1] = ITOF(0);
    ptr->material_diffuse[2] = ITOF(0);
    ptr->material_diffuse[3] = ITOF(1);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_AMBIENT, buf);

    buf[0] = ITOF(1)/8;
    buf[1] = ITOF(1)/8;
    buf[2] = ITOF(1)/8;
    buf[3] = ITOF(1);

    ptr->position[0] = ITOF(1)/2;
    ptr->position[1] = ITOF(1)/2;
    ptr->position[2] = ITOF(1);
    ptr->position[3] = ITOF(1);

    for (i = GL_LIGHT0; i <= GL_LIGHT7; i++) {
	glLighttv(i, GL_DIFFUSE, buf);
	glLighttv(i, GL_POSITION, ptr->position);
    }
}

static void Set(void *data, ftype *ref)
{
    difMatRec *ptr = (difMatRec *)data;

    ptr->material_diffuse[ptr->color] = *ref;
    glMaterialtv(GL_FRONT_AND_BACK, GL_DIFFUSE, ptr->material_diffuse);

    DrawPoint2f(ITOF(1)/2, ITOF(1)/2);

    *ref += ptr->step;
}

static long Test(void *data, ftype ref)
{
    difMatRec *ptr = (difMatRec *)data;
    char str[240];
    GLftype buf[3]; 
    ftype value, delta;
    long i;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 0; i < 3; i++) {
        if (buf[i] != ITOF(0) && i != ptr->color) {
            ColorError_RGBZero(str, ptr->color, buf);
            ErrorReport(__FILE__, __LINE__, str);
            return ERROR;
	}
    }

    value = buf[ptr->color];
    if (ptr->lastDelta == -ITOF(2)) {
	ptr->lastDelta = -ITOF(1);
	ptr->lastValue = value;
    } else if (ptr->lastDelta == -ITOF(1)) {
	if (value < ptr->lastValue) {
            ColorError_RGBMonotonic(str, ptr->color, value, ptr->lastValue);
            ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    if (ABS(delta-ptr->delta) > epsilon.color[ptr->color]/4) {
                ColorError_RGBStep(str, ptr->color, delta, ptr->delta);
                ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    ptr->lastDelta = delta;
	}
    } else {
	if (value < ptr->lastValue) {
            ColorError_RGBMonotonic(str, ptr->color, value, ptr->lastValue);
            ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    if (ABS(delta-ptr->delta) > epsilon.color[ptr->color]/4) {
                ColorError_RGBStep(str, ptr->color, delta, ptr->delta);
                ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    if (ABS(delta-ptr->lastDelta) > epsilon.zero) {
                ColorError_RGBDelta(str, ptr->color, delta, ptr->lastDelta);
                ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    ptr->lastDelta = delta;
	}
    }

    if (ref >= ITOF(1) && ABS(value-ITOF(1)) > epsilon.zero) {
        ColorError_RGBClamp(str, ptr->color, value);
        ErrorReport(__FILE__, __LINE__, str);
	return ERROR;
    }

    return NO_ERROR;
}

long DifMatExec(void)
{
    difMatRec data;
    long max;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_LIGHTING);

    max = (machine.pathLevel == 0) ? 3 : 1;
    for (data.color = 0; data.color < max; data.color++) {
	if (RampUtil(ITOF(0), ITOF(1), Init, Set, Test, &data) == ERROR) {
	    return ERROR;
	}
    }
    return NO_ERROR;
}
