/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/l_se.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** l_se.c
** Light - Specular Exponent Test.
**
** Description -
**    This test analyses the effect that varying the Normal and
**    the Specular Exponent together have on lighting. To analyse only
**    these parameters, the Specular_Material and Specular_Light
**    1, 1) and the Specular_Material is set to maxIndex. Other
**    lighting parameters are chosen so that the Diffuse,
**    Ambient, and Emissive summands of the lighting
**    are set to zero. Values are chosen so the Attenuation and
**    the Spotlight factor are both 1.0.
**    
**    The test repeatedly renders and reads the same point at the
**    lower left corner of the screen. The third component of the
**    Normal and the Specular Exponent are both incremented in
**    such a way as to create a uniform change in the theoretical
**    Specular Exponent factor within the Specular summand.
**    (This controlled incrementation takes place in the routine
**    called Set() below). The other two components of the Normal
**    are set so that the total length of the Normal is 1. The
**    eye position and the light position are set so that the
**    Half_Angle vector, H_hat, is (0, 0, 1), with these settings
**    (Normal dot H_hat) = Normal[2].
**    
**    The implementation fails if blue or green are
**    not 0.0. In both modes one light is enabled at a time, all
**    individual lights are tested. The observed results are also
**    checked for monotonicity, and the change is checked for
**    uniformity. The observed change should be less than 1 shade
**    from the theoretical change (delta - trueDelta) < delta.
**    When the expected value reaches 1.0 the observed color is
**    checked to insure that the point has rendered with a fully
**    saturated red.
**    
** Error Explanation -
**    other_color > 0.0.
**    value < lastValue.
**    maxObserved < 1.0.
**    (thisDelta - correctDelta) > 2 shades (see NOTE below).
**    (thisDelta - lastDelta) > 1 shade.
**
** NOTE - fixed-point 16.16 errors in Set introduced when computing
**  x = sqrt(1-z^2) result in precision loss and jittering in the
**  sampling of the specular exponent function, which can cause the
**  final color value to increase by two color shades instead of the
**  allowed one. For the time being we have relaxed the accuracy
**  requirement to 2 shades instead of 1, but Set() should be
**  fixed to address this problem.
**
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilons, 0.25*color_epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Logicop, Stencil.
**        Not allowed = Alias, Dither, Fog, Shade.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _specExpRec {
    ftype step, numColor, lastValue, lastDelta, delta;
    GLenum light;
    GLftype normal[3], position[4]; 
} specExpRec;


/*****************************************************************************/

static void Set(void *data, ftype *ref)
{
    specExpRec *ptr = (specExpRec *)data;
    GLftype exponent;
    ftype tmp; 

    tmp = (*ref > ITOF(0)) ? *ref : ptr->step;
    tmp = (tmp <= ITOF(1)) ? tmp : ITOF(1);
    exponent = tmp * 128;
    glMaterialtv(GL_FRONT_AND_BACK, GL_SHININESS, &exponent);

    tmp = DIV(LOG(tmp), exponent);
    ptr->normal[2] = EXP(tmp);
    tmp = ptr->normal[2]; 
    ptr->normal[0] = SQRT(ITOF(1) - MUL(tmp, tmp));

    glNormal3t(ptr->normal[0], ptr->normal[1], ptr->normal[2]);
    DrawPoint2f(ITOF(1)/2, ITOF(1)/2);

    *ref += ptr->step;
}

/****************************************************************************/

static void Init(void *data)
{
    specExpRec *ptr = (specExpRec *)data;
    GLftype buf[4];
    ftype step;
    GLenum i;

    ptr->lastValue = -ITOF(1);
    ptr->lastDelta = -ITOF(2);

    step = ITOF(2);
    ptr->delta = DIV(ITOF(1), LDEXP(ITOF(1), buffer.colorBits[0]) - ITOF(1)); 
    ptr->step = DIV(ptr->delta, step);

    for (i = GL_LIGHT0; i <= GL_LIGHT7; i++) {
	glDisable(i);
    }
    glEnable(ptr->light);

    ptr->normal[0] = ITOF(0);
    ptr->normal[1] = ITOF(0);
    ptr->normal[2] = ITOF(0);

    ptr->position[0] = ITOF(1)/2;
    ptr->position[1] = ITOF(1)/2;
    ptr->position[2] = ITOF(1);
    ptr->position[3] = ITOF(1);
    glLighttv(ptr->light, GL_POSITION, ptr->position);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_AMBIENT, buf);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_DIFFUSE, buf);

    buf[0] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_SHININESS, buf);

    buf[0] = ITOF(1);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glLighttv(ptr->light, GL_SPECULAR, buf);
    glMaterialtv(GL_FRONT_AND_BACK, GL_SPECULAR, buf);

    buf[0] = ITOF(1);
    buf[1] = ITOF(1);
    buf[2] = ITOF(1);
    buf[3] = ITOF(1);
    glLighttv(ptr->light, GL_DIFFUSE, buf);
}

static long Test(void *data, ftype ref)
{
    specExpRec *ptr = (specExpRec *)data;
    char str[240];
    GLftype buf[3]; 
    ftype value, delta;
    long i;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 1; i < 3; i++) {
	if (buf[i] != ITOF(0)) {
            ColorError_RGBZero(str, 0, buf);
            ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
    }

    value = buf[0];
    if (ptr->lastDelta == -ITOF(2)) {
	ptr->lastDelta = -ITOF(1);
	ptr->lastValue = value;
    } else if (ptr->lastDelta == -ITOF(1)) {
	if (value < ptr->lastValue) {
            ColorError_RGBMonotonic(str, 0, value, ptr->lastValue);
            ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    if (ABS(delta-ptr->delta) > epsilon.color[0]) {
                ColorError_RGBStep(str, 0, delta, ptr->delta);
                ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    ptr->lastDelta = delta;
	}
    } else {
	if (value < ptr->lastValue) {
            ColorError_RGBMonotonic(str, 0, value, ptr->lastValue);
            ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    if (ABS(delta-ptr->delta) > epsilon.color[0]) {
                ColorError_RGBStep(str, 0, delta, ptr->delta);
                ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
            /* See NOTE at head of l_se.c - originally > epsilon.zero */
            if (ABS(delta-ptr->lastDelta) > epsilon.color[0] + epsilon.zero) {
                ColorError_RGBDelta(str, 0, delta, ptr->lastDelta);
                ErrorReport(__FILE__, __LINE__, str);
		return ERROR;
	    }
	    ptr->lastDelta = delta;
	}
    }

    if (ref >= ITOF(1) && ABS(value-ITOF(1)) > epsilon.zero) {
        ColorError_RGBClamp(str, 0, value);
        ErrorReport(__FILE__, __LINE__, str);
	return ERROR;
    }

    return NO_ERROR;
}

long SpecExpExec(void)
{
    specExpRec data;
    long max;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_LIGHTING);

    max = (machine.pathLevel == 0) ? GL_LIGHT7 : GL_LIGHT0;
    /*
    ** The exponentials in this test may not work well for 12 bit deep lights 
    ** with weak math libraries.
    */
    for (data.light = GL_LIGHT0; (long)data.light <= max; data.light++) {
	if (RampUtil(ITOF(0), ITOF(1), Init, Set, Test, &data) == ERROR) {
	    return ERROR;
	}
    }
    return NO_ERROR;
}
