/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/l_sed.c,v 1.5 2005/05/11 15:21:40 artlab Exp $
*/

/*
** l_sed.c
** Light - Spot Exponent And Direction Test.
**
** Description -
**    This test analyses the effect that varying the
**    Spotlight_Direction has on the color produced by lighting.
**    To analyse only this one parameter, the Ambient summand is
**    used. The test repeats itself with different Spot_Exponents,
**    and different Cutoff_Angles.
**    
**    The Ambient_Material is set so that one color
**    channel is fully saturated, and the other channels are set
**    to zero. All three channels are tested. The implementation
**    fails if any color channel not being tested is not observed
**    to be 0.0. The Ambient_Light is set to (1, 1, 1, 1). Other
**    lighting parameters are chosen so that the Diffuse,
**    Specular, and Emissive summands of the lighting are set
**    to 0.0.
**    
**    The test repeatedly renders and reads the same point at the
**    lower left corner of the screen. The third component of the
**    Spotlight_Direction is incremented by a an amount sensitive
**    to the Spotlight_Exponent being tested. The increment is
**    controlled so that the observed resulting color would
**    theoretically increment uniformly. (This controlled
**    incrementing takes place in the routine called Set()
**    below). The other two components of the Spotlight_Direction
**    are adjusted so that the total length of the
**    Spotlight_Direction vector is 1. The rendered point's eye
**    position and the Spotlight_Position are set so that the
**    vector from the Spotlight_Position to the rendered point's
**    eye position, PV, is (0, 0, 1). With these settings
**    (SpotDirection dot PV) = SpotDirection[2]).
**    
**    If the theoretically calculated Spot_Exponent
**    factor indicates that the Cutoff_Angle would cause the
**    object to render black, the implementation fails if the
**    observed color is not black. If the expected value
**    indicates that the observed color should not be "cutoff"
**    the implementation fails if the color is more than 5% away
**    from the theoretical correct value. A margin around the
**    cutoff angle of 1% is used. One light is
**    enabled at a time, all individual lights are tested. The
**    observed results are compared to the theoretical value.
**    The error here is 10% for bright shades or one colorEpsilon
**    for dark shades. This test will not run with less than 3
**    bits of color per channel. When Spotlight_Position[2]
**    reaches 1.0, the color is finally checked to insure that
**    the point has rendered with a fully saturated color. The
**    observed results are also checked for monoticity.
**    
** Error Explanation -
**    other_color > 0.0.
**    value < lastValue.
**    maxObserved < 1.0.
**    ABS(observed - expected) > 0.1.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilons, 0.25*color_epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Logicop, Stencil.
**        Not allowed = Alias, Dither, Fog, Shade.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _spotExpDirRec {
    long color;
    ftype step, numColor, lastValue, lastDelta, firstDelta;
    GLenum light, expIndex; 
    ftype cosCutoff, cutoffMargin;
    GLftype exp, angle, position[4], direction[3];
    ftype errConst;
} spotExpDirRec;


/*****************************************************************************/

static void Set(void *data, ftype *ref)
{
    spotExpDirRec *ptr = (spotExpDirRec *)data;

    GLftype refValue = POW(*ref, DIV(ITOF(1), ptr->exp));

    ptr->direction[0] = SQRT(ITOF(1) - MUL(refValue, refValue));
    ptr->direction[2] = refValue;

    glLighttv(ptr->light, GL_SPOT_DIRECTION, ptr->direction);

    DrawPoint2f(ITOF(1)/2, ITOF(1)/2);

    /*
    ** This incrementing is sensitive to the different functions
    ** which result from using different Spotlight_Exponents.
    */
    *ref += ptr->step;
}

/****************************************************************************/

static void Init(void *data)
{
    spotExpDirRec *ptr = (spotExpDirRec *)data;
    GLftype buf[4];
    ftype tmp;
    GLenum i;

    ptr->lastValue = -ITOF(1);
    ptr->lastDelta = -ITOF(2);

    if (machine.pathLevel == 0) {
	ptr->step = DIV(ITOF(1)/8, ptr->numColor);
    } else {
	ptr->step = ITOF(1)/4;
    }

    ptr->errConst = ATOF("0.1");

    ptr->light = GL_LIGHT0 + (ptr->expIndex % 8);
    for (i = GL_LIGHT0; i <= GL_LIGHT7; i++) {
	glDisable(i);
    }
    glEnable(ptr->light);

    ptr->angle = OESRandom(ITOF(0), ITOF(90));
    glLighttv(ptr->light, GL_SPOT_CUTOFF, &ptr->angle);
    tmp = ptr->angle;
    ptr->cosCutoff = COS(DIV(MUL(tmp, PI), ITOF(180)));
    ptr->cutoffMargin = ATOF("0.01");

    ptr->direction[0] = ITOF(0);
    ptr->direction[1] = ITOF(0);
    ptr->direction[2] = ITOF(0);

    ptr->position[0] = ITOF(1)/2;
    ptr->position[1] = ITOF(1)/2;
    ptr->position[2] = -ITOF(1);
    ptr->position[3] = ITOF(1);
    glLighttv(ptr->light, GL_POSITION, ptr->position);

    glLighttv(ptr->light, GL_SPOT_EXPONENT, &ptr->exp);

    buf[0] = ITOF(1);
    buf[1] = ITOF(1);
    buf[2] = ITOF(1);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_AMBIENT, buf);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    buf[ptr->color] = ITOF(1);
    glLighttv(ptr->light, GL_AMBIENT, buf);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_DIFFUSE, buf);
    glMaterialtv(GL_FRONT_AND_BACK, GL_SPECULAR, buf);
    glLightModeltv(GL_LIGHT_MODEL_AMBIENT, buf);
}

static long Test(void *data, ftype ref)
{
    spotExpDirRec *ptr = (spotExpDirRec *)data;
    GLftype buf[3]; 
    ftype value, expected, delta, error;
    long i, cutoff, noCutoff;
    GLftype refValue;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 1; i < 3; i++) {
	if (buf[i] != ITOF(0) && i != ptr->color) {
	    return ERROR;
	}
    }

    refValue = POW(ref, DIV(ITOF(1), ptr->exp));
    expected = ref;

    /*
    ** These errors will not work for a badly implemented lookup table.
    */
    if (ptr->numColor > 0) {
	if (expected < DIV(ITOF(1), ptr->numColor)) {
	    error = epsilon.color[ptr->color];
	} else {
	    error = ptr->errConst;
	    if (epsilon.color[ptr->color] > error) {
		error = epsilon.color[ptr->color];
	    }
	}
    } else {
	error = epsilon.color[ptr->color];
    }

    /*
    ** The monotonicity requirement may be too stringent for machines
    ** with extreme color granularity, but I doubt it.
    */
    value = buf[ptr->color];

    cutoff = (refValue < ptr->cosCutoff-ptr->cutoffMargin);
    noCutoff = (refValue > ptr->cosCutoff+ptr->cutoffMargin);

    if (cutoff) {
	if (value > epsilon.color[ptr->color]/2) {
	    return ERROR;
	}
    } else {
	if (noCutoff) {
	    if (ABS(value-expected) > error) {
		return ERROR;
	    }
	}
    }

    /*
    ** The linearity test may be relaxed, but due to variable sampling
    ** this should be linear.
    */
    if (ptr->lastDelta == -ITOF(2)) {
	ptr->lastDelta = -ITOF(1);
	ptr->lastValue = value;
    } else if (ptr->lastDelta == -ITOF(1)) {
	if (value < ptr->lastValue) {
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    } else {
	if (value < ptr->lastValue) {
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    }
    if (refValue >= ITOF(1) && ABS(value-ITOF(1)) > epsilon.zero) {
	return ERROR;
    }
    return NO_ERROR;
}

long SpotExpDirExec(void)
{
    static const ftype expSlow[] = {
	ITOF(1), ITOF(2), ITOF(3), ITOF(4), ITOF(5), ITOF(6), ITOF(7),
        ITOF(12), ITOF(15), ITOF(19), ITOF(26), ITOF(37), ITOF(49),
	ITOF(64), ITOF(81), ITOF(107), ITOF(116), ITOF(125), ITOF(126),
        ITOF(127), ITOF(128)
    };
    static const ftype expFast[] = {
	ITOF(1), ITOF(64), ITOF(128)
    };
    spotExpDirRec data;
    ftype const *expPtr;
    long max;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_LIGHTING);

    max = (machine.pathLevel == 0) ? 21 : 3;
    expPtr = (machine.pathLevel == 0) ? expSlow : expFast;
    for (data.expIndex = 0; (long)data.expIndex < max; data.expIndex++) {
	data.color = (long)data.expIndex % 3;
	data.exp = expPtr[data.expIndex];
	data.numColor = POW(ITOF(2), ITOF(buffer.colorBits[data.color])) - ITOF(1);
	if (RampUtil(ITOF(0), ITOF(1), Init, Set, Test, &data) == ERROR) {
	    ErrorReport(__FILE__, __LINE__, 0);
	    return ERROR;
	}
    }
    return NO_ERROR;
}
