/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/l_sen.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** l_sen.c
** Light - Specular Exponent Normal Test.
**
** Description -
**    This test analyses the effect that varying the Normal has
**    on the color produced by lighting, by way of its effect on
**    the Specular summand.
**    To analyse only this one parameter, the
**    Specular_Material is set to color channel at a time, and
**    Specular_Light is set to (1, 1, 1, 1).
**    Other lighting parameters are
**    chosen so that the Diffuse, Ambient, and Emissive summands
**    of the lighting are set to 0.0. Values are chosen
**    so the Attenuation and the Spotlight factor are both 1.0.
**    This test repeats itself with different Specular_Exponents
**    (GL_SHININESS), and with different color channels.
**    
**    The test repeatedly renders and reads the same point at the
**    lower left corner of the screen. The third component of the
**    Normal is incremented by a an amount sensitive to whichever
**    Specular_Exponent is being tested. The increment is
**    controlled so that the observed resulting color would
**    theoretically increment uniformly. (This controlled
**    incrementation takes place in the routine called Set()
**    below). The other two components of the Normal are set so
**    that the total length of the Normal is 1. The eye position
**    and the light position are set so that the Half_Angle
**    vector, H_hat, is (0, 0, 1), with these settings (Normal
**    dot H_hat) = Normal[2].
**    
**    The implementation fails if any color channel
**    not being tested is not 0.0. In both modes one light is
**    enabled at a time, all individual lights are tested. The
**    observed results are compared to the theoretical value.
**    The error here is 5% for bright colors and 1% for dark
**    shades. This test will not run with less than 3 bits of
**    color per channel. When Light_Position[2] reaches 1.0, the
**    color is finally checked to insure that the point has
**    rendered with a fully saturated color.
**    
** Error Explanation -
**    other_color > 0.0.
**    value < lastValue.
**    maxObserved < 1.0.
**    ABS(observed - expected) > 0.05 for highly saturated colors.
**    ABS(observed - expected) > 0.01 for dark shades.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilons, 0.25*color_epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Logicop, Stencil.
**        Not allowed = Alias, Dither, Fog, Shade.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _specExpNormRec {
    long  color;
    ftype step, lastValue, lastDelta;
    GLenum  light, expIndex;
    GLftype exp, normal[3];
    ftype numColor, errConstSat, errConstDark;
} specExpNormRec;


/*****************************************************************************/

static void Set(void *data, ftype *ref)
{
    specExpNormRec *ptr = (specExpNormRec *)data;

    ptr->normal[0] = SQRT(ITOF(1) - MUL(*ref, *ref));
    ptr->normal[2] = *ref;
    glNormal3t(ptr->normal[0], ptr->normal[1], ptr->normal[2]);

    DrawPoint2f(ITOF(1)/2, ITOF(1)/2);

    /*
    ** This increment controls that the theoretical increment of the
    ** observed color should be uniform.
    */
    *ref = POW(*ref, ptr->exp);

    *ref += ptr->step;
    *ref = POW(*ref, DIV(ITOF(1), ptr->exp));
}

/****************************************************************************/

static void Init(void *data)
{
    specExpNormRec *ptr = (specExpNormRec *)data;
    GLftype buf[4];
    GLenum i;

    ptr->lastValue = -ITOF(1);
    ptr->lastDelta = -ITOF(2);

    if (machine.pathLevel == 0) {
	ptr->step = DIV(ITOF(1)/2, ptr->numColor);
    } else {
	ptr->step = ITOF(1)/4;
    }

    i = 1;
    do {
	ptr->errConstSat = DIV(ITOF(i), ptr->numColor);
	i++;
    } while (ptr->errConstSat < ATOF("0.05") && i < 1000);

    i = 1;
    do {
	ptr->errConstDark = DIV(ITOF(i), ptr->numColor);
	i++;
    } while (ptr->errConstDark < ATOF("0.01") && i < 1000);

    ptr->light = GL_LIGHT0 + (ptr->expIndex % 8);
    for (i = GL_LIGHT0; i <= GL_LIGHT7; i++) {
	glDisable(i);
    }
    glEnable(ptr->light);

    ptr->normal[0] = ITOF(0);
    ptr->normal[1] = ITOF(0);
    ptr->normal[2] = ITOF(0);

    buf[0] = ITOF(1)/2;
    buf[1] = ITOF(1)/2;
    buf[2] = ITOF(1);
    buf[3] = ITOF(1);
    glLighttv(ptr->light, GL_POSITION, buf);

    glMaterialtv(GL_FRONT_AND_BACK, GL_SHININESS, &ptr->exp);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_AMBIENT, buf);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_DIFFUSE, buf);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    buf[ptr->color] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_SPECULAR, buf);

    buf[0] = ITOF(1);
    buf[1] = ITOF(1);
    buf[2] = ITOF(1);
    buf[3] = ITOF(1);
    glLighttv(ptr->light, GL_SPECULAR, buf);
}

static long Test(void *data, ftype ref)
{
    specExpNormRec *ptr = (specExpNormRec *)data;
    GLftype buf[3]; 
    ftype value, expected,  delta, error;
    long i;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 1; i < 3; i++) {
	if (buf[i] != ITOF(0) && i != ptr->color) {
	    return ERROR;
	}
    }

    value = buf[ptr->color];
    expected = POW(ref, ptr->exp);

    /*
    ** We should observe colors within 5% of the expected value
    ** for the numbers reflecting some saturation, and we expect only 
    ** color resolution error near 0;
    ** another possibility error = expected * ptr->errConst;
    */
    if (ref == ITOF(0) || ptr->exp == ITOF(0)) {
	error = ptr->errConstDark + epsilon.zero;
    } else if (DIV(expected, ref) < DIV(ATOF("0.01"), ptr->exp)) {
	error = ptr->errConstDark + epsilon.zero;
    } else {
	error = ptr->errConstSat + epsilon.zero;
    }

    if (ABS(value-expected) > error) {
	return ERROR;
    }
    /*
    ** The monotonicity requirement may be too stringent for machines
    ** with extreme color granularity. I doubt it, but we don't include it.
    ** Also due to variable sampling this should be linear, but we do not 
    ** test for uniform delta.
    */
    if (ptr->lastDelta == -ITOF(2)) {
	ptr->lastDelta = -ITOF(1);
	ptr->lastValue = value;
    } else if (ptr->lastDelta == -ITOF(1)) {
	if (value < ptr->lastValue) {
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    } else {
	if (value < ptr->lastValue) {
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    }
    if (ref >= ITOF(1) && ABS(value-ITOF(1)) > epsilon.zero) {
	return ERROR;
    }
    return NO_ERROR;
}

long SpecExpNormExec(void)
{
    static const ftype expSlow[] = {
	ITOF(1), ITOF(2), ITOF(3), ITOF(4), ITOF(5), ITOF(6), ITOF(7),
	ITOF(12), ITOF(15), ITOF(19), ITOF(26), ITOF(37), ITOF(49),
	ITOF(64), ITOF(81), ITOF(107), ITOF(116), ITOF(125), ITOF(126), ITOF(127), ITOF(128)
    };
    static const ftype expFast[] = {
	ITOF(1), ITOF(66), ITOF(128)
    };
    specExpNormRec data;
    ftype const *expPtr;
    long max;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_LIGHTING);

    max = (machine.pathLevel == 0) ? 21 : 3;
    expPtr = (machine.pathLevel == 0) ? expSlow : expFast;
    for (data.expIndex = 0; (long)data.expIndex < max; data.expIndex++) {
	data.color = (long)data.expIndex % 3;
	data.numColor = LDEXP(ITOF(1), buffer.colorBits[data.color]) - ITOF(1);
	data.exp = expPtr[data.expIndex];
	if (RampUtil(ITOF(0), ITOF(1), Init, Set, Test, &data) == ERROR) {
	    ErrorReport(__FILE__, __LINE__, 0);
	    return ERROR;
	}
    }
    return NO_ERROR;
}
