/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/l_sn.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** l_sn.c
** Light - Specular Normal Test.
**
** Description -
**    This test
**    analyses the effect that varying the Normal has on the
**    color produced by lighting, by way of its effect on the
**    Specular summand. To analyse only this one parameter, the
**    Specular_Lighting factor is set to a completely saturated
**    red. Lighting parameters are chosen so that the Diffuse,
**    Ambient, and Emissive summands of the lighting equation are
**    set to zero and the Attenuation and the Spotlight factor
**    are both 1.0. The Specular_Exponent is always set to 1.0.
**    
**    The test repeatedly renders and reads the same point at the
**    lower left corner of the screen. The third component of the
**    Normal is incremented by a uniform amount. The other two
**    components of the Normal are set so that the total length
**    of the Normal is 1. The Light_Position is set so that the
**    vector from the eye position of the rendered point to the
**    Light_position is (0, 0, 1). With the viewer set at
**    infinity. With these settings the normalized Half_Angle
**    vector is (0, 0, 1) so (Normal dot H_hat) = Normal[2]. The
**    lighting equation is reduced to color = red * Normal[2].
**    
**    Only the red color channel is tested. The
**    implementation fails if any other color channel is not
**    black. One light is enabled at a time, all
**    individual lights are tested.
**    
**    The uniformity of the change is tested, and the
**    implementation fails if consecutive changes are too
**    different. (thisDelta - lastDelta) > 1/4 shade. The
**    observed results are also checked for monotonicity. When
**    Normal[2] reaches 1.0 the color is finally checked to
**    insure that the point has rendered with a fully saturated
**    red.
**    
** Error Explanation -
**    other_color > 0.0.
**    value < lastValue.
**    maxObserved < 1.0.
**    ABS(thisDelta - lastDelta) > 1/4 shade.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilons, 0.25*color_epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Logicop, Stencil.
**        Not allowed = Alias, Dither, Fog, Shade.
*/

#include "conform.h"
#include "util.h"
#include "utilru.h"


typedef struct _specNormRec {
    ftype step, numColor, lastValue, lastDelta, delta;
    GLenum light;
    GLftype normal[3];
} specNormRec;


/*****************************************************************************/

static void Set(void *data, ftype *ref)
{
    specNormRec *ptr = (specNormRec *)data;

    ptr->normal[0] = SQRT(ITOF(1) - MUL(*ref, *ref));
    ptr->normal[2] = *ref;

    glNormal3t(ptr->normal[0], ptr->normal[1], ptr->normal[2]);
    DrawPoint2f(ITOF(1)/2, ITOF(1)/2);

    *ref += ptr->step;
}

/****************************************************************************/

static void Init(void *data)
{
    specNormRec *ptr = (specNormRec *)data;
    GLftype buf[4];
    ftype step;
    GLenum i;

    ptr->lastValue = -ITOF(1);
    ptr->lastDelta = -ITOF(2);

    step = (machine.pathLevel == 0) ? ITOF(4) : ITOF(2);
    ptr->delta = DIV(ITOF(1), LDEXP(ITOF(1), buffer.colorBits[0]) - ITOF(1));
    ptr->step = DIV(ptr->delta, step);

    for (i = GL_LIGHT0; i <= GL_LIGHT7; i++) {
	glDisable(i);
    }
    glEnable(ptr->light);

    ptr->normal[0] = ITOF(0);
    ptr->normal[1] = ITOF(0);
    ptr->normal[2] = ITOF(0);

    buf[0] = ITOF(1)/2;
    buf[1] = ITOF(1)/2;
    buf[2] = ITOF(1);
    buf[3] = ITOF(1);
    glLighttv(ptr->light, GL_POSITION, buf);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_AMBIENT, buf);

    buf[0] = ITOF(0);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_DIFFUSE, buf);

    buf[0] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_SHININESS, buf);

    buf[0] = ITOF(1);
    buf[1] = ITOF(0);
    buf[2] = ITOF(0);
    buf[3] = ITOF(1);
    glMaterialtv(GL_FRONT_AND_BACK, GL_SPECULAR, buf);

    buf[0] = ITOF(1);
    buf[1] = ITOF(1);
    buf[2] = ITOF(1);
    buf[3] = ITOF(1);
    glLighttv(ptr->light, GL_SPECULAR, buf);
}

static long Test(void *data, ftype ref)
{
    specNormRec *ptr = (specNormRec *)data;
    char str[240];
    GLftype buf[3]; 
    ftype value, delta;
    long i;

    ReadScreen(0, 0, 1, 1, GL_RGB, buf);

    for (i = 1; i < 3; i++) {
	if (buf[i] != ITOF(0)) {
            ColorError_RGBZero(str, 0, buf);
            ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
    }

    value = buf[0];
    if (ptr->lastDelta == -ITOF(2)) {
	ptr->lastDelta = -ITOF(1);
	ptr->lastValue = value;
    } else if (ptr->lastDelta == -ITOF(1)) {
	if (value < ptr->lastValue) {
            ColorError_RGBMonotonic(str, 0, value, ptr->lastValue);
            ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    } else {
	if (value < ptr->lastValue) {
            ColorError_RGBMonotonic(str, 0, value, ptr->lastValue);
            ErrorReport(__FILE__, __LINE__, str);
	    return ERROR;
	}
	delta = value - ptr->lastValue;
	ptr->lastValue = value;
	if (ABS(delta) > epsilon.zero) {
	    ptr->lastDelta = delta;
	}
    }

    if (ref >= ITOF(1) && ABS(value-ITOF(1)) > epsilon.zero) {
        ColorError_RGBClamp(str, 0, value);
        ErrorReport(__FILE__, __LINE__, str);
	return ERROR;
    }

    return NO_ERROR;
}

long SpecNormExec(void)
{
    specNormRec data;
    long max;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_LIGHTING);

    max = (machine.pathLevel == 0) ? GL_LIGHT7 : GL_LIGHT0;
    for (data.light = GL_LIGHT0; (long)data.light <= max; data.light++) {
	if (RampUtil(ITOF(0), ITOF(1), Init, Set, Test, &data) == ERROR) {
	    return ERROR;
	}
    }
    return NO_ERROR;
}
