/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/lineaa.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** lineaa.c
** Line Anti-Aliasing Test.
**
** Description -
**    The first part tests position independence of anti-aliased
**    lines. The coverage of a random length line is calculated.
**    This line is then translated. Regardless of its new
**    position, the coverage should remain the same, modulo
**    difference resulting from subpixel positioning. The test
**    is repeated for a several line lengths, each taken through
**    a several randomly generated translations. Lines of width
**    start at 1.0 and are incremented by implementation dependent
**    granularity up to the implementation range maximum.
**    
**    The second part tests the subsetting restriction for
**    anti-aliasing. A line is drawn at a fixed position with a
**    series of decreasing lengths and widths. The coverage for
**    each successive line should be less than or equal to the
**    coverage for the previous, longer or wider, line.
**
** Error Explanation -
**    For the first section of the test:
**        The error allowance is based on a minimum of 4
**        subpixel sampling. This yields 0.125 per pixel.
**        One half of a color epsilon is added per pixel to
**        account for the rounding from a coverage value to
**        a shade. This sum is multiplied by the number of
**        pixels along the perimeter (twice width for ends,
**        plus twice length).
**    
**    For the second section of the test:
**        The error allowance is epsilon.zero times line length.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        (under error explanation above).
**    Paths:
**        Allowed = Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Alpha, Blend, Dither.
*/

#include "conform.h"
#include "util.h"


static char errStr[160];
static char errStrFmt[][80] = {
    "Line primitive did not draw.",
    "Coverage is %g, initial coverage was %g. Error margin is %g.",
    "Failure in subset test. Coverage for line is %g, should be less than %g."
};

static void DrawLine(GLftype *vertex0, GLftype *vertex1)
{
    GLftype vertices[2][2];
    vertices[0][0] = vertex0[0];
    vertices[0][1] = vertex0[1];
    vertices[1][0] = vertex1[0];
    vertices[1][1] = vertex1[1];

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FTYPE, 0, &vertices[0][0]);
    glDrawArrays(GL_LINES, 0, 2);
    glDisableClientState(GL_VERTEX_ARRAY);
}

long LineAntiAliasExec(void)
{
    GLftype *buf;
    GLftype v1[2], v2[2];
    GLftype angle, shiftX, shiftY, sum, saveSum = ITOF(0), errorMargin;
    GLftype width;
    GLftype widthGranularity;
    GLint widthRange[2];
    long i, j, k, max;

    /*
    ** There is no specified coverage computation for anti-aliasing,
    ** therefore we test that the sum of the coverage values is
    ** fairly constant, independent of translation in the window. 
    */

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*3*sizeof(GLftype));

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glColor4t(ITOF(1), ITOF(0), ITOF(0), ITOF(1));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    glGetIntegerv(GL_SMOOTH_LINE_WIDTH_RANGE, widthRange);
    widthGranularity = ITOF(1)/8;
    width = ITOF(1);

    max = (machine.pathLevel == 0) ? 9 : 2;

    for (i = 0; i < max; i++) {
	v1[0] = ITOF(0);
	v1[1] = ITOF(0);
	v2[0] = OESRandom(ITOF(1), ITOF(WINDSIZEX)/3);
	v2[1] = ITOF(0);
	angle = OESRandom(ITOF(0), ITOF(360));

	glLineWidtht(width);

	for (k = 0; k < 8; k++) {
	    glClear(GL_COLOR_BUFFER_BIT);

	    glPushMatrix();

	    shiftX = OESRandom(ITOF(WINDSIZEX)/3, ITOF(WINDSIZEX)*2/3);
	    shiftY = OESRandom(ITOF(WINDSIZEY)/3, ITOF(WINDSIZEY)*2/3);
	    glTranslatet(shiftX, shiftY, ITOF(0));
	    glRotatet(angle, ITOF(0), ITOF(0), ITOF(1));

	    DrawLine(v1, v2);

	    glPopMatrix();
	    glFlush();

	    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);

	    sum = 0;
	    for (j = 0; j < WINDSIZEY*WINDSIZEY*3; j += 3) {
		sum += buf[j];
	    }

	    if (k == 0) {
		saveSum = sum;
	    } else {
		if (sum == ITOF(0)) {
		    ErrorReport(__FILE__, __LINE__, errStrFmt[0]);
		    FREE(buf);
		    return ERROR;
		}

		/*
		**  Error allowance based on minimum of 4 subpixel sampling.
		**  The error margin is multiplied by the number of pixels
		**  on the perimeter of the primitive.
		*/

		errorMargin = 2 * MUL(ITOF(1)/8 + epsilon.color[0]/2,
			      v2[0] + width);
		if (ABS(saveSum-sum) > errorMargin) {
		    StrMake(errStr, errStrFmt[1], sum, saveSum, errorMargin);
		    ErrorReport(__FILE__, __LINE__, errStr);
		    FREE(buf);
		    return ERROR;
		}
	    }
	}

	width = MIN(width + widthGranularity, ITOF(widthRange[1]));
    }

    /*
    ** Test subsetting rule.
    */

    width = ITOF(1);
    for (i = 0; i < max; i++) {
	v1[0] = ITOF(0);
	v1[1] = ITOF(0);
	v2[0] = OESRandom(ITOF(15), ITOF(WINDSIZEX-10));
	v2[1] = ITOF(0);
	angle = OESRandom(ITOF(0), ITOF(90));

	width = MIN(width + 6*widthGranularity, ITOF(widthRange[1]));
	glLineWidtht(width);

	for (k = 0; k < 8; k++) {
	    glClear(GL_COLOR_BUFFER_BIT);

	    glPushMatrix();
	    /*
	    ** Translate away from corner so wide lines don't get clipped.
	    */
	    glTranslatet(ITOF(5), ITOF(5), ITOF(0));
	    glRotatet(angle, 0, 0, 1);

	    glLineWidtht(width);
	    DrawLine(v1, v2);

	    glPopMatrix();
	    glFlush();

	    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);

	    sum = 0;
	    for (j = 0; j < WINDSIZEY*WINDSIZEY*3; j += 3) {
		sum += buf[j];
	    }

	    if (k == 0) {
		saveSum = sum;
	    } else {
		if (sum == ITOF(0)) {
		    ErrorReport(__FILE__, __LINE__, errStrFmt[0]);
		    FREE(buf);
		    return ERROR;
		}

		/*
		**  The line coverage must be a subset of the
		**  coverage of the previous (longer or wider) line.
		**  Allow numerical error proportional to line length.
		*/
		if ((sum-saveSum) > MUL(v2[0], epsilon.zero)) {
		    StrMake(errStr, errStrFmt[2], sum, saveSum);
		    ErrorReport(__FILE__, __LINE__, errStr);
		    FREE(buf);
		    return ERROR;
		}
	    }
	    saveSum = sum;

	    /*
	    ** Decrease length or width with each iteration.
	    */
	    if (k%2 == 0) {
		v2[0] -= ATOF("1.1");
	    } else {
		width = (width > ITOF(1) + widthGranularity) ? 
			(width - widthGranularity) : ITOF(1);
	    }
	}
    }

    FREE(buf);
    return NO_ERROR;
}
