/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/linehv.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** linehv.c
** Horizontal And Vertical Line Test.
**
** Description -
**    Tests the rasterization of horizontal and vertical lines.
**    Even though the OpenGL ES specification is very lacks on
**    defining exact fragment location for lines, an
**    implementation that does not properly handle horizontal or
**    vertical with well behaved endpoints would be a poor OpenGL ES
**    library. The lines are integer widths, within ALIASED_LINE_WIDTH_RANGE. 
**    The vertices of the lines are pixel centered. The lines are
**    tested at exact pixel locations.
**
**    There is an allowance for half-open lines (for left or right
**    ends but not both ends).
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include <limits.h>


static char errStr[160];
static char srrStrFmt[] = "Line (%d, %d) (%d, %d) (width of %d) is not %s.";


static void XClearDraw(GLint *v1, GLint *v2, GLsizei width)
{
    GLftype vertices[2][2];
    vertices[0][0] = ITOF(v1[0]) + ITOF(1)/2;
    vertices[0][1] = ITOF(v1[1]) + ((width % 2) ? ITOF(1)/2 : ITOF(0));
    vertices[1][0] = ITOF(v2[0]) + ITOF(1)/2;
    vertices[1][1] = ITOF(v2[1]) + ((width % 2) ? ITOF(1)/2 : ITOF(0));

    glClear(GL_COLOR_BUFFER_BIT);

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FTYPE, 0, &vertices[0][0]);
    glDrawArrays(GL_LINES, 0, 2);
    glDisableClientState(GL_VERTEX_ARRAY);
}

static void YClearDraw(GLint *v1, GLint *v2, GLsizei width)
{
    GLftype vertices[2][2];
    vertices[0][0] = ITOF(v1[0]) + ((width % 2) ? ITOF(1)/2 : ITOF(0));
    vertices[0][1] = ITOF(v1[1]) + ITOF(1)/2;
    vertices[1][0] = ITOF(v2[0]) + ((width % 2) ? ITOF(1)/2 : ITOF(0));
    vertices[1][1] = ITOF(v2[1]) + ITOF(1)/2;

    glClear(GL_COLOR_BUFFER_BIT);

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FTYPE, 0, &vertices[0][0]);
    glDrawArrays(GL_LINES, 0, 2);
    glDisableClientState(GL_VERTEX_ARRAY);
}

static long XLines(GLint *v1, GLint *v2, GLsizei len, GLsizei width,
		   GLftype *buf)
{
    GLftype *ptr;
    GLint endPoints[2], x, y, i, j;

    /* Initialize to invalid values. */
    endPoints[0] = INT_MAX;
    endPoints[1] = INT_MAX;

    XClearDraw(v1, v2, width);

    if (v1[0] <= v2[0]) {
	x = v1[0] - 1;
    } else {
	x = v2[0] - 1;
    }
    y = v1[1] - (width / 2) - 1;
    ReadScreen(x, y, len+3, width+2, GL_AUTO_COLOR, buf);

    ptr = buf;
    for (i = 0; i < len+3; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    return ERROR;
	}
    }
    for (i = 0; i < width; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    return ERROR;
	}
	if (AutoColorCompare(*ptr++, COLOR_ON) == GL_TRUE) {
	    endPoints[0] = GL_TRUE;
	} else {
	    endPoints[0] = GL_FALSE;
	}
	for (j = 0; j < len-1; j++) {
	    if (AutoColorCompare(*ptr++, COLOR_ON) == GL_FALSE) {
		return ERROR;
	    }
	}
	if (AutoColorCompare(*ptr++, COLOR_ON) == GL_TRUE) {
	    endPoints[1] = GL_TRUE;
	} else {
	    endPoints[1] = GL_FALSE;
	}
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    return ERROR;
	}
    }
    for (i = 0; i < len+3; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    return ERROR;
	}
    }

    if (endPoints[0] == GL_FALSE && endPoints[1] == GL_FALSE) {
	return ERROR;
    }

    return NO_ERROR;
}

static long YLines(GLint *v1, GLint *v2, GLsizei len, GLsizei width,
		   GLftype *buf)
{
    GLftype *ptr;
    GLint *endPoints, x, y, i, j;

    YClearDraw(v1, v2, width);

    x = v1[0] - (width / 2) - 1;
    if (v1[1] <= v2[1]) {
	y = v1[1]-1;
    } else {
	y = v2[1]-1;
    }
    ReadScreen(x, y, width+2, len+3, GL_AUTO_COLOR, buf);

    ptr = buf;
    for (i = 0; i < width+2; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    return ERROR;
	}
    }
    endPoints = (GLint *)MALLOC(width*sizeof(GLint));
    if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	FREE(endPoints);
	return ERROR;
    }
    for (i = 0; i < width; i++) {
	if (AutoColorCompare(*ptr++, COLOR_ON) == GL_TRUE) {
	    endPoints[i] = GL_TRUE;
	} else {
	    endPoints[i] = GL_FALSE;
	}
    }
    if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	FREE(endPoints);
	return ERROR;
    }
    for (i = 0; i < len-1; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    FREE(endPoints);
	    return ERROR;
	}
	for (j = 0; j < width; j++) {
	    if (AutoColorCompare(*ptr++, COLOR_ON) == GL_FALSE) {
		FREE(endPoints);
		return ERROR;
	    }
	}
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    FREE(endPoints);
	    return ERROR;
	}
    }
    if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	FREE(endPoints);
	return ERROR;
    }
    for (i = 0; i < width; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_TRUE) {
	    if (endPoints[i] == GL_FALSE) {
		FREE(endPoints);
		return ERROR;
	    }
	}
    }
    if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	FREE(endPoints);
	return ERROR;
    }
    FREE(endPoints);
    for (i = 0; i < width+2; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    return ERROR;
	}
    }

    return NO_ERROR;
}

long LineHVExec(void)
{
    GLftype *buf;
    GLint widthRange[2];
    long max;
    GLint v1[2], v2[2];
    GLsizei len, i;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    glGetIntegerv(GL_ALIASED_LINE_WIDTH_RANGE, widthRange);
    if (widthRange[1] >= MIN(WINDSIZEX/2, WINDSIZEY/2)) {
	widthRange[1] = MIN(WINDSIZEX/2, WINDSIZEY/2) - 1;
    }
    max = (machine.pathLevel == 0) ? widthRange[1] : 1;

    for (i = 1; i <= max; i++) {
	glLineWidtht(ITOF(i));

	len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
	v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
	v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] + len;
	v2[1] = v1[1];
	if (XLines(v1, v2, len, i, buf) == ERROR) {
	    StrMake(errStr, srrStrFmt, v1[0], v1[1], v2[0], v2[1], i,
		    "horizontal");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}

	len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
	v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
	v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0];
	v2[1] = v1[1] + len;
	if (YLines(v1, v2, len, i, buf) == ERROR) {
	    StrMake(errStr, srrStrFmt, v1[0], v1[1], v2[0], v2[1], i,
		    "vertical");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}

	len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
	v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
	v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] - len;
	v2[1] = v1[1];
	if (XLines(v1, v2, len, i, buf) == ERROR) {
	    StrMake(errStr, srrStrFmt, v1[0], v1[1], v2[0], v2[1], i,
		    "horizontal");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}

	len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
	v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
	v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0];
	v2[1] = v1[1] - len;
	if (YLines(v1, v2, len, i, buf) == ERROR) {
	    StrMake(errStr, srrStrFmt, v1[0], v1[1], v2[0], v2[1], i,
		    "vertical");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}
    }

    FREE(buf);
    return NO_ERROR;
}
