/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/linerast.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** linerast.c
** Line Rasterization Test.
**
** Description -
**    Tests the rasterization of single width lines. A
**    specification of the OpenGL ES states that there can be only
**    one pixel per column for an X-major line and one pixel per
**    row for a Y-major line. This test will draw a series of
**    lines of random lengths and directions to verify this.
**    Special code for checking near the endpoints of the line is
**    used since the endpoints may stray up to one pixel in any
**    direction from their calculated locations.
**
** Error Explanation -
**    Failure occurs if there is more than one pixel in the
**    cross-section of the line. Also, an error will result if an
**    endpoint strays more than one pixel from its calculated
**    location.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"


static char errStr[240];
static char errStrFmt[][160] = {
    "Line (%d, %d) (%d, %d) (%s-major line) has bad endpoints.",
    "Line (%d, %d) (%d, %d) (%s-major line) has more then one pixel per minor axis.",
    "Line (%d, %d) (%d, %d) (%s-major line) has missing pixels per minor axis."
};


static void ClearSetRead(GLint *v1, GLint *v2, GLint len, GLftype *buf)
{
    GLftype vertices[2][2];
    vertices[0][0] = ITOF(v1[0]) + ITOF(1)/2;
    vertices[0][1] = ITOF(v1[1]) + ITOF(1)/2;
    vertices[1][0] = ITOF(v2[0]) + ITOF(1)/2;
    vertices[1][1] = ITOF(v2[1]) + ITOF(1)/2;

    glClear(GL_COLOR_BUFFER_BIT);

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FTYPE, 0, &vertices[0][0]);
    glDrawArrays(GL_LINES, 0, 2);
    glDisableClientState(GL_VERTEX_ARRAY);

    ReadScreen((v2[0]>=v1[0])?(v1[0]-2):(v2[0]-2),
	       (v2[1]>=v1[1])?(v1[1]-2):(v2[1]-2),
	       len+4, len+4, GL_AUTO_COLOR, buf);
}

static long XLines(GLint *v1, GLint *v2, GLint len, GLftype *buf)
{
    long i, j, count;

    ClearSetRead(v1, v2, len, buf);

    for (i = 0; i < len+4; i++) {
	if (AutoColorCompare(buf[i*(len+4)], COLOR_OFF) == GL_FALSE) {
	    StrMake(errStr, errStrFmt[0], v1[0], v1[1], v2[0], v2[1], "x");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }

    for (i = 1; i <= len+2; i++) {
	count = 0;
	for (j = 0; j < len+4; j++) {
	    if (AutoColorCompare(buf[j*(len+4)+i], COLOR_ON) == GL_TRUE) {
		count++;
	    }
	}
	if (i <= 3 || i >= len) {
	    if (count > 1) {
		StrMake(errStr, errStrFmt[0], v1[0], v1[1], v2[0], v2[1], "x");
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	} else {
	    if (count > 1) {
		StrMake(errStr, errStrFmt[1], v1[0], v1[1], v2[0], v2[1], "x");
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    } else if (count == 0) {
		StrMake(errStr, errStrFmt[2], v1[0], v1[1], v2[0], v2[1], "x");
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	}
    }

    for (i = 0; i < len+4; i++) {
	if (AutoColorCompare(buf[i*(len+4)+(len+3)], COLOR_OFF) == GL_FALSE) {
	    StrMake(errStr, errStrFmt[0], v1[0], v1[1], v2[0], v2[1], "x");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }

    return NO_ERROR;
}

static long YLines(GLint *v1, GLint *v2, GLint len, GLftype *buf)
{
    long i, j, count;

    ClearSetRead(v1, v2, len, buf);

    for (i = 0; i < len+4; i++) {
	if (AutoColorCompare(buf[i], COLOR_OFF) == GL_FALSE) {
	    StrMake(errStr, errStrFmt[0], v1[0], v1[1], v2[0], v2[1], "y");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }

    for (i = 1; i <= len+2; i++) {
	count = 0;
	for (j = 0; j < len+4; j++) {
	    if (AutoColorCompare(buf[i*(len+4)+j], COLOR_ON) == GL_TRUE) {
		count++;
	    }
	}
	if (i <= 3 || i >= len) {
	    if (count > 1) {
		StrMake(errStr, errStrFmt[0], v1[0], v1[1], v2[0], v2[1], "y");
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	} else {
	    if (count > 1) {
		StrMake(errStr, errStrFmt[1], v1[0], v1[1], v2[0], v2[1], "y");
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    } else if (count == 0) {
		StrMake(errStr, errStrFmt[2], v1[0], v1[1], v2[0], v2[1], "y");
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	}
    }

    for (i = 0; i < len+4; i++) {
	if (AutoColorCompare(buf[(len+3)*(len+4)+i], COLOR_OFF) == GL_FALSE) {
	    StrMake(errStr, errStrFmt[0], v1[0], v1[1], v2[0], v2[1], "y");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }

    return NO_ERROR;
}

long LineRasterExec(void)
{
    GLftype *buf;
    GLint v1[2], v2[2], len, max, i;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
    max = (machine.pathLevel == 0) ? len : 1;
    for (i = 0; i < max; i++) {
	v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
	v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] + len;
	v2[1] = v1[1] + i;
        if (XLines(v1, v2, len, buf) == ERROR) {
	    FREE(buf);
	    return ERROR;
	}
    }
    len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
    max = (machine.pathLevel == 0) ? len : 1;
    for (i = 0; i < max; i++) {
        v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] + i;
	v2[1] = v1[1] + len;
        if (YLines(v1, v2, len, buf) == ERROR) {
	    FREE(buf);
	    return ERROR;
	}
    }
    len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
    max = (machine.pathLevel == 0) ? len : 1;
    for (i = 0; i < max; i++) {
        v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] - len;
	v2[1] = v1[1] + i;
        if (XLines(v1, v2, len, buf) == ERROR) {
	    FREE(buf);
	    return ERROR;
	}
    }
    len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
    max = (machine.pathLevel == 0) ? len : 1;
    for (i = 0; i < max; i++) {
        v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] - i;
	v2[1] = v1[1] + len;
        if (YLines(v1, v2, len, buf) == ERROR) {
	    FREE(buf);
	    return ERROR;
	}
    }
    len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
    max = (machine.pathLevel == 0) ? len : 1;
    for (i = 0; i < max; i++) {
        v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] - len;
	v2[1] = v1[1] - i;
        if (XLines(v1, v2, len, buf) == ERROR) {
	    FREE(buf);
	    return ERROR;
	}
    }
    len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
    max = (machine.pathLevel == 0) ? len : 1;
    for (i = 0; i < max; i++) {
        v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] - i;
	v2[1] = v1[1] - len;
        if (YLines(v1, v2, len, buf) == ERROR) {
	    FREE(buf);
	    return ERROR;
	}
    }
    len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
    max = (machine.pathLevel == 0) ? len : 1;
    for (i = 0; i < max; i++) {
        v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] + len;
	v2[1] = v1[1] - i;
        if (XLines(v1, v2, len, buf) == ERROR) {
	    FREE(buf);
	    return ERROR;
	}
    }
    len = FTOI(OESRandom(ITOF(3), ITOF(WINDSIZEX)/8));
    max = (machine.pathLevel == 0) ? len : 1;
    for (i = 0; i < max; i++) {
        v1[0] = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        v1[1] = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	v2[0] = v1[0] + i;
	v2[1] = v1[1] - len;
        if (YLines(v1, v2, len, buf) == ERROR) {
	    FREE(buf);
	    return ERROR;
	}
    }

    FREE(buf);
    return NO_ERROR;
}
