/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/logicop.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** logicop.c
** Logicop Test.
**
** Description -
**    Tests the functionality of logical operations. There are 16
**    logicop functions: GL_AND, GL_AND_INVERTED, GL_AND_REVERSE,
**    GL_CLEAR, GL_COPY, GL_COPY_INVERTED, GL_EQUIV, GL_INVERT,
**    GL_NAND, GL_NOOP, GL_NOR, GL_OR, GL_OR_INVERTED,
**    GL_OR_REVERSE, GL_SET and GL_XOR. Each one is tested by
**    choosing a random source and destination color, drawing a
**    point using the destination color with logicop disabled,
**    drawing the same point using the source color with logicop
**    enabled, then reading the resulting color and comparing it
**    to a calculated value.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Since logical operations are bitwise operations, no epsilon is used.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Shade, Stencil.
**        Not allowed = Alias, Dither, Logicop.
*/

#include "conform.h"
#include "util.h"


static char errStr[240];
static char errStrFmt[] = "Logic Op is %s. Source color channel is %d, destination color channel is %d. Result is %d, should be %d.";


static GLuint TrueResult(GLenum op, GLuint srcColor, GLuint destColor,
			   GLuint mask)
{
    GLuint result = 0;

    switch (op) {
      case GL_CLEAR:
	result = 0;
	break;
      case GL_AND:
	result = srcColor & destColor;
	break;
      case GL_AND_REVERSE:
	result = srcColor & (~destColor);
	break;
      case GL_COPY:
	result = srcColor;
	break;
      case GL_AND_INVERTED:
	result = (~srcColor) & destColor;
	break;
      case GL_NOOP:
	result = destColor;
	break;
      case GL_XOR:
	result = srcColor ^ destColor;
	break;
      case GL_OR:
	result = srcColor | destColor;
	break;
      case GL_NOR:
	result = ~(srcColor | destColor);
	break;
      case GL_EQUIV:
	result = ~(srcColor ^ destColor);
	break;
      case GL_INVERT:
	result = ~destColor;
	break;
      case GL_OR_REVERSE:
	result = srcColor | (~destColor);
	break;
      case GL_COPY_INVERTED:
	result = ~srcColor;
	break;
      case GL_OR_INVERTED:
	result = (~srcColor) | destColor;
	break;
      case GL_NAND:
	result = ~(srcColor & destColor);
	break;
      case GL_SET:
	result = ~0;
	break;
    }
    return result & mask;
}

static long Test(const GLftype pmax[4], GLenum op)
{
    char tmp[40];
    GLftype buf[4], max[4];
    GLuint trueResult[4], result[4], masks[4];
    GLint srcColor[4], destColor[4];
    GLint x, y, m[4];
    int i;

    /* Make sure scale factor max[i] is not zero, which leads to
     *  divide by zero errors in the glColor4t() calls below.
     */
    for (i = 0; i < 4; i++) {
	if (pmax[i] == 0)
	    max[i] = 1;
	else
	    max[i] = pmax[i];
    }

    glGetIntegerv(GL_RED_BITS, &m[0]);
    glGetIntegerv(GL_GREEN_BITS, &m[1]);
    glGetIntegerv(GL_BLUE_BITS, &m[2]);
    glGetIntegerv(GL_ALPHA_BITS, &m[3]);

    for (i = 0; i < 4; i++) {
	masks[i] = (1 << m[i]) - 1;
    }

    glDisable(GL_COLOR_LOGIC_OP);

    for (i = 0; i < 4; i++) {
	srcColor[i] = FTOI(OESRandom(ITOF(0), max[i]));
	destColor[i] = FTOI(OESRandom(ITOF(0), max[i]));
    }

    glColor4t(
	DIV(ITOF(srcColor[0]), max[0]), DIV(ITOF(srcColor[1]), max[1]),
	DIV(ITOF(srcColor[2]), max[2]), DIV(ITOF(srcColor[3]), max[3]));
    DrawPoint2f(ITOF(1)/2, ITOF(1)/2);
    glColor4t(
	DIV(ITOF(destColor[0]), max[0]), DIV(ITOF(destColor[1]), max[1]),
	DIV(ITOF(destColor[2]), max[2]), DIV(ITOF(destColor[3]), max[3]));
    DrawPoint2f(ITOF(3)/2, ITOF(1)/2);

    ReadScreen(0, 0, 1, 1, GL_RGBA, &buf[0]);
    for (i = 0; i < 4; i++) {
	srcColor[i] = FTOI(MUL(buf[i], ITOF(masks[i])) + ITOF(1)/2);
    }
    ReadScreen(1, 0, 1, 1, GL_RGBA, &buf[0]);
    for (i = 0; i < 4; i++) {
	destColor[i] = FTOI(MUL(buf[i], ITOF(masks[i])) + ITOF(1)/2);
    }

    for (i = 0; i < 4; i++) {
	trueResult[i] = TrueResult(op, srcColor[i], destColor[i], masks[i]);
    }

    x = FTOI(OESRandom(ITOF(0), ITOF(WINDSIZEX-1)));
    y = FTOI(OESRandom(ITOF(0), ITOF(WINDSIZEY-1)));

    glColor4t(
        DIV(ITOF(destColor[0]), max[0]), DIV(ITOF(destColor[1]), max[1]),
        DIV(ITOF(destColor[2]), max[2]), DIV(ITOF(destColor[3]), max[3]));
    DrawPoint2f(ITOF(x) + ITOF(1)/2, ITOF(y) + ITOF(1)/2);

    glEnable(GL_COLOR_LOGIC_OP);
    glLogicOp(op);

    glColor4t(
        DIV(ITOF(srcColor[0]), max[0]), DIV(ITOF(srcColor[1]), max[1]),
        DIV(ITOF(srcColor[2]), max[2]), DIV(ITOF(srcColor[3]), max[3]));
    DrawPoint2f(ITOF(x) + ITOF(1)/2, ITOF(y) + ITOF(1)/2);

    ReadScreen(x, y, 1, 1, GL_RGBA, &buf[0]);
    for (i = 0; i < 4; i++) {
	result[i] = FTOI(MUL(buf[i], ITOF(masks[i])) + ITOF(1)/2);
    }

    for (i = 0; i < 4; i++) {
	if (result[i] != trueResult[i]) {
	    GetEnumName(op, tmp);
	    StrMake(errStr, errStrFmt, tmp, srcColor[i], destColor[i], result[i],
		    trueResult[i]);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }
    return NO_ERROR;
}

long LogicOpExec(void)
{
    GLftype max[4];
    int i;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    for (i = 0; i < 4; i++) {
	max[i] = LDEXP(ITOF(1), buffer.colorBits[i]) - ITOF(1);
    }
    if (Test(max, GL_AND) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_AND_INVERTED) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_AND_REVERSE) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_CLEAR) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_COPY) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_COPY_INVERTED) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_EQUIV) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_INVERT) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_NAND) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_NOOP) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_NOR) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_OR) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_OR_INVERTED) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_OR_REVERSE) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_SET) == ERROR) {
	return ERROR;
    }
    if (Test(max, GL_XOR) == ERROR) {
	return ERROR;
    }
    return NO_ERROR;
}
