/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/mask.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** mask.c
** Mask Test.
**
** Description -
**    A point is rendered using a mask and a color.
**    The rendered result is read back and compared to the
**    expected result. The implementation fails if the color read
**    back differs from the expected color. The individual color
**    channels of any color used are either 0.0 or 1.0. All
**    possible combinations of color channels are tested, each
**    using all possible masks.
**    
** Technical Specification -
**    Buffer requirements:
**        No buffer requirements.
**    Color requirements:
**        No color requirements.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        epsilon.zero
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"

#define REDBIT   1
#define GREENBIT 2
#define BLUEBIT  4
#define ALPHABIT 8

static GLftype errColorExpected[4], errColorObserved[4];
static char errStr[160];
static char errStrFmt[80] =
    "Color expected = (%g %g %g %g), color observed = (%g %g %g %g). Mask %x.";

static void SaveErrColor(GLftype from[4], GLftype to[4])
{

    to[0] = from[0];
    to[1] = from[1];
    to[2] = from[2];
    to[3] = from[3];
}

static void Draw(long componentMask, long RGBMask, GLftype shade)
{
    GLboolean rMask, gMask, bMask, aMask;
    GLftype color[4];

    glClear(GL_COLOR_BUFFER_BIT);

    color[0] = (componentMask & REDBIT) ? shade : ITOF(0);
    color[1] = (componentMask & GREENBIT) ? shade : ITOF(0);
    color[2] = (componentMask & BLUEBIT) ? shade : ITOF(0);
    color[3] = (componentMask & ALPHABIT) ? shade : ITOF(0);
    glColor4t(color[0], color[1], color[2], color[3]);

    rMask = (REDBIT & RGBMask) ? GL_TRUE : GL_FALSE;
    gMask = (GREENBIT & RGBMask) ? GL_TRUE : GL_FALSE;
    bMask = (BLUEBIT & RGBMask) ? GL_TRUE : GL_FALSE;
    aMask = (ALPHABIT & RGBMask) ? GL_TRUE : GL_FALSE;
    glColorMask(rMask, gMask, bMask, aMask);

    DrawPoint2f(ITOF(1)/2, ITOF(1)/2);

    SaveErrColor(color, errColorExpected);
}

static long Check(long drawn, GLftype color)
{

    if (drawn == GL_FALSE) {
	if (ABS(color) > epsilon.zero) {
	    return ERROR;
	}
    } else {
	if (ABS(color) < epsilon.zero) {
	    return ERROR;
	}
    }
    return NO_ERROR;
}

static long Test(long componentMask, long RGBMask)
{
    GLftype buf[4];
    long i, shouldDrawMask;

    ReadScreen(0, 0, 1, 1, GL_RGBA, buf);

    shouldDrawMask = componentMask & RGBMask;

    for (i = 0; i < 4; i++) {
	if (buffer.colorBits[i] == 0) {
            if (ABS(buf[i]-(i == 3 ? ITOF(1) : ITOF(0))) > epsilon.zero) {
        	errColorExpected[i] = (i == 3 ? ITOF(1) : ITOF(0));
        	SaveErrColor(buf, errColorObserved);
        	return ERROR;
            } else {
        	return NO_ERROR;
            }
	} else if (Check(shouldDrawMask&(1 << i), buf[i]) == ERROR) {
            SaveErrColor(buf, errColorObserved);
            return ERROR;
	}
    }

    return NO_ERROR;
}

long MaskExec(void)
{
    long componentMask, RGBMask;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    for (componentMask = 0; componentMask < 16; componentMask++) {
	for (RGBMask = 0; RGBMask < 16; RGBMask++) {
	    Draw(componentMask, RGBMask, ITOF(1));
	    if (Test(componentMask, RGBMask) == ERROR) {
		StrMake(errStr, errStrFmt, errColorExpected[0], 
			errColorExpected[1], errColorExpected[2], 
			errColorExpected[3], errColorObserved[0], 
			errColorObserved[1], errColorObserved[2], 
			errColorObserved[3], RGBMask);
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	}
    }
    return NO_ERROR;
}
