/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
**
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/mget.c,v 1.4 2005/06/03 08:37:40 artlab Exp $
*/
 

/**
 * \file
 * \brief	Matrix get test.
 * \author	petri@hybrid.fi
 *//*-------------------------------------------------------------------*/

/*
** mget.c
** Matrix Get Test.
**
** Description -
**    Tests the functionality of glGetFloat*()/glGetFixed*() with the
**    three matrix queries (GL_MODELVIEW_MATRIX, GL_PROJECTION_MATRIX
**    and GL_TEXTURE_MATRIX). All the different texture units are also
**    tested.
**
**    The test consists of two sub-tests: matrix load test and matrix
**    multiply test.
**
**    In the matrix load tests, random matrices are generated and then
**    loaded into OpenGL ES after which they are queried back and
**    compared against the original values. If all the values are inside
**    the error epsilon, the test passes and otherwise it fails. This
**    operation is then repeated with a new random matrix.
**
**    The matrix multiply test works somewhat like the matrix load test
**    except that after loading one random matrix, it generates another
**    one and uses glMultMatrix*() to multiply the two matrices. The
**    same operation is also computed in the test and then the results
**    are compared as in the previous test.
**
** Error Explanation -
**    The difference between the loaded and queried values must not
**    be greater than zero epsilon times the reference value. If an
**    overflowing value is generated as a result of the matrix
**    multiplication, the given value is ignored and thus cannot fail
**    the test.
**
** Technical Specification -
**    Buffer requirements:
**        No buffer requirements.
**    Color requirements:
**        No color requirements.
**    States requirements:
**        No state restrictions.
**    Error epsilon:
**        Zero epsilon.
**    Paths:
**       Allowed = Alias, Alpha, Blend, Depth, Dither, Fog, Logicop, Shade,
**                 Stencil.
*/

#include "conform.h"
#include "util.h"

#ifdef GL_OES_VERSION_1_1

static char errStr[320];
static char *errStrFmt[2] =
{
    "Matrix mode is %s.\nLoad test failed.\nMatrix element at %d is %g, should be %g",
    "Matrix mode is %s.\nMultiply test failed.\nMatrix element at %d is %g, should be %g"
};

/*-------------------------------------------------------------------*//*!
 * \brief	Generate a random matrix.
 * \param	mtx	Pointer to matrix.
 *//*-------------------------------------------------------------------*/

static void matrixRandomize (GLftype mtx[16])
{
	int i;
	for (i = 0; i < 16; i++)
	{
		GLftype f;

		switch(RAND() % 6)
		{
			case 0:
				f = ITOF(0);
				break;

			case 1:
				f = ITOF(1);
				break;

#if defined(COMMON)
			case 2: /* [2^-30..2^30] */
				f = MUL(OESRandom(ITOF(0), ITOF(1)), POW(ITOF(2), OESRandom(ITOF(-30), ITOF(30))));
				break;
#endif

			default: /* [0..65535] */
				f = OESRandom(ITOF(0), ITOF(65535));
				break;
		}

		/* Use positive values 67% of the time. */
		if ((RAND() % 3) == 0)
			f = -f;

		mtx[i] = f;
	}
}

/*-------------------------------------------------------------------*//*!
 * \brief	Epsilon comparison of two values.
 * \param	a		Pointer to first matrix.
 * \param	b		Pointer to second matrix.
 * \param	epsilon	Pointer to matrix containing epsilon values.
 * \return	Returns ERROR if values are not within the epsilon of each other.
 *//*-------------------------------------------------------------------*/

static int valueCompare (const GLftype a, const GLftype b, GLftype epsilon)
{
	if (ABS(a - b) > epsilon)
		return ERROR;

	return NO_ERROR;
}

/*-------------------------------------------------------------------*//*!
 * \brief	Scale matrix with a constant.
 * \param	dst		Pointer to destination matrix.
 * \param	src		Pointer to source matrix.
 * \param	scale	Scale value.
 *//*-------------------------------------------------------------------*/

static void matrixScale (GLftype dst[16], const GLftype src[16], GLftype scale)
{
	int i;
	for (i = 0; i < 16; i++)
		dst[i] = MUL(src[i], scale);
}

/*-------------------------------------------------------------------*//*!
 * \brief	Converts a matrix to epsilon matrix.
 * \param	dst		Pointer to destination matrix.
 * \param	src		Pointer to source matrix.
 *
 * The original matrix is scaled by the EPSILON_SCALE and then each
 * component is clamped to a minimum of EPSILON.
 *//*-------------------------------------------------------------------*/

static void matrixToEpsilon (GLftype dst[16], const GLftype src[16])
{
	int i;

	matrixScale(dst, src, DIV(ITOF(1), CalcEpsilon(15)));

	for (i = 0; i < 16; i++)
	{
		GLftype f = ABS(dst[i]);
		if (f < epsilon.zero)
			f = epsilon.zero;
		dst[i] = f;
	}
}

/*-------------------------------------------------------------------*//*!
 * \brief	Matrix multiply.
 * \param	dst		Pointer to destination matrix.
 * \param	a		Pointer to first matrix.
 * \param	b		Pointer to second matrix.
 *//*-------------------------------------------------------------------*/

static void matrixMultiply (GLftype dst[16], const GLftype a[16], const GLftype b[16])
{
	int i;
	for (i = 0; i < 16; i++)
	{
		GLftype	v = ITOF(0);
		int		j;

		for (j = 0; j < 4; j++)
			v += MUL(a[(j<<2) + (i&3)], b[(i&(3<<2)) + j]);

		dst[i] = v;
	}
}

/*-------------------------------------------------------------------*//*!
 * \brief	Epsilon matrix multiply.
 * \param	dst		Pointer to destination matrix.
 * \param	a		Pointer to first matrix.
 * \param	b		Pointer to second matrix.
 *
 * Performs a matrix multiply which sums the absolute values of the terms
 * of the dot product, thus computing the scale of each result value.
 *//*-------------------------------------------------------------------*/

static void matrixMultiplyEpsilon (GLftype dst[16], const GLftype a[16], const GLftype b[16])
{
	int i;
	for (i = 0; i < 16; i++)
	{
		GLftype	v = 0.0f;
		int		j;

		for (j = 0; j < 4; j++)
			v += ABS(MUL(a[(j<<2) + (i&3)], b[(i&(3<<2)) + j]));

		dst[i] = v;
	}
}

/*-------------------------------------------------------------------*//*!
 * \brief	Test matrix getters.
 *//*-------------------------------------------------------------------*/

long MatrixGetTestExec (void)
{
	int		count = (machine.pathLevel == 0) ? 1000 : 100;
	GLint	numTextureUnits;

	glGetIntegerv(GL_MAX_TEXTURE_UNITS, &numTextureUnits);

	while (count--)
	{
		GLenum	matrixMode;
		GLenum	matrixGet;
		char	matrixName[40];

		switch (RAND() % 3)
		{
			default:
			case 0:
				matrixMode	= GL_MODELVIEW;
				matrixGet	= GL_MODELVIEW_MATRIX;
				break;

			case 1:
				matrixMode	= GL_PROJECTION;
				matrixGet	= GL_PROJECTION_MATRIX;
				break;

			case 2:
				matrixMode	= GL_TEXTURE;
				matrixGet	= GL_TEXTURE_MATRIX;

				glActiveTexture(GL_TEXTURE0 + (RAND() % numTextureUnits));
				break;
		}

		glMatrixMode(matrixMode);
		GetEnumName(matrixMode, matrixName);

		/*---------------------------------------------------------------------
		 * LOAD TEST.
		 * 1. Generate a random matrix.
		 * 2. Load the matrix to ES implementation (using glLoadMatrix()).
		 * 3. Query the matrix back.
		 * 4. Compare against the original matrix (using epsilon matrix).
		 *-------------------------------------------------------------------*/
		{
			GLftype	mtxOrig[16];
			GLftype	mtxTest[16];
			GLftype	epsilon[16];
			int		i;

			matrixRandomize(mtxOrig);
			matrixToEpsilon(epsilon, mtxOrig);

			glLoadMatrixt(mtxOrig);
			glGetFloatvt(matrixGet, mtxTest);

			for (i = 0; i < 16; i++)
			{
				if (valueCompare(mtxOrig[i], mtxTest[i], epsilon[i]) == ERROR)
				{
					StrMake(errStr, errStrFmt[0], matrixName, i, mtxTest[i], mtxOrig[i]);
					ErrorReport(__FILE__, __LINE__, errStr);
					return ERROR;
				}
			}
		}

		/*---------------------------------------------------------------------
		 * MULTIPLY TEST.
		 * 1. Generate two random matrices.
		 * 2. Compute the reference matrix multiply and epsilon matrix.
		 * 3. Load the first matrix to ES (using glLoadMatrix()).
		 * 4. Multiply the second matrix in ES (using glMultMatrix()).
		 * 5. Query the matrix back.
		 * 6. Compare reference and ES multiply results (using epsilon matrix).
		 *-------------------------------------------------------------------*/
		{
			GLftype	mtxA[16];
			GLftype	mtxB[16];
			GLftype	mtxMult[16];
			GLftype	mtxTest[16];
			GLftype	epsilon[16];
			int		i;

			matrixRandomize(mtxA);
			matrixRandomize(mtxB);

			matrixMultiply(mtxMult, mtxA, mtxB);
			matrixMultiplyEpsilon(epsilon, mtxA, mtxB);
			matrixToEpsilon(epsilon, epsilon);

			glLoadMatrixt(mtxA);
			glMultMatrixt(mtxB);
			glGetFloatvt(matrixGet, mtxTest);

			for (i = 0; i < 16; i++)
			{
#if !defined(COMMON)
				/* don't let overflows break common lite implementations */
				if (ABS(mtxTest[i]) > ITOF(65535))
					continue;
#endif

				if (valueCompare(mtxMult[i], mtxTest[i], epsilon[i]) == ERROR)
				{
					StrMake(errStr, errStrFmt[1], matrixName, i, mtxTest[i], mtxMult[i]);
					ErrorReport(__FILE__, __LINE__, errStr);
					return ERROR;
				}
			}
		}
	}

	return NO_ERROR;
}

#endif /* GL_OES_VERSION_1_1 */
