/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)()(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
*/

/*
** miplevels.c
** Mipmap Levels Test.
**
** Description -
**    Valid Draw Test:
**      Creates a set of mipmaps with random contents
**      Calls glTexImage2D a variety of times to
**         complete the texture (i.e. sets all levels)
**         The calls are made in different orders
**      Draws a quad at each mipmap level resolution
**         and checks the mipmap level is drawn correctly
**    Invalid Draw Test:
**      Creates a set of mipmaps with random contents
**      Calls glTexImage2D a variety of times but never
**         completes the texture (i.e. never sets all 
**         levels or mixes up formats)
**      Attempts to draw a quad and expects the draw to
**         occur as if texturing is DISABLED
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        None.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color epsilon. Zero epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
**	  
**	  
*/

#include "conform.h"
#include "util.h"
#include "utilm.h"

/* Defines the number of mipmap levels to create and test for */
#define MIPDATA_LEVEL_NO 6

/* Holds the order textures are bound in for error reports.
 * The maximum number of binds is seen by the ValidDrawTest
 * that binds every level, then up to half of the levels are
 * bound to different formats, then the different formats are
 * rebound to the original format.
 */
#define BINDORDER_SIZE 32*MIPDATA_LEVEL_NO*2
static char BindOrder[BINDORDER_SIZE];

/* Holds the GL reference to the texture used */
static GLenum tex;

/* Structure for holding test data for all possible format/type combos */
typedef struct 
{
    /* Holds the GL format */
    GLenum format;

    /* Holds the GL type */
    GLenum type;

    /* Holds the number of components for that format */
    int components;

    /* Holds the number of bytes required to store that type */
    int byteSize;

    /* Holds the index of the alpha component, or -1 if there is none */
    int alphaComponent;

    /* Holds the processed floating point data for comparison */
    GLftype *fData;

} MipMapTestData; 

/* Holds the format/type of each level which is set 
 * after a glTexImage call. This is so it can be 
 * compared correctly when drawn
 */
static int *levelCombos;

/* Holds the raw test data used */
static int *mmData;

/* Holds the floating point version of the test data for all GL_UNSIGNED_BYTE comparisons */
static GLftype *fData;

/* Sets up the possible combinations with components and sizes */
static MipMapTestData testData[] = { {GL_ALPHA,           GL_UNSIGNED_BYTE,          1, 1,  0, 0},
                                     {GL_LUMINANCE,       GL_UNSIGNED_BYTE,          1, 1, -1, 0},
                                     {GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE,          2, 2,  1, 0},
                                     {GL_RGB,             GL_UNSIGNED_BYTE,          3, 3, -1, 0},
                                     {GL_RGBA,            GL_UNSIGNED_BYTE,          4, 4,  3, 0},
                                     {GL_RGB,             GL_UNSIGNED_SHORT_5_6_5,   3, 2, -1, 0},
                                     {GL_RGBA,            GL_UNSIGNED_SHORT_4_4_4_4, 4, 2,  3, 0},
                                     {GL_RGBA,            GL_UNSIGNED_SHORT_5_5_5_1, 4, 2,  3, 0} };

/* Sets up the offsets required for the different mipmap levels */
static int mipMapOffsets[] = {0, 1, 5, 21, 85, 341, 1365, 5461, 21845};

/* We need to know the alpha size for the current config to know if we check
 * the alpha value equal to the test data's alpha value (if the config has
 * an alpha component) or equal to 1 (if the config does not have an alpha
 * component).
 */
static int alphaSize;

/* Get a random int from lo to hi */
static GLint OESRandomInt(GLint lo, GLint hi)
{
    GLint x = FTOI(OESRandom(ITOF(lo),ITOF(hi+1)));
    if (x > hi) x = hi;
    return x;
}

/* Generates random data for mipmap levels for all format/types */
static void SetupTestData(void)
{
    int Count_A=0, Count_B;
    int dataSize, dataElementSize;
    int i;

    /* Find the largest size data element used */
    dataElementSize = testData[0].byteSize;
    for (i = 1; i < sizeof(testData) / sizeof(testData[0]); i++)
        if (testData[i].byteSize > dataElementSize)
            dataElementSize = testData[i].byteSize;

    /* Create the random data and floating 
     * point version of it (for GL_UNSIGNED_BYTE)
     */
    dataSize = mipMapOffsets[MIPDATA_LEVEL_NO] * dataElementSize;
    mmData = (int *)MALLOC(dataSize * sizeof(GLubyte));
    fData = (GLftype *)MALLOC(dataSize * sizeof(GLftype));
    for (Count_A=0; Count_A<dataSize; Count_A++)
    {
        ((GLubyte *)mmData)[Count_A] = (GLubyte)OESRandomInt(0,255);
        fData[Count_A] = DIV(ITOF(((int)((GLubyte *)mmData)[Count_A])),ITOF(255));
    }

    /* Assign the floating point data to the test data, making any 
     * necessary conversions
     */
    for (i = 0; i < sizeof(testData) / sizeof(testData[0]); i++)
    {
        if (testData[i].type == GL_UNSIGNED_BYTE)
        {
            /* Already have the floating point data, no conversion necessary */
            testData[i].fData = fData;
            continue;
        }

        /* Allocate the memory for the custom floating point representation */
        dataSize = testData[i].components * mipMapOffsets[MIPDATA_LEVEL_NO];
        testData[i].fData = (GLftype *)MALLOC(dataSize * sizeof(GLftype));

        switch(testData[i].type)
        {
            case GL_UNSIGNED_SHORT_5_6_5:
                Count_B = 0;
                for (Count_A=0; Count_A<dataSize; Count_A+=3)
                {
                    testData[i].fData[Count_A]   = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0xF800) >> 8), ITOF(255));
                    testData[i].fData[Count_A+1] = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0x07E0) >> 3), ITOF(255));
                    testData[i].fData[Count_A+2] = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0x001F) << 3), ITOF(255));
                    Count_B++;
                }
                break;
            case GL_UNSIGNED_SHORT_4_4_4_4:    
                Count_B = 0;
                for (Count_A=0; Count_A<dataSize; Count_A+=4)
                {
                    testData[i].fData[Count_A]   = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0xF000) >> 8), ITOF(255));
                    testData[i].fData[Count_A+1] = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0x0F00) >> 4), ITOF(255));
                    testData[i].fData[Count_A+2] = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0x00F0)     ), ITOF(255));
                    testData[i].fData[Count_A+3] = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0x000F) << 4), ITOF(255));
                    Count_B++;
                }
                break;
            case GL_UNSIGNED_SHORT_5_5_5_1:
                Count_B = 0;
                for (Count_A=0; Count_A<dataSize; Count_A+=4)
                {
                    testData[i].fData[Count_A]   = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0xF800) >> 8), ITOF(255));
                    testData[i].fData[Count_A+1] = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0x07C0) >> 3), ITOF(255));
                    testData[i].fData[Count_A+2] = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0x003E) << 2), ITOF(255));
                    testData[i].fData[Count_A+3] = DIV(ITOF((((GLushort *)mmData)[Count_B] & 0x0001) << 7), ITOF(255));
                    Count_B++;
                }
                break;
        }
    }

    /* Ignore what component is the alpha component if the render target
     * has an alpha component.  We only skip the alpha component if the
     * render target has no alpha.  So here we need to find if the render
     * target has an alpha component (alphaSize > 0).
     */
    alphaSize = buffer.colorBits[3];
}

/* Sets up the mipmap level with the relevant data for the format/type/level */
static void BindMipMap(int size, int level, int totalLevels, int combo)
{
    int mipSize, dataLevel;
    char buf[100];

    dataLevel = totalLevels - level - 1;
    if (strlen(BindOrder) == 0)
        sprintf(buf, "%i {0x%x,0x%x}", level, testData[combo].format, testData[combo].type);
    else
        sprintf(buf, ", %i {0x%x,0x%x}", level, testData[combo].format, testData[combo].type);
    strcat(BindOrder, buf);   
    
    mipSize = size >> level;
    
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexImage2D(GL_TEXTURE_2D, level, testData[combo].format, mipSize, mipSize, 0, testData[combo].format, testData[combo].type, &((GLubyte *)mmData)[mipMapOffsets[dataLevel] * testData[combo].byteSize]);

    /* Save the format to check later */
    levelCombos[level] = combo;
}

/* Check the texture is drawn correctly */
static long CheckDrawResult(int drawSize, int level, int totalLevels)
{
    int i, j, screenSize, dataLevel, combo;
    GLenum format, type;
    GLsizei texBits[4];
    GLftype texEpsilon[4];
    GLftype *screenData, *texData;
    int noAlpha;
    GLftype componentVal, epsilonVal;

    combo = levelCombos[level];
    format = testData[combo].format;
    type = testData[combo].type;

    dataLevel = totalLevels - level - 1;

    GetComponentBits(format, type, texBits);
    CalcTextureEpsilon(texBits, texEpsilon);
  	SwizzleTextureEpsilon(format, type, texEpsilon);
    
    screenSize = drawSize * drawSize;
    
    screenData = (GLftype *)MALLOC(screenSize * testData[combo].components * sizeof(GLftype));
    ReadScreen(0,0,drawSize,drawSize,format,screenData);
    texData = &testData[combo].fData[mipMapOffsets[dataLevel] * testData[combo].components];

    noAlpha = (testData[combo].alphaComponent != -1 && alphaSize == 0);
    
    for (i=0; i<screenSize; i++)
        for (j=0; j<testData[combo].components; j++)
        {
            /* If we're checking the alpha component on a config that does
             * not have an alpha component, this component should be 1.
             */
            if (noAlpha && j == testData[combo].alphaComponent)
            {
                componentVal = ITOF(1);
                epsilonVal = epsilon.zero;
            }
            else
            {
                componentVal = texData[i * testData[combo].components + j];
                epsilonVal = texEpsilon[j];
            }

            if ( ABS(SUB(screenData[i * testData[combo].components + j],componentVal)) > epsilonVal )
            {
                FREE(screenData);
                return ERROR;
            }
        }
    
    FREE(screenData);
    return NO_ERROR;
}

/* Checks that no textured draw has occured */
static long CheckNoDrawResult(int drawSize, int level)
{
    int i, success, screenSize;
    GLftype *screenData;

    screenSize = drawSize * drawSize * 4;
    
    screenData = (GLftype *)MALLOC(screenSize * sizeof(GLftype));
    ReadScreen(0,0,drawSize,drawSize,GL_RGBA,screenData);
    
    for (i=0; i<screenSize; i+=4)
    {
        success = 1;
        
        /* Expect the screen to be all white (due to draw with no texture) */
        success &= ABS(SUB(screenData[i  ], ITOF(1)))      < epsilon.zero;
        success &= ABS(SUB(screenData[i+1], ITOF(1)))      < epsilon.zero;
        success &= ABS(SUB(screenData[i+2], ITOF(1)))      < epsilon.zero;
        success &= ABS(SUB(screenData[i+3], ITOF(1)))      < epsilon.zero;
        
        if (!success)
        {
            FREE(screenData);
            return ERROR;
        }
    }
    
    FREE(screenData);
    return NO_ERROR;
}

/* Frees all the allocated memory in the test */
static void FreeTextureData(void)
{
    int combo;
    for (combo=0; combo<sizeof(testData)/sizeof(testData[0]); combo++)
        if(testData[combo].type != GL_UNSIGNED_BYTE)
            FREE(testData[combo].fData);

    FREE(mmData);
    FREE(fData);
}

/* Sets a few GL settings */
static void SetupGL(void)
{
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    
    glEnable(GL_TEXTURE_2D);
}

/* Produces an error string relevant for the fail */
static void CreateErrString(char* errStr, int mode, int texSize, int failLevel, int totalLevels, int combo)
{
    switch (mode)
    {
    case 0: /* Failed on invalid draw (incomplete levels) */
        sprintf(errStr, "Expecting untextured draw due to incomplete texture (incomplete levels). Textured draw occured. Square Tex Size:%i. Mipmap level failure: %i. Total mipmap level no: %i. Format: 0x%x. Type: 0x%x. Bind Order: %s.", texSize, failLevel, totalLevels, testData[combo].format, testData[combo].type, BindOrder);
        break;
    case 1: /* Failed on invalid draw (incompatible levels) */
        sprintf(errStr, "Expecting untextured draw due to incomplete texture (incompatible level formats). Textured draw occured. Square Tex Size:%i. Mipmap level failure: %i. Total mipmap level no: %i. Format: 0x%x. Type: 0x%x. Bind Order: %s.", texSize, failLevel, totalLevels, testData[combo].format, testData[combo].type, BindOrder);
        break;
    case 2: /* Failed on valid draw (ascending mip levels) */
        sprintf(errStr, "Textured draw unsuccessful for mipmap levels defined in ascending order. Square Tex Size:%i. Mipmap level failure: %i. Total mipmap level no: %i. Format: 0x%x. Type: 0x%x. Bind Order: %s.", texSize, failLevel, totalLevels, testData[combo].format, testData[combo].type, BindOrder);
        break;
    case 3: /* Failed on valid draw (descending mip levels) */
        sprintf(errStr, "Textured draw unsuccessful for mipmap levels defined in descending order. Square Tex Size:%i. Mipmap level failure: %i. Total mipmap level no: %i. Format: 0x%x. Type: 0x%x. Bind Order: %s.", texSize, failLevel, totalLevels, testData[combo].format, testData[combo].type, BindOrder);
        break;
    case 4: /* Failed on valid draw (random mip levels) */
        sprintf(errStr, "Textured draw unsuccessful for mipmap levels defined in random order. Square Tex Size:%i. Mipmap level failure: %i. Total mipmap level no: %i. Format: 0x%x. Type: 0x%x. Bind Order: %s.", texSize, failLevel, totalLevels, testData[combo].format, testData[combo].type, BindOrder);
        break;
    }
}

long InvalidDrawTest(int errType, int mipLevelNo, int combo)
{   
    
    int texSize, level;
    int randLevel, randCombo;
    int completeFlags, incompleteFlags, mipLevelFlags;
    int mipSize;
    char errStr[1024 + BINDORDER_SIZE];

    /*******************
     * SETUP TEST DATA *
     *******************/
    
    glGenTextures(1, &tex);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex);
    
    texSize = 1 << (mipLevelNo-1);
    
    levelCombos = (int *)MALLOC(mipLevelNo * sizeof(int));
    
    for (level=0; level<mipLevelNo; level++)
        levelCombos[level] = -1;
    
    /*******************
     * FILL IN MIPMAPS *
     *******************/

    strcpy(BindOrder, "");
    if (errType == 0)
    {
        completeFlags = (1 << mipLevelNo) - 1; /* Flags for all levels */
        incompleteFlags = OESRandomInt(0, completeFlags-1); /* Target flags */
        mipLevelFlags = 0;
        
        /* Randomly pick mip levels to set up until target set up is reached */
        while (mipLevelFlags < incompleteFlags)
        {
            randLevel = OESRandomInt(0,mipLevelNo-1);
            
            if ((incompleteFlags & (1 << randLevel)) &&
                !(mipLevelFlags & (1 << randLevel)))
            {
                mipLevelFlags |= 1 << randLevel;
                
                BindMipMap(texSize, randLevel, mipLevelNo, combo);
            }
            else
            {
                /* If the randomly picked level is already bound, just do the
                 * lowest level next so that this loop is only iterated through
                 * a finite number of times
                 */

                /* Find the levels that have not yet been bound */
                int notLoaded = incompleteFlags - mipLevelFlags;

                /* Find the lowest level number */
                for (randLevel = 0; randLevel < mipLevelNo; randLevel++)
                    if (notLoaded & (1 << randLevel))
                        break;

                mipLevelFlags |= 1 << randLevel;

                BindMipMap(texSize, randLevel, mipLevelNo, combo);
            }
        }
    }
    else if (errType == 1)
    {
        int numCombos = sizeof(testData) / sizeof(testData[0]);
        /* Load any level with a combo other than the given one */
        randLevel = OESRandomInt(1,mipLevelNo-1);
        /* Take from the upper half 50% of the time */
        if ((combo < numCombos-1 && OESRandomInt(0,1)) || combo == 0) 
            randCombo = OESRandomInt(combo+1, numCombos-1);
        else 
            randCombo = OESRandomInt(0, combo-1);

        /* Load all the levels with the given combo except the random one */
        for (level=0; level<mipLevelNo; level++)
            BindMipMap(texSize, level, mipLevelNo, (level==randLevel) ? randCombo : combo);
    }
    
    /*************
     * TEST DRAW *
     *************/
    
    mipSize = texSize;
    for (level=0; level<mipLevelNo; level++)
    {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glViewport(0,0,mipSize,mipSize);
        
        DrawTexQuad2i2f(-1,-1,1,1, ITOF(0),ITOF(0),ITOF(1),ITOF(1));
        
        /* skip if level combo is -1 */
        if(levelCombos[level] != -1)
        {
            if (CheckNoDrawResult(mipSize, level) == ERROR)
            {
                CreateErrString(errStr, errType, texSize, level, mipLevelNo, combo);
                ErrorReport(__FILE__, __LINE__, errStr);

                goto MIPMAP_INVAL_ERR;
            }
        }

        mipSize >>= 1;
    }
    
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, &tex);
    FREE(levelCombos);

    return NO_ERROR;
    
MIPMAP_INVAL_ERR:
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, &tex);   
    FREE(levelCombos);

    return ERROR;
}

/*
** Test for a valid texture draw
**
** mipMapOrder - defines which order the mipmap levels are sent
**                0=ascending, 1=descending, 2=overlapping 3=randomly
** format - defines the format to be used for all mipmap levels
** type - defines the type to be used for all mipmap levels
**        or 0 (which means use a random type for each level)
*/
static long ValidDrawTest(int mipMapOrder, int mipLevelNo, int combo)
{   
    int texSize, level;
    int randLevel, randCombo;
    int mipLevelFlags, expectedValue;
    int changedLevelNo, i;
    int mipSize;
    char errStr[1024 + BINDORDER_SIZE];

    /*******************
     * SETUP TEST DATA *
     *******************/
    
    glGenTextures(1, &tex);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex);
    
    texSize = 1 << (mipLevelNo-1);
    
    levelCombos = (int *)MALLOC(mipLevelNo * sizeof(int));
    
    for (level=0; level<mipLevelNo; level++)
        levelCombos[level] = -1;
    
    /*******************
     * FILL IN MIPMAPS *
     *******************/

    strcpy(BindOrder, "");
    if (mipMapOrder == 0) /* Ascending Order */
    {
        for (level=0; level<mipLevelNo; level++)
        {
        	BindMipMap(texSize, level, mipLevelNo, combo);
            
        }
    }
    else if (mipMapOrder == 1) /* Descending Order */
    {
        for (level=mipLevelNo-1; level>=0; level--)
        {
        	BindMipMap(texSize, level, mipLevelNo, combo);
            
        }
    }
    else if (mipMapOrder == 2) /* Overlapping Order */
    {
        /* Define all the mip levels as required */
        for (i=0; i<mipLevelNo; i++)
            BindMipMap(texSize, i, mipLevelNo, combo);

        /* Then change some of them to incorrect formats/types
         * (Divide by 2 because this will limit the number of bind
         * operations to maximum 2*mipLevelNo)
         */
        changedLevelNo = OESRandomInt(0, mipLevelNo/2);
        for (i=0; i<changedLevelNo; i++)
        {
            randLevel = OESRandomInt(0,mipLevelNo-1);
            randCombo = OESRandomInt(0, 7);

            BindMipMap(texSize, randLevel, mipLevelNo, randCombo);
        }

        /* Then fix the incorrect levels */
        for (level=0; level<mipLevelNo; level++)
        {
            if (levelCombos[level] != combo)
            {
                BindMipMap(texSize, level, mipLevelNo, combo);
            }
        }

    }
    else /* Random order */
    {
        mipLevelFlags = 0;
        expectedValue = (1 << mipLevelNo) - 1;
        
        /* Randomly pick mip levels to set up until target set up is reached */
        while (mipLevelFlags < expectedValue)
        {
            randLevel = OESRandomInt(0,mipLevelNo-1);
            
            if (!(mipLevelFlags & (1 << randLevel)))
            {
                mipLevelFlags |= 1 << randLevel;
                
                BindMipMap(texSize, randLevel, mipLevelNo, combo);
            }
            else
            {
                /* If the randomly picked level is already bound, just do the
                 * lowest level next so that this loop is only iterated through
                 * a finite number of times
                 */

                /* Find the lowest level number */
                for (randLevel = 0; randLevel < mipLevelNo; randLevel++)
                    if (!(mipLevelFlags & (1 << randLevel)))
                        break;

                mipLevelFlags |= 1 << randLevel;

                BindMipMap(texSize, randLevel, mipLevelNo, combo);
            }
        }
    }
    
    /*************
     * TEST DRAW *
     *************/
    
    /* Draw each mipmap level and check it is correct */
    mipSize = texSize;
    for (level=0; level<mipLevelNo; level++)
    {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glViewport(0,0,mipSize,mipSize);
        
        DrawTexQuad2i2f(-1,-1,1,1, ITOF(0),ITOF(0),ITOF(1),ITOF(1));
        
        if (CheckDrawResult(mipSize, level, mipLevelNo) == ERROR)
        {
            CreateErrString(errStr, mipMapOrder+2, texSize, level, mipLevelNo, combo);
            ErrorReport(__FILE__, __LINE__, errStr);

            goto MIPMAP_VAL_ERR;
        }
        
        mipSize >>= 1;
    }
    
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, &tex);
    FREE(levelCombos);

    return NO_ERROR;
    
MIPMAP_VAL_ERR:
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, &tex);
    FREE(levelCombos);

    return ERROR;
}

long MipLevelsExec(void)
{
    int errType, mmOrder, mipLevelNo, combo;

    SetupGL();
    SetupTestData();
    
    /*********************
      INVALID DRAW TESTS
    *********************/
    
    for (errType=0; errType<2; errType++)
        for (combo=0; combo<4; combo++)
            if (InvalidDrawTest(errType,MIPDATA_LEVEL_NO-1,combo) == ERROR)
                goto MIPLEVELS_ERROR;
            
    /*******************
      VALID DRAW TESTS
    *******************/
    
    for (mipLevelNo=1; mipLevelNo<MIPDATA_LEVEL_NO; mipLevelNo++)
        for (mmOrder=0; mmOrder<4; mmOrder++)
            for (combo=0; combo<8; combo++)
                if (ValidDrawTest(mmOrder,mipLevelNo,combo) == ERROR)
                    goto MIPLEVELS_ERROR;

    FreeTextureData();
    return NO_ERROR;
MIPLEVELS_ERROR:

    FreeTextureData();
    return ERROR;
}

#undef MIPDATA_LEVEL_NO
