/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/mipsel.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** mipsel.c
** Mipmap Selection Test.
**
** Description -
**    Tests the selection of levels with various magnification
**    and minification filters. A seven level deep mipmap is set
**    up, each level distinguished by a different color. A series
**    of shrinking square polygons are drawn with texture
**    coordinates set so that each polygon uses a different level
**    of the mipmap. This is done by matching the number of
**    fragments of a particular polygon to the number of texels
**    in the corresponding level. The rendered polygon is read
**    back and compared to the correct mipmap level color.
**
** Technical Specification -
**    Buffer requirements:
**	  Color buffer.
**    Color requirements:
**	  Full color.
**    States requirements:
**	  Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**    Paths:
**	  Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**	  Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilt.h"


static long errSizeX, errSizeY;
static char errMagFilter[40], errMinFilter[40];
static char errStr[240];
static char errStrFmt[] = "Map size is %d by %d. Magnification filter is %s. Minification filter is %s. Color is (%g, %g, %g), should be (%g, %g, %g).";


static long Compare(GLsizei size, GLsizei color, GLftype *buf, GLftype levelErr)
{
    long i;

    ReadScreen(0, 0, size, size, GL_RGB, buf);

    for (i = 0; i < size*size*3; i += 3) {
	if (ABS(buf[i]-colorMap[color][0]) > levelErr+epsilon.color[0] ||
	    ABS(buf[i+1]-colorMap[color][1]) > levelErr+epsilon.color[1] ||
	    ABS(buf[i+2]-colorMap[color][2]) > levelErr+epsilon.color[2]) {
	    StrMake(errStr, errStrFmt, errSizeX, errSizeY, errMagFilter,
		    errMinFilter, buf[i], buf[i+1], buf[i+2],
		    colorMap[color][0], colorMap[color][1], colorMap[color][2]);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }

    return NO_ERROR;
}

/*
** OpenGL ES implementations are allowed some latitude in the method they use
** to determine rho, as long as max(m1, m2) <= rho <= m1 + m2.
** In this test m1 = |du/dx| and m2 = |dv/dy|. This means that for
** GL_*_MIPMAP_LINEAR filters, the color observed may actually be
** result from interpolation between two levels.
*/

static GLftype GetRhoMipMapLevelError(GLftype s, GLftype t, GLftype size,
				      GLint filter)
{
    GLftype error = ITOF(0), m1, m2;

    switch (filter) {
      case GL_NEAREST_MIPMAP_NEAREST:
      case GL_LINEAR_MIPMAP_NEAREST:
	error = ITOF(0);
	break;
      case GL_NEAREST_MIPMAP_LINEAR:
      case GL_LINEAR_MIPMAP_LINEAR:
	m1 = DIV(32 * s, size);
	m2 = DIV(32 * t, size);
	error = LOG2(m1+m2) - LOG2(m2);
	break;
    }
    return error;
}

static long Test(GLenum filter, GLftype *buf)
{
    GLint size, i;
    GLftype levelErr, s, t;

    GetEnumName(filter, errMinFilter);

    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter);

    s = ATOF("0.0625");
    t = ITOF(1);
    size = 32;

    for (i = 1; i <= 6; i++) {
	errSizeX = size;
	errSizeY = size;
	levelErr = GetRhoMipMapLevelError(s, t, ITOF(32), filter);

	DrawTexQuad2i2f(0, 0, size, size, ITOF(0), ITOF(0), s, t);

	if (Compare(size, i, buf, levelErr) == ERROR) {
	    return ERROR;
	}

	/*
	** This choice of size and s fixes m1 = 1.0 / 16.0,
	** up to representational and floating point error.
	*/
	size /= 2;
	s /= 2;
    }

    return NO_ERROR;
}

long MipSelectExec(void)
{
    GLenum x;
    GLftype *buf;
    GLubyte *texBuf;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*3*sizeof(GLftype));
    texBuf = (GLubyte *)MALLOC(WINDSIZEX*WINDSIZEY*3*sizeof(GLubyte));

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_TEXTURE_2D);

    x = GL_DECAL;
    glTexEnvt(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, x);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    MakeMipmaps(texBuf);

    x = GL_NEAREST;
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, x);
    GetEnumName(x, errMagFilter);
    if (Test(GL_NEAREST_MIPMAP_NEAREST, buf) == ERROR) {
	FREE(buf);
        FREE(texBuf);
	return ERROR;
    }
    if (Test(GL_LINEAR_MIPMAP_NEAREST, buf) == ERROR) {
	FREE(buf);
        FREE(texBuf);
	return ERROR;
    }
    if (Test(GL_NEAREST_MIPMAP_LINEAR, buf) == ERROR) {
	FREE(buf);
        FREE(texBuf);
	return ERROR;
    }
    if (Test(GL_LINEAR_MIPMAP_LINEAR, buf) == ERROR) {
	FREE(buf);
        FREE(texBuf);
	return ERROR;
    }

    x = GL_LINEAR;
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, x);
    GetEnumName(x, errMagFilter);
    if (Test(GL_NEAREST_MIPMAP_NEAREST, buf) == ERROR) {
	FREE(buf);
        FREE(texBuf);
	return ERROR;
    }
    if (Test(GL_LINEAR_MIPMAP_NEAREST, buf) == ERROR) {
	FREE(buf);
        FREE(texBuf);
	return ERROR;
    }
    if (Test(GL_NEAREST_MIPMAP_LINEAR, buf) == ERROR) {
	FREE(buf);
        FREE(texBuf);
	return ERROR;
    }
    if (Test(GL_LINEAR_MIPMAP_LINEAR, buf) == ERROR) {
	FREE(buf);
        FREE(texBuf);
	return ERROR;
    }

    FREE(buf);
    FREE(texBuf);
    return NO_ERROR;
}
