/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
**
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/mpalette.c,v 1.6 2005/05/17 13:12:33 artlab Exp $
*/

/*
** mpalette.c
** Matrix Palette Test.
**
** Description -
**    Tests the functionality of matrix palettes.
**
**    The test creates a random matrix palette that is composed from
**    scales and shears in x and y.
**
**    In a series of iterations, a random subset of matrices with associated
**    weights is chosen. The test first computes the expected location itself
**    and verifies that a point at that location can be successfully drawn.
**    Then the matrix palette is enabled and the point is rendered again.
**    Again, the test verifies that the point is correctly drawn.
**
**    This test is only executed if the extension OES_matrix_palette
**    is present.
**
** Error Explanation -
**    Since values may fall on pixel boundaries, and
**    transformations have numerical error, allowance is made for
**    pixel to be off by one in any direction from expected
**    location.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        The routine looks for the transformed point in a 3x3 grid of pixels
**        around the expected location, to allow a margin for matrix
**        operations.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilm.h"

#ifdef GL_OES_matrix_palette

static char errStr[240];
static char errRender[] = "Could not render point at (%1.2f, %1.2f).";
static char errTransform[] = "Original point at (%1.2f, %1.2f), transformed point expected at (%1.2f, %1.2f).";

/* Forward declarations of local functions */

static long ExecuteMatrixPalette(GLint maxPaletteMatrices, GLint maxVertexUnits);
static void MakePalette(GLftype palette[][16], GLint maxPaletteMatrices);
static void RandomPermutation(GLubyte *indices, GLint maxVertexUnits);
static void PartitionOne(GLftype * weights, GLint units);
static void MakeIdentity(GLftype * matrix);
static void MultiplyAdd(const GLftype * matrix, const GLftype * input, GLftype weight, GLftype * output);
static void TransformPalette(const GLftype palette[][16], const GLftype *weights,
    const GLubyte *indices, GLint units, const GLftype * input, GLftype * output);
static void DrawWeightedPoint4fv(const GLftype vertex[4], const GLftype *weights,
    const GLubyte *indices, GLint units);
static GLint OESRandomInt(GLint lo, GLint hi);

long MatrixPaletteExec(void)
{
    GLint maxPaletteMatrices, maxVertexUnits;

    /*
    ** 1. Verify that the extension is properly identified as part of the
    **    extension string
    */
    if (!tkCheckExtension("GL_OES_matrix_palette"))
    {
        return NO_ERROR;
    }

    /*
    ** 2. Retrieve MAX_PALETTE_MATRICES_OES and MAX_VERTEX_UNITS_OES
    **    Validate the retrieved values exist and are greater or equal
    **      to 9 and 3.
    */

    glGetIntegerv(GL_MAX_PALETTE_MATRICES_OES, &maxPaletteMatrices);
    glGetIntegerv(GL_MAX_VERTEX_UNITS_OES, &maxVertexUnits);

    if (glGetError() != GL_NO_ERROR)
    {
        ErrorReport(__FILE__, __LINE__, "Could not retrieve implementation limits");
        return ERROR;
    }

    if (maxPaletteMatrices < 9)
    {
        ErrorReport(__FILE__, __LINE__, "Implementation does not support enough palette matrices");
        return ERROR;
    }

    if (maxVertexUnits < 3)
    {
        ErrorReport(__FILE__, __LINE__, "Implementation does not support minimum number of vertex units");
        return ERROR;
    }

    /*
    ** 3. Core test:
    **      Fill all matrices of the palette with generated matrices (how?)
    **      for <iteration>
    **        Select a number of units 1..max units
    **        Select indices and weights for each unit
    **        Calculate the resulting transformation matrix; draw the point
    **        without palette turned on, verify correct drawing
    **        Turn palette on
    **        Draw point using the selected number of units, weights and indices
    **        Verify correct drawing 
    **        Turn palette off
    **
    **    Additional error checking:
    **      Ensure no INVALID_VALUE after MatrixIndexPointerOES,
    **      WeightPointerOES, or CurrentPaletteMatrixOES
    */
    return ExecuteMatrixPalette(maxPaletteMatrices, maxVertexUnits);
}

/*
** Execute the core test:
**
**      Fill all matrices of the palette with generated matrices (how?)
**      for <iteration>
**        Select a number of units 1..max units
**        Select indices and weights for each unit
**        Calculate the resulting transformation matrix; draw the point
**        without palette turned on, verify correct drawing
**        Turn palette on
**        Draw point using the selected number of units, weights and indices
**        Verify correct drawing 
**        Turn palette off
**
**    Additional error checking:
**      Ensure no INVALID_VALUE after MatrixIndexPointerOES,
**      WeightPointerOES, or CurrentPaletteMatrixOES
*/
static long ExecuteMatrixPalette(GLint maxPaletteMatrices, GLint maxVertexUnits)
{
    long result = NO_ERROR; 
    GLftype (*palette)[16];
    GLftype *weights;
    GLubyte *indices;

    GLftype *buf;
    GLftype point[4], xFormPoint[4];
    long i, j, outerCount, innerCount;

    palette = (GLftype (*)[16])MALLOC(maxPaletteMatrices * 16 * sizeof(GLftype));
    weights = (GLftype *)MALLOC(maxVertexUnits * sizeof(GLftype));
    indices = (GLubyte *)MALLOC(maxPaletteMatrices * sizeof(GLubyte));
    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    if (machine.pathLevel == 0) {
        outerCount = 10;
        innerCount = 10;
    } else {
        outerCount = 2;
        innerCount = 4;
    }
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    for (i = 0; i < outerCount; ++i)
    {
        MakePalette(palette, maxPaletteMatrices);

        for (j = 0; j < innerCount; ++j)
        {
            int units;
            GLftype range;

            /* select number of units to use */
            units = OESRandomInt(1, maxVertexUnits); 
            range = DIV(1, ITOF(4 * units));

            point[0] = OESRandom(-range, range);
            point[1] = OESRandom(-range, range);
            point[2] = ITOF(0);
            point[3] = ITOF(1);

            /* select weights and indices */

            PartitionOne(weights, units);
            RandomPermutation(indices, maxPaletteMatrices);

            /* compute resulting transformation and verify */
            /* correct rendering */
            TransformPalette((const GLftype (*)[16])palette,
                     weights, indices, units, point,
                     xFormPoint);
            
            glClear(GL_COLOR_BUFFER_BIT);
            glMatrixMode(GL_MODELVIEW);
            glLoadIdentity();
            DrawPoint4fv(xFormPoint);
            ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

            if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
                StrMake(errStr, errRender, xFormPoint[0], xFormPoint[1]);
                ErrorReport(__FILE__, __LINE__, errStr);
                result = ERROR;
                goto error;
            }

            /* render using palette and verify correct rendering */

            glClear(GL_COLOR_BUFFER_BIT);
            glEnable(GL_MATRIX_PALETTE_OES);
            DrawWeightedPoint4fv(point, weights, indices, units);
            glDisable(GL_MATRIX_PALETTE_OES);
            ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

            if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
                StrMake(errStr, errTransform, point[0], point[1],
                        xFormPoint[0], xFormPoint[1]);
                ErrorReport(__FILE__, __LINE__, errStr);
                result = ERROR;
                goto error;
            }
        }
    }

error:
    FREE(buf);
    FREE(weights);
    FREE(indices);
    FREE(palette);

    return result;
}

static void MakePalette(GLftype palette[][16], GLint maxPaletteMatrices)
{
    int i, row, column;

    for (i = 0; i < maxPaletteMatrices; ++i)
    {
        GLftype increment;

        MakeIdentity(palette[i]);

        row = platform_rand() >= (RANDOM_MAX / 2);
        column = platform_rand() >= (RANDOM_MAX / 2);

        increment = OESRandom(ITOF(0), ITOF(2));
        palette[i][row + column * 4] = ADD(palette[i][row + column * 4], increment); 

        if (i & 1)
        {
            /* For odd numbered matrices, load the matrix directly */

            glMatrixMode(GL_MATRIX_PALETTE_OES);
            glCurrentPaletteMatrixOES(i);
            glLoadMatrixt(palette[i]);
        }
        else
        {
            /* For even numbered matrices, load indirectly through the modelview matrix */

            glMatrixMode(GL_MODELVIEW);
            glPushMatrix();
            glLoadMatrixt(palette[i]);
            glCurrentPaletteMatrixOES(i);
            glLoadPaletteFromModelViewMatrixOES();
            glPopMatrix();
        }
    }
}

static void MakeIdentity(GLftype * matrix)
{
    GLint i;

    for (i = 0; i < 16; ++i)
        matrix[i] = ITOF(0);

    matrix[0] = matrix[5] = matrix[10] = matrix[15] = ITOF(1);
}


static void MultiplyAdd(const GLftype * matrix, const GLftype * input, GLftype weight, GLftype * output)
{
    GLint index;

    for (index = 0; index < 4; ++index)
    {
        int coordIndex;
        GLftype sum = ITOF(0);

        for (coordIndex = 0; coordIndex < 4; ++coordIndex)
        {
            GLftype element = matrix[index + coordIndex * 4];
            GLftype coord = input[coordIndex];
            GLftype product = MUL(element, coord);
            sum = ADD(sum, product);
        }

        output[index] = ADD(output[index], MUL(weight, sum));
    }
}


static void TransformPalette(const GLftype palette[][16], const GLftype *weights,
    const GLubyte *indices,    GLint units, const GLftype * input, GLftype * output)
{
    GLint index;

    for (index = 0; index < 4; ++index)
    {
        output[index] = ITOF(0);
    }

    for (index = 0; index < units; ++index)
    {
        GLubyte matrixIndex = indices[index];
        MultiplyAdd(palette[matrixIndex], input, weights[index], output);
    }
}

static void PartitionOne(GLftype * weights, GLint units)
{
    GLint i, j;

    if (units < 1)
        return;

    if (units == 1)
    {
        weights[0] = ITOF(1);
        return;
    }

    /* Draw random interval seprators */

    for (i = 0; i < units - 1; ++i)
    {
        weights[i] = OESRandom(ITOF(0), ITOF(1));
    }

    weights[units - 1] = ITOF(1);

    /* Bubble sort the separators, then replace them by the deltas */

    for (i = 0; i < units - 1; ++i)
    {
        for (j = i + 1; j < units; ++j)
        {
            if (weights[i] > weights[j])
            {
                GLftype tmp = weights[i];
                weights[i] = weights[j];
                weights[j] = tmp;
            }
        }
    }

    /* Replace separators by interval size, i.e. the difference to the previous boundary */

    for (i = units - 1; i > 0; --i)
    {
        weights[i] -= weights[i - 1];
    }
}

static void    RandomPermutation(GLubyte *indices, GLint maxVertexUnits)
{
    GLint i;

    for (i = 0; i < maxVertexUnits; ++i)
        indices[i] = i;

    for (i = maxVertexUnits - 1; i > 0; --i) 
    {
        GLint j = OESRandomInt(0, i);

        GLubyte tmp = indices[i];
        indices[i] = indices[j];
        indices[j] = tmp;

    }
}

static void DrawWeightedPoint4fv(const GLftype vertex[4], const GLftype *weights,
    const GLubyte *indices,    GLint units)
{
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_MATRIX_INDEX_ARRAY_OES);
    glEnableClientState(GL_WEIGHT_ARRAY_OES);
    glVertexPointer(4, GL_FTYPE, 0, &vertex[0]);
    glMatrixIndexPointerOES(units, GL_UNSIGNED_BYTE, 0, indices);
    glWeightPointerOES(units, GL_FTYPE, 0, weights);
    glDrawArrays(GL_POINTS, 0, 1);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_MATRIX_INDEX_ARRAY_OES);
    glDisableClientState(GL_WEIGHT_ARRAY_OES);
}


static GLint OESRandomInt(GLint lo, GLint hi)
{
    GLint   x = FTOI(OESRandom(ITOF(lo),ITOF(hi+1)));
    if (x > hi) x = hi;
    return x;
}

#endif /* GL_OES_matrix_palette */
