/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/mquery.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** mquery.c
** Matrix Query Test.
**
** Description -
**    Tests the functionality of glQueryMatrixxOES() on the three
**    matrix modes (GL_MODELVIEW_MATRIX, GL_PROJECTION_MATRIX
**    and GL_TEXTURE_MATRIX).
**
**    In the common lite profile, random matrices are loaded
**    with glLoadMatrix(). The values are queried and compared.
**    
**    In the common profile, random matrices with overflow values
**    are loaded. If the implementation supports overflow tracking
**    (determined by probing), the status bits are compared against
**    the overflowing elements of the matrix.
**
**    This test is only executed if the extension OES_query_matrix
**    is present.
**
** Error Explanation -
**    The difference between the loaded and queried values must not
**    be greater than zero epsilon. If an overflowing value is loaded,
**    the corresponding status bit must be set or the status must always
**    be 0.
**
** Technical Specification -
**    Buffer requirements:
**        No buffer requirements.
**    Color requirements:
**        No color requirements.
**    States requirements:
**        No state restrictions.
**    Error epsilon:
**        Zero epsilon.
**    Paths:
**       Allowed = Alias, Alpha, Blend, Depth, Dither, Fog, Logicop, Shade,
**                 Stencil.
*/

#include "conform.h"
#include "util.h"

#ifdef GL_OES_query_matrix

static char errStr[320];
static char *errStrFmt[4] = {
    "Matrix mode is %s.\nStatus is %X, should be 0.",
    "Matrix mode is %s.\nMatrix element at %d is %g, should be %g",
    "Matrix mode is %s.\nImplementation seems not to track overflow, but status was %X.",
    "Matrix mode is %s.\nMatrix element at %d is %g, overflow bit should be %s."
};


static void MakeMatrix(GLftype *buf)
{
    long i;

    for (i = 0; i < 16; i++) {
        buf[i] = OESRandom(XTOF(0x80000000), XTOF(0x7FFFFFFF));
    }
}

static long TestInRange(const char* tmp)
{
    GLftype matrix[16];
    GLfixed mantissa[16];
    GLint exponent[16];
    GLbitfield status;
    long i;

    MakeMatrix(matrix);

    glLoadMatrixt(matrix);
    status = glQueryMatrixxOES(mantissa, exponent);

    if (status != 0) {
        StrMake(errStr, errStrFmt[0], tmp, status);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    for (i = 0; i != 16; ++i) {
        GLftype value = LDEXP(XTOF(mantissa[i]), exponent[i]);

        if (ABS(value - matrix[i]) > ABS(MUL(epsilon.zero, matrix[i]))) {
            StrMake(errStr, errStrFmt[1], tmp, i, value, matrix[i]);
            ErrorReport(__FILE__, __LINE__, errStr);
            return ERROR;
        }
    }

    return NO_ERROR;
}

#ifdef COMMON
static void MakeOverflowMatrix(GLfloat *buf)
{
    long i;

    for (i = 0; i < 16; i++) {
        /* can't use NaN here, as comparison of NaNs is not portable */
        buf[i] = OESRandom(ITOF(0), ITOF(2)) < 1 ? 0 : LDEXP(1.0, 20);
    }
}

static long ProbeOverflow(void)
{
    GLfloat matrix[16];
    GLfixed mantissa[16];
    GLint exponent[16];
    GLbitfield status;
    long i, tracks_overflow;

    tracks_overflow = 0;

    for (i = 0; i != 100; ++i) {
        MakeOverflowMatrix(matrix);

        glLoadMatrixt(matrix);
        status = glQueryMatrixxOES(mantissa, exponent);

        if (status) {
            tracks_overflow = 1;
        }
    }

    return tracks_overflow;
}

static long WouldOverflow(float value)
{
    return !(value <= XTOF(0x7FFFFFFF) && value >= XTOF((GLfixed) 0x80000000));
}

static long TestOverflow(const char *tmp, int tracks_overflow)
{
    GLfloat matrix[16];
    GLfixed mantissa[16];
    GLint exponent[16];
    GLbitfield status;
    long i;

    MakeOverflowMatrix(matrix);

    glLoadMatrixt(matrix);
    status = glQueryMatrixxOES(mantissa, exponent);

    if (status != 0 && !tracks_overflow) {
        StrMake(errStr, errStrFmt[2], tmp, status);
        ErrorReport(__FILE__, __LINE__, errStr);
        return ERROR;
    }

    for (i = 0; i != 16; ++i) {

        if (WouldOverflow(matrix[i])) {
            if (!tracks_overflow) {
                continue;
            }

            if ((status & (1 << i)) == 0) {
                StrMake(errStr, errStrFmt[3], tmp, i, matrix[i], "set");
                ErrorReport(__FILE__, __LINE__, errStr);
                return ERROR;
            }
        } else {
            if ((status & (1 << i)) != 0) {
                StrMake(errStr, errStrFmt[3], tmp, i, matrix[i], "not set");
                ErrorReport(__FILE__, __LINE__, errStr);
                return ERROR;
            }
        }
    }

    return NO_ERROR;
}
#endif /* COMMON */

long MatrixQueryExec(void)
{
    static const GLenum modes[] = {GL_PROJECTION, GL_TEXTURE, GL_MODELVIEW};
    char tmp[40];
    long i, j, count;
#ifdef COMMON
    long tracks_overflow;
#endif

    /* Skip the test if extension is not available */
    if (!tkCheckExtension("GL_OES_query_matrix"))
    {
        return NO_ERROR;
    }

    if (machine.pathLevel == 0) {
        count = 100;
    } else {
        count = 10;
    }

#ifdef COMMON
    tracks_overflow = ProbeOverflow();
#endif

    for (i = 0; i != 3; ++i) {
        glMatrixMode(modes[i]);
        GetEnumName(modes[i], tmp);

        for (j = 0; j != count; ++j) {
            if (TestInRange(tmp) == ERROR) {
                return ERROR;
            }
#ifdef COMMON
            if (TestOverflow(tmp, tracks_overflow) == ERROR) {
                return ERROR;
            }
#endif
        }
    }

    return NO_ERROR;
}

#endif /* GL_OES_query_matrix */
