/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/packpix.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** packpix.c
** Packed Pixels Test.
**
** Description -
**    Draws a rectangle in the lower left hand corner of the
**    window using glDrawPixels() for every packed pixel format
**    and type allowed.  For each, a solid rectangle appears.
**
**    Rectangles are drawn with fully saturated colors. The
**    outcome of each glDrawPixels() is checked by examining one
**    pixel in the interior of the upper half of the rectangle,
**    and one in the lower half. The results should be the color
**    set in the initial data.
**
**    The rectangle is read back from the framebuffer.
**    The result should match the original image.
**
**    The test is structured as follows:
**	  For each packed data type:
**	      For each legal pixel format:
**		  SetupData() - create raw image data.
**		  ConvertCompData() - convert image data.
**		  glDrawPixels().
**		  TestPixRGB().
**		  glReadPixels()
**		  TestPackedPixels()
** Technical Specification -
**    Buffer requirements:
**	  Color buffer.
**    Color requirements:
**	  Full color.
**    States requirements:
**	  Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**	  Color epsilon. Zero epsilons.
**    Paths:
**	  Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**	  Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilp.h"

#define W 32
#define H 32
#define FREE_AND_RETURN(err) {			    \
    FREE(dataBuf);				    \
    FREE(convertBuf);				    \
    FREE(buf);					    \
    return(err);				    \
}
#define MAKE_ERROR_STRING(t, f) {			\
    GetEnumName(t, typeStr);				\
    GetEnumName(f, formatStr);				\
    StrMake(errStr, errStrFmt, typeStr, formatStr);	\
}

static char errStr[160];
static char errStrFmt[] = "DrawPixels type is %s, format is %s.";

static long TestPixRGB(GLftype *buf, packPixRec const *format,
		       ftype r, ftype g, ftype b, long split)
{
    ftype midX, lowY, hiY, packepsilon[3];
    long i;

    /*
    ** Compute epsilons based on both the size of the packed field and
    ** the color buffer depth.
    */
    for (i = 0; i < 3; i++) {
	packepsilon[i] = CalcEpsilon(format->size[i]);
	if (epsilon.color[i] > packepsilon[i])
	    packepsilon[i] = epsilon.color[i];
    }

    /*
    ** Test one pixel in the center of the lower half of the region drawn.
    ** Split indicates that the other half should remain black (for alpha,
    ** depth tests, etc).
    */

    lowY = ITOF(H)/4;
    hiY = ITOF(H)*3/4;
    midX = ITOF(W)/2;

    ReadScreen(0, 0, W, H, GL_RGB, buf);

    i = FTOI((W * lowY + midX) * 3);
    if (!(ABS(buf[i]-r) < packepsilon[0] &&
	  ABS(buf[i+1]-g) < packepsilon[1] &&
	  ABS(buf[i+2]-b) < packepsilon[2])) {
	return ERROR;
    }

    if (split == GL_TRUE) {
	i = FTOI((W * hiY + midX) * 3);
	if (!(buf[i] < packepsilon[0] &&
	      buf[i+1] < packepsilon[1] &&
	      buf[i+2] < packepsilon[2])) {
	    return ERROR;
	}
    }
    return NO_ERROR;
}

static void SetupData(GLftype *fPtr, long format)
{
    long size, i;

    size = W * H;
    switch (format) {
      case GL_RGB:
	for (i = 0; i < size; i++) {
	     *fPtr++ = ITOF(9)/10;
	     *fPtr++ = ITOF(1)/10;
	     *fPtr++ = ITOF(1)/2;
	     *fPtr++ = ITOF(0);
	}
	break;
      case GL_RGBA:
	for (i = 0; i < size; i++) {
	     *fPtr++ = ITOF(9)/10;
	     *fPtr++ = ITOF(1)/10;
	     *fPtr++ = ITOF(1)/2;
	     *fPtr++ = (i < size/2) ? ITOF(1) : ITOF(0);
	}
	break;
    }
}

/* 3 and 4 component packed formats to test */
static GLenum rgbType[] = {
    GL_UNSIGNED_SHORT_5_6_5
};
static const GLsizei rgbType_count = sizeof(rgbType) / sizeof(rgbType[0]);

static GLenum rgbaType[] = {
    GL_UNSIGNED_SHORT_4_4_4_4,
    GL_UNSIGNED_SHORT_5_5_5_1
};
static const GLsizei rgbaType_count = sizeof(rgbaType) / sizeof(rgbaType[0]);

long PackedPixelsExec(void)
{
    GLftype *dataBuf, *buf;
    GLubyte *convertBuf;
    GLftype color[4];
    long i;
    unsigned size;
    char typeStr[40], formatStr[40];

    size = W * H;
    dataBuf = (GLftype *)MALLOC(size*4*sizeof(GLftype));
    convertBuf = (GLubyte *)MALLOC(size*4*sizeof(GLftype));
    buf = (GLftype *)MALLOC(size*4*sizeof(GLftype));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glColor4t(ITOF(0), ITOF(0), ITOF(0), ITOF(1));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    /*
    ** Test RGB for 3 component packed formats.
    */
    SetupData(dataBuf, GL_RGB);
    for (i = 0; i != rgbType_count; i++) {
	packPixRec const *format = packPixInfo(rgbType[i]);

	ConvertRGBAtoPackedImage(dataBuf, convertBuf, format, W*H);
	ConvertPackedtoRGBA(convertBuf, color, format);
	glClear(GL_COLOR_BUFFER_BIT);
	DrawPixels(0, 0, W, H, GL_RGB, format->type, convertBuf);
	if (TestPixRGB(buf, format, color[0], color[1], color[2], GL_FALSE)
	    == ERROR) {
	    MAKE_ERROR_STRING(format->type, GL_RGB);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE_AND_RETURN(ERROR);
	}
    }

    /*
    ** Test RGBA for 4 component packed formats.
    */
    glEnable(GL_ALPHA_TEST);
    glAlphaFunct(GL_GREATER, ITOF(1)/2);

    SetupData(dataBuf, GL_RGBA);
    for (i = 0; i != rgbaType_count; i++) {
	packPixRec const *format = packPixInfo(rgbaType[i]);

	ConvertRGBAtoPackedImage(dataBuf, convertBuf, format, W*H);
	ConvertPackedtoRGBA(convertBuf, color, format);
	glClear(GL_COLOR_BUFFER_BIT);
	DrawPixels(0, 0, W, H, GL_RGBA, format->type, convertBuf);
	if (TestPixRGB(buf, format, color[0], color[1], color[2], GL_TRUE)
	    == ERROR) {
	    MAKE_ERROR_STRING(format->type, GL_RGBA);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE_AND_RETURN(ERROR);
	}
    }

    glDisable(GL_ALPHA_TEST);
    FREE_AND_RETURN(NO_ERROR);
}
