/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/pntaa.c,v 1.5 2005/05/11 15:21:40 artlab Exp $
*/

/*
** pntaa.c
** Point Anti-Aliasing Test.
**
** Description -
**    Tests position independence of anti-aliased points. It
**    chooses a point of random size from the implementation dependent
**    range, and computes the resulting coverage.
**    The point is then drawn at various offsets, each time a new
**    coverage is computed and compared with the original value.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        The error allowance is based on a minimum of 4 subpixel sampling.
**        This yields 0.25 per pixel on the circumference of the point.
**        Thus 0.25 is multiplied by 2*pi*r or 2*pointsize, to create the
**        allowable error margin.
**    Paths:
**        Allowed = Depth, Fog, Shade, Stencil.
**        Not allowed = Alias, Alpha, Logicop, Blend, Dither.
*/

#include "conform.h"
#include "util.h"


static char errStr[240];
static char errStrFmt[][80] = {
    "Point primative did not draw.",
    "Coverage is %g, should be %g. Error margin is %g."
};


static long Test(ftype dirX, ftype dirY, GLftype *buf, ftype errorMargin)
{
    GLftype sum, saveSum = ITOF(0), size, offset, step, x, y;
    GLint sizeRange[2]; 
    long i;

    glGetIntegerv(GL_SMOOTH_POINT_SIZE_RANGE, sizeRange);

    /* 
    ** Limit point sizes tested to half the window size.
    */
    sizeRange[1] = (sizeRange[1] > WINDSIZEX/2) ? WINDSIZEX/2 : sizeRange[1];

    size = OESRandom(ITOF(sizeRange[0]), ITOF(sizeRange[1]));
    glPointSizet(size);

    errorMargin = MUL(errorMargin, MUL(PI, size));

    x = OESRandom(size+ITOF(10), ITOF(WINDSIZEX-1)-size-ITOF(10));
    y = OESRandom(size+ITOF(10), ITOF(WINDSIZEY-1)-size-ITOF(10));

    step = (machine.pathLevel == 0) ? ATOF("0.055") : ATOF("0.5");
    for (offset = ITOF(0); offset <= ITOF(2); offset += step) {
	glClear(GL_COLOR_BUFFER_BIT);

	{
	    GLftype vertex[2];
	    if (dirX == ITOF(0)) {
		vertex[0] = x;
		vertex[1] = MUL(offset, dirY) + y;
	    } else {
		vertex[0] = MUL(offset, dirX) + x;
		vertex[1] = y;
	    }

	    glEnableClientState(GL_VERTEX_ARRAY);
	    glVertexPointer(2, GL_FTYPE, 0, &vertex[0]);
	    glDrawArrays(GL_POINTS, 0, 1);
	    glDisableClientState(GL_VERTEX_ARRAY);
	}

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);

	sum = ITOF(0);
	for (i = 0; i < WINDSIZEX*WINDSIZEY*3; i += 3) {
	    sum = ADD(sum, buf[i]);
	}

	if (offset < step) {
	    saveSum = sum;
	} else {
	    if (sum == ITOF(0)) {
		ErrorReport(__FILE__, __LINE__, errStrFmt[0]);
		return ERROR;
	    } else if (ABS(saveSum-sum) > errorMargin) {
		StrMake(errStr, errStrFmt[1], sum, saveSum, errorMargin);
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	}
    }

    return NO_ERROR;
}

long PointAntiAliasExec(void)
{
    GLftype *buf;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*3*sizeof(GLftype));

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glColor4t(ITOF(1), ITOF(0), ITOF(0), ITOF(1));
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    if (Test(ITOF(1), ITOF(0), buf, ITOF(1)/4) == ERROR) {
	FREE(buf);
	return ERROR;
    }
    if (Test(ITOF(-1), ITOF(0), buf, ITOF(1)/4) == ERROR) {
	FREE(buf);
	return ERROR;
    }
    if (Test(ITOF(0), ITOF(1), buf, ITOF(1)/4) == ERROR) {
	FREE(buf);
	return ERROR;
    }
    if (Test(ITOF(0), ITOF(-1), buf, ITOF(1)/4) == ERROR) {
	FREE(buf);
	return ERROR;
    }

    FREE(buf);
    return NO_ERROR;
}
