/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
**
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/pntsprt.c,v 1.6 2005/05/17 13:12:33 artlab Exp $
*/

/*
** pntsprt.c
**
** Point Sprite Test.
**
** Description -
**
** Clear the background to black.
** For a number of points, starting at a size of 1 and doubling in size
**		- Define a texture map of that size with random, non zero values.
**		- Set the point-size equal to the texture map size.
**		- Draw a single point, centred on (pointSize/2, pointSize/2).
**		- Read back the colour buffer, where the point should have drawn.
**		- Ascertain if the point that was drawn matches the point size.
**		- If the point size and drawn size match, compare the contents of
**		  the colour buffer and the texture map. Note that the comparison
**		  takes into account the inversion of the point sprite t coordinate.
**		- If the drawn size is not that of the requested size, then see if
**		  the requested point size is valid or not.
**
** It isn't possible to query an implementation about its maximum supported
** point size where point sprites are concerned.  An application may or may
** not clamp to the maximum anti-aliased point size.  This test looks to see
** if something was drawn that is of the requested point size and if so, will
** check what has been drawn.  This enables some implementation testing where
** the point-size is limited, but the maximum sprite point size is not.
**
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Multiple colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"

#ifdef GL_OES_VERSION_1_1

static char errEnv[40], errMagFilter[40], errMinFilter[40];
static char errStr[320];
static char errStrFmt[] =
"Point size (%d). Location (%d, %d).\
 Texture environment mode is %s.\
 Magnification filter is %s,\
 minification filter is %s.\
 Color is (%d, %d, %d), should be (%d, %d, %d).";

static void Make(GLsizei size, GLubyte *buf)
{
	GLint level, i;

	GLftype	high;
	GLftype low;

	/*  Set upnpack alignment... */
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	low  = ITOF(6)/10;	/* == 0.6 */
	high = ITOF(1);		/* == 1.0 */
	for (i = 0; i < size*size*3; i++)
	{
	    /* I don't want to have any 0 component values (background) */
	    /* in the texture map. */
		buf[i] = (GLubyte) FTOI((255*OESRandom(low, high)));
	}
	level = 0;
	while (size > 0)
	{
		glTexImage2D(GL_TEXTURE_2D, level++, GL_RGB, size, size, 0,
					 GL_RGB, GL_UNSIGNED_BYTE, (GLvoid *)buf);
		size /= 2;
	}
}

long PointSpriteDraw(int pointSize, int *drawnPointSize, int possibleMaxPointSize)
{
    GLubyte *textureBuf, *workBuf;
    GLftype x;
    GLsizei textureBits[4];
    GLftype textureEpsilon[4];
    long i, j, k, l;
	int gap = 0;

    textureBuf = (GLubyte *)MALLOC(pointSize*pointSize*3*sizeof(GLubyte));
    workBuf    = (GLubyte *)MALLOC(WINDSIZEX*WINDSIZEY*4*sizeof(GLubyte));

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glClear(GL_COLOR_BUFFER_BIT);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
	glEnable(GL_TEXTURE_2D);

    x = GL_DECAL;
    glTexEnvtv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, &x);
    GetEnumName((GLenum)x, errEnv);

    x = GL_NEAREST;
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, x);
    GetEnumName((GLenum)x, errMinFilter);

    x = GL_NEAREST;
    glTexParametert(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, x);
    GetEnumName((GLenum)x, errMagFilter);

    Make(pointSize, textureBuf);

    GetComponentBits(GL_RGB, GL_UNSIGNED_BYTE, textureBits);
    CalcTextureEpsilon(textureBits, textureEpsilon);

    /* Draw a textured point sprite with no distortion.. */
	{
		GLftype	vertexArray[1][2];

		vertexArray[0][0] = ITOF(pointSize)/2;
		vertexArray[0][1] = ITOF(pointSize)/2;

		glVertexPointer(2, GL_FTYPE, 0, vertexArray);
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnable(GL_POINT_SPRITE_OES);

		x = GL_TRUE;
	    glTexEnvtv(GL_POINT_SPRITE_OES, GL_COORD_REPLACE_OES, &x);

		glPointSizet(ITOF(pointSize));
		glDrawArrays(GL_POINTS, 0, 1);

		x = GL_FALSE;
	    glTexEnvtv(GL_POINT_SPRITE_OES, GL_COORD_REPLACE_OES, &x);
		glDisable(GL_POINT_SPRITE_OES);
		glDisableClientState(GL_VERTEX_ARRAY);
		glDisable(GL_TEXTURE_2D);
	}
    glReadPixels(0, 0, pointSize, pointSize, GL_RGBA, GL_UNSIGNED_BYTE, workBuf);

        /* See how wide a point was drawn. */
	i = (pointSize >> 1) * pointSize * 4;
	for (j = 0; j < pointSize; j++, i += 4)
	{
		if (workBuf[i] == 0)
			gap++;
	}
	*drawnPointSize = pointSize - gap;

	/* If the drawn pointsize is not equal to the requested point size, */
	/* it may be that the point has been clamped to it's maximum. If */
	/* this is the case, then it is not an error, though the */
	/* 'undistorted' comparison will fail. */
	if (pointSize <= possibleMaxPointSize)
	{
		for (i = 0; i < pointSize; i++)
		{
			for (j = 0; j < pointSize; j++)
			{
			    /* k is inverted because point sprite 't' coord */
			    /* is upside-down! */
				k = ((pointSize -i -1) * pointSize + j) * 4;
				l = (i * pointSize + j) * 3;
				if (ABS(ITOF(workBuf[k+0]-textureBuf[l+0]))/255 > 2 * textureEpsilon[0]	||
					ABS(ITOF(workBuf[k+1]-textureBuf[l+1]))/255 > 2 * textureEpsilon[1] ||
					ABS(ITOF(workBuf[k+2]-textureBuf[l+2]))/255 > 2 * textureEpsilon[2])
				{
					StrMake(errStr, errStrFmt, pointSize, i, j, errEnv, errMagFilter,
						errMinFilter, workBuf[k],    workBuf[k+1],
						workBuf[k+2], textureBuf[l], textureBuf[l+1],
						textureBuf[l+2]);
					ErrorReport(__FILE__, __LINE__, errStr);
					FREE(textureBuf);
					FREE(workBuf);
					return ERROR;
				}
			}
		}
    }
    FREE(textureBuf);
    FREE(workBuf);
    return NO_ERROR;
}

long PointSpriteExec(void)
{
	long err = NO_ERROR;
	int drawnPointSize;
	int pntRange[2];
	int limit = (WINDSIZEX < WINDSIZEY ? WINDSIZEX : WINDSIZEY);
	int i;

    glGetIntegerv(GL_ALIASED_POINT_SIZE_RANGE, pntRange);

	for (i = 1; i <= limit; i*=2)
	{
		err = PointSpriteDraw(i, &drawnPointSize, pntRange[1]);
		if (err != NO_ERROR)
			return(err);
	}
	return(err);
}

#endif
