/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
**
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/pntszary.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** pntszary.c
**
** Point Size Array Test.
**
** Description -
**
** This test checks the operation of using an array to define independent sizes
** for points rendered through a vertex array call.   The range of point sizes
** supported through the array must be a superset of those that are provided for
** when the point size array is not enabled.  Points that are drawn, with and
** without the point size array mechanism should be identical.
**
** The first part of the test renders a set of randomly sized points one at a
** time using the point size array to define the size of the point.  Each point
** is checked for correctness along the lines of the pntrast.c test.
**
** The second part of the test compares a set of points with varying sizes that
** are rendered without using the point-size array to the same set of points
** rendered using the point size array. The two should produce identical
** results.
**
** The third part of the tests is a repeat of the above, except that the points
** are rendered usung DrawElements as opposed to DrawArrays to ensure indexed
** points select the correct point size.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"

#ifdef GL_OES_VERSION_1_1

/*
** Defines how many points to place in the vertex arrays.
*/
#define	POINT_SIZE_ARRAY_SIZE	50

static char errStr[160];
static char errStrFmt[][80] = {
    "Point did not draw correctly when using Point Size Array. Size = %d.",
    "Non point-size array and point-size array using DrawArrays is different",
    "Non point-size array and point-size array using DrawElements is different",
};

int PointHasDrawnCorrectly(int size, GLftype *ptr)
{
    GLint i, j;

	for (i = 0; i < size+2; i++)
	    if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE)
			return(ERROR);

	for (i = 0; i < size; i++)
	{
	    if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE)
			return(ERROR);

	    for (j = 0; j < size; j++)
		{
			if (AutoColorCompare(*ptr++, COLOR_ON) == GL_FALSE)
				return(ERROR);
	    }
	    if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE)
			return(ERROR);
	}

	for (i = 0; i < size+2; i++)
	    if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE)
			return(ERROR);

	return(NO_ERROR);
}

long PointSizeArrayExec(void)
{
    GLftype *buf, *stdbuf, *arybuf, *sbuf, *abuf;
    GLint count, pntRange[2];

	GLftype	vertexArray[POINT_SIZE_ARRAY_SIZE][2];
	GLftype	colorArray[POINT_SIZE_ARRAY_SIZE][4];
	GLftype	pointSizeArray[POINT_SIZE_ARRAY_SIZE];
	GLushort indices[POINT_SIZE_ARRAY_SIZE];

    if ((buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype))) == NULL)
	{
		return(ERROR);
	}
	if ((stdbuf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype)*3)) == NULL)
	{
		FREE(buf);
		return(ERROR);
	}
    if ((arybuf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype)*3)) == NULL)
	{
		FREE(buf);
		FREE(stdbuf);
		return(ERROR);
	}

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    glGetIntegerv(GL_ALIASED_POINT_SIZE_RANGE, pntRange);
    /*
    ** Limit point sizes tested to half the window size.
    */
	if (pntRange[1] > WINDSIZEX/2)
		pntRange[1] = WINDSIZEX/2;
	if (pntRange[1] > WINDSIZEY/2)
		pntRange[1] = WINDSIZEY/2;

    /*
	** Create an array containing a random set of point-sizes. Also set
	** up the appropriate vertex array and index list used later.
    */
	for (count = 0; count < POINT_SIZE_ARRAY_SIZE; count++)
	{
		int thisPointSize = FTOI(OESRandom(ITOF(pntRange[0]), ITOF(pntRange[1])));

		pointSizeArray[count] = ITOF(thisPointSize);
	    vertexArray[count][0] = ITOF(thisPointSize)/2 + ITOF(1);
	    vertexArray[count][1] = ITOF(thisPointSize)/2 + ITOF(1);
		indices[count] = POINT_SIZE_ARRAY_SIZE - 1 - count;
	}
	glVertexPointer(2, GL_FTYPE, 0, vertexArray);
	glColorPointer(4, GL_FTYPE, 0, colorArray);
	glPointSizePointerOES(GL_FTYPE, 0, pointSizeArray);

	/*
	** Draw assorted sized points one at a time using the point size array and
	** check that they have been drawn correctly.
	*/
	glPointSizet(ITOF(1));
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_POINT_SIZE_ARRAY_OES);

	for (count = 0; count < POINT_SIZE_ARRAY_SIZE; count++)
	{
		int thisPointSize = FTOI(pointSizeArray[count]);

		glClear(GL_COLOR_BUFFER_BIT);

		glDrawArrays(GL_POINTS, count, 1);

		ReadScreen(0, 0, thisPointSize+2, thisPointSize+2, GL_AUTO_COLOR, buf);

		if (PointHasDrawnCorrectly(thisPointSize, buf) == ERROR)
		{
		    glDisableClientState(GL_VERTEX_ARRAY);
		    glDisableClientState(GL_POINT_SIZE_ARRAY_OES);
			FREE(buf);
			FREE(stdbuf);
			FREE(arybuf);

			StrMake(errStr, errStrFmt[0], thisPointSize);
			ErrorReport(__FILE__, __LINE__, errStr);
			return(ERROR);
		}
	}
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_POINT_SIZE_ARRAY_OES);

	/*
	** Draw a set of points to random places using random colours and different
	** point sizes without using the point vertex array.
	*/
	for (count = 0; count < POINT_SIZE_ARRAY_SIZE; count++)
	{
		GLftype vmin = -ITOF(pntRange[1]);
		GLftype vmax = ITOF(WINDSIZEX) + ITOF(pntRange[1]);
		/*
		** Generate random vertices (including some that will get clipped).
		*/
	    vertexArray[count][0] = OESRandom(vmin, vmax);
	    vertexArray[count][1] = OESRandom(vmin, vmax);
		/*
		** Generate random colors..
		*/
		colorArray[count][0] = OESRandom(ITOF(0), ITOF(1));
		colorArray[count][1] = OESRandom(ITOF(0), ITOF(1));
		colorArray[count][2] = OESRandom(ITOF(0), ITOF(1));
		colorArray[count][3] = OESRandom(ITOF(0), ITOF(1));
	}
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	for (count = 0; count < POINT_SIZE_ARRAY_SIZE; count++)
	{
		glPointSizet(pointSizeArray[count]);
		glColor4t(colorArray[count][0],
				  colorArray[count][1],
				  colorArray[count][2],
				  colorArray[count][3]);

		glEnableClientState(GL_VERTEX_ARRAY);
		glDrawArrays(GL_POINTS, count, 1);
		glDisableClientState(GL_VERTEX_ARRAY);
	}
	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, stdbuf);

	/*
	** Now draw all the points in one go using the array.
	*/
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_POINT_SIZE_ARRAY_OES);

	glDrawArrays(GL_POINTS, 0, count);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_POINT_SIZE_ARRAY_OES);

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, arybuf);
	/*
	** Check to see if they drew the same.
	*/
	sbuf = stdbuf;
	abuf = arybuf;
	for (count = 0; count < WINDSIZEX*WINDSIZEY*3; count++, sbuf++, abuf++)
	{
		if (*sbuf != *abuf)
		{
			StrMake(errStr, errStrFmt[1]);
			ErrorReport(__FILE__, __LINE__, errStr);
			return(ERROR);
		}
	}

	/*
	** Now repreat, drawing in reverse and using an inverted index list to
	** drive the drawing.
	*/
	glClear(GL_COLOR_BUFFER_BIT);
	for (count = POINT_SIZE_ARRAY_SIZE-1; count >= 0; count--)
	{
		glPointSizet(pointSizeArray[count]);
		glColor4t(colorArray[count][0],
				  colorArray[count][1],
				  colorArray[count][2],
				  colorArray[count][3]);

		glEnableClientState(GL_VERTEX_ARRAY);
		glDrawArrays(GL_POINTS, count, 1);
		glDisableClientState(GL_VERTEX_ARRAY);
	}

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, stdbuf);
	/*
	** Now draw all the points in one go using the array.
	*/
	glClear(GL_COLOR_BUFFER_BIT);
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_POINT_SIZE_ARRAY_OES);

	glDrawElements(GL_POINTS, POINT_SIZE_ARRAY_SIZE, GL_UNSIGNED_SHORT, indices);

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_POINT_SIZE_ARRAY_OES);

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, arybuf);
	/*
	** Check to see if they drew the same.
	*/
	sbuf = stdbuf;
	abuf = arybuf;
	for (count = 0; count < WINDSIZEX*WINDSIZEY*3; count++, sbuf++, abuf++)
	{
		if (*sbuf != *abuf)
		{
			StrMake(errStr, errStrFmt[2]);
			ErrorReport(__FILE__, __LINE__, errStr);
			return(ERROR);
		}
	}
	FREE(buf);
	FREE(stdbuf);
	FREE(arybuf);
	return (NO_ERROR);
}

#endif


