/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/scissor.c,v 1.5 2005/05/11 15:21:40 artlab Exp $
*/

/*
** scissor.c
** Scissor Test.
**
** Description -
**    A scissor box is set up with a glScissor() call. A large
**    polygon is then drawn over the box. Only the region defined
**    by the box should render. Scissor boxes of various sizes
**    (including zero width and height) are used to insure
**    completeness of the test.
**
** Error Explanation -
**    Failure occurs if the scissor box does not scissor to the
**    correct size. Special attention is paid along the bottom
**    and right edges.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"


static char errStr[240];
static char errStrFmt[][160] = {
    "Scissor box with zero width and height did not scissor out polygon.",
    "glScissor(%d, %d, %d, %d) did not scissor correctly along the %s edge.",
    "glScissor(%d, %d, %d, %d) did not scissor correctly within the scissor box.",
};


static long TestClear(GLftype *buf)
{
    long i;

    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

    for (i = 0; i < WINDSIZEX*WINDSIZEY; i++) {
	if (AutoColorCompare(buf[i], COLOR_OFF) == GL_FALSE) {
	    return ERROR;
	}
    }
    return NO_ERROR;
}

static long Test(GLint x, GLint y, GLsizei sizeX, GLsizei sizeY, GLftype *buf)
{
    GLftype *ptr;
    long i, j;

    ReadScreen(x-1, y-1, sizeX+2, sizeY+2, GL_AUTO_COLOR, buf);

    ptr = buf;
    for (i = 0; i < sizeX+2; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    StrMake(errStr, errStrFmt[1], x, y, sizeX, sizeY, "bottom");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }
    for (i = 0; i < sizeY; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    StrMake(errStr, errStrFmt[1], x, y, sizeX, sizeY, "left");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}

	for (j = 0; j < sizeX; j++) {
	    if (AutoColorCompare(*ptr++, COLOR_ON) == GL_FALSE) {
		StrMake(errStr, errStrFmt[2], x, y, sizeX, sizeY);
		ErrorReport(__FILE__, __LINE__, errStr);
		return ERROR;
	    }
	}

	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    StrMake(errStr, errStrFmt[1], x, y, sizeX, sizeY, "right");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }
    for (i = 0; i < sizeX+2; i++) {
	if (AutoColorCompare(*ptr++, COLOR_OFF) == GL_FALSE) {
	    StrMake(errStr, errStrFmt[1], x, y, sizeX, sizeY, "top");
	    ErrorReport(__FILE__, __LINE__, errStr);
	    return ERROR;
	}
    }

    return NO_ERROR;
}

long ScissorExec(void)
{
    static const GLftype vertices[4][2] = {
        {ITOF(0), ITOF(0)},
        {ITOF(WINDSIZEX), ITOF(0)},
        {ITOF(0), ITOF(WINDSIZEY)},
        {ITOF(WINDSIZEX), ITOF(WINDSIZEY)}
    };
    GLftype *buf;
    GLint x, y, max, i;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    glDisable(GL_SCISSOR_TEST);
    glClear(GL_COLOR_BUFFER_BIT);

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FTYPE, 0, &vertices[0][0]);

    x = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
    y = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
    glScissor(x, y, 0, 0);
    glEnable(GL_SCISSOR_TEST);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    if (TestClear(buf) == ERROR) {
        glDisableClientState(GL_VERTEX_ARRAY);
	ErrorReport(__FILE__, __LINE__, errStrFmt[0]);
	FREE(buf);
	return ERROR;
    }

    max = (machine.pathLevel == 0) ? (WINDSIZEX / 8) : 2;
    for (i = 1; i < max; i++) {
	glDisable(GL_SCISSOR_TEST);
	glClear(GL_COLOR_BUFFER_BIT);

        x = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        y = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	glScissor(x, y, WINDSIZEX/8, i);
	glEnable(GL_SCISSOR_TEST);
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

        if (Test(x, y, WINDSIZEX/8, i, buf) == ERROR) {
            glDisableClientState(GL_VERTEX_ARRAY);
	    FREE(buf);
	    return ERROR;
	}
    }

    max = (machine.pathLevel == 0) ? (WINDSIZEY / 8) : 2;
    for (i = 1; i < max; i++) {
	glDisable(GL_SCISSOR_TEST);
	glClear(GL_COLOR_BUFFER_BIT);

        x = FTOI(OESRandom(ITOF(WINDSIZEX)/4, ITOF(WINDSIZEX)*3/4));
        y = FTOI(OESRandom(ITOF(WINDSIZEY)/4, ITOF(WINDSIZEY)*3/4));
	glScissor(x, y, i, WINDSIZEY/8);
	glEnable(GL_SCISSOR_TEST);
        glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

        if (Test(x, y, i, WINDSIZEY/8, buf) == ERROR) {
            glDisableClientState(GL_VERTEX_ARRAY);
	    FREE(buf);
	    return ERROR;
	}
    }

    glDisableClientState(GL_VERTEX_ARRAY);
    FREE(buf);
    return NO_ERROR;
}
