/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/trirast.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** trirast.c
** Triangle Rasterization Test.
**
** Description - 
**    This test renders GL_TRIANGLES, and GL_TRIANGLE_STRIPS, and
**    GL_TRIANGLE_FANS. It checks if those fragments which are
**    colored the primitive's color have centers which are within
**    the convex hull formed by the vertices, and if those
**    fragments which are colored the background color are
**    outside the convex hull of formed by the vertices.
**    
** Error Explanation - 
**    Those fragments which are found to be in error are then
**    checked to determine their distance from the nearest
**    primitive edge. If this distance is less than 1/2 of a
**    pixel width the error is ignored.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilg.h"


typedef struct _edgeEqnRec {
    slopeRec slope;
    interceptRec intercept;
    vertexRec vOp;
    long (*Inequality)(ftype, ftype, slopeRec *, interceptRec *);
    ftype length;
} edgeEqnRec;

typedef struct _triRec {
    edgeEqnRec edge01, edge02, edge12;
    vertexRec v0, v1, v2;
} triRec;


static ftype errX, errY;
static char *errExpectedDrawn, *errObservedDrawn;
static char errStr[320];
static char errStrFmt[] = "While rendering a %s with vertices (%g, %g), (%g, %g), (%g, %g), pixel at (%g, %g) should be the %s color, but is rendered the %s color.";


static void MakeEdge(edgeEqnRec *e, vertexRec *v0, vertexRec *v1,
		     vertexRec *vOp)
{

    GetSlope(&e->slope, v0, v1);
    GetIntercept(&e->intercept, v0, &e->slope);

    e->length = SQRT(MUL(e->slope.dx, e->slope.dx) + MUL(e->slope.dy, e->slope.dy));

    e->vOp.x = vOp->x;
    e->vOp.y = vOp->y;
    e->vOp.z = vOp->z;

    if (LessThan(vOp->x, vOp->y, &e->slope, &e->intercept)) {
	e->Inequality = LessThan;
    } else {
	e->Inequality = MoreThan;
    }
}

static void MakeVertices(vertexRec *v)
{
    long i;

    for (i = 0; i < 3; i++) {
	v[i].x = OESRandom(ITOF(2), ITOF(WINDSIZEX-2));
	v[i].y = OESRandom(ITOF(2), ITOF(WINDSIZEY-2));
	v[i].z = OESRandom(ITOF(-1), ITOF(1));
    }
}

static void MakeTri(triRec *t, vertexRec *v)
{

    MakeEdge(&t->edge01, &v[0], &v[1], &v[2]);
    MakeEdge(&t->edge02, &v[0], &v[2], &v[1]);
    MakeEdge(&t->edge12, &v[1], &v[2], &v[0]);

    t->v0.x = v[0].x;
    t->v0.y = v[0].y;
    t->v0.z = v[0].z;

    t->v1.x = v[1].x;
    t->v1.y = v[1].y;
    t->v1.z = v[1].z;

    t->v2.x = v[2].x;
    t->v2.y = v[2].y;
    t->v2.z = v[2].z;
}

static void RenderTriangle(triRec *t, GLenum mode)
{
    GLftype vertices[3][3];
    vertices[0][0] = t->v0.x; vertices[0][1] = t->v0.y; vertices[0][2] = t->v0.z;
    vertices[1][0] = t->v1.x; vertices[1][1] = t->v1.y; vertices[1][2] = t->v1.z;
    vertices[2][0] = t->v2.x; vertices[2][1] = t->v2.y; vertices[2][2] = t->v2.z;

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FTYPE, 0, &vertices[0][0]);
    glDrawArrays(mode, 0, 3);
    glDisableClientState(GL_VERTEX_ARRAY);
}

static void RenderTriSeperate(triRec *t)
{

    RenderTriangle(t, GL_TRIANGLES);
}

static void RenderTriStrip(triRec *t)
{

    RenderTriangle(t, GL_TRIANGLE_STRIP);
}

static void RenderTriFan(triRec *t)
{

    RenderTriangle(t, GL_TRIANGLE_FAN);
}

static ftype Error(edgeEqnRec *e, ftype x, ftype y)
{
    ftype offBy;

    offBy = y * e->slope.dx - x * e->slope.dy;
    offBy = e->intercept.bDx - offBy;

    if (e->length != ITOF(0)) {
	offBy /= e->length;
    } else {
	/*
	**  Everything is on the correct side of a trivial line.
	*/
	offBy = ITOF(0);
    }
    offBy = ABS(offBy);
    return offBy;
}

static long CheckIfIn(ftype x, ftype y, triRec *t, ftype error)
{
    long e1, e2, e3;
    ftype offBy1, offBy2, offBy3;

    e1 = t->edge01.Inequality(x, y, &t->edge01.slope, &t->edge01.intercept);
    e2 = t->edge02.Inequality(x, y, &t->edge02.slope, &t->edge02.intercept);
    e3 = t->edge12.Inequality(x, y, &t->edge12.slope, &t->edge12.intercept);

    if (e1 && e2 && e3) {
	return GL_TRUE;
    } else if (error == ITOF(0)) {
	return GL_FALSE;
    } else {
	if (!e1) {
	    offBy1 = Error(&t->edge01, x, y);
	    if (offBy1 >= error) {
		return GL_FALSE;
	    }
	}
	if (!e2) {
	    offBy2 = Error(&t->edge02, x, y);
	    if (offBy2 >= error) {
		return GL_FALSE;
	    }
	
	}
	if (!e3) {
	    offBy3 = Error(&t->edge12, x, y);
	    if (offBy3 >= error) {
		return GL_FALSE;
	    }
	}
    }
    return GL_TRUE;
}

static long CheckIfOut(ftype x, ftype y, triRec *t, ftype error)
{
    long e1, e2, e3;
    ftype offBy1, offBy2, offBy3, minError;

    minError = ITOF(0);

    e1 = !t->edge01.Inequality(x, y, &t->edge01.slope, &t->edge01.intercept);
    e2 = !t->edge02.Inequality(x, y, &t->edge02.slope, &t->edge02.intercept);
    e3 = !t->edge12.Inequality(x, y, &t->edge12.slope, &t->edge12.intercept);

    if (e1 || e2 || e3) {
	return GL_TRUE;
    } else if (error == ITOF(0)) {
	return GL_FALSE;
    } else {
	/*
	** Error is always positive
	*/
	offBy1 = Error(&t->edge01, x, y);
	offBy2 = Error(&t->edge02, x, y);
	offBy3 = Error(&t->edge12, x, y);

	minError = offBy1;
	minError = (offBy2 < minError) ? offBy2 : minError;
	minError = (offBy3 < minError) ? offBy3 : minError;
    }
    return (minError < error);
}

static long Compare(triRec *t, GLftype *buf)
{
    long flag, i, j;

    flag = NO_ERROR;
    for (i = 0; i < WINDSIZEY; i++) {
	for (j = 0; j < WINDSIZEX; j++) {
	    /*
	    ** If it is rendered, is it close to the ideal triangle?
	    ** If it is not rendered, is it sufficiently outside the
	    ** ideal triangle?
	    */
	    if (AutoColorCompare(buf[i*WINDSIZEX+j], COLOR_OFF) == GL_TRUE) {
		if (CheckIfOut(ITOF(j)+ITOF(1)/2, ITOF(i)+ITOF(1)/2, t, ITOF(1)/2) == GL_FALSE) {
		    errX = ITOF(j)+ITOF(1)/2; 
		    errY = ITOF(i)+ITOF(1)/2;
		    errExpectedDrawn = "primitive";
		    errObservedDrawn = "background";
		    flag = ERROR; } 
	    } else {
		if (CheckIfIn(ITOF(j)+ITOF(1)/2, ITOF(i)+ITOF(1)/2, t, ITOF(1)/2) == GL_FALSE) {
		    errX = ITOF(j)+ITOF(1)/2; 
		    errY = ITOF(i)+ITOF(1)/2;
		    errExpectedDrawn = "background";
		    errObservedDrawn = "primitive";
		    flag = ERROR;
		}
	    }
	}
    }
    return flag;
}

long TriRasterExec(void)
{
    GLftype *buf;
    triRec t;
    vertexRec v[3];
    long max, flag, i;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));
    flag = NO_ERROR;

    Ortho2D(ITOF(0), ITOF(WINDSIZEX), ITOF(0), ITOF(WINDSIZEY));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    max = (machine.pathLevel == 0) ? 50 : 5;
    for (i = 0; i < max; i++) {
	glClear(GL_COLOR_BUFFER_BIT);
	MakeVertices(v);
	MakeTri(&t, v);
	RenderTriSeperate(&t);
	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);
	if (Compare(&t, buf) == ERROR) {
	    StrMake(errStr, errStrFmt, "GL_TRIANGLES", t.v0.x, t.v0.y, t.v1.x,
		    t.v1.y, t.v2.x, t.v2.y, errX, errY, errExpectedDrawn,
		    errObservedDrawn);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    flag = ERROR;
	    break;
	}

	glClear(GL_COLOR_BUFFER_BIT);
	RenderTriStrip(&t);
	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);
	if (Compare(&t, buf) == ERROR) {
	    StrMake(errStr, errStrFmt, "GL_TRIANGLE_STRIP", t.v0.x, t.v0.y,
		    t.v1.x, t.v1.y, t.v2.x, t.v2.y, errX, errY,
		    errExpectedDrawn, errObservedDrawn);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    flag = ERROR;
	    break;
	}

	glClear(GL_COLOR_BUFFER_BIT);
	RenderTriFan(&t);
	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);
	if (Compare(&t, buf) == ERROR) {
	    StrMake(errStr, errStrFmt, "GL_TRIANGLE_FAN", t.v0.x, t.v0.y,
		    t.v1.x, t.v1.y, t.v2.x, t.v2.y, errX, errY,
		    errExpectedDrawn, errObservedDrawn);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    flag = ERROR;
	    break;
	}
    }

    FREE(buf);
    return flag;
}
