/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
**
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/userclip.c,v 1.7 2005/06/03 07:42:18 artlab Exp $
*/

/**
 * \file
 * \brief   User clip plane test.
 * \author  jasin@hybrid.fi
 *//*-------------------------------------------------------------------*/

/*
** userclip.c
** User Clip Plane Test
**
** Description -
**    The first test tests the user clip planes by drawing a point inside
**    the half space of a plane and checking that it got drawn. After this
**    the plane is flipped and the point is drawn again and it is checked
**    that nothing got drawn.
**
**    The second test tests the planes by drawing a triangle which fills
**    the entire screen with one user clip plane enabled using a gray
**    color and an additive blend mode. Then flips the plane and draws the
**    triangle again without clearing the screen. At this point the
**    entire screen should be gray. If there are black pixels, there have
**    been gaps. If there are white pixels, there has been some overlap.
**    
**    
**
** Error Explanation -
**    Failure occurs if the difference between the expected (gray) color
**    and a value in the color buffer is greater than color epsilon.
**
** Technical Specification -
**    Buffer requirements:
**        None.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilons.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"

#ifdef GL_OES_VERSION_1_1

static const GLftype expColor[4] =
{
    ITOF(1) / 2,
    ITOF(1) / 2,
    ITOF(1) / 2,
    ITOF(1) / 2
};

static char errStr[100];
static const char errStrFmt[] = "Found %s at location (%d,%d).";

static long Test(void)
{
    GLftype* buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*3*sizeof(GLftype));
    GLftype* ptr;
    int x, y;

    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);
    
    ptr = buf;

    /* Check that there are not white (overlap) or black (gap) pixels */
    for (y = 0; y < WINDSIZEY; y++)
        for (x = 0; x < WINDSIZEX; x++)
        {
            if (ABS(ptr[0]-expColor[0]) > epsilon.color[0] ||
                ABS(ptr[1]-expColor[1]) > epsilon.color[1] ||
                ABS(ptr[2]-expColor[2]) > epsilon.color[2])
            {
                StrMake(errStr, errStrFmt,
                    (ptr[0] < epsilon.zero ||
                     ptr[1] < epsilon.zero ||
                     ptr[2] < epsilon.zero) ? "a gap" : "overlap",
                     x,
                     y);
                ErrorReport(__FILE__, __LINE__, errStr);
                FREE(buf);
                return ERROR;
            }

            ptr += 3;
        }

    FREE(buf);
    return NO_ERROR;
}

static void DrawSplitTriangle(GLenum plane)
{
    /* 
    ** A large triangle which fills the whole screen. The z coordinates
    ** are randomized later.
    */
    GLftype vertices[3][3] =
    {
        {ITOF(-3), ITOF(-1)},
        {ITOF( 0), ITOF( 2)},
        {ITOF( 3), ITOF(-1)}
    };

    GLftype colors[3][4];
    GLftype eq[4];

    static const GLftype zMin = ITOF(-100);
    static const GLftype zMax = ITOF( 100);
    
    /* Randomize z coordinates */
    vertices[0][2] = OESRandom(zMin, zMax);
    vertices[1][2] = OESRandom(zMin, zMax);
    vertices[2][2] = OESRandom(zMin, zMax);

    colors[0][0] = expColor[0];
    colors[0][1] = expColor[1];
    colors[0][2] = expColor[2];
    colors[0][3] = expColor[3];

    colors[1][0] = expColor[0];
    colors[1][1] = expColor[1];
    colors[1][2] = expColor[2];
    colors[1][3] = expColor[3];

    colors[2][0] = expColor[0];
    colors[2][1] = expColor[1];
    colors[2][2] = expColor[2];
    colors[2][3] = expColor[3];


    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FTYPE, 0, &vertices[0][0]);

    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(4, GL_FTYPE, 0, &colors[0][0]);


    /* Generate plane equation. */
    eq[0] = OESRandom(-ITOF(2), ITOF(2));
    eq[1] = OESRandom(-ITOF(2), ITOF(2));
    eq[2] = (RAND() % 4 != 0) ? ITOF(0) : OESRandom(-ITOF(1)/10, ITOF(1)/10);
    eq[3] = (RAND() % 4 != 0) ? ITOF(0) : OESRandom(-ITOF(1)/100, ITOF(1)/100);
    
    glEnable(plane);
    glClipPlanet(plane, eq);

    glClear(GL_COLOR_BUFFER_BIT);

    /* Render the triangle */
    glDrawArrays(GL_TRIANGLES, 0, 3);
    
    /* Negate plane */
    eq[0] = -eq[0];
    eq[1] = -eq[1];
    eq[2] = -eq[2];
    eq[3] = -eq[3];

    glClipPlanet(plane, eq);

    /* Render again */
    glDrawArrays(GL_TRIANGLES, 0, 3);
    glDisable(plane);
    
    glDisableClientState(GL_COLOR_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
}

static int CountVisiblePixels(GLftype* buf, int xSize, int ySize)
{
    int i;
    int pixelCount = 0;

    for(i = 0; i < xSize*ySize; i++)
    {
        if (buf[0] > epsilon.zero ||
            buf[1] > epsilon.zero ||
            buf[2] > epsilon.zero)
            pixelCount ++;

        buf += 3;
    }

    return pixelCount;
}

static long TestHalfSpace(GLenum plane)
{
    GLftype vertex[3];
    GLftype color[4] = { ITOF(1), ITOF(1), ITOF(1), ITOF(1) };

    GLftype xyBound = DIV(ITOF(99), ITOF(100));
    GLftype zBound = ITOF(99);
    GLftype eq[4];

    GLftype* buf = (GLftype*)MALLOC(WINDSIZEX*WINDSIZEY*3*sizeof(GLftype));
    int pixelCount;

    /*
    ** Randomize a pixel, which isn't at the origin of the xy-plane but
    ** still visible.
    */
    do
    {
        vertex[0] = OESRandom(-xyBound, xyBound);
        vertex[1] = OESRandom(-xyBound, xyBound);
        vertex[2] = OESRandom(-zBound,  zBound);
    } while (vertex[0] == ITOF(0) && vertex[1] == ITOF(0));

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FTYPE, 0, &vertex[0]);

    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(4, GL_FTYPE, 0, &color[0]);

    /*
    ** A plane which should contain the pixel
    */
    eq[0] = vertex[0];
    eq[1] = vertex[1];
    eq[2] = vertex[2];
    eq[3] = ITOF(0);

    glEnable(plane);
    glClipPlanet(plane, eq);
    
    glClear(GL_COLOR_BUFFER_BIT);
    glDrawArrays(GL_POINTS, 0, 1);

    /*
    ** Check that exactly one pixel is drawn.
    */
    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);
    pixelCount = CountVisiblePixels(buf, WINDSIZEX, WINDSIZEY);
    if (pixelCount != 1)
    {
        FREE(buf);
        return ERROR;
    }

    /*
    ** Negate the plane so that the pixel doesn't get drawn
    */
    eq[0] = -eq[0];
    eq[1] = -eq[1];
    eq[2] = -eq[2];
    eq[3] = ITOF(0);

    glClipPlanet(plane, eq);
    glClear(GL_COLOR_BUFFER_BIT);

    glDrawArrays(GL_POINTS, 0, 1);

    /*
    ** Check that no pixels were drawn.
    */
    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);
    pixelCount = CountVisiblePixels(buf, WINDSIZEX, WINDSIZEY);
    if (pixelCount != 0)
    {
        FREE(buf);
        return ERROR;
    }

    FREE(buf);

    glDisable(plane);

    glDisableClientState(GL_COLOR_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);

    return NO_ERROR;
}

long UserClipExec(void)
{
    long i;
    long p;
    GLint numPlanes;

    long pStart;
    long max;

    glGetIntegerv(GL_MAX_CLIP_PLANES, &numPlanes);

    Ortho3D(-ITOF(1), ITOF(1), -ITOF(1), ITOF(1), -ITOF(100), ITOF(100));
    
    glClearColort(ITOF(0), ITOF(0), ITOF(0), ITOF(0));
    glShadeModel(GL_FLAT);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);
    glDisable(GL_MULTISAMPLE);
    glDisable(GL_DITHER);

    max     = (machine.pathLevel == 0) ? 30 : 3;
    pStart  = (machine.pathLevel == 0) ? 0 : (numPlanes - 1);

    for (p = pStart; p < numPlanes; p ++)
        for (i = 0; i < max; i ++)
        {
            if (TestHalfSpace(GL_CLIP_PLANE0 + p) == ERROR)
            {
                glDisable(GL_BLEND);
                glBlendFunc(GL_ONE, GL_ZERO);
                return ERROR;
            }

            DrawSplitTriangle(GL_CLIP_PLANE0 + p);
            if (Test() == ERROR)
            {
                glDisable(GL_BLEND);
                glBlendFunc(GL_ONE, GL_ZERO);
                return ERROR;
            }
        }

    glDisable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ZERO);
    
    return NO_ERROR;
}

#endif /* GL_OES_VERSION_1_1 */
