/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/vorder.c,v 1.3 2005/05/11 15:21:40 artlab Exp $
*/

/*
** vorder.c
** Vertex Order Test.
**
** Description -
**    Test that triangle strips and triangle fans 
**    preserve proper vertex order when folded back on
**    themselves. Front and back are assigned different colors
**    and drawn counter-clockwise.
**    
**    The test algorithm:
**        - Use glLightModel to assign different colors to
**        front and back.
**        - For both GL_TRIANGLE_STRIP and GL_TRIANGLE_FAN:
**            - Generate vertices of a strip/fan which folds
**            back on itself. Each primitive consists of
**            approximately 16 vertices. The folded-back region
**            is constructed so as to not obscure front-facing
**            vertices, except on the crease edge.
**        - Draw the strip/fan.
**        - Examine framebuffer at vertices to be sure
**        appropriate front or back color appeared. Since
**        vertex location may be on pixel boundaries, an
**        allowance is made for the vertex to be off one
**        pixel in any direction.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color and zero epsilons.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"


static char errStr[240];
static char errStrFmt[] = "Primitive type %s. Error at vertex %i at (%1.1f, %1.1f).";


static long Compare(long x, long y, long color, GLftype *buf)
{
    long i;
    long cmp;

    i = 3 * (WINDSIZEX * y + x);
    cmp = (ABS(buf[i]-colorMap[color][0]) <= epsilon.color[0] &&
	   ABS(buf[i+1]-colorMap[color][1]) <= epsilon.color[1] &&
	   ABS(buf[i+2]-colorMap[color][2]) <= epsilon.color[2]);
    return cmp;
}

static long TestBuf(GLftype *buf, GLftype *point, long index)
{
    long i = 0, j = 0;

    i = FTOI((MUL(ITOF(1) + point[0], ITOF(WINDSIZEX)) / 2) - ITOF(1)/2);
    j = FTOI((MUL(ITOF(1) + point[1], ITOF(WINDSIZEY)) / 2) - ITOF(1)/2);

    /*
    **  Test the 9 pixels around the target.
    */
    if (!(Compare(i  , j  , index, buf) ||
          Compare(i-1, j+1, index, buf) ||
          Compare(i  , j+1, index, buf) ||
          Compare(i+1, j+1, index, buf) ||
          Compare(i-1, j  , index, buf) ||
          Compare(i+1, j  , index, buf) ||
          Compare(i-1, j-1, index, buf) ||
          Compare(i  , j-1, index, buf) ||
          Compare(i+1, j-1, index, buf))) {
        return ERROR;
    }
    return NO_ERROR;
}

static void GeneratePoints(GLftype points[10][4], GLenum primType)
{
    GLftype deltaX, deltaY, rad;
    int i;

    switch (primType) {
      case GL_TRIANGLE_STRIP:
	deltaX = 5 * DIV(ITOF(2), ITOF(WINDSIZEX));
	deltaY = 20 * DIV(ITOF(2), ITOF(WINDSIZEY));

	for (i = 0; i < 8; i += 2) {
	    points[i][0] = deltaX * (i + 1) - ITOF(1);
	    points[i][1] = ITOF(0);
	    points[i][2] = ITOF(0);
	    points[i][3] = ITOF(1);
	    points[i+1][0] = deltaX * (i + 1) - ITOF(1);
	    points[i+1][1] = -deltaY;
	    points[i+1][2] = ITOF(0);
	    points[i+1][3] = ITOF(1);
	}
	deltaY = 3 * DIV(ITOF(2), ITOF(WINDSIZEY));
	for (; i < 16; i += 2) {
	    points[i][0] = (-ITOF(1) + deltaX) + 8 * deltaX;
	    points[i][1] = (i - 8) * deltaY;
	    points[i][2] = ITOF(0);
	    points[i][3] = ITOF(1);
	    points[i+1][0] = (-ITOF(1) + deltaX) + 6 * deltaX;
	    points[i+1][1] = (i - 8) * deltaY;
	    points[i+1][2] = ITOF(0);
	    points[i+1][3] = ITOF(1);
	}
	break;
      case GL_TRIANGLE_FAN:
	rad = ITOF(1)/2;
	points[0][0] = ITOF(0);
	points[0][1] = ITOF(0);
	points[0][2] = ITOF(0);
	points[0][3] = ITOF(1);
	for (i = 1; i < 8; i++) {
	    points[i][0] = MUL(rad, COS((i-1)*MUL(ITOF(50), PI)/180));
	    points[i][1] = MUL(rad, SIN((i-1)*MUL(ITOF(50), PI)/180));
	}
	rad = ATOF("0.8");
	for (; i < 16; i++) {
	    points[i][0] = MUL(rad, COS(MUL(ITOF(350-i*6), PI)/180));
	    points[i][1] = MUL(rad, SIN(MUL(ITOF(350-i*6), PI)/180));
	}
	break;
    }
}

long VertexOrderExec(void)
{
    static GLenum primType[3] = {
	GL_TRIANGLE_STRIP, GL_TRIANGLE_FAN
    };
    GLftype *buf;
    GLftype points[20][4];
    GLint mode, type, i;
    char typeStr[20];

    /*
    ** Test that triangle strips and triangle fans preserve
    ** proper vertex order when folded back on themselves. Front and
    ** back are assigned different colors and drawn counter-clockwise.
    */

    buf = (GLftype *)MALLOC(3*WINDSIZEX*WINDSIZEY*sizeof(GLftype));

    AutoClearColor(BLACK);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);
    glEnable(GL_CULL_FACE);

    for (type = 0; type < 2; type++) {
	GeneratePoints(points, primType[type]);
 
	for (mode = 0; mode < 1; mode++) {
	    glClear(GL_COLOR_BUFFER_BIT);

	    glCullFace(GL_BACK);

	    AutoColor(RED);
	    glEnableClientState(GL_VERTEX_ARRAY);
	    glVertexPointer(2, GL_FTYPE, 4*sizeof(GLftype), &points[0][0]);
	    glDrawArrays(primType[type], 0, 16);
	    glDisableClientState(GL_VERTEX_ARRAY);

	    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);

	    /*
	    ** Some crease vertices (and the trifan center) will be overwritten
	    ** by back facing polygons, so don't test them for front color.
	    */

	    for (i = 1; i < 6; i++) {
		if (TestBuf(buf, &points[i][0], RED) == ERROR) {
		    GetEnumName(primType[type], typeStr);
		    StrMake(errStr, errStrFmt, typeStr, i,
		     	    points[i][0], points[i][1]);
		    ErrorReport(__FILE__, __LINE__, errStr);
		    FREE(buf);
		    return ERROR;
		}
	    }
	    for (i = 8; i < 16; i++) {
		if (TestBuf(buf, &points[i][0], BLACK) == ERROR) {
		    GetEnumName(primType[type], typeStr);
		    StrMake(errStr, errStrFmt, typeStr, i,
		     	    points[i][0], points[i][1]);
		    ErrorReport(__FILE__, __LINE__, errStr);
		    FREE(buf);
		    return ERROR;
		}
	    }

	    glClear(GL_COLOR_BUFFER_BIT);

            glCullFace(GL_FRONT);

	    AutoColor(RED);
            glEnableClientState(GL_VERTEX_ARRAY);
            glVertexPointer(2, GL_FTYPE, 4*sizeof(GLftype), &points[0][0]);
            glDrawArrays(primType[type], 0, 16);
            glDisableClientState(GL_VERTEX_ARRAY);

            ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);

            /*
            ** Some crease vertices (and the trifan center) will be overwritten
            ** by back facing polygons, so don't test them for front color.
            */

            for (i = 1; i < 6; i++) {
                if (TestBuf(buf, &points[i][0], BLACK) == ERROR) {
                    GetEnumName(primType[type], typeStr);
                    StrMake(errStr, errStrFmt, typeStr, i,
                            points[i][0], points[i][1]);
                    ErrorReport(__FILE__, __LINE__, errStr);
                    FREE(buf);
                    return ERROR;
                }
            }
            for (i = 8; i < 16; i++) {
                if (TestBuf(buf, &points[i][0], RED) == ERROR) {
                    GetEnumName(primType[type], typeStr);
                    StrMake(errStr, errStrFmt, typeStr, i,
                            points[i][0], points[i][1]);
                    ErrorReport(__FILE__, __LINE__, errStr);
                    FREE(buf);
                    return ERROR;
                }
            }
	}
    }

    glDisable(GL_CULL_FACE);

    FREE(buf);
    return NO_ERROR;
}
