/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/xform.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** xform.c
** Transform Test.
**
** Description -
**    Tests glFrustum(), glOrtho(), glRotate(), glScale() and
**    glTranslate(). A single
**    point is drawn under various transformations and its new
**    position checked. The order of matrix multiplication is
**    also checked by executing pairs of rotates and translates.
**    The test also verifies that no assumption of matrix mode
**    GL_PROJECTION are tied to glOrtho() and glFrustum(); they
**    should be applied with the current matrix mode. Similarly,
**    glRotate(), glTranslate() and glScale() should not assume
**    GL_MODELVIEW.
**    
**    The sections of the test contain the following:
**    
**        Tests single tranformations. In each iteration of a loop,
**        creates either a translation, rotation, or scale call using
**        random data then checks resulting point location.
**    
**        Tests the order of matrix multiplications.
**        Calls a translation then a scale, checks results,
**        then calls a scale followed by a translation and
**        checks results.
**    
**        Test that glOrtho() and glFrustum() do not assume
**        GL_PROJECTION transformation. In GL_MODELVIEW mode, make
**        calls to glOrtho() and glFrumtum(). Check that glGet()
**        returns tranformed GL_MODELVIEW matrix. Check that
**        glGet() shows GL_PROJECTION is still identity. For
**        glOrtho(), overwrite GL_PROJECTION matrix identity and
**        verify the point drawn correctly demonstrates the
**        GL_MODELVIEW transform.
**    
**        Test that glScale(), glTranlate() and glRotate() do not
**        assume GL_MODELVIEW transformation. In GL_PROJECTION
**        mode, make calls to translate, rotate and scale. Check
**        that Get returns proper GL_PROJECTION matrix. Check that
**        Get shows GL_MODELVIEW matrix is still identity. For each
**        transformation, overwrite GL_MODELVIEW matrix with
**        identity and verify that the drawn point demonstrates new
**        GL_PROJECTION matrix.
**
** Error Explanation -
**    Since values may fall on pixel boundaries, and
**    transformations have numerical error, allowance is made for
**    pixel to be off by one in any direction from expected
**    location.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        The routine looks for the transformed point in a 3x3 grid of pixels
**        around the expected location, to allow a margin for matrix
**        operations.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilm.h"


static char errStr[240];
static char errStrFmt[] = "Original point at (%1.2f, %1.2f), transformed point expected at (%1.2f, %1.2f).";

long XFormExec(void)
{
    static const GLftype xformData[3][4] = {
	{ITOF(1)/10, ITOF(2)/10, ITOF(0), ITOF(0)},  /* translate */
	{ITOF(3), ITOF(2), ITOF(1), ITOF(0)},  /* scale     */
	{ITOF(0), ITOF(0), ITOF(1), ITOF(37)}  /* rotate    */
    };
    GLftype *buf;
    xFormRec data, translateData, scaleData;
    GLftype iMatrix[16];
    GLftype point[4], xFormPoint[4];
    GLftype primWidth;
    long i, type, max;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    glMatrixMode(GL_MODELVIEW);

    primWidth = DIV(ITOF(2), ITOF(WINDSIZEX));

    /*
    ** Test position of point after various transformations.
    */

    point[0] = OESRandom(-ITOF(1)/4, ITOF(1)/4);
    point[1] = OESRandom(-ITOF(1)/4, ITOF(1)/4);
    point[2] = ITOF(0);
    point[3] = ITOF(1);

    max = (machine.pathLevel == 0) ? 40 : 2;
    for (i = 1; i < max; i++) {
	glClear(GL_COLOR_BUFFER_BIT);

	glPushMatrix();

	XFormMake(&type, &data);
	if (XFormValid(type, &data) == ERROR) {
	    continue;
	}
	XFormCalcGL(type, &data);

	point[2] = (type >= XFORM_ORTHO) ? ITOF(-2) : ITOF(0);
	DrawPoint4fv(point);

	glPopMatrix();

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

	XFormCalcTrue(type, &data, point, xFormPoint);

	if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
	    StrMake(errStr, errStrFmt, point[0], point[1],
	            xFormPoint[0], xFormPoint[1]);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}
    }

    /*
    ** Test order of matrix multiplication.
    */

    glClear(GL_COLOR_BUFFER_BIT);

    glPushMatrix();

    point[0] = primWidth;
    point[1] = primWidth;
    point[2] = ITOF(0);

    translateData.x = 10 * primWidth;
    translateData.y = 10 * primWidth;
    translateData.z = ITOF(0);
    XFormCalcGL(XFORM_TRANSLATE, &translateData);

    scaleData.x = ITOF(1);
    scaleData.y = ITOF(2);
    scaleData.z = ITOF(1);
    XFormCalcGL(XFORM_SCALE, &scaleData);

    DrawPoint4fv(point);

    glPopMatrix();

    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);
    XFormCalcTrue(XFORM_SCALE, &scaleData, point, xFormPoint);
    XFormCalcTrue(XFORM_TRANSLATE, &translateData, xFormPoint, xFormPoint);
    if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
	StrMake(errStr, errStrFmt, point[0], point[1],
		xFormPoint[0], xFormPoint[1]);
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    glClear(GL_COLOR_BUFFER_BIT);

    glPushMatrix();

    point[0] = primWidth;
    point[1] = primWidth;

    scaleData.x = ITOF(1);
    scaleData.y = ITOF(2);
    scaleData.z = ITOF(1);
    XFormCalcGL(XFORM_SCALE, &scaleData);

    translateData.x = 10 * primWidth;
    translateData.y = 10 * primWidth;
    translateData.z = ITOF(0);
    XFormCalcGL(XFORM_TRANSLATE, &translateData);

    DrawPoint4fv(point);

    glPopMatrix();

    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);
    XFormCalcTrue(XFORM_TRANSLATE, &translateData, point, xFormPoint);
    XFormCalcTrue(XFORM_SCALE, &scaleData, xFormPoint, xFormPoint);
    if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
	StrMake(errStr, errStrFmt, point[0], point[1],
		xFormPoint[0], xFormPoint[1]);
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    /*
    ** Make sure that Ortho and Frustum do not assume PROJECTION transformation:
    ** - For Ortho and Frustum, overwrite Projection matrix identity and verify
    **   the point drawn correctly demonstrates the ModelView transform.
    */

    glClear(GL_COLOR_BUFFER_BIT);

    MakeIdentMatrix(iMatrix);
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixt(iMatrix);

    glPushMatrix();

    data.d1 = ITOF(33);
    data.d2 = ITOF(7);
    data.d3 = ITOF(40);

    XFormCalcGL(XFORM_ORTHO, &data);

    glMatrixMode(GL_PROJECTION);
    glLoadMatrixt(iMatrix);

    DrawPoint4fv(point);

    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);
    XFormCalcTrue(XFORM_ORTHO, &data, point, xFormPoint);
    if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
	glMatrixMode(GL_MODELVIEW);
	StrMake(errStr, errStrFmt, point[0], point[1],
		xFormPoint[0], xFormPoint[1]);
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    glMatrixMode(GL_MODELVIEW);

    glPopMatrix();
    glPushMatrix();

    data.d1 = -ITOF(4);
    data.d2 = ITOF(3);
    data.d3 = -ITOF(1);
    data.d4 = ITOF(3);
    data.d5 = ITOF(1);
    data.d6 = ITOF(6);

    XFormCalcGL(XFORM_FRUSTUM, &data);

    point[0] = OESRandom(-ITOF(1)/2, ITOF(1)/2);
    point[1] = OESRandom(-ITOF(1)/2, ITOF(1)/2);
    point[2] = OESRandom(-ITOF(2), -ITOF(5));

    glClear(GL_COLOR_BUFFER_BIT);

    DrawPoint4fv(point);

    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

    XFormCalcTrue(XFORM_FRUSTUM, &data, point, xFormPoint);
    if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
	StrMake(errStr, errStrFmt, point[0], point[1],
	xFormPoint[0], xFormPoint[1]);
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    glPopMatrix();

    /*
    ** Make sure Scale, Translate and Rotate do not assume MODELVIEW
    ** transformation:
    ** - Check that Get returns proper Projection matrix.
    ** - Check that Get shows ModelView matrix is still identity.
    ** - For each transformation, overwrite ModelView matrix with identity
    **   and verify that the drawn point demonstrates new Projection matrix.
    */

    point[0] = primWidth;
    point[1] = primWidth;
    point[2] = ITOF(0);

    for (type = XFORM_TRANSLATE; type < XFORM_ORTHO; type++) {
	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();

	data.x = xformData[type][0];
	data.y = xformData[type][1];
	data.z = xformData[type][2];
	data.angle = xformData[type][3];

	XFormCalcGL(type, &data);

	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixt(iMatrix);

	DrawPoint4fv(point);

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);
	XFormCalcTrue(type, &data, point, xFormPoint);
	if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
	    StrMake(errStr, errStrFmt, point[0], point[1],
		    xFormPoint[0], xFormPoint[1]);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
    }

    glMatrixMode(GL_MODELVIEW);

    FREE(buf);
    return NO_ERROR;
}
