/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/xformmix.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** xformmix.c
** Matrix Stack Mixing Test.
**
** Description -
**    Tests that matrix stacks can be mixed. A series of
**    transformations are applied while the matrix mode is
**    switched with each glPushMatrix() and glPopMatrix(). After
**    a series of transformations, a small quad is drawn and its
**    position is checked.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        The routine looks for the transformed point in a 3x3 grid of pixels
**        around the expected location, to allow a margin for matrix
**        operations.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"


static char errStr[240];
static char errStrFmt[] = "Tranformation matrix has diagonals set to %2.2f, and internal variable (pattern) for matrices applied is %d.";


static void Make(GLftype value, GLftype *buf)
{
    long i;

    for (i = 0; i < 16; i++) {
	buf[i] = ITOF(0);
    }
    for (i = 0; i < 3; i++) {
	buf[i*4+i] = value;
    }
    buf[15] = ITOF(1);
}

static long Compare(long x, long y, GLint index, GLftype *buf)
{
    long i;

    i = WINDSIZEX * y + x;
    return !AutoColorCompare(buf[i], index);
}

static long TestBuffer(long size, GLftype *buf)
{
    long midX, midY;

    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

    midX = WINDSIZEX / 2;
    midY = WINDSIZEY / 2;

    if (Compare(midX-size-2, midY-size-2, COLOR_OFF, buf) ||
	Compare(midX-size-2, midY-size+2, COLOR_OFF, buf) ||
	Compare(midX-size+2, midY-size+2, COLOR_ON,  buf) ||
	Compare(midX-size+2, midY-size-2, COLOR_OFF, buf)) {
	return ERROR;
    }
    if (Compare(midX-size-2, midY+size-2, COLOR_OFF, buf) ||
	Compare(midX-size-2, midY+size+2, COLOR_OFF, buf) ||
	Compare(midX-size+2, midY+size+2, COLOR_OFF, buf) ||
	Compare(midX-size+2, midY+size-2, COLOR_ON,  buf)) {
	return ERROR;
    }
    if (Compare(midX+size-2, midY+size-2, COLOR_ON,  buf) ||
	Compare(midX+size-2, midY+size+2, COLOR_OFF, buf) ||
	Compare(midX+size+2, midY+size+2, COLOR_OFF, buf) ||
	Compare(midX+size+2, midY+size-2, COLOR_OFF, buf)) {
	return ERROR;
    }
    if (Compare(midX+size-2, midY-size-2, COLOR_OFF, buf) ||
	Compare(midX+size-2, midY-size+2, COLOR_ON,  buf) ||
	Compare(midX+size+2, midY-size+2, COLOR_OFF, buf) ||
	Compare(midX+size+2, midY-size-2, COLOR_OFF, buf)) {
	return ERROR;
    }

    return NO_ERROR;
}

static long MixAndCompare(long pattern, GLenum *mode, GLftype *buf)
{
    static const GLftype vertices[4][2] = {
	{-ITOF(2) / WINDSIZEX, -ITOF(2) / WINDSIZEY},
	{+ITOF(2) / WINDSIZEX, -ITOF(2) / WINDSIZEY},
	{-ITOF(2) / WINDSIZEX, +ITOF(2) / WINDSIZEY},
	{+ITOF(2) / WINDSIZEX, +ITOF(2) / WINDSIZEY}
    };
    GLftype matrix[16], x;

    glClear(GL_COLOR_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);
    x = OESRandom(ITOF(4), MIN(WINDSIZEX, WINDSIZEY)*ATOF("0.30"));
    Make(x, matrix);
    glLoadMatrixt(matrix);

    if (pattern & 1) {
	glMatrixMode(mode[0]);
	Make(OESRandom(ITOF(1), ITOF(50)), matrix);
	glPushMatrix();
	glLoadMatrixt(matrix);
    }
    if (pattern & 2) {
	glMatrixMode(mode[1]);
	Make(OESRandom(ITOF(1), ITOF(50)), matrix);
	glPushMatrix();
	glLoadMatrixt(matrix);
    }
    if (pattern & 4) {
	glMatrixMode(mode[2]);
	Make(OESRandom(ITOF(1), ITOF(50)), matrix);
	glPushMatrix();
	glLoadMatrixt(matrix);
    }
    if (pattern & 1) {
	glMatrixMode(mode[0]);
	glPopMatrix();
    }
    if (pattern & 2) {
	glMatrixMode(mode[1]);
	glPopMatrix();
    }
    if (pattern & 4) {
	glMatrixMode(mode[2]);
	glPopMatrix();
    }

    glMatrixMode(GL_MODELVIEW);
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(2, GL_FTYPE, 0, &vertices[0][0]);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    glDisableClientState(GL_VERTEX_ARRAY);

    if (TestBuffer(FTOI(x), buf) == ERROR) {
	StrMake(errStr, errStrFmt, x, pattern);
	return ERROR;
    }

    return NO_ERROR;
}

static long Test(GLenum *mode, GLftype *buf)
{
    long i;

    if (machine.pathLevel == 0) {
	for (i = 1; i <= 7; i++) {
	    if (MixAndCompare(i, mode, buf) == ERROR) {
		return ERROR;
	    }
	}
    } else {
	if (MixAndCompare(FTOI(OESRandom(ITOF(1), ITOF(7))), mode, buf) == ERROR) {
	    return ERROR;
	}
    }

    return NO_ERROR;
}

long XFormMixExec(void)
{
    GLftype *buf;
    GLenum mode[3];

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    mode[0] = GL_MODELVIEW;
    mode[1] = GL_PROJECTION;
    mode[2] = GL_TEXTURE;
    if (Test(mode, buf) == ERROR) {
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    mode[0] = GL_PROJECTION;
    mode[1] = GL_MODELVIEW;
    mode[2] = GL_TEXTURE;
    if (Test(mode, buf) == ERROR) {
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    mode[0] = GL_MODELVIEW;
    mode[1] = GL_TEXTURE;
    mode[2] = GL_PROJECTION;
    if (Test(mode, buf) == ERROR) {
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    mode[0] = GL_TEXTURE;
    mode[1] = GL_PROJECTION;
    mode[2] = GL_MODELVIEW;
    if (Test(mode, buf) == ERROR) {
	ErrorReport(__FILE__, __LINE__, errStr);
	FREE(buf);
	return ERROR;
    }

    FREE(buf);
    return NO_ERROR;
}
