/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/xformn.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** xformn.c
** Transform Normal Test.
**
** Description -
**    Tests that the normal vector is correctly multiplied by the
**    inverse of the modelview matrix. To test this, a polygon is
**    drawn with lighting parameters set in such a way that most
**    values drop out and the resulting pixel color is a function
**    of the normal vector.
**    
**    The equation reduces to c = max(0, n dot VP)
**    and (n dot VP) reduces to cos(theta).
**    
**    The polygon is rotated around various axes. Initially the
**    light is fixed while the polygon rotates. At each rotation,
**    the color of the fragment at the center of the window
**    (position 0, 0, 0) is read and compared with expected
**    results, from the equations above. In the second section,
**    the light source rotated about an axis, always at a
**    distance of ITOF(1) from the origin. This section is repeated,
**    taking both the light and the primitive through rotations
**    about different axes.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Color epsilon.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"


/* Range of angles to run the tests over */
#define ROTANGLE 87

static char errStr[240];
static char errStrFmtRGB[] = "Angle between normal and light is %4.2f. Expected color is (%1.2f, %1.2f, %1.2f).";


static long TestNormalRGB(GLftype angle, GLftype *buf)
{
    GLftype cosine;
    long midX, midY, i;

    midX = WINDSIZEX / 2;
    midY = WINDSIZEY / 2;
   
    i = WINDSIZEX * midY * 3 + midX * 3;
    cosine = COS(MUL(angle, PI)/180);
    cosine = (cosine < ITOF(0)) ? ITOF(0) : cosine;

    if (ABS(buf[i]-cosine) < epsilon.color[0] &&
	ABS(buf[i+1]-cosine) < epsilon.color[1] &&
	ABS(buf[i+2]-cosine) < epsilon.color[2]) {
	return NO_ERROR;
    }

    StrMake(errStr, errStrFmtRGB, angle, cosine, cosine, cosine);
    return ERROR;
}

long XFormNormalExec(void)
{
    static const GLftype triangle[3][2] = {
	{-ITOF(7)/10, ITOF(6)/10},
	{-ITOF(7)/10, -ITOF(6)/10},
	{ITOF(7)/10, ITOF(0)}
    };
    static const GLftype quad[4][2] = {
	{-ITOF(7)/10, -ITOF(7)/10},
	{+ITOF(7)/10, -ITOF(7)/10},
	{-ITOF(7)/10, +ITOF(7)/10},
	{+ITOF(7)/10, +ITOF(7)/10}
    };
    static const GLftype quad_45[4][3] = {
	{-ITOF(1)/2, -ITOF(3)/10, -ITOF(1)/2},
	{+ITOF(1)/2, -ITOF(3)/10, -ITOF(1)/2},
	{-ITOF(1)/2, +ITOF(3)/10, +ITOF(1)/2},
	{+ITOF(1)/2, +ITOF(3)/10, +ITOF(1)/2}
    };
    static const GLftype front_mat_ambient[] = {ITOF(0), ITOF(0), ITOF(0), ITOF(1)};
    static const GLftype front_mat_diffuse[] = {ITOF(1), ITOF(1), ITOF(1), ITOF(1)};
    static const GLftype front_mat_specular[] = {ITOF(0), ITOF(0), ITOF(0), ITOF(1)};
    static GLftype position[] = {ITOF(0), ITOF(0), ITOF(1), ITOF(0)};
    GLftype *buf;
    GLftype polyAngle, lightAngle, step;
    GLint i;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*3*sizeof(GLftype));

    glViewport(0, 0, WINDSIZEX, WINDSIZEY);

    AutoClearColor(BLACK);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    /*
    ** To test value of normal, set lighting equation parameters so that
    ** most values drop out.  The equation reduces to
    **	    c = max(0, n dot VP)
    **      and (n dot VP) reduces to cos(theta),
    ** where theta is the angle  between the normal and the vector from
    ** the vertex to the light.
    */
    glMaterialtv(GL_FRONT_AND_BACK, GL_AMBIENT, front_mat_ambient);
    glMaterialtv(GL_FRONT_AND_BACK, GL_SPECULAR, front_mat_specular);
    glMaterialtv(GL_FRONT_AND_BACK, GL_DIFFUSE, front_mat_diffuse);

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    /*
    ** Test the normal after rotation by rotating the polygon about the
    ** y-axis and examining the resulting color of at a point on the y-axis.
    */
    step = (machine.pathLevel == 0) ? ITOF(12) : ITOF(45);
    for (polyAngle = -ITOF(ROTANGLE); polyAngle < ITOF(ROTANGLE); polyAngle += step) {
	glClear(GL_COLOR_BUFFER_BIT);

	glPushMatrix();

	glRotatet(polyAngle, ITOF(0), ITOF(1), ITOF(0));

	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(2, GL_FTYPE, 0, &triangle[0][0]);
	glDrawArrays(GL_TRIANGLES, 0, 3);
	glDisableClientState(GL_VERTEX_ARRAY);

	glPopMatrix();

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);
	if (TestNormalRGB(polyAngle, buf) == ERROR) {
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}
    }

    /*
    ** Move the light source to locations on a circle of radius one in the xz
    ** plane. Rotate the polygon around the y-axis.
    */
    for (i = 0; i < 4; i++) {
	lightAngle = OESRandom(-ITOF(ROTANGLE), ITOF(ROTANGLE));
	position[0] = SIN(MUL(lightAngle, PI)/180);
	position[1] = ITOF(0);
	position[2] = COS(MUL(lightAngle, PI)/180);
	glLighttv(GL_LIGHT0, GL_POSITION, position);

	step = (machine.pathLevel == 0) ? ITOF(12) : ITOF(45);
	for (polyAngle = -ITOF(ROTANGLE); polyAngle < ITOF(ROTANGLE); polyAngle += step) {
	    glClear(GL_COLOR_BUFFER_BIT);

	    glPushMatrix();

	    glRotatet(polyAngle, ITOF(0), ITOF(1), ITOF(0));

	    glEnableClientState(GL_VERTEX_ARRAY);
	    glVertexPointer(2, GL_FTYPE, 0, &triangle[0][0]);
	    glDrawArrays(GL_TRIANGLES, 0, 3);
	    glDisableClientState(GL_VERTEX_ARRAY);

	    glPopMatrix();

	    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);
	    if (TestNormalRGB(lightAngle-polyAngle, buf) == ERROR) {
		ErrorReport(__FILE__, __LINE__, errStr);
		FREE(buf);
		return ERROR;
	    }
	}
    }

    /*
    ** Move the light source to locations on a circle of radius one in the yz
    ** plane. Rotate the polygon around the x-axis.
    */
    for (i = 0; i < 4; i++) {
	lightAngle = OESRandom(-ITOF(ROTANGLE), ITOF(ROTANGLE));
	position[0] = ITOF(0);
	position[1] = -SIN(MUL(lightAngle, PI)/180);
	position[2] = COS(MUL(lightAngle, PI)/180);
	glLighttv(GL_LIGHT0, GL_POSITION, position);

	step = (machine.pathLevel == 0) ? ITOF(12) : ITOF(45);
	for (polyAngle = -ITOF(ROTANGLE); polyAngle < ITOF(ROTANGLE); polyAngle += step) {
	    glClear(GL_COLOR_BUFFER_BIT);

	    glPushMatrix();

	    glRotatet(polyAngle, ITOF(1), ITOF(0), ITOF(0));

	    glEnableClientState(GL_VERTEX_ARRAY);
	    glVertexPointer(2, GL_FTYPE, 0, &quad[0][0]);
	    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	    glDisableClientState(GL_VERTEX_ARRAY);

	    glPopMatrix();

	    ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);
	    if (TestNormalRGB(lightAngle-polyAngle, buf) == ERROR) {
		ErrorReport(__FILE__, __LINE__, errStr);
		FREE(buf);
		return ERROR;
	    }
	}
    }

    /*
    ** Repeat the test on a 45 degree axis.
    ** The light is fixed at a distance 1 from the origin, tilted 45 degrees
    ** down in the yz plane. The polygon is a rectangle tilted 45 degrees about
    ** the x-axis, rotating about the (0, 1, 1) axis.
    */
    position[0] = ITOF(0);
    position[1] = -SIN(45*PI/180);
    position[2] = COS(45*PI/180);
    glLighttv(GL_LIGHT0, GL_POSITION, position);

    step = (machine.pathLevel == 0) ? ITOF(12) : ITOF(45);
    for (polyAngle = -ITOF(60); polyAngle < ITOF(60); polyAngle += step) {
	glClear(GL_COLOR_BUFFER_BIT);

	glPushMatrix();

	glRotatet(polyAngle, ITOF(0), ITOF(1), ITOF(1));

	glNormal3t(position[0], position[1], position[2]);
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FTYPE, 0, &quad_45[0][0]);
	glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
	glDisableClientState(GL_VERTEX_ARRAY);

	glPopMatrix();

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_RGB, buf);
	if (TestNormalRGB(polyAngle, buf) == ERROR) {
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}
    }

    FREE(buf);
    return NO_ERROR;
}
