/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/xformvp.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** xformvp.c
** Viewport Transformation Test.
**
** Description -
**    Tests that viewport computations are correct. The test
**    randomly selects values to define a viewport within the
**    window. A point is drawn with this new viewport and its
**    position in the window is examined.
**
** Error Explanation -
**    An allowance is made for the point position to be off by
**    one pixel in any direction.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        Full color.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        The routine looks for the transformed point in a 3x3 grid of pixels
**        around the expected location, to allow a margin for matrix
**        operations.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilm.h"


static char errStr[240];
static char errStrFmt[] = "Error during test iteration %d, with glViewport parameters x = %d, y = %d, w = %d, h = %d.";


typedef struct _vpDataRec {
    GLint x, y, w, h;
    GLftype n, f;
} vpDataRec;


static long Compare(long x, long y, GLint index, GLftype *buf)
{
    long i;

    i = WINDSIZEX * y + x;
    return AutoColorCompare(buf[i], index);
}

static long TestBuf(GLftype *buf, GLftype *point)
{
    long i, j;

    i = FTOI(point[0]);
    j = FTOI(point[1]);

    /*
    **  Test the 9 pixels around the target.
    */
    if (!(Compare(i  , j  , COLOR_ON, buf) ||
	  Compare(i-1, j+1, COLOR_ON, buf) ||
          Compare(i  , j+1, COLOR_ON, buf) ||
          Compare(i+1, j+1, COLOR_ON, buf) ||
          Compare(i-1, j  , COLOR_ON, buf) ||
          Compare(i+1, j  , COLOR_ON, buf) ||
          Compare(i-1, j-1, COLOR_ON, buf) ||
          Compare(i  , j-1, COLOR_ON, buf) ||
          Compare(i+1, j-1, COLOR_ON, buf))) {
        return ERROR;
    }
    return NO_ERROR;
}

#if 1
/*
** using 1 << zBits causes an overflow for fixed point type.
** It's ok as z values are not used anyway.
*/
static void ViewportXForm(GLftype *point, GLftype *xformPoint, vpDataRec *data)
{
    GLftype midX, midY;

    midX = ITOF(data->x) + ITOF(data->w) / 2;
    midY = ITOF(data->y) + ITOF(data->h) / 2;

    xformPoint[0] = MUL(ITOF(data->w) / 2, point[0]) + midX;
    xformPoint[1] = MUL(ITOF(data->h) / 2, point[1]) + midY;
    xformPoint[2] = 42;
}
#else
static void ViewportXForm(GLftype *point, GLftype *xformPoint, vpDataRec *data)
{
    ftype maxZ;
    GLftype midX, midY, zNear, zFar;

    midX = ITOF(data->x) + ITOF(data->w) / 2;
    midY = ITOF(data->y) + ITOF(data->h) / 2;

    maxZ = LDEXP(ITOF(1), buffer.zBits - 2) - ITOF(1);

    zFar = MUL(maxZ, data->f);
    zNear = MUL(maxZ, data->n);
    xformPoint[0] = MUL(ITOF(data->w) / 2, point[0]) + midX;
    xformPoint[1] = MUL(ITOF(data->h) / 2, point[1]) + midY;
    xformPoint[2] = MUL((zFar - zNear) / 2, point[2]) + (zFar + zNear) / 2;
}
#endif

long XFormViewportExec(void)
{
    GLftype *buf;
    vpDataRec data;
    GLftype point[4], xformPoint[4];
    long flag, max, i;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));
    flag = NO_ERROR;

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    max = (machine.pathLevel == 0) ? 20 : 1;
    for (i = 0; i < max; i++) {
	glClear(GL_COLOR_BUFFER_BIT);

	data.x = FTOI(OESRandom(ATOF("1.1"), ITOF(WINDSIZEX) - ATOF("2.1")));
	data.y = FTOI(OESRandom(ATOF("1.1"), ITOF(WINDSIZEY) - ATOF("2.1")));
	data.w = FTOI(OESRandom(ATOF("1.1"), ITOF(WINDSIZEX-data.x)));
	data.h = FTOI(OESRandom(ATOF("1.1"), ITOF(WINDSIZEY-data.y)));
	data.n = ITOF(0);
	data.f = ITOF(1);

	point[0] = OESRandom(-ITOF(1), ITOF(1));
	point[1] = OESRandom(-ITOF(1), ITOF(1));
	point[2] = OESRandom(-ITOF(1), ITOF(1));
	point[3] = OESRandom(ITOF(1), ITOF(10));

	glViewport(data.x, data.y, data.w, data.h);

	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(4, GL_FTYPE, 0, &point[0]);
	glDrawArrays(GL_POINTS, 0, 1);
	glDisableClientState(GL_VERTEX_ARRAY);

	/*
	** Create normalized device coordinates.
	*/
	point[0] = DIV(point[0], point[3]);
	point[1] = DIV(point[1], point[3]);
	point[2] = DIV(point[2], point[3]);

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

	ViewportXForm(point, xformPoint, &data);
	if (TestBuf(buf, xformPoint) == ERROR) {
	    StrMake(errStr, errStrFmt, i, data.x, data.y, data.w, data.h); 
	    ErrorReport(__FILE__, __LINE__, errStr);
	    flag = ERROR;
	    break;
	}
    }

    FREE(buf);
    return flag;
}
