/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/conform/conform/xformw.c,v 1.4 2005/05/11 15:21:40 artlab Exp $
*/

/*
** xform.c
** Transform Homogenous Coordinates Test.
**
** Description -
**    This test determines that the implementation can properly handle 
**    four-component (homogeneous) vertex position input properly. This 
**    is important when OpenGL ES is used only as a rasterizer, i.e., vertex 
**    transformation and lighting has been performed by the application 
**    itself.
**
**    The test consists of two subtests:
**    
**    Test 1:
**        A simple test to see if the clipper and rasterizer can fulfill these
**        conditions (note that this does not test the transformation pipeline
**        itself). Modelview and projection matrices are set to identity.
**        Random 16.16 fixed point input vertex positions are generated, such
**        that max3(abs(x),abs(y),abs(z)) is smaler than w. This condition
**        guarantees that the transformed vertex always falls inside the view
**        frustum. A point is rendered with this vertex and verified that it
**        was actually drawn.
**
**    Test 2:
**        Another property of homogeneous coordinates is that if all
**        components (including W) are multiplied by a positive non-zero
**        scalar so that no overflows happen in the multiplication itself,
**        the vertex should transform into the same position on the screen
**        (after division by W).
**        The tolerance of the drawn pixel position is +/- 1 pixel. A bigger
**        difference could indicate an overflow.
**        The scalar multiplier is chosen in the range [0.25, 50.0] such that
**        fixed-point implementations would not lose too many bits.
**
** Technical Specification -
**    Buffer requirements:
**        Color buffer.
**    Color requirements:
**        2 colors.
**    States requirements:
**        Disabled = GL_DITHER, GL_MULTISAMPLE.
**    Error epsilon:
**        Test 2 looks for the transformed point in a 3x3 grid of pixels
**        around the expected location, to allow a margin for matrix
**        operations.
**    Paths:
**        Allowed = Alpha, Blend, Depth, Fog, Logicop, Shade, Stencil.
**        Not allowed = Alias, Dither.
*/

#include "conform.h"
#include "util.h"
#include "utilm.h"


static char errStr[240];
static char *errStrFmt[2] = {
    "Number of pixels drawn was %d, should be 1.\n",
    "Point expected at (%1.2f, %1.2f)."
};

static long TestBuf(GLftype const *buf, GLint color, long *pixel_count)
{
    long i, j, k;

    k = 0;
    *pixel_count = 0;

    for (i = 0; i < WINDSIZEY; i++) {
	for (j = 0; j < WINDSIZEX; j++) {
	    if (AutoColorCompare(buf[k], color)) {
		(*pixel_count)++;
	    }
	    k++;
	}
    }

    if (*pixel_count == 1)
	return NO_ERROR;
    else
	return ERROR;
}

long XFormHomogenousExec(void)
{
    GLftype *buf;
    GLftype point[4], xFormPoint[4], scale;
    long i, max, pixel_count;

    buf = (GLftype *)MALLOC(WINDSIZEX*WINDSIZEY*sizeof(GLftype));

    AutoClearColor(COLOR_OFF);
    AutoColor(COLOR_ON);
    glDisable(GL_DITHER);
    glDisable(GL_MULTISAMPLE);

    /*
    ** Test homogenous coordinates.
    */

    max = (machine.pathLevel == 0) ? 40 : 2;
    for (i = 0; i < max; i++) {
	glClear(GL_COLOR_BUFFER_BIT);

	point[0] = OESRandom(-ITOF(1024), ITOF(1024));
	point[1] = OESRandom(-ITOF(1024), ITOF(1024));
	point[2] = OESRandom(-ITOF(1024), ITOF(1024));
	point[3] = MAX(MAX(ABS(point[0]), ABS(point[1])), ABS(point[2]))*9/8;

	DrawPoint4fv(point);

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

	if (TestBuf(buf, COLOR_ON, &pixel_count) == ERROR) {
	    StrMake(errStr, errStrFmt[0], pixel_count); 
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}
    }

    /*
    ** Test scaled homogenous coordinates.
    */

    max = (machine.pathLevel == 0) ? 40 : 2;
    for (i = 0; i < max; i++) {
	glClear(GL_COLOR_BUFFER_BIT);

	point[0] = OESRandom(-ITOF(128), ITOF(128));
	point[1] = OESRandom(-ITOF(128), ITOF(128));
	point[2] = OESRandom(-ITOF(128), ITOF(128));
	point[3] = MAX(MAX(ABS(point[0]), ABS(point[1])), ABS(point[2]))*9/8;

        xFormPoint[0] = DIV(point[0], point[3]);
        xFormPoint[1] = DIV(point[1], point[3]);
        xFormPoint[2] = DIV(point[2], point[3]);
        xFormPoint[3] = ITOF(1);

	scale = OESRandom(ITOF(1)/4, ITOF(50));

	point[0] = MUL(point[0], scale);
	point[1] = MUL(point[1], scale);
	point[2] = MUL(point[2], scale);
	point[3] = MUL(point[3], scale);

	DrawPoint4fv(point);

	ReadScreen(0, 0, WINDSIZEX, WINDSIZEY, GL_AUTO_COLOR, buf);

	if (XFormTestBuf(buf, xFormPoint, COLOR_ON) == ERROR) {
	    StrMake(errStr, errStrFmt[1], xFormPoint[0], xFormPoint[1]);
	    ErrorReport(__FILE__, __LINE__, errStr);
	    FREE(buf);
	    return ERROR;
	}
    }

    FREE(buf);
    return NO_ERROR;
}
