/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
*/

/*
** Copyright 1999-2005, Silicon Graphics, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of Silicon Graphics, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of Silicon Graphics, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/fixed/fixed.c,v 1.5 2005/05/11 15:22:46 artlab Exp $
*/

#include <assert.h>
#ifdef CHECK_ERRORS
#include <errno.h>
#endif
#include "int32.h"
#include "int64.h"
#include "float64.h"
#include "fixed.h"

#ifdef CHECK_ERRORS

static const char *
errorToString(const int error)
{
    switch (error) {
    case FIXED_NONE:
        return "no error";

    case FIXED_DOMAIN:
        return "domain";

    case FIXED_SINGULARITY:
        return "singularity";

    case FIXED_OVERFLOW:
        return "overflow";

    case FIXED_UNDERFLOW:
        return "underflow";

    case FIXED_TOTAL_LOSS:
        return "total loss of precision";

    case FIXED_PARTIAL_LOSS:
        return "partial loss of precision";

    default:
        return "<unknown error>";
    }
}

static int
typeToErrno(const int type)
{
    assert(type != FIXED_NONE);

    switch (type) {
    default:
        return 0;

    case FIXED_DOMAIN:
    case FIXED_SINGULARITY:
        return EDOM;

    case FIXED_OVERFLOW:
    case FIXED_UNDERFLOW:
    case FIXED_TOTAL_LOSS:
    case FIXED_PARTIAL_LOSS:
        return ERANGE;
    }
}

static int
default_error_handler(struct xmath_exception * const exception)
{
    assert(exception != NULL);

    switch (exception->type) {
    case FIXED_UNDERFLOW:
    case FIXED_OVERFLOW:
        /* ignore underflows and overflows by default */
        return 1;
    }

    return 0;
}

static xmatherr error_handler = default_error_handler;

xmatherr
xsetmatherr(xmatherr const handler)
{
    xmatherr old_handler;

    assert(handler);

    old_handler = error_handler;
    error_handler = handler;
    return old_handler;
}

static fixed
throwException(char const * const name, const int type, const fixed arg1, const fixed arg2,
    const fixed retval)
{
    struct xmath_exception exception;

    exception.type = type;
    exception.name = name;
    exception.arg1 = arg1;
    exception.arg2 = arg2;
    exception.retval = retval;

    if (error_handler(&exception))
        return exception.retval;

    errno = typeToErrno(type);
    xfprintf(stderr, "%s: %s, first = %f, second = %f, result = %f\n",
        exception.name, errorToString(exception.type), arg1, arg2, retval);

    return retval;
}

#endif /* ! CHECK_ERRORS */

int
xprintf(char const * const format, ...)
{
    VA_LIST arg;
    int length;

    VA_START(arg, format);
    length = xvfprintf(stdout, format, arg);
    VA_END(arg);

    return length;
}

int
xfprintf(FILEHANDLE const file, char const * const format, ...)
{
    VA_LIST arg;
    int length;

    VA_START(arg, format);
    length = xvfprintf(file, format, arg);
    VA_END(arg);

    return length;
}

int
xsprintf(char * const string, char const * const format, ...)
{
    VA_LIST arg;
    int length;

    VA_START(arg, format);
    length = xvsprintf(string, format, arg);
    VA_END(arg);

    return length;
}

int
xvprintf(char const * const format, VA_LIST arg)
{
    return xvfprintf(stdout, format, arg);
}

int
xvfprintf(FILEHANDLE const file, char const * const format, VA_LIST arg)
{
    char string[1024];

    xvsprintf(string, format, arg);
    return FPRINTF1(file, "%s", string);
} 

static float64 xtof64(fixed const f);

int
xvsprintf(char * const string, char const * const format, VA_LIST arg)
{
    int i, j, k, percent;
    char * buffer;

    buffer = (char *)MALLOC((strlen(format) + 1) * sizeof(char));

    assert(string);
    assert(format);
    assert(buffer);

    percent = 0;
    k = 0;

    for (i = j = 0; format[i] != '\0'; ++i) {

        if (!percent) {
            if (format[i] == '%') {
                percent = 1;
		k = 0;
                buffer[k++] = format[i];
            } else {
                string[j++] = format[i];
            }
            continue;
        }

        if (format[i] == 'n') {
            *VA_ARG(arg, int *) = j;
            continue;
        }

        buffer[k++] = format[i];
        percent = 0;

        switch (format[i]) {
        default:
            percent = 1;
            break;

        case '%':
            buffer[k] = '\0';
            j += SPRINT(&string[j], buffer);
            break;

        case 'c':
        case 'd':
        case 'i':
        case 'o':
        case 'u':
        case 'x':
        case 'X':
            buffer[k] = '\0';
            switch (format[i - 1]) {
            case 'I':
            case 'L':
                j += SPRINTF1(&string[j], buffer, VA_ARG(arg, long));
                break;

            case 'h':
            default:
                j += SPRINTF1(&string[j], buffer, VA_ARG(arg, int));
                break;
            }
            break;

        case 's':
            buffer[k] = '\0';
            j += SPRINTF1(&string[j], buffer, VA_ARG(arg, char *));
            break;

        case 'p':
            buffer[k] = '\0';
            j += SPRINTF1(&string[j], buffer, VA_ARG(arg, void *));
            break;

        case 'e':
        case 'E':
        case 'f':
        case 'g':
        case 'G':
            buffer[k] = '\0';
            j += fsprintf64(&string[j], buffer, xtof64(VA_ARG(arg, fixed)));
            break;
        }
    }

    FREE(buffer);

    string[j] = '\0';
    return j;
}

static fixed
f64tox(float64 const f64)
{
    fixed f;
    int count;

#ifdef CHECK_ERRORS
    float64 FLOAT_MAX, FLOAT_MIN;
    FLOAT_MAX.mantissa = FIXED_MAX;
    FLOAT_MAX.exponent = -16;
    FLOAT_MIN.mantissa = FIXED_MIN;
    FLOAT_MIN.exponent = -16;

    if (fcompare64(f64, FLOAT_MAX) > 0)
        return throwException("f64tox", FIXED_OVERFLOW, FIXED_MAX, 0, FIXED_MAX);

    if (fcompare64(f64, FLOAT_MIN) < 0)
        return throwException("f64tox", FIXED_OVERFLOW, FIXED_MIN, 0, FIXED_MIN);
#endif

    count = f64.exponent + 16;
    f = ishift32(f64.mantissa, count);

    /* round positive values */
    if (count < 0 && (f64.mantissa & (1 << (-count - 1)))) {
        if (f > 0)
            f = xadd(f, FIXED_EPSILON);
    }

#ifdef CHECK_ERRORS
    if (f == 0 && f64.mantissa != 0)
        return throwException("f64tox", FIXED_UNDERFLOW, 0, 0, 0);
#endif

    return f;
}

static float64
xtof64(fixed const f)
{
    float64 f64;

    f64.mantissa = f;
    f64.exponent = -16;

    return f64;
}

fixed
itox(int const i)
{
#ifdef CHECK_ERRORS
    if (i > ishift32(FIXED_MAX, -16))
        return throwException("itox", FIXED_OVERFLOW, i, 0, FIXED_MAX);

    if (i < ishift32(FIXED_MIN, -16))
        throwException("itox", FIXED_OVERFLOW, i, 0, FIXED_MIN);
#endif

    return i << 16;
}

int
xtoi(fixed const x)
{
    return ishift32(x, -16);
}

fixed
atox(char const * const s)
{
    return f64tox(atof64(s));
}

fixed
xneg(fixed const f)
{
#ifdef CHECK_ERRORS
    if (f == FIXED_MIN)
        return throwException("xneg", FIXED_OVERFLOW, 0, 0, -f);
#endif

    return -f;
}

fixed
xabs(fixed const f)
{
    if (f >= 0)
        return f;
    else
        return xneg(f);
}

fixed
xadd(fixed const x, fixed const y)
{
#ifdef CHECK_ERRORS
    if (x < 0 && y < 0) {
        if (FIXED_MIN - x > y)
            return throwException("xadd", FIXED_OVERFLOW, x, y, FIXED_MIN);
    }

    if (x > 0 && y > 0) {
        if (FIXED_MAX - x < y)
            return throwException("xadd", FIXED_OVERFLOW, x, y, FIXED_MAX);
    }
#endif

    return x + y;
}

fixed
xsub(fixed const x, fixed const y)
{
#ifdef CHECK_ERRORS
    if (x < 0 && y > 0) {
        if (FIXED_MIN + y > x)
            return throwException("xsub", FIXED_OVERFLOW, x, y, FIXED_MIN);
    }

    if (x > 0 && y < 0) {
        if (FIXED_MAX + y < x)
            return throwException("xsub", FIXED_OVERFLOW, x, y, FIXED_MAX);
    }
#endif

    return x - y;
}

fixed
xmul(fixed const x, fixed const y)
{
    int64 z;

    z = imul64(x, y);

#ifdef CHECK_ERRORS
    if (((z.hi & 0xFFFF8000) != 0) && ((z.hi & 0xFFFF8000) != 0xFFFF8000)) {
        if (z.hi & (1 << 31))
            return throwException("xmul", FIXED_OVERFLOW, x, y, FIXED_MIN);
        else
            return throwException("xmul", FIXED_OVERFLOW, x, y, FIXED_MAX);
    }
#endif

    z = ishift64(z, 16);

#ifdef CHECK_ERRORS
    if (z.hi == 0 && x != 0 && y != 0)
            return throwException("xmul", FIXED_UNDERFLOW, x, y, 0);
#endif

    return z.hi;
} 

#if 1

fixed
xdiv(fixed const x_, fixed const y_)
{
    float64 x, y, z;

#ifdef CHECK_ERRORS
    if (y_ == 0)
        return throwException("xdiv", FIXED_SINGULARITY, x_, y_,
            x_ < 0 ? FIXED_MIN : FIXED_MAX);
#endif

    x = xtof64(x_);
    y = xtof64(y_);

    z = fdiv64(x, y);

    return f64tox(z);
}

#else

/* not very accurate */
fixed
xdiv(fixed const x_, fixed const y)
{
    int64 x, z;

#ifdef CHECK_ERRORS
    if (y == 0)
        return throwException("xdiv", FIXED_SINGULARITY, x_, y,
            x_ < 0 ? FIXED_MIN : FIXED_MAX);
#endif

    x.hi = x_;
    x.lo = 0;

    z = idiv64(x, y);

#ifdef CHECK_ERRORS
    if (((z.hi & 0xFFFF8000) != 0) && ((z.hi & 0xFFFF8000) != 0xFFFF8000)) {
        if (z.hi & (1 << 31))
            return throwException("xdiv", FIXED_OVERFLOW, x_, y, FIXED_MIN);
        else
            return throwException("xdiv", FIXED_OVERFLOW, x_, y, FIXED_MAX);
    }
#endif

    z = ishift64(z, 16);

#ifdef CHECK_ERRORS
    if (z.hi == 0 && x_ != 0)
            return throwException("xdiv", FIXED_UNDERFLOW, x_, y, 0);
#endif

    return z.hi;
}

#endif

fixed
xsqrt(fixed const x)
{
    uint32_t root, rh, rl, tdiv, count;

#ifdef CHECK_ERRORS
    if (x < 0)
        return throwException("xsqrt", FIXED_DOMAIN, x, 0, FIXED_NAN);
#endif

    root = 0;
    rh = 0;
    rl = x;
    count = 15 + (16 >> 1);

    do {
        rh = (rh << 2) | (rl >> 30);
        rl <<= 2;
        root <<= 1;
        tdiv = (root << 1) + 1;
        if (rh >= tdiv) {
            rh -= tdiv;
            root += 1;
        }
    } while (count-- != 0);

    return root;
}

fixed
xldexp(fixed const value, int const exp)
{

    if (exp > 0) {
#ifdef CHECK_ERRORS
        if (value < 0) {
            if (ishift32(FIXED_MIN, -exp) > value)
                return throwException("xldexp", FIXED_OVERFLOW, value, itox(exp), FIXED_MIN);
        } else {
            if (ishift32(FIXED_MAX, -exp) < value)
                return throwException("xldexp", FIXED_OVERFLOW, value, itox(exp), FIXED_MAX);
        }
#endif
        return ishift32(value, exp);
    }

    if (exp < 0) {
         const int result = ishift32(value, exp);

#ifdef CHECK_ERRORS
         if (value != 0 && result == 0)
            return throwException("xldexp", FIXED_UNDERFLOW, value, itox(exp), result);
#endif

         return result;
    }

    return value;
}

fixed
xmodx(const fixed value, fixed * const ip)
{
    assert(ip != NULL);

    if (value == FIXED_MIN) {
        *ip = 0xFFFF0000;
        return FIXED_MIN - *ip;
    }

    if (value < 0) {
        *ip = - (fixed)(-value & 0xFFFF0000);
        return - (fixed)(-value & 0x0000FFFF);

    } else {
        *ip = (fixed)(value & 0xFFFF0000);
        return (fixed)(value & 0x0000FFFF);
    }
}

fixed
xmod(fixed const x, fixed const y)
{
    fixed result, ip;

#ifdef CHECK_ERRORS
    if (y == 0)
        return throwException("xmod", FIXED_DOMAIN, x, y, FIXED_NAN);
#endif

    result = xdiv(x, y);
    xmodx(result, &ip);

    return x - y * xtoi(ip);
}

fixed
xfrexp(fixed const value, int * const exp)
{
    fixed abs_value;

    assert(exp);

    if (value == FIXED_ZERO) {
        *exp = 0;
        return FIXED_ZERO;
    }

    if (value == FIXED_MIN) {
        *exp = 16;
        return 0xFFFF8000;
    }

    abs_value = xabs(value);
    *exp = 0;
    if (abs_value >= FIXED_ONE) {
        while (abs_value >= FIXED_ONE) {
            abs_value >>= 1;
            ++ *exp;
        }
    } else if (abs_value < FIXED_HALF) {
        while (abs_value < FIXED_HALF) {
            abs_value <<= 1;
            -- *exp;
        }
    }

    if (value < 0)
        return -abs_value;
    else
        return abs_value;
}

fixed
xceil(fixed const x)
{
    if (x & 0xFFFF) {
        return xadd(x & 0xFFFF0000, FIXED_ONE);
    } else {
        return x & 0xFFFF0000;
    }
}

fixed
xfloor(fixed const x)
{
    return x & 0xFFFF0000;
}

#if 0
/* limited to (-1, Inf) */
static fixed
xexp_naive(const fixed value)
{
    fixed s, sn, a;
    int i;

    assert(value > itox(-1));

    sn = a = itox(1);
    i = 0;

    do {
        ++i;
        s = sn;
        a = xmul(a, value);
        a = xdiv(a, itox(i));
        sn += a;
    } while (sn != s);

    return sn;
}

fixed
xexp(fixed const value)
{
    fixed sn, z;
    int i, m;

    m = 0;
    z = value;

    while (xabs(z) > FIXED_ONE) {
        ++m;
        z /= 2;
    }
    /* {z in (-1, +1)} */

    sn = xexp_naive(z);

    for (i = 1; i <= m; ++i) {
        sn = xmul(sn, sn);
    }

    return sn;
}

#else

fixed
xexp(fixed const x)
{
    return f64tox(fexp64(xtof64(x)));
}

#endif

fixed
xlog(fixed const x)
{
#ifdef CHECK_ERRORS
    if (x <= 0)
        return throwException("xlog", FIXED_DOMAIN, x, 0, FIXED_NAN);
#endif

    return f64tox(flog64(xtof64(x)));
}

fixed
xlog10(fixed const x)
{
#ifdef CHECK_ERRORS
    if (x <= 0)
        return throwException("xlog10", FIXED_DOMAIN, x, 0, FIXED_NAN);
#endif

    return f64tox(flog1064(xtof64(x)));
}

fixed
xpow(fixed const x, fixed const y)
{
    return f64tox(fpow64(xtof64(x), xtof64(y)));
}

/* limited to [0, FIXED_PI_2] */
static fixed
xsin_naive(fixed const angle)
{
    fixed sn, s, a, angle2;
    int i;

    assert(angle >= FIXED_ZERO && angle <= FIXED_PI_2);

    i = 1;
    angle2 = xmul(angle, angle);
    a = angle;
    sn = a;

    do {
        i += 2;
        s = sn;
        a = xmul(a, angle2);
        a = xdiv(a, itox((i - 1) * i));
        if (((i / 2) % 2) == 0)
            sn += a;
        else
            sn -= a;

    } while (sn != s);

    return sn; 
}

fixed
xsin(fixed const angle)
{
    fixed new_angle;
    int negate;

    new_angle = xmod(angle, FIXED_PI * 2);

    if (new_angle < 0) {
        new_angle += FIXED_PI * 2;
    }

    assert(new_angle >= FIXED_ZERO && new_angle <= FIXED_PI * 2);

    if (new_angle <= FIXED_PI_2) {
        negate = 0;

    } else if (new_angle <= FIXED_PI) {
        new_angle = FIXED_PI - new_angle;
        negate = 0;

    } else if (new_angle <= FIXED_PI * 3 / 2) {
        new_angle = new_angle - FIXED_PI;
        negate = 1;

    } else {
        new_angle = FIXED_PI * 2 - new_angle;
        negate = 1;
    }

    if (negate)
        return xneg(xsin_naive(new_angle));
    else
        return xsin_naive(new_angle);
}

fixed
xcos(fixed const angle)
{
    return xsin(angle + FIXED_PI_2);
}

fixed
xtan(fixed const angle)
{
    return xdiv(xsin(angle), xcos(angle));
}

fixed
xsinh(fixed const arg)
{
    return (xexp(arg) - xexp(-arg)) / 2;
}

fixed
xcosh(fixed const arg)
{
    return (xexp(arg) + xexp(-arg)) / 2;
}

fixed
xtanh(fixed const arg)
{
    return xdiv(xsinh(arg), xcosh(arg));
}

fixed
xatan(fixed const x)
{
    fixed sn, a, xn, x2;
    int i;

    if (x < 0)
        return -xatan(x);

    if (x > FIXED_HALF) {
        fixed const n = xdiv(FIXED_ONE, x);

        return
            xatan(xdiv(FIXED_ONE, n + FIXED_ONE)) +
            xatan(xdiv(FIXED_ONE, xmul(n, n + FIXED_ONE) + FIXED_ONE));
    }

    i = 1;
    x2 = xmul(x, x);
    xn = x;
    a = xn;
    sn = a;

    do {
        i += 2;
        xn = xmul(xn, x2);
        a = xn;
        a = xdiv(a, itox(i));
        if (((i / 2) % 2) == 0)
            sn += a;
        else
            sn -= a;

    } while (a != 0);

    return sn;
}

fixed
xatan2(fixed const y, fixed const x)
{
    if (y < 0)
        return -xatan2(-y, x);

    if (x < 0)
        return FIXED_PI - xatan2(y, -x);

    if (x == 0 && y == 0)
        return 0;

    return xatan(xdiv(y, x));
}

fixed
xasin(fixed const x)
{
#ifdef CHECK_ERRORS
    if (xabs(x) > FIXED_ONE)
        return throwException("xasin", FIXED_DOMAIN, x, 0, FIXED_NAN);
#endif
        
    return FIXED_PI_2 - xacos(x);
}

fixed
xacos(fixed const x)
{
#ifdef CHECK_ERRORS
    if (xabs(x) > FIXED_ONE)
        return throwException("xacos", FIXED_DOMAIN, x, 0, FIXED_NAN);
#endif

    return xatan2(xsqrt(FIXED_ONE - xmul(x, x)), x);
}
