/*
** Copyright (c) 2008 The Khronos Group Inc.
** 
** All Rights Reserved.  This code is protected by copyright laws and 
** contains material proprietary to the Khronos Group, Inc.  This is 
** UNPUBLISHED PROPRIETARY SOURCE CODE that may not be disclosed in whole 
** or in part to third parties, and may not be reproduced, republished, 
** distributed, transmitted, displayed, broadcast or otherwise exploited 
** in any manner without the express prior written permission of Khronos 
** Group.
** 
** The receipt or possession of this code does not convey any rights to 
** reproduce, disclose, or distribute its contents, or to manufacture, 
** use, or sell anything that it may describe, in whole or in part other 
** than under the terms of the Khronos Adopters Agreement or Khronos 
** Conformance Test Source License Agreement as executed between Khronos 
** and the recipient.
** 
** For the avoidance of doubt, this code when provided: a) under the 
** Khronos Conformance Test Source License Agreement is for the sole 
** purpose of creating conformance tests for delivery to Khronos and does 
** not provide for formally testing products or use of Khronos trademarks 
** on conformant products; b) under the Khronos Adopters Agreement is for 
** the sole purpose of formally administering tests to products pursuant 
** to the Khronos Conformance Process Document.
** 
** Khronos, OpenKODE, OpenVG, OpenWF, glFX, OpenMAX and OpenSL ES are 
** trademarks of the Khronos Group Inc.  COLLADA is a trademark of Sony 
** Computer Entertainment Inc. used by permission by Khronos.  OpenGL and 
** OpenML are registered trademarks and the OpenGL ES logo is a trademark 
** of Silicon Graphics Inc. used by permission by Khronos. 
** 
** Use, duplication or disclosure by the Government is subject to 
** restrictions as set forth in subdivision (c)(1)(ii) of the Rights in 
** Technical Data and Computer Software clause at DFARS 252.227-7013, and/
** or in similar or successor clauses in the FAR, DOD or NASA FAR 
** Supplement. Unpublished rights reserved under the Copyright Laws of 
** the United States and other countries.
**
**
** $Header: /home/cvs/OpenGL-ES/1.0/ogles/conform/fixed/hgdiv64.c,v 1.2 2005/05/11 15:22:46 artlab Exp $
*/

/*------------------------------------------------------------------------
 *
 * Signed 64:32 integer division
 * -----------------------------------------
 *
 * Description:     Routine for performing signed 64:32 integer division
 *
 * Note:            This routine is a replacement for the idiv64() routine
 *                  in OpenGL ES conformance tests. The code is written
 *                  in portable C and could be easily further optimized
 *                  by converting the primitive 64-bit ops to use
 *                  assembler sequences (see Hybrid's 64-bit integer
 *                  library for fully optimized implementations).
 *                  Also, if we know that the platform has a shifter with 
 *                  modulo > 32 (such as ARM), many of the shifter ops 
 *                  could be simplified.
 *
 *                  The code does not use external libraries or
 *                  functions (i.e. it does not use the 64-bit primitive
 *                  functions in the conformance test package). It relies
 *                  only on the definitions of int32_t, uint32_t and
 *                  int64.
 *
 * Comments & Bugs: Report to wili@hybrid.fi
 *
 * $Id: hgdiv64.c,v 1.2 2005/05/11 15:22:46 artlab Exp $
 * $Date: 2005/05/11 15:22:46 $
 * $Author: artlab $ *
 *
 *----------------------------------------------------------------------*/

#include "int64.h"

/*-------------------------------------------------------------------*//*!
 * \brief   Performs 64:32 signed division
 * \param   x   64-bit signed integer to be divided
 * \param   y   32-bit signed integer divisor
 * \return  (x/y) (64-bit signed)
 * \note    If y is zero, the return value is zero 
 *//*-------------------------------------------------------------------*/

int64 idiv64 (
    int64   x,
    int32_t y)
{
    /*-----------------------------------------------------------------------
     * Comparison macros (these could be considerably improved for platforms
     * that lack predicates by converting the code into logic ops).
     *---------------------------------------------------------------------*/

    #define hgLessThan64u(a,b)     ((a.hi < b.hi) || ((a.hi == b.hi) && (a.lo < b.lo)))
    #define hgLessThan64_32u(a,b)  ((!(a.hi)) && ((a.lo) < (b)))

    /*-----------------------------------------------------------------------
     * Unsigned 64-bit integer
     *---------------------------------------------------------------------*/

    typedef struct
    {
        uint32_t    hi;                     /*!< High 32 bits   */
        uint32_t    lo;                     /*!< Low 32 bits    */
    } HGuint64;

    /*-----------------------------------------------------------------------
     * Compute sign of result, then take abs() of x and y so that we can
     * perform an unsigned division.
     *---------------------------------------------------------------------*/

    HGuint64    a, c, z, o;
    int32_t     sign    = (x.hi ^ y);                       /* negative if result signed */
    uint32_t    b       = (uint32_t)((y < 0) ? -y : y);     /* abs(y) */

    if (!y)                                                 /* division by zero! */
    {   
        x.hi = x.lo = 0;
        return x;
    }

    a.hi = x.hi, a.lo = x.lo;

    if ((int32_t)(x.hi) < 0)                    /* negate? */
    {
        int32_t mask    = -(int32_t)( (uint32_t)(x.lo | (-(int32_t)x.lo)) >> 31 );
        a.hi            = ~a.hi - (~mask);
        a.lo            = (uint32_t)-(int32_t)x.lo;
    }

    /*-----------------------------------------------------------------------
     * Count leading zeroes of a and b, and align the values (this minimizes
     * the number of iterations needed later). Note that this could be
     * done by the clz (or bsr) instruction if assembler was used.
     *---------------------------------------------------------------------*/

    {
        uint32_t bb = b, aa = a.hi, m = 32;

        #define OP(Y) if (aa<(1u<<(Y))) m-=(Y); else aa >>= (Y); if (bb<(1u<<(Y))) m+=(Y); else bb >>= (Y);

        OP(16); 
        OP(8); 
        OP(4); 
        OP(2); 

        m += (2 >> bb) - (2 >> aa);

        #undef OP

        if (m >= 32)    /* 64-bit shift left */
        {
            z.hi = b << (m-32);
            o.hi = 1 << (m-32);
            z.lo = 0;
            o.lo = 0;
        } else
        {
            z.hi = m ? (b >> (32-m)) : 0;
            o.hi = 0;
            z.lo = b << m;
            o.lo = 1 << m;
        }
    }

    /*-----------------------------------------------------------------------
     * Iterations [0,63]. Note that the inner loop can be executed at
     * most 63 times *in total*.
     *---------------------------------------------------------------------*/

    c.hi = c.lo = 0;
    
    do
    {
        int iter = 0;

        while (hgLessThan64u(a,z))
        {
            z.lo = (z.lo >> 1) | (z.hi << 31);              /* 64-bit lsr 1 */
            z.hi = (z.hi >> 1);
            iter++;
        }

        a.hi -= z.hi + ((a.lo < z.lo) ? 1 : 0);             /* 64-bit sub */
        a.lo -= z.lo;

        if (iter >= 32)                                     /* 64-bit lsr */
        {
            o.lo = o.hi >> (iter-32);
            o.hi = 0;
        } else if (iter)
        {
            o.lo = (o.lo >> iter) | (o.hi << (32-iter));
            o.hi >>= iter;
        }

        c.hi |= o.hi;                                       /* 64-bit or */
        c.lo |= o.lo;

    } while (!(hgLessThan64_32u(a,b)));

    /*-----------------------------------------------------------------------
     * If the result of the division should be negative, negate the result.
     *---------------------------------------------------------------------*/

    x.hi = c.hi, x.lo = c.lo;

    if (sign < 0)                                           /* apply sign? */
    {
        int32_t mask    = -(int32_t)( ((uint32_t)(x.lo | (-(int32_t)x.lo))) >> 31 );
        x.hi            = ~x.hi - (~mask);
        x.lo            = -(int32_t)x.lo;
    }

    /*-----------------------------------------------------------------------
     * Done
     *---------------------------------------------------------------------*/

    return x;

    #undef hgLessThan64u
    #undef hgLessThan64_32u
}

/*----------------------------------------------------------------------*/

